@@ -12,17 +12,17 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from collections.abc import Sequence
+import functools
 from functools import partial
 import logging
-from typing import (Any, Callable, FrozenSet, List, Optional, Sequence, Tuple,
-                    Union)
+from typing import Any, Callable, Optional, Union
 import types
 
 import numpy as np
 
-import jax
-from jax.tree_util import tree_flatten, tree_unflatten, tree_structure, keystr
 from jax._src import ad_util
+from jax._src import api
 from jax._src import core
 from jax._src import dispatch
 from jax._src import linear_util as lu
@@ -31,6 +31,7 @@ from jax._src import source_info_util
 from jax._src import traceback_util
 from jax._src import util
 from jax._src.api_util import flatten_fun, shaped_abstractify
+from jax._src.config import config
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
@@ -39,6 +40,7 @@ from jax._src.lax import lax as lax_internal
 from jax._src.lax import convolution as lax_convolution
 from jax._src.lib.mlir.dialects import hlo
 from jax._src.traceback_util import api_boundary
+from jax._src.tree_util import tree_flatten, tree_unflatten, tree_structure, keystr
 from jax._src.util import (unzip2, wraps, split_list, partition_list, safe_map,
                            safe_zip, merge_lists, weakref_lru_cache)
 
@@ -50,8 +52,6 @@ zip = safe_zip
 
 logger = logging.getLogger(__name__)
 
-allowed_effects: effects.EffectTypeSet = effects.remat_allowed_effects
-
 ### Policies
 
 def everything_saveable(*_, **__) -> bool:
@@ -132,7 +132,7 @@ checkpoint_policies = types.SimpleNamespace(
 @api_boundary
 def checkpoint(fun: Callable, *, prevent_cse: bool = True,
                policy: Optional[Callable[..., bool]] = None,
-               static_argnums: Union[int, Tuple[int, ...]] = (),
+               static_argnums: Union[int, tuple[int, ...]] = (),
                ) -> Callable:
   """Make ``fun`` recompute internal linearization points when differentiated.
 
@@ -346,9 +346,14 @@ class WrapHashably:
 # See api_benchmark.py:bench_remat_eager_retracing_overheads_static_argnums.
 # On that benchmark, including this caching makes a ~10x difference (which can
 # be made arbitrary large by involving larger functions to be traced).
-@weakref_lru_cache
-def _dyn_args_fun(fun: Callable, static_argnums: FrozenSet[int],
-                  static_args: Tuple[WrapHashably, ...], nargs: int):
+def _dyn_args_fun(fun: Callable, static_argnums: frozenset[int],
+                  static_args: tuple[WrapHashably, ...], nargs: int):
+  if any(isinstance(x.val, core.Tracer) for x in static_args):
+    return _dyn_args_fun_uncached(fun, static_argnums, static_args, nargs)
+  return _dyn_args_fun_cached(fun, static_argnums, static_args, nargs)
+
+def _dyn_args_fun_uncached(fun: Callable, static_argnums: frozenset[int],
+                           static_args: tuple[WrapHashably, ...], nargs: int):
   def new_fun(*dyn_args, **kwargs):
     static_args_, dyn_args_ = iter(static_args), iter(dyn_args)
     full_args = [next(static_args_).val if i in static_argnums
@@ -356,6 +361,8 @@ def _dyn_args_fun(fun: Callable, static_argnums: FrozenSet[int],
     return fun(*full_args, **kwargs)
   return new_fun
 
+_dyn_args_fun_cached = weakref_lru_cache(_dyn_args_fun_uncached)
+
 # This helper is similar to those in control_flow/common.py, but with
 # remat-specific errors.
 @weakref_lru_cache
@@ -382,14 +389,14 @@ def _trace_to_jaxpr(fun, in_tree, in_avals):
 
 ### Utilities
 
-def saved_residuals(f, *args, **kwargs) -> List[Tuple[core.AbstractValue, str]]:
+def saved_residuals(f, *args, **kwargs) -> list[tuple[core.AbstractValue, str]]:
   in_leaves, in_tree = tree_flatten((args, kwargs))
 
   def f_(*args):
     args, kwargs = tree_unflatten(in_tree, args)
     return f(*args, **kwargs)
 
-  out = jax.make_jaxpr(lambda *args: jax.linearize(f_, *args)[1],
+  out = api.make_jaxpr(lambda *args: api.linearize(f_, *args)[1],
                        return_shape=True)(*in_leaves)
   assert isinstance(out, tuple)
   jaxpr_, out_shape = out
@@ -400,7 +407,7 @@ def saved_residuals(f, *args, **kwargs) -> List[Tuple[core.AbstractValue, str]]:
   arg_info = pe.arg_info_all(dbg)
   return _saved_residuals(jaxpr, arg_info)
 
-def _saved_residuals(jaxpr, arg_info) -> List[Tuple[core.AbstractValue, str]]:
+def _saved_residuals(jaxpr, arg_info) -> list[tuple[core.AbstractValue, str]]:
   res_lits = [x for x in jaxpr.outvars if     isinstance(x, core.Literal)]
   res_vars = {x for x in jaxpr.outvars if not isinstance(x, core.Literal)}
 
@@ -474,11 +481,11 @@ def remat_jvp(primals, tangents, jaxpr, prevent_cse, differentiated, policy):
   return out_primals, out_tangents
 ad.primitive_jvps[remat_p] = remat_jvp
 
-allowed_effects.add_type(lax_internal.InOutFeedEffect)
+effects.remat_allowed_effects.add_type(lax_internal.InOutFeedEffect)
 
 def remat_partial_eval(trace, *tracers, jaxpr, **params):
   assert not jaxpr.constvars
-  disallowed_effects = allowed_effects.filter_not_in(jaxpr.effects)
+  disallowed_effects = effects.remat_allowed_effects.filter_not_in(jaxpr.effects)
   if disallowed_effects:
     raise NotImplementedError(
         'Effects not supported in partial-eval of `checkpoint`/`remat`: '
@@ -522,7 +529,7 @@ def remat_partial_eval(trace, *tracers, jaxpr, **params):
     res_invars, _ = partition_list(staged_unk, jaxpr_unknown.invars[num_res:])
     res_outvars = jaxpr_known.outvars[len(jaxpr_known.outvars) - num_res:]
     body_res = _saved_residuals(jaxpr_known.replace(outvars=res_outvars), None)
-    logger.log(logging.WARNING if jax.config.jax_log_checkpoint_residuals
+    logger.log(logging.WARNING if config.jax_log_checkpoint_residuals
                else logging.DEBUG,
                'remat-decorated function ' +
                'saving inputs with shapes:\n' * bool(res_invars) +
@@ -570,7 +577,7 @@ ad.reducing_transposes[remat_p] = remat_transpose
 def transpose_jaxpr(jaxpr: core.ClosedJaxpr, in_linear: Union[bool, Sequence[bool]],
                     out_zeros: Union[bool, Sequence[bool]],
                     reduce_axes: Sequence[core.AxisName],
-                    ) -> Tuple[core.ClosedJaxpr, List[bool]]:
+                    ) -> tuple[core.ClosedJaxpr, list[bool]]:
   if type(in_linear) is bool:
     in_linear = (in_linear,) * len(jaxpr.in_avals)
   if type(out_zeros) is bool:
@@ -631,8 +638,8 @@ batching.axis_primitive_batchers[remat_p] = partial(remat_vmap, None)
 batching.spmd_axis_primitive_batchers[remat_p] = remat_vmap
 
 # TODO(mattjj,sharadmv): de-duplicate with pe.dce_jaxpr_call_rule
-def remat_dce(used_outputs: List[bool], eqn: core.JaxprEqn
-              ) -> Tuple[List[bool], Optional[core.JaxprEqn]]:
+def remat_dce(used_outputs: list[bool], eqn: core.JaxprEqn
+              ) -> tuple[list[bool], Optional[core.JaxprEqn]]:
   new_jaxpr, used_inputs = pe.dce_jaxpr(eqn.params['jaxpr'], used_outputs)
   new_params = dict(eqn.params, jaxpr=new_jaxpr)
   if not any(used_inputs) and not any(used_outputs) and not new_jaxpr.effects:
@@ -652,7 +659,7 @@ def remat_lowering(*args, jaxpr: core.Jaxpr, prevent_cse: bool,
   assert not jaxpr.constvars
 
   if differentiated and prevent_cse:
-    if jax.config.jax_remat_opt_barrier:
+    if config.jax_remat_opt_barrier:
       translation_rule = _remat_translation_using_opt_barrier
     elif is_gpu_platform:
       translation_rule = _remat_translation_using_while
@@ -661,7 +668,7 @@ def remat_lowering(*args, jaxpr: core.Jaxpr, prevent_cse: bool,
   else:
     translation_rule = lambda *args, jaxpr: core.eval_jaxpr(jaxpr, (), *args)
 
-  return jax.named_call(translation_rule, name="remat")(*args, jaxpr=jaxpr)
+  return api.named_call(translation_rule, name="remat")(*args, jaxpr=jaxpr)
 
 def _remat_translation_using_opt_barrier(*args, jaxpr: core.Jaxpr):
   args = _optimization_barrier(args)
@@ -670,9 +677,9 @@ def _remat_translation_using_opt_barrier(*args, jaxpr: core.Jaxpr):
 # TODO(mattjj): add core utility for 'create dummy value for this type'?
 def _dummy_like(aval: core.AbstractValue) -> Any:
   if aval is core.abstract_token:
-    return jax.lax.create_token()
+    return lax_internal.create_token()
   elif isinstance(aval, (core.ShapedArray, core.DShapedArray)):
-    return jax.lax.broadcast(lax_internal.empty(aval.dtype), aval.shape)  # type: ignore
+    return lax_internal.broadcast(lax_internal.empty(aval.dtype), aval.shape)  # type: ignore
   else:
     raise ValueError(aval)
 
@@ -682,11 +689,13 @@ def _remat_translation_using_while(*args, jaxpr: core.Jaxpr):
   #     result = eval_jaxpr(*args)
   #  }
   # The loop carry is a tuple: (counter, result, args)
+  from jax._src.lax import control_flow as lax_control_flow
+
   avals_out = tuple(v.aval for v in jaxpr.outvars)
   carry_init = (np.int32(0), tuple(map(_dummy_like, avals_out)), args)
   def cond(carry):
     counter, _, _ = carry
-    unif = jax.lax.rng_uniform(np.int32(1), np.int32(2), shape=())
+    unif = lax_internal.rng_uniform(np.int32(1), np.int32(2), shape=())
     return counter < unif
 
   def body(carry):
@@ -694,7 +703,7 @@ def _remat_translation_using_while(*args, jaxpr: core.Jaxpr):
     results = core.eval_jaxpr(jaxpr, (), *args)
     return (counter + 1, tuple(results), args)
 
-  carry_res = jax.lax.while_loop(cond, body, carry_init)
+  carry_res = lax_control_flow.while_loop(cond, body, carry_init)
   return carry_res[1]
 
 def _remat_translation_using_cond(*args, jaxpr: core.Jaxpr):
@@ -703,6 +712,8 @@ def _remat_translation_using_cond(*args, jaxpr: core.Jaxpr):
   #    return eval_jaxpr(*args)
   #  else:
   #    return 0
+  from jax._src.lax import control_flow as lax_control_flow
+
   avals_out = tuple(v.aval for v in jaxpr.outvars)
 
   def remat_comp(*args):
@@ -710,8 +721,8 @@ def _remat_translation_using_cond(*args, jaxpr: core.Jaxpr):
   def dummy_comp(*args):
     return tuple(map(_dummy_like, avals_out))
 
-  unif = jax.lax.rng_uniform(np.float32(0), np.float32(1), shape=())
-  return jax.lax.cond(unif < np.float32(2), remat_comp, dummy_comp, *args)
+  unif = lax_internal.rng_uniform(np.float32(0), np.float32(1), shape=())
+  return lax_control_flow.cond(unif < np.float32(2), remat_comp, dummy_comp, *args)
 
 mlir.register_lowering(
     remat_p, mlir.lower_fun(remat_lowering, multiple_results=True))
@@ -760,3 +771,63 @@ def name_batcher(args, dims, *, name):
   (x,), (d,) = args, dims
   return name_p.bind(x, name=name), d
 batching.primitive_batchers[name_p] = name_batcher
+
+
+@functools.wraps(checkpoint)
+def checkpoint_wrapper(
+    fun: Callable,
+    *,
+    concrete: bool = False,
+    prevent_cse: bool = True,
+    static_argnums: Union[int, tuple[int, ...]] = (),
+    policy: Optional[Callable[..., bool]] = None,
+) -> Callable:
+  if concrete:
+    msg = ("The 'concrete' option to jax.checkpoint / jax.remat is deprecated; "
+           "in its place, you can use its `static_argnums` option, and if "
+           "necessary the `jax.ensure_compile_time_eval()` context manager.\n"
+           "\n"
+           "For example, if using `concrete=True` for an `is_training` flag:\n"
+           "\n"
+           "  from functools import partial\n"
+           "\n"
+           "  @partial(jax.checkpoint, concrete=True)\n"
+           "  def foo(x, is_training):\n"
+           "    if is_training:\n"
+           "      return f(x)\n"
+           "    else:\n"
+           "      return g(x)\n"
+           "\n"
+           "replace it with a use of `static_argnums`:\n"
+           "\n"
+           "  @partial(jax.checkpoint, static_argnums=(1,))\n"
+           "  def foo(x, is_training):\n"
+           "    ...\n"
+           "\n"
+           "If jax.numpy operations need to be performed on static arguments, "
+           "we can use the `jax.ensure_compile_time_eval()` context manager. "
+           "For example, we can replace this use of `concrete=True`\n:"
+           "\n"
+           "  @partial(jax.checkpoint, concrete=True)\n"
+           "  def foo(x, y):\n"
+           "    if y > 0:\n"
+           "      return f(x)\n"
+           "    else:\n"
+           "      return g(x)\n"
+           "\n"
+           "with this combination of `static_argnums` and "
+           "`jax.ensure_compile_time_eval()`:\n"
+           "\n"
+           "  @partial(jax.checkpoint, static_argnums=(1,))\n"
+           "  def foo(x, y):\n"
+           "    with jax.ensure_compile_time_eval():\n"
+           "      y_pos = y > 0\n"
+           "    if y_pos:\n"
+           "      return f(x)\n"
+           "    else:\n"
+           "      return g(x)\n"
+           "\n"
+           "See https://jax.readthedocs.io/en/latest/jep/11830-new-remat-checkpoint.html\n")
+    raise NotImplementedError(msg)
+  return checkpoint(fun, prevent_cse=prevent_cse, policy=policy,
+                    static_argnums=static_argnums)
