@@ -14,24 +14,23 @@
 from __future__ import annotations
 
 import types
-from typing import Any, Callable, Dict, TypeVar, Union, cast
-
-from jax.tree_util import register_pytree_node
+from typing import Any, Callable, TypeVar, cast
 
 from jax._src import core
+from jax._src import traceback_util
 from jax._src.core import (lattice_join, Primitive, valid_jaxtype,
                            raise_to_shaped, get_aval)
-from jax._src.util import safe_map
+from jax._src.tree_util import register_pytree_node
 from jax._src.typing import Array, ArrayLike
+from jax._src.util import safe_map
 
-from jax._src import traceback_util
 traceback_util.register_exclusion(__file__)
 
 T = TypeVar('T')
 
 map = safe_map
 
-jaxval_adders: Dict[type, Callable[[ArrayLike, ArrayLike], Array]] = {}
+jaxval_adders: dict[type, Callable[[ArrayLike, ArrayLike], Array]] = {}
 
 def add_jaxvals(x: ArrayLike, y: ArrayLike) -> Array:
   return add_jaxvals_p.bind(x, y)
@@ -47,9 +46,9 @@ def add_impl(xs, ys):
 def add_abstract(xs, ys):
   return lattice_join(xs, ys)
 
-jaxval_zeros_likers: Dict[type, Callable[[Any], Array]] = {}
+jaxval_zeros_likers: dict[type, Callable[[Any], Array]] = {}
 
-def instantiate(z: Union[Zero, Array]) -> Array:
+def instantiate(z: Zero | Array) -> Array:
   if type(z) is Zero:
     return zeros_like_aval(z.aval)
   return cast(Array, z)
@@ -57,7 +56,7 @@ def instantiate(z: Union[Zero, Array]) -> Array:
 def zeros_like_aval(aval: core.AbstractValue) -> Array:
   return aval_zeros_likers[type(aval)](aval)
 
-aval_zeros_likers: Dict[type, Callable[[Any], Array]] = {}
+aval_zeros_likers: dict[type, Callable[[Any], Array]] = {}
 
 def zeros_like_jaxval(val: ArrayLike) -> Array:
   return zeros_like_p.bind(val)
@@ -123,9 +122,9 @@ class SymbolicZero:
 JaxTypeOrTracer = Any
 
 def replace_internal_symbolic_zeros(
-    x: Union[JaxTypeOrTracer, Zero]) -> Union[JaxTypeOrTracer, SymbolicZero]:
+    x: JaxTypeOrTracer | Zero) -> JaxTypeOrTracer | SymbolicZero:
   return SymbolicZero(x.aval) if type(x) is Zero else x
 
 def replace_rule_output_symbolic_zeros(
-    x: Union[JaxTypeOrTracer, SymbolicZero]) -> Union[JaxTypeOrTracer, Zero]:
+    x: JaxTypeOrTracer | SymbolicZero) -> JaxTypeOrTracer | Zero:
   return Zero(x.aval) if type(x) is SymbolicZero else x
