@@ -14,39 +14,41 @@
 
 from __future__ import annotations
 
+from collections import defaultdict
+import enum
 import math
 import operator as op
 import numpy as np
 import functools
-from typing import (Sequence, Tuple, Callable, Union, Optional, cast, List, Set,
-                    TYPE_CHECKING)
+from typing import Any, Callable, cast, TYPE_CHECKING
+from collections.abc import Sequence
 
-import jax
 from jax._src import abstract_arrays
+from jax._src import api
 from jax._src import api_util
 from jax._src import basearray
 from jax._src import core
 from jax._src import dispatch
 from jax._src import dtypes
 from jax._src import profiler
+from jax._src import tree_util
+from jax._src import xla_bridge
 from jax._src.config import config
-from jax._src.util import use_cpp_class, use_cpp_method
 from jax._src.lib import xla_client as xc
-from jax._src import api
-from jax._src.typing import ArrayLike
-from jax.interpreters import mlir
+from jax._src.interpreters import mlir
 from jax._src.interpreters import pxla
 from jax._src.interpreters import xla
 from jax._src.sharding import Sharding
 from jax._src.sharding_impls import (
     SingleDeviceSharding, XLACompatibleSharding, PmapSharding,
     device_replica_id_map, hashed_index)
+from jax._src.typing import ArrayLike
+from jax._src.util import safe_zip, unzip3, use_cpp_class, use_cpp_method
 
-Shape = Tuple[int, ...]
+Shape = tuple[int, ...]
 Device = xc.Device
-DeviceArray = xc.Buffer
-Index = Tuple[slice, ...]
-
+Index = tuple[slice, ...]
+PRNGKeyArrayImpl = Any  # TODO(jakevdp): fix cycles and import this.
 
 class Shard:
   """A single data shard of an Array.
@@ -61,7 +63,7 @@ class Shard:
   """
 
   def __init__(self, device: Device, sharding: Sharding, global_shape: Shape,
-               data: Optional[ArrayImpl] = None):
+               data: None | ArrayImpl | PRNGKeyArrayImpl = None):
     self._device = device
     self._sharding = sharding
     self._global_shape = global_shape
@@ -108,29 +110,6 @@ def _reconstruct_array(fun, args, arr_state, aval_state):
   return jnp_value
 
 
-def _single_device_array_from_buf(buf, committed) -> ArrayImpl:
-  if isinstance(buf, ArrayImpl) and buf._committed == committed:  # type: ignore
-    return buf
-  db = dispatch._set_aval(buf)
-  return ArrayImpl(db.aval, SingleDeviceSharding(db.device()), [db],
-                   committed=committed, _skip_checks=True)
-
-
-def _is_reduced_on_dim(idx):
-  # TODO(yashkatariya): This handles very narrow use case where we know XLA will
-  # not return an output with uneven sharding. Remove this after we have the
-  # ability to catch uneven shardings in lower_sharding_computation and raise
-  # a special exception for that which can be caught here to fallback to
-  # bouncing via host.
-  if not isinstance(idx, tuple):
-    idx = (idx,)
-  return all(isinstance(i, int) or
-             (isinstance(i, slice) and  i == slice(None)) or
-             (isinstance(i, (np.ndarray, jax.Array)) and not i.shape and
-              np.issubdtype(i.dtype, np.integer))
-             for i in idx)
-
-
 @functools.lru_cache(maxsize=4096)
 def _cached_index_calc(s, shape):
   map_ = s.addressable_devices_indices_map(shape)
@@ -157,13 +136,13 @@ def _create_copy_plan(arrays, s: Sharding, shape: Shape):
 @functools.lru_cache(maxsize=4096)
 def _process_has_full_value_in_mcjax(s, shape):
   # Return False for single host as a fast path.
-  if jax.process_count() == 1:
+  if xla_bridge.process_count() == 1:
     return False
 
   num_unique_indices = len(
-      set(hashed_index(v) for v in s.devices_indices_map(shape).values()))
+      {hashed_index(v) for v in s.devices_indices_map(shape).values()})
   num_addressable_unique_indices = len(
-      set(hashed_index(v) for v in s.addressable_devices_indices_map(shape).values()))
+      {hashed_index(v) for v in s.addressable_devices_indices_map(shape).values()})
   return num_unique_indices == num_addressable_unique_indices
 
 
@@ -172,26 +151,20 @@ class ArrayImpl(basearray.Array):
 
   aval: core.ShapedArray
   _sharding: Sharding
-  _arrays: List[DeviceArray]
+  _arrays: list[ArrayImpl]
   _committed: bool
   _skip_checks: bool
-  _npy_value: Optional[np.ndarray]
+  _npy_value: np.ndarray | None
 
   @use_cpp_method()
   def __init__(self, aval: core.ShapedArray, sharding: Sharding,
-               arrays: Union[Sequence[DeviceArray], Sequence[ArrayImpl]],
+               arrays: Sequence[ArrayImpl],
                committed: bool, _skip_checks: bool = False):
     # NOTE: the actual implementation of the constructor is moved to C++.
 
     self.aval = aval
     self._sharding = sharding
-    # Extract DeviceArrays from arrays with `SingleDeviceSharding` to keep the
-    # code handling `self._arrays` simpler.
-    # TODO(yashkatariya): This will be slower as it will happen during
-    # `__init__` on single controller environment. Make it lazy.
-    self._arrays = [a if isinstance(a, DeviceArray) else a._arrays[0] for a in arrays]
-    # See https://jax.readthedocs.io/en/latest/faq.html#controlling-data-and-computation-placement-on-devices
-    # for what committed means.
+    self._arrays = [a._arrays[0] for a in arrays]
     self._committed = committed
     self._npy_value = None
 
@@ -219,13 +192,13 @@ class ArrayImpl(basearray.Array):
           f"got {len(self._arrays)}")
 
     array_device_ids = set(device_id_to_buffer.keys())
-    addressable_device_ids = set(d.id for d in addressable_dev)
+    addressable_device_ids = {d.id for d in addressable_dev}
     # Calculate a symmetric difference because the device ids between sharding
     # and _arrays should match.
-    diff = set(array_device_ids) ^ set(addressable_device_ids)
+    diff = array_device_ids ^ addressable_device_ids
     if diff:
-      dev_in_sharding_not_in_arrays = set(addressable_device_ids) - set(array_device_ids)
-      dev_in_arrays_not_in_sharding = set(array_device_ids) - set(addressable_device_ids)
+      dev_in_sharding_not_in_arrays = addressable_device_ids - array_device_ids
+      dev_in_arrays_not_in_sharding = array_device_ids - addressable_device_ids
       err_msg = (
           "Addressable devices and per-device arrays devices do not match.")
       if dev_in_sharding_not_in_arrays:
@@ -240,8 +213,9 @@ class ArrayImpl(basearray.Array):
     for db in self._arrays:
       if db.shape != ss:
         raise ValueError(
-            f"Expected shard shape {ss} doesn't match the buffer "
-            f"shape {db.shape} for buffer: {db}")
+            f"Expected shard shape {ss} doesn't match the single device array "
+            f"shape {db.shape}. Shape of Array is "
+            f"{self.aval.str_short()} with sharding {self.sharding}")
 
     # Rearrange arrays based on the device assignment.
     if isinstance(self.sharding, XLACompatibleSharding):
@@ -282,29 +256,33 @@ class ArrayImpl(basearray.Array):
       raise TypeError("len() of unsized object") from err  # same as numpy error
 
   def __bool__(self):
-    return bool(self._value)
-
-  def __nonzero__(self):
+    # deprecated 2023 September 18.
+    # TODO(jakevdp) change to warn_on_empty=False
+    core.check_bool_conversion(self, warn_on_empty=True)
     return bool(self._value)
 
   def __float__(self):
+    core.check_scalar_conversion(self)
     return self._value.__float__()
 
   def __int__(self):
+    core.check_scalar_conversion(self)
     return self._value.__int__()
 
   def __complex__(self):
+    core.check_scalar_conversion(self)
     return self._value.__complex__()
 
   def __hex__(self):
-    assert self.ndim == 0, 'hex only works on scalar values'
+    core.check_integer_conversion(self)
     return hex(self._value)  # type: ignore
 
   def __oct__(self):
-    assert self.ndim == 0, 'oct only works on scalar values'
+    core.check_integer_conversion(self)
     return oct(self._value)  # type: ignore
 
   def __index__(self):
+    core.check_integer_conversion(self)
     return op.index(self._value)
 
   def tobytes(self, order="C"):
@@ -336,8 +314,10 @@ class ArrayImpl(basearray.Array):
         except ValueError:
           arr_idx = None
         if arr_idx is not None:
-          arr = self._arrays[arr_idx]
-          return _single_device_array_from_buf(arr, committed=False)
+          a = self._arrays[arr_idx]
+          return ArrayImpl(
+              a.aval, SingleDeviceSharding(a.device()), [a], committed=False,
+              _skip_checks=True)
       return lax_numpy._rewriting_take(self, idx)
     else:
       return lax_numpy._rewriting_take(self, idx)
@@ -358,7 +338,7 @@ class ArrayImpl(basearray.Array):
 
   @property
   def is_fully_replicated(self) -> bool:
-    return self.shape == self._arrays[0].shape
+    return self.sharding.is_fully_replicated
 
   def __repr__(self):
     prefix = 'Array('
@@ -379,16 +359,59 @@ class ArrayImpl(basearray.Array):
     else:
       return f"{prefix}{self.shape}, {dtype_str}"
 
-  @functools.cached_property
+  @property
   def is_fully_addressable(self) -> bool:
+    """Is this Array fully addressable?
+
+    A jax.Array is fully addressable if the current process can address all of
+    the devices named in the :class:`Sharding`. ``is_fully_addressable`` is
+    equivalent to "is_local" in multi-process JAX.
+
+    Note that fully replicated is not equal to fully addressable i.e.
+    a jax.Array which is fully replicated can span across multiple hosts and is
+    not fully addressable.
+    """
     return self.sharding.is_fully_addressable
 
   def __array__(self, dtype=None, context=None):
     return np.asarray(self._value, dtype=dtype)
 
-  def __dlpack__(self):
-    from jax.dlpack import to_dlpack  # pylint: disable=g-import-not-at-top
-    return to_dlpack(self)
+  def __dlpack__(self, *, stream: int | Any | None = None):
+    if len(self._arrays) != 1:
+      raise ValueError("__dlpack__ only supported for unsharded arrays.")
+    from jax._src.dlpack import to_dlpack  # pylint: disable=g-import-not-at-top
+    return to_dlpack(self, stream=stream)
+
+  def __dlpack_device__(self) -> tuple[enum.Enum, int]:
+    if len(self._arrays) != 1:
+      raise ValueError("__dlpack__ only supported for unsharded arrays.")
+
+    from jax._src.dlpack import DLDeviceType  # pylint: disable=g-import-not-at-top
+
+    if self.platform() == "cpu":
+      return DLDeviceType.kDLCPU, 0
+
+    elif self.platform() == "gpu":
+      platform_version = self.device().client.platform_version
+      if "cuda" in platform_version:
+        dl_device_type = DLDeviceType.kDLCUDA
+      elif "rocm" in platform_version:
+        dl_device_type = DLDeviceType.kDLROCM
+      else:
+        raise ValueError("Unknown GPU platform for __dlpack__: "
+                         f"{platform_version}")
+
+      local_hardware_id = self.device().local_hardware_id
+      if local_hardware_id is None:
+        raise ValueError("Couldn't get local_hardware_id for __dlpack__")
+
+      return dl_device_type, local_hardware_id
+
+    else:
+      raise ValueError(
+          "__dlpack__ device only supported for CPU and GPU, got platform: "
+          f"{self.platform()}"
+      )
 
   def __reduce__(self):
     fun, args, arr_state = self._value.__reduce__()  # type: ignore
@@ -428,7 +451,7 @@ class ArrayImpl(basearray.Array):
     raise ValueError('Length of devices is greater than 1. '
                      'Please use `.devices()`.')
 
-  def devices(self) -> Set[Device]:
+  def devices(self) -> set[Device]:
     self._check_if_deleted()
     return self.sharding.device_set
 
@@ -438,7 +461,7 @@ class ArrayImpl(basearray.Array):
   def device_buffer(self) -> ArrayImpl:
     self._check_if_deleted()
     if len(self._arrays) == 1:
-      return _single_device_array_from_buf(self._arrays[0], self._committed)
+      return self._arrays[0]
     raise ValueError('Length of buffers is greater than 1. Please use '
                      '`.device_buffers` instead.')
 
@@ -447,22 +470,20 @@ class ArrayImpl(basearray.Array):
   @property
   def device_buffers(self) -> Sequence[ArrayImpl]:
     self._check_if_deleted()
-    return [_single_device_array_from_buf(a, self._committed)
-            for a in self._arrays]
+    return cast(Sequence[ArrayImpl], self._arrays)
 
   def addressable_data(self, index: int) -> ArrayImpl:
     self._check_if_deleted()
-    return _single_device_array_from_buf(self._arrays[index], self._committed)
+    if self.is_fully_replicated:
+      return self._fully_replicated_shard()
+    return self._arrays[index]
 
   @functools.cached_property
   def addressable_shards(self) -> Sequence[Shard]:
     self._check_if_deleted()
     out = []
-    for db in self._arrays:
-      # Wrap the device arrays in `Array` until C++ returns an Array instead
-      # of a DA.
-      array = _single_device_array_from_buf(db, self._committed)
-      out.append(Shard(db.device(), self.sharding, self.shape, array))
+    for a in self._arrays:
+      out.append(Shard(a.device(), self.sharding, self.shape, a))
     return out
 
   @property
@@ -477,11 +498,10 @@ class ArrayImpl(basearray.Array):
       return self.addressable_shards
 
     out = []
-    device_id_to_buffer = {db.device().id: db for db in self._arrays}
+    device_id_to_buffer = {a.device().id: a for a in self._arrays}
     for global_d in self.sharding.device_set:
       if device_id_to_buffer.get(global_d.id, None) is not None:
-        array = _single_device_array_from_buf(
-            device_id_to_buffer[global_d.id], self._committed)
+        array = device_id_to_buffer[global_d.id]
       else:
         array = None
       out.append(Shard(global_d, self.sharding, self.shape, array))
@@ -507,7 +527,8 @@ class ArrayImpl(basearray.Array):
 
   def _check_if_deleted(self):
     if self.is_deleted():
-      raise RuntimeError("Array has been deleted.")
+      raise RuntimeError(
+          f"Array has been deleted with shape={self.aval.str_short()}.")
 
   @use_cpp_method()
   def block_until_ready(self):
@@ -581,7 +602,7 @@ setattr(ArrayImpl, "__array_priority__", 100)
 
 def make_array_from_callback(
     shape: Shape, sharding: Sharding,
-    data_callback: Callable[[Optional[Index]], ArrayLike]) -> ArrayImpl:
+    data_callback: Callable[[Index | None], ArrayLike]) -> ArrayImpl:
   """Returns a ``jax.Array`` via data fetched from ``data_callback``.
 
   ``data_callback`` is used to fetch the data for each addressable shard of the
@@ -627,6 +648,8 @@ def make_array_from_callback(
       for device in sharding.addressable_devices
   ]
   aval = core.ShapedArray(shape, arrays[0].dtype, weak_type=False)
+  if dtypes.issubdtype(aval.dtype, dtypes.extended):
+    return aval.dtype._rules.make_sharded_array(aval, sharding, arrays, committed=True)
   return ArrayImpl(aval, sharding, arrays, committed=True)
 
 
@@ -635,9 +658,8 @@ def make_array_from_single_device_arrays(
 ) -> ArrayImpl:
   r"""Returns a ``jax.Array`` from a sequence of ``jax.Array``\s on a single device.
 
-  ``jax.Array`` on a single device is analogous to a ``DeviceArray``. You can use
-  this function if you have already ``jax.device_put`` the value on a single
-  device and want to create a global Array. The smaller ``jax.Array``\s should be
+  You can use this function if you have already ``jax.device_put`` the value on
+  a single device and want to create a global Array. The smaller ``jax.Array``\s should be
   addressable and belong to the current process.
 
   Args:
@@ -656,24 +678,39 @@ def make_array_from_single_device_arrays(
     >>> from jax.sharding import PartitionSpec as P
     >>> import numpy as np
     ...
-    >>> shape = (8, 8)
+    >>> global_shape = (8, 8)
     >>> global_mesh = Mesh(np.array(jax.devices()).reshape(2, 4), ('x', 'y'))
     >>> sharding = jax.sharding.NamedSharding(global_mesh, P('x', 'y'))
-    >>> inp_data = np.arange(math.prod(shape)).reshape(shape)
+    >>> inp_data = np.arange(math.prod(global_shape)).reshape(global_shape)
     ...
     >>> arrays = [
     ...     jax.device_put(inp_data[index], d)
-    ...     for d, index in sharding.addressable_devices_indices_map(shape).items()]
+    ...     for d, index in sharding.addressable_devices_indices_map(global_shape).items()]
     ...
-    >>> arr = jax.make_array_from_single_device_arrays(shape, sharding, arrays)
+    >>> arr = jax.make_array_from_single_device_arrays(global_shape, sharding, arrays)
     >>> arr.addressable_data(0).shape
     (4, 2)
+
+    In multi-process case, if the input is process local and data parallel
+    i.e. each process receives a different part of the data, then you can use
+    `make_array_from_single_device_arrays` to create a global jax.Array
+
+    >>> local_shape = (8, 2)
+    >>> global_shape = (jax.process_count() * local_shape[0], ) + local_shape[1:]
+    >>> local_array = np.arange(math.prod(local_shape)).reshape(local_shape)
+    >>> arrays = jax.device_put(
+    ...   np.split(local_array, len(global_mesh.local_devices), axis = 0), global_mesh.local_devices)
+    >>> sharding = jax.sharding.NamedSharding(global_mesh, P(('x', 'y'), ))
+    >>> arr = jax.make_array_from_single_device_arrays(global_shape, sharding, arrays)
+    >>> arr.addressable_data(0).shape
+    (1, 2)
   """
   # All input arrays should be committed. Checking it is expensive on
   # single-controller systems.
   aval = core.ShapedArray(shape, arrays[0].dtype, weak_type=False)
-  # TODO(phawkins): ideally the cast() could be checked. Revisit this after
-  # removing DeviceArray.
+  if dtypes.issubdtype(aval.dtype, dtypes.extended):
+    return aval.dtype._rules.make_sharded_array(aval, sharding, arrays, committed=True)
+  # TODO(phawkins): ideally the cast() could be checked.
   return ArrayImpl(aval, sharding, cast(Sequence[ArrayImpl], arrays),
                    committed=True)
 
@@ -686,12 +723,87 @@ api_util._shaped_abstractify_handlers[ArrayImpl] = op.attrgetter('aval')
 basearray.Array.register(ArrayImpl)
 
 
-def _array_mlir_constant_handler(val, canonicalize_types=True):
-  return mlir.ir_constants(val._value,
-                           canonicalize_types=canonicalize_types)
+def _array_mlir_constant_handler(val):
+  return mlir.ir_constants(val._value)
 mlir.register_constant_handler(ArrayImpl, _array_mlir_constant_handler)
 
 
+# NOTE(skye): we could refactor to generate _multi_slice parameters directly
+# from the input ShardingSpec, rather than the indices. However, this would
+# require duplicating the ordering logic of spec_to_indices, which is more
+# subtle and more likely to change than the index logic we have to support here.
+def as_slice_indices(arr: Any, idx: Index) -> tuple[
+    tuple[int, ...], tuple[int, ...], tuple[int, ...]]:
+  """Returns start_indices, limit_indices, removed_dims"""
+  start_indices = [0] * arr.ndim
+  limit_indices = list(arr.shape)
+  removed_dims = []
+
+  tuple_idx = idx if isinstance(idx, tuple) else (idx,)
+  for dim, sub_idx in enumerate(tuple_idx):
+    if isinstance(sub_idx, int):
+      start_indices[dim] = sub_idx
+      limit_indices[dim] = sub_idx + 1
+      removed_dims.append(dim)
+    elif sub_idx == slice(None):
+      continue
+    else:
+      assert isinstance(sub_idx, slice), sub_idx
+      assert isinstance(sub_idx.start, int), sub_idx
+      assert isinstance(sub_idx.stop, int), sub_idx
+      start_indices[dim] = sub_idx.start
+      limit_indices[dim] = sub_idx.stop
+
+  return tuple(start_indices), tuple(limit_indices), tuple(removed_dims) # type: ignore
+
+
+def shard_device_array(x, devices, indices, sharding):
+  start_indices, limit_indices, removed_dims = unzip3(
+      as_slice_indices(x, idx) for idx in indices)
+  if sharding.is_fully_replicated:
+    shards = [x] * len(devices)
+  else:
+    shards = x._multi_slice(start_indices, limit_indices, removed_dims)
+  aval = api_util.shaped_abstractify(x)
+  return pxla.batched_device_put(aval, sharding, shards, devices)
+
+def _hashable_index(idx):
+  return tree_util.tree_map(
+      lambda x: (x.start, x.stop) if type(x) == slice else x, idx)
+
+# The fast path is handled directly in shard_args().
+def shard_sharded_device_array_slow_path(x, devices, indices, sharding):
+  candidates = defaultdict(list)
+  if isinstance(x, ArrayImpl):
+    bufs = [buf.data for buf in x.addressable_shards]
+    arr_indices = tuple(x.sharding.devices_indices_map(x.shape).values())
+  else:
+    bufs = x.device_buffers
+    arr_indices = x.indices
+  for buf, idx in safe_zip(bufs, arr_indices):
+    candidates[_hashable_index(idx)].append(buf)
+
+  bufs = []
+  for idx, device in safe_zip(indices, devices):
+    # Look up all buffers that contain the correct slice of the logical array.
+    candidates_list = candidates[_hashable_index(idx)]
+    if not candidates_list:
+      # This array isn't sharded correctly. Reshard it via host roundtrip.
+      # TODO(skye): more efficient reshard?
+      return pxla.shard_arg(x._value, devices, indices, sharding,
+                            canonicalize=False)
+    # Try to find a candidate buffer already on the correct device,
+    # otherwise copy one of them.
+    for buf in candidates_list:
+      if buf.device() == device:
+        bufs.append(buf)
+        break
+    else:
+      bufs.append(buf)
+
+  return pxla.batched_device_put(x.aval, sharding, bufs, devices)
+
+
 def _array_shard_arg(x, devices, indices, sharding):
   x._check_if_deleted()
 
@@ -708,10 +820,9 @@ def _array_shard_arg(x, devices, indices, sharding):
           x, list(devices), sharding)
     # Resharding starts here:
     if dispatch.is_single_device_sharding(x.sharding):
-      return pxla.shard_device_array(x, devices, indices, sharding)
+      return shard_device_array(x, devices, indices, sharding)
     else:
-      return pxla.shard_sharded_device_array_slow_path(
-          x, devices, indices, sharding)
+      return shard_sharded_device_array_slow_path(x, devices, indices, sharding)
 
 
 pxla.shard_arg_handlers[ArrayImpl] = _array_shard_arg
@@ -721,7 +832,7 @@ def _array_global_result_handler(global_aval, out_sharding, committed,
                                  is_out_sharding_from_xla):
   if global_aval.dtype == dtypes.float0:
     return lambda _: np.zeros(global_aval.shape, dtypes.float0)  # type: ignore
-  if core.is_opaque_dtype(global_aval.dtype):
+  if dtypes.issubdtype(global_aval.dtype, dtypes.extended):
     return global_aval.dtype._rules.global_sharded_result_handler(
         global_aval, out_sharding, committed, is_out_sharding_from_xla)
   return xc.array_result_handler(
@@ -736,7 +847,7 @@ pxla.global_result_handlers[core.AbstractToken] = lambda *_: lambda *_: core.tok
 def _array_local_result_handler(aval, sharding, indices):
   if aval.dtype == dtypes.float0:
     return lambda _: np.zeros(aval.shape, dtypes.float0)  # type: ignore
-  if core.is_opaque_dtype(aval.dtype):
+  if dtypes.issubdtype(aval.dtype, dtypes.extended):
     return aval.dtype._rules.local_sharded_result_handler(
         aval, sharding, indices)
   return xc.array_result_handler(
