@@ -13,15 +13,16 @@
 # limitations under the License.
 from __future__ import annotations
 
+from collections.abc import Sequence
 import dataclasses
 import functools
 import itertools as it
-from typing import (Union, Optional, Callable, Dict, Tuple, TypeVar,
-                    FrozenSet, Type, Set, List, Sequence, Any)
+from typing import Union, Callable, TypeVar, Any
 
 import numpy as np
 
 import jax.numpy as jnp
+from jax import dtypes
 from jax import lax
 
 from jax._src import api
@@ -30,7 +31,6 @@ from jax._src import core
 from jax._src import custom_derivatives
 from jax._src import effects
 from jax._src import pjit
-from jax._src import prng
 from jax._src import sharding_impls
 from jax._src import source_info_util
 from jax._src import traceback_util
@@ -57,8 +57,8 @@ zip, unsafe_zip = safe_zip, zip
 
 Bool = Union[bool, Array]
 Int = Union[int, Array]
-ErrorCategory = Type['JaxException']
-Payload = List[Union[np.ndarray, Array]]
+ErrorCategory = type['JaxException']
+Payload = list[Union[np.ndarray, Array]]
 PyTreeDef = jtu.PyTreeDef
 Out = TypeVar('Out')
 
@@ -102,10 +102,10 @@ class JaxException(Exception):
 @functools.total_ordering
 @dataclasses.dataclass(eq=True, frozen=True)
 class ErrorEffect(effects.Effect):
-  error_type: Type[JaxException]
-  shape_dtypes: Tuple[api.ShapeDtypeStruct, ...]
+  error_type: type[JaxException]
+  shape_dtypes: tuple[api.ShapeDtypeStruct, ...]
 
-  def __lt__(self, other: 'ErrorEffect'):
+  def __lt__(self, other: ErrorEffect):
     shape_dtypes = lambda x: tuple((sd.shape, str(sd.dtype))  # dtype is not comparable
                                    for sd in x.shape_dtypes)
     unpack = lambda x: (str(x.error_type), shape_dtypes(x))
@@ -117,7 +117,7 @@ effects.lowerable_effects.add_type(ErrorEffect)
 class DivisionByZeroError(JaxException):
 
   def __str__(self):
-    return f'division by zero at:\n\n{self.traceback_info}'
+    return 'division by zero'
 
   def get_effect_type(self):
     return ErrorEffect(DivisionByZeroError, ())
@@ -139,7 +139,7 @@ class NaNError(JaxException):
     return ErrorEffect(NaNError, ())
 
   def __str__(self):
-    return f'nan generated by primitive: {self.prim} at:\n\n{self.traceback_info}'
+    return f'nan generated by primitive: {self.prim}.'
 
 class OOBError(JaxException):
 
@@ -160,8 +160,7 @@ class OOBError(JaxException):
     return (f'out-of-bounds indexing for array of '
             f'shape {self.operand_shape}: '
             f'index {self._payload[0]} is out of bounds for axis '
-            f'{self._payload[1]} with size {self._payload[2]}. '
-            f'Failed at:\n\n{self.traceback_info}')
+            f'{self._payload[1]} with size {self._payload[2]}. ')
 
   def get_effect_type(self):
     return ErrorEffect(OOBError, (api.ShapeDtypeStruct((3,), jnp.int32),))
@@ -185,7 +184,7 @@ class FailedCheckError(JaxException):
 
   def __str__(self):
     return (self.fmt_string.format(*self.args, **self.kwargs)
-            + f' `check` failed at:\n\n{self.traceback_info}')
+            + ' (`check` failed)')
 
   def get_effect_type(self):
     vals = jtu.tree_leaves((self.args, self.kwargs))
@@ -195,7 +194,7 @@ class FailedCheckError(JaxException):
 
 @dataclasses.dataclass
 class BatchedError(JaxException):
-  error_mapping: Dict[Tuple[int, ...], JaxException]
+  error_mapping: dict[tuple[int, ...], JaxException]
 
   def __post_init__(self):
     traceback_info = list(self.error_mapping.values())[0].traceback_info
@@ -212,19 +211,19 @@ class BatchedError(JaxException):
 @jtu.register_pytree_node_class
 @dataclasses.dataclass(frozen=True)
 class Error:
-  _pred: Dict[ErrorEffect, Bool]
-  _code: Dict[ErrorEffect, Int]
-  _metadata: Dict[Int, PyTreeDef]  # mapping of code to JaxException treedef.
-  _payload: Dict[ErrorEffect, Payload]
+  _pred: dict[ErrorEffect, Bool]
+  _code: dict[ErrorEffect, Int]
+  _metadata: dict[Int, PyTreeDef]  # mapping of code to JaxException treedef.
+  _payload: dict[ErrorEffect, Payload]
 
-  def get(self) -> Optional[str]:
+  def get(self) -> str | None:
     """Returns error message if error happened, None if no error happened."""
     exp = self.get_exception()
     if exp is not None:
       return str(exp)
     return None
 
-  def get_exception(self) -> Optional[JaxException]:
+  def get_exception(self) -> JaxException | None:
     """Returns Python exception if error happened, None if no error happened."""
     if any(map(np.shape, self._pred.values())):
       return self._get_batched_exception()
@@ -250,7 +249,7 @@ class Error:
 
   # Internal helpers
 
-  def _get_batched_exception(self) -> Optional[BatchedError]:
+  def _get_batched_exception(self) -> BatchedError | None:
     shape = np.shape(list(self._pred.values())[0])
     error_mapping = {}
     for idx in np.ndindex(*shape):
@@ -278,7 +277,7 @@ class Error:
     new_metadata = {**self._metadata, **metadata}
     return Error(new_errs, new_codes, new_metadata, new_payload)
 
-  def _add_placeholder_effects(self, effects: Set[ErrorEffect]):
+  def _add_placeholder_effects(self, effects: set[ErrorEffect]):
     """Fill out Error with `effects` and np.ones arrays of their payloads."""
     new_err = self._pred.copy()
     new_code = self._code.copy()
@@ -337,7 +336,7 @@ def _flatten_and_get_error_metadata_thunk(*invals):
 
 def default_checkify_rule(primitive: core.Primitive, error: Error,
                           enabled_errors, *invals: core.Value,
-                          **params: Any) -> Tuple[Error, Sequence[core.Value]]:
+                          **params: Any) -> tuple[Error, Sequence[core.Value]]:
   """Default rule for primitives in `checkify` interpreter."""
   if 'call_jaxpr' not in params:
     # Default non-HOP case: just call primitive and don't update error.
@@ -389,18 +388,20 @@ def get_shaped_aval(val):
   return core.raise_to_shaped(core.get_aval(val))
 
 def checkify_jaxpr(jaxpr: core.ClosedJaxpr, enabled_errors,
-                   error: Error, *args) -> Tuple[Error, List[core.Value]]:
+                   error: Error, *args) -> tuple[Error, list[core.Value]]:
   err_vals, err_tree = jtu.tree_flatten(error)
   return checkify_jaxpr_flat(jaxpr.jaxpr, jaxpr.consts,
                              enabled_errors, err_tree, *err_vals, *args)
 
 def checkify_jaxpr_flat(jaxpr: core.Jaxpr, consts: Sequence[core.Value],
                         enabled_errors, err_tree: PyTreeDef,
-                        *args: core.Value) -> Tuple[Error, List[Any]]:
-  env: Dict[core.Var, Any] = {}
+                        *args: core.Value) -> tuple[Error, list[Any]]:
+  env: dict[core.Var, Any] = {}
   err_vals, in_args = split_list(args, [err_tree.num_leaves])
   error = jtu.tree_unflatten(err_tree, err_vals)
 
+  last_used = core.last_used(jaxpr)
+
   def read_env(var: core.Atom):
     if isinstance(var, core.Literal):
       return var.val
@@ -426,6 +427,7 @@ def checkify_jaxpr_flat(jaxpr: core.Jaxpr, consts: Sequence[core.Value],
       map(write_env, eqn.outvars, outvals)
     else:
       write_env(eqn.outvars[0], outvals)
+    core.clean_up_dead_vars(eqn, env, last_used)
 
   return error, map(read_env, jaxpr.outvars)
 
@@ -459,6 +461,21 @@ def _reduce_any_error(error: Error):
 check_p = core.Primitive('check')
 check_p.multiple_results = True  # zero results
 
+
+def _pp_check(eqn, context, settings) -> core.pp.Doc:
+  annotation = (source_info_util.summarize(eqn.source_info)
+                if settings.source_info else None)
+  name_stack_annotation = (f'[{eqn.source_info.name_stack}]'
+                           if settings.name_stack else None)
+  trimmed_params = sorted((k, v) for (k, v) in eqn.params.items()
+                          if k != "err_tree")
+  rhs = [core.pp.text(eqn.primitive.name, annotation=name_stack_annotation),
+         core.pp_kv_pairs(trimmed_params, context, settings),
+         core.pp.text(" ") + core.pp_vars(eqn.invars, context)]
+  return core.pp.concat([core.pp.text("", annotation), *rhs])
+
+core.pp_eqn_rules[check_p] = _pp_check
+
 # TODO(lenamartens): inherit from Exception instead of ValueError.
 class JaxRuntimeError(ValueError):
   pass
@@ -471,6 +488,9 @@ def check_impl(*args, err_tree, debug):
   error = tree_unflatten(err_tree, args)
   exc = error.get_exception()
   if exc:
+    filtered_tb = traceback_util.filter_traceback(
+        exc.traceback_info.as_python_traceback())
+    exc.with_traceback(filtered_tb)
     raise JaxRuntimeError(str(exc)) from exc
   return []
 
@@ -540,13 +560,11 @@ ad.primitive_jvps[check_p] = check_jvp_rule
 ## checkify rules
 
 ErrorCheckRule = Callable  # (Error, FrozenSet[ErrorCategory], *in_vals, **params) -> (Any, Error)
-error_checks: Dict[core.Primitive, ErrorCheckRule] = {}
+error_checks: dict[core.Primitive, ErrorCheckRule] = {}
 
 
-def summary() -> str:
-  return str(source_info_util.summarize(
-      source_info_util.current(),
-      config.read("jax_tracer_error_num_traceback_frames")))
+def get_traceback():
+  return source_info_util.current().traceback
 
 def nan_error_check(prim, error, enabled_errors, *in_vals, **params):
   out = prim.bind(*in_vals, **params)
@@ -558,13 +576,13 @@ def check_nans(prim, error, enabled_errors, out):
     return error
 
   def isnan(x):
-    if isinstance(x, prng.PRNGKeyArray):
+    if jnp.issubdtype(x.dtype, dtypes.prng_key):
       return False
     return jnp.any(jnp.isnan(x))
 
   any_nans = (jnp.any(jnp.array([isnan(x) for x in out]))
               if prim.multiple_results else isnan(out))
-  return assert_func(error, any_nans, NaNError(summary(), prim.name))
+  return assert_func(error, any_nans, NaNError(get_traceback(), prim.name))
 
 
 # All primitives which can generate a NaN.
@@ -600,7 +618,7 @@ def dynamic_slice_error_check(error, enabled_errors, operand, *start_indices, sl
   oob_mask = (start_indices < 0) | (start_indices + slice_sizes > operand_dims)
 
   payload = oob_payload(oob_mask, start_indices, range(operand.ndim), operand.shape)
-  error = assert_func(error, jnp.any(oob_mask), OOBError(summary(), "dynamic_slice", operand.shape, payload))
+  error = assert_func(error, jnp.any(oob_mask), OOBError(get_traceback(), "dynamic_slice", operand.shape, payload))
   return error, out
 error_checks[lax.dynamic_slice_p] = dynamic_slice_error_check
 
@@ -616,7 +634,7 @@ def dynamic_update_slice_error_check(error, enabled_errors, operand, update, *st
   oob_mask = (start_indices < 0) | (start_indices + update_dims > operand_dims)
 
   payload = oob_payload(oob_mask, start_indices, range(operand.ndim), operand.shape)
-  error = assert_func(error, jnp.any(oob_mask), OOBError(summary(), "dynamic_update_slice", operand.shape, payload))
+  error = assert_func(error, jnp.any(oob_mask), OOBError(get_traceback(), "dynamic_update_slice", operand.shape, payload))
   return error, out
 error_checks[lax.dynamic_update_slice_p] = dynamic_update_slice_error_check
 
@@ -642,7 +660,7 @@ def gather_error_check(error, enabled_errors, operand, start_indices, *,
   oob_mask = (start_indices < 0) | (start_indices > upper_bound.astype(start_indices.dtype))
 
   payload = oob_payload(oob_mask, start_indices, dnums.start_index_map, operand.shape)
-  error = assert_func(error, jnp.any(oob_mask), OOBError(summary(), "gather", operand.shape, payload))
+  error = assert_func(error, jnp.any(oob_mask), OOBError(get_traceback(), "gather", operand.shape, payload))
   return error, out
 error_checks[lax.gather_p] = gather_error_check
 
@@ -650,7 +668,7 @@ def div_error_check(error, enabled_errors, x, y):
   """Checks for division by zero and NaN."""
   if DivisionByZeroError in enabled_errors:
     any_zero = jnp.any(jnp.equal(y, 0))
-    error = assert_func(error, any_zero, DivisionByZeroError(summary()))
+    error = assert_func(error, any_zero, DivisionByZeroError(get_traceback()))
   return nan_error_check(lax.div_p, error, enabled_errors, x, y)
 error_checks[lax.div_p] = div_error_check
 
@@ -703,7 +721,7 @@ def scatter_error_check(prim, error, enabled_errors, operand, indices, updates,
     return error, out
 
   out_of_bounds, payload = scatter_oob(operand, indices, updates, dimension_numbers)
-  oob_error = OOBError(summary(), prim.name, operand.shape, payload)
+  oob_error = OOBError(get_traceback(), prim.name, operand.shape, payload)
   error = assert_func(error, out_of_bounds, oob_error)
   error = check_nans(prim, error, enabled_errors, out)
   return error, out
@@ -722,7 +740,7 @@ error_checks[lax.scatter_max_p] = functools.partial(scatter_error_check,
 @weakref_lru_cache
 def jaxpr_to_checkify_jaxpr(
     jaxpr: core.ClosedJaxpr, enabled_errors, err_tree: PyTreeDef,
-    *flat_err_and_in_vals) -> Tuple[core.ClosedJaxpr, PyTreeDef, Set[ErrorEffect]]:
+    *flat_err_and_in_vals) -> tuple[core.ClosedJaxpr, PyTreeDef, set[ErrorEffect]]:
   checkify_jaxpr_partial = functools.partial(checkify_jaxpr_flat, jaxpr.jaxpr,
                                              jaxpr.consts, enabled_errors,
                                              err_tree)
@@ -787,12 +805,11 @@ def scan_error_check(error, enabled_errors, *in_flat, reverse, length, jaxpr,
   checked_jaxpr_, out_tree, _ = jaxpr_to_checkify_jaxpr(jaxpr, enabled_errors,
                                                         err_tree, *new_in_aval)
 
-  new_in_flat = [*consts, *err_vals, *carry, *xs]
-  new_linear = (*[False] * len(err_vals), *linear)
   tomove = ([False] * len(err_vals) + [True] * len(consts)
             + [False] * (len(carry) + len(xs)))
   checked_jaxpr = pe.move_binders_to_front(checked_jaxpr_, tomove)
   new_in_flat = [*consts, *err_vals, *carry, *xs]
+  new_linear = (*[False] * len(err_vals), *linear)
   err_and_out = lax.scan_p.bind(
       *new_in_flat, reverse=reverse, length=length, jaxpr=checked_jaxpr,
       num_consts=len(consts), num_carry=len(carry)+len(err_vals),
@@ -805,7 +822,7 @@ error_checks[lax.scan_p] = scan_error_check
 def checkify_while_body_jaxpr(
     cond_jaxpr: core.ClosedJaxpr, body_jaxpr: core.ClosedJaxpr,
     enabled_errors, error: Error,
-    c_consts_num: int) -> Tuple[core.ClosedJaxpr, PyTreeDef, Set[ErrorEffect]]:
+    c_consts_num: int) -> tuple[core.ClosedJaxpr, PyTreeDef, set[ErrorEffect]]:
   cond_f = core.jaxpr_as_fun(cond_jaxpr)
   body_f = core.jaxpr_as_fun(body_jaxpr)
   def new_body_f(*c_consts_and_vals):
@@ -1044,8 +1061,8 @@ all_checks = automatic_checks | user_checks
 
 
 def checkify(f: Callable[..., Out],
-             errors: FrozenSet[ErrorCategory] = user_checks
-             ) -> Callable[..., Tuple[Error, Out]]:
+             errors: frozenset[ErrorCategory] = user_checks
+             ) -> Callable[..., tuple[Error, Out]]:
   """Functionalize `check` calls in `fun`, and optionally add run-time error checks.
 
   Run-time errors are either user-added :func:`~check` assertions, or
@@ -1161,7 +1178,7 @@ def _check(pred, msg, debug, *fmt_args, **fmt_kwargs):
       raise TypeError('Formatting arguments to checkify.check need to be '
                       'PyTrees of arrays, but got '
                       f'{repr(arg)} of type {type(arg)}.')
-  new_error = FailedCheckError(summary(), msg, *fmt_args, **fmt_kwargs)
+  new_error = FailedCheckError(get_traceback(), msg, *fmt_args, **fmt_kwargs)
   error = assert_func(init_error, jnp.logical_not(pred), new_error)
   _check_error(error, debug=debug)
 
