@@ -12,8 +12,9 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from collections.abc import Sequence
 import logging
-from typing import List, Optional, Type, Sequence, Tuple
+from typing import Optional
 from jax._src.cloud_tpu_init import running_in_cloud_tpu_vm
 
 logger = logging.getLogger(__name__)
@@ -28,7 +29,7 @@ class ClusterEnv:
   :class:`ClusterEnv` subclasses are automatically detected when imported.
   """
 
-  _cluster_types: List[Type['ClusterEnv']] = []
+  _cluster_types: list[type['ClusterEnv']] = []
 
   def __init_subclass__(cls, **kwargs):
     super().__init_subclass__(**kwargs)
@@ -41,7 +42,7 @@ class ClusterEnv:
                                            num_processes: Optional[int],
                                            process_id: Optional[int],
                                            local_device_ids: Optional[Sequence[int]]
-                                          ) -> Tuple[Optional[str], Optional[int], Optional[int],
+                                          ) -> tuple[Optional[str], Optional[int], Optional[int],
                                                      Optional[Sequence[int]]]:
     if all(p is not None for p in (coordinator_address, num_processes,
       process_id, local_device_ids)):
@@ -81,7 +82,7 @@ class ClusterEnv:
     """Returns address and port used by JAX to bootstrap.
 
     Process id 0 will open a tcp socket at "hostname:port" where
-    all the proccesses will connect to initialize the distributed JAX service.
+    all the processes will connect to initialize the distributed JAX service.
     The selected port needs to be free.
     :func:`get_coordinator_address` needs to return the same hostname and port on all the processes.
 
