@@ -12,6 +12,9 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from __future__ import annotations
+
+from collections.abc import Hashable, Iterator
 import contextlib
 import functools
 import itertools
@@ -19,15 +22,19 @@ import logging
 import os
 import sys
 import threading
-from typing import Any, List, Callable, Hashable, NamedTuple, Iterator, Optional
+from typing import Any, Callable, Generic, NamedTuple, Optional, TypeVar
 
 from jax._src import lib
 from jax._src.lib import jax_jit
 from jax._src.lib import transfer_guard_lib
 from jax._src.lib import xla_client
+from jax._src import logging_config
+from jax._src.lib import xla_extension_version
 
 logger = logging.getLogger(__name__)
 
+_T = TypeVar('_T')
+
 
 def bool_env(varname: str, default: bool) -> bool:
   """Read an environment variable and interpret it as a boolean.
@@ -63,6 +70,16 @@ UPGRADE_BOOL_HELP = (
 UPGRADE_BOOL_EXTRA_DESC = " (transient)"
 
 
+class FlagHolder(Generic[_T]):
+  def __init__(self, flags: NameSpace, name: str):
+    self._flags = flags
+    self._name = name
+
+  @property
+  def value(self) -> _T:
+    return getattr(self._flags, self._name)
+
+
 class Config:
   _HAS_DYNAMIC_ATTRIBUTES = True
 
@@ -111,26 +128,31 @@ class Config:
     if name not in self.values:
       raise AttributeError(f"Unrecognized config option: {name}")
 
-  def DEFINE_bool(self, name, default, *args, **kwargs):
+  def DEFINE_bool(self, name, default, *args, **kwargs) -> FlagHolder[bool]:
     update_hook = kwargs.pop("update_hook", None)
     self.add_option(name, default, bool, args, kwargs, update_hook=update_hook)
+    return FlagHolder(self.FLAGS, name)
 
-  def DEFINE_integer(self, name, default, *args, **kwargs):
+  def DEFINE_integer(self, name, default, *args, **kwargs) -> FlagHolder[int]:
     update_hook = kwargs.pop("update_hook", None)
     self.add_option(name, default, int, args, kwargs, update_hook=update_hook)
+    return FlagHolder(self.FLAGS, name)
 
-  def DEFINE_float(self, name, default, *args, **kwargs):
+  def DEFINE_float(self, name, default, *args, **kwargs) -> FlagHolder[float]:
     update_hook = kwargs.pop("update_hook", None)
     self.add_option(name, default, float, args, kwargs, update_hook=update_hook)
+    return FlagHolder(self.FLAGS, name)
 
-  def DEFINE_string(self, name, default, *args, **kwargs):
+  def DEFINE_string(self, name, default, *args, **kwargs) -> FlagHolder[str]:
     update_hook = kwargs.pop("update_hook", None)
     self.add_option(name, default, str, args, kwargs, update_hook=update_hook)
+    return FlagHolder(self.FLAGS, name)
 
-  def DEFINE_enum(self, name, default, *args, **kwargs):
+  def DEFINE_enum(self, name, default, *args, **kwargs) -> FlagHolder[str]:
     update_hook = kwargs.pop("update_hook", None)
     self.add_option(name, default, 'enum', args, kwargs,
                     update_hook=update_hook)
+    return FlagHolder(self.FLAGS, name)
 
   def config_with_absl(self):
     # Run this before calling `app.run(main)` etc
@@ -152,7 +174,15 @@ class Config:
 
   def complete_absl_config(self, absl_flags):
     for name, _ in self.values.items():
-      flag = absl_flags.FLAGS[name]
+      try:
+        flag = absl_flags.FLAGS[name]
+      except KeyError:
+        # This can happen if a new flag was added after config_with_absl() was
+        # called, but before complete_absl_config was run. We could in principle
+        # add code to DEFINE_... to register any newly added flags with ABSL
+        # if config_with_absl() has already been called, but arguably the user
+        # should have called config_with_absl() later.
+        continue
       if flag.present:
         self.update(name, flag.value)
 
@@ -246,7 +276,7 @@ class Config:
                                 default_value=True)
 
   def define_enum_state(
-      self, name: str, enum_values: List[str], default: Optional[str],
+      self, name: str, enum_values: list[str], default: Optional[str],
       help: str, update_global_hook: Optional[Callable[[str], None]] = None,
       update_thread_local_hook: Optional[Callable[[Optional[str]], None]] \
         = None):
@@ -296,8 +326,6 @@ class Config:
       name: string, converted to lowercase to define the name of the config
         option (and absl flag). It is converted to uppercase to define the
         corresponding shell environment variable.
-      enum_values: list of strings representing the possible values for the
-        option.
       default: optional int, default value.
       help: string, used to populate the flag help information as well as the
         docstring of the returned context manager.
@@ -337,8 +365,6 @@ class Config:
       name: string, converted to lowercase to define the name of the config
         option (and absl flag). It is converted to uppercase to define the
         corresponding shell environment variable.
-      enum_values: list of strings representing the possible values for the
-        option.
       default: optional float, default value.
       help: string, used to populate the flag help information as well as the
         docstring of the returned context manager.
@@ -468,9 +494,11 @@ class Config:
     return (axis_env_state, mesh_context_manager, self.x64_enabled,
             self.jax_numpy_rank_promotion, self.jax_default_matmul_precision,
             self.jax_dynamic_shapes, self.jax_numpy_dtype_promotion,
-            self.jax_default_device, self.jax_array,
+            self.jax_default_device,
             self.jax_threefry_partitionable,
             self.jax_softmax_custom_jvp,
+            self.jax_enable_memories,
+            self.jax_disable_jit,
             # Technically this affects jaxpr->MHLO lowering, not tracing.
             self.jax_hlo_source_file_canonicalization_regex)
 
@@ -550,6 +578,22 @@ config = Config()
 flags = config
 FLAGS = flags.FLAGS
 
+def DEFINE_bool(name, default, *args, **kwargs):
+  return flags.DEFINE_bool(name, default, *args, **kwargs)
+
+def DEFINE_integer(name, default, *args, **kwargs):
+  return flags.DEFINE_integer(name, default, *args, **kwargs)
+
+def DEFINE_float(name, default, *args, **kwargs):
+  return flags.DEFINE_float(name, default, *args, **kwargs)
+
+def DEFINE_string(name, default, *args, **kwargs):
+  return flags.DEFINE_string(name, default, *args, **kwargs)
+
+def DEFINE_enum(name, default, *args, **kwargs):
+  return flags.DEFINE_enum(name, default, *args, **kwargs)
+
+
 already_configured_with_absl = False
 
 
@@ -616,51 +660,6 @@ def update_thread_local_jit_state(**kw):
   tls.extra_jit_context = _thread_local_state_cache.canonicalize(tmp)
 
 
-flags.DEFINE_integer(
-    'jax_tracer_error_num_traceback_frames',
-    int_env('JAX_TRACER_ERROR_NUM_TRACEBACK_FRAMES', 5),
-    help='Set the number of stack frames in JAX tracer error messages.'
-)
-
-flags.DEFINE_bool(
-    'jax_pprint_use_color',
-    bool_env('JAX_PPRINT_USE_COLOR', True),
-    help='Enable jaxpr pretty-printing with colorful syntax highlighting.'
-)
-
-flags.DEFINE_bool(
-    'jax_host_callback_inline',
-    bool_env('JAX_HOST_CALLBACK_INLINE', False),
-    help='Inline the host_callback, if not in a staged context.'
-)
-flags.DEFINE_integer(
-    'jax_host_callback_max_queue_byte_size',
-    int_env('JAX_HOST_CALLBACK_MAX_QUEUE_BYTE_SIZE', int(256 * 1e6)),
-    help=('The size in bytes of the buffer used to hold outfeeds from each '
-          'device. When this capacity is reached consuming outfeeds from the '
-          'device is paused, thus potentially pausing the device computation, '
-          'until the Python callback consume more outfeeds.'),
-    lower_bound=int(16 * 1e6)
-)
-flags.DEFINE_bool(
-    'jax_host_callback_outfeed',
-    bool_env('JAX_HOST_CALLBACK_OUTFEED', False),
-    help=(
-        'Use outfeed implementation for host_callback, even on CPU and GPU. '
-        'If false, use the CustomCall implementation. '
-        'Has no effect on TPU, since only the outfeed mechanism is implemented.'
-    )
-)
-flags.DEFINE_bool(
-    'jax_host_callback_ad_transforms',
-    bool_env('JAX_HOST_CALLBACK_AD_TRANSFORMS', False),
-    help=(
-        'Enable support for jvp/vjp for the host_callback primitives. Default is '
-        'False, which means that host_callback operates only on primals. '
-        'The flag exists only temporarily, for backward compatibility.'
-    )
-)
-
 # TODO(b/214340779): remove flag when XLA:CPU is improved.
 jax2tf_associative_scan_reductions = config.define_bool_state(
     name='jax2tf_associative_scan_reductions',
@@ -679,7 +678,7 @@ jax2tf_associative_scan_reductions = config.define_bool_state(
 
 jax2tf_default_native_serialization = config.define_bool_state(
     name='jax2tf_default_native_serialization',
-    default=bool_env('JAX2TF_DEFAULT_NATIVE_SERIALIZATION', False),
+    default=bool_env('JAX2TF_DEFAULT_NATIVE_SERIALIZATION', True),
     help=(
         'Sets the default value of the native_serialization parameter to '
         'jax2tf.convert. Prefer using the parameter instead of the flag, '
@@ -687,6 +686,21 @@ jax2tf_default_native_serialization = config.define_bool_state(
     )
 )
 
+jax_serialization_version = config.define_int_state(
+    name='jax_serialization_version',
+    # Note: bump the default serialization version at least one month after
+    # we update XlaCallModule to support the new version, so that serialized
+    # modules are forward compatible with deployed versions of XlaCallModule.
+    # Version 7 of XlaCallModule is supported since July 12th, 2023.
+    default=int_env('JAX_SERIALIZATION_VERSION', 7),
+    help=(
+        'The version number to use for native serialization. This must be '
+        'within the range of versions supported by the tf.XlaCallModule '
+        'used in your deployment environment. '
+        'See https://github.com/google/jax/blob/main/jax/experimental/jax2tf/README.md#native-serialization-versions.'
+    )
+)
+
 jax_platforms = config.define_string_state(
     name='jax_platforms',
     default=None,
@@ -754,61 +768,40 @@ parallel_functions_output_gda = config.define_bool_state(
     default=False,
     help='If True, pjit will output GDAs.')
 
-def _update_jax_array_global(val):
-  if val is not None and not val:
-    raise ValueError(
-        'jax.config.jax_array cannot be disabled after jax 0.4.7 release.'
-        ' Please downgrade to jax and jaxlib 0.4.6 if you want to disable'
-        ' jax.config.jax_array.')
-
-def _update_jax_array_thread_local(val):
-  if val is not None and not val:
-    raise ValueError(
-        'jax.config.jax_array cannot be disabled after jax 0.4.7 release.'
-        ' Please downgrade to jax and jaxlib 0.4.6 if you want to disable'
-        ' jax.config.jax_array.')
-
-jax_array = config.define_bool_state(
-    name='jax_array',
-    default=True,
-    upgrade=True,
-    update_global_hook = _update_jax_array_global,
-    update_thread_local_hook = _update_jax_array_thread_local,
-    help=('If True, new pjit behavior will be enabled and `jax.Array` will be '
-          'used.'))
-
-
-jit_pjit_api_merge = config.define_bool_state(
-    name='jax_jit_pjit_api_merge',
-    default=True,
-    upgrade=True,
-    help=('If True, jit and pjit API will be merged. You can only disable it via '
-          "the environment variable i.e. `os.environ['JAX_JIT_PJIT_API_MERGE'] = '0'`. "
-          "The merge must be disabled via an environment variable since it "
-          "affects JAX at import time so it needs to be disabled before jax is "
-          "imported."))
-
-
 pmap_shmap_merge = config.define_bool_state(
     name='jax_pmap_shmap_merge',
     default=False,
     upgrade=True,
     help='If True, pmap and shard_map API will be merged.')
 
+def _update_jax_memories_global(val):
+  if xla_extension_version >= 190:
+    lib.jax_jit.global_state().enable_memories = val
+
+def _update_jax_memories_thread_local(val):
+  if xla_extension_version >= 190:
+    lib.jax_jit.thread_local_state().enable_memories = val
+
+enable_memories = config.define_bool_state(
+    'jax_enable_memories',
+    default=True,
+    upgrade=True,
+    update_global_hook=_update_jax_memories_global,
+    update_thread_local_hook=_update_jax_memories_thread_local,
+    help=("If True, will allow fetching memory kinds available on executable "
+          "and annotate Shardings with it."))
 
 spmd_mode = config.define_enum_state(
     name='jax_spmd_mode',
-    enum_values=['allow_all', 'allow_jit', 'allow_pjit'],
+    enum_values=['allow_all', 'allow_jit'],
     default='allow_jit',
     help=("Decides whether Math on `jax.Array`'s that are not fully addressable "
           "(i.e. spans across multiple processes) is allowed. The options are: "
-          "* allow_pjit: Default, only `pjit` computations are allowed to "
-          "    execute on non-fully addressable `jax.Array`s\n"
-          "* allow_jit: `pjit` and `jax.jit` computations are allowed to "
-          "    execute on non-fully addressable `jax.Array`s\n"
+          "* allow_jit: Default, `pjit` and `jax.jit` computations are allowed "
+          "    to execute on non-fully addressable `jax.Array`s\n"
           "* allow_all: `jnp`, normal math (like `a + b`, etc), `pjit`, "
-          "     `jax.jit` and all other operations are allowed to "
-          "     execute on non-fully addresable `jax.Array`s."))
+          "    `jax.jit` and all other operations are allowed to "
+          "    execute on non-fully addressable `jax.Array`s."))
 
 
 distributed_debug = config.define_bool_state(
@@ -818,6 +811,13 @@ distributed_debug = config.define_bool_state(
           'computations. Logging is performed with `logging` at WARNING '
           'level.'))
 
+legacy_prng_key = config.define_enum_state(
+    name='jax_legacy_prng_key',
+    enum_values=['allow', 'warn', 'error'],
+    default='allow',
+    help=('Specify the behavior when raw PRNG keys are passed to '
+          'jax.random APIs.')
+)
 
 enable_custom_prng = config.define_bool_state(
     name='jax_enable_custom_prng',
@@ -838,10 +838,9 @@ threefry_partitionable = config.define_bool_state(
     default=False,
     upgrade=True,
     help=('Enables internal threefry PRNG implementation changes that '
-          'render it automatically partitionable in some cases. For use '
-          'with pjit and/or jax_array=True. Without this flag, using the '
-          'standard jax.random pseudo-random number generation may result '
-          'in extraneous communication and/or redundant distributed '
+          'render it automatically partitionable in some cases. Without this '
+          'flag, using the standard jax.random pseudo-random number generation '
+          'may result in extraneous communication and/or redundant distributed '
           'computation. With this flag, the communication overheads disappear '
           'in some cases.'),
     update_global_hook=lambda val: _update_global_jit_state(
@@ -917,6 +916,14 @@ include_full_tracebacks_in_locations = config.define_bool_state(
     ),
 )
 
+use_original_compilation_cache_key_generation = config.define_bool_state(
+    name='jax_use_original_compilation_cache_key_generation',
+    default=True,
+    help="If true, use the original cache-key generation algorithm. This is "
+         "a transient flag; once the new cache-key generation algorithm is "
+         "deployed, this flag and the original cache-key generation algorithm "
+         "will be removed.")
+
 config.define_enum_state(
     name='jax_default_dtype_bits',
     enum_values=['32', '64'],
@@ -1046,17 +1053,21 @@ default_matmul_precision = config.define_enum_state(
 
 traceback_filtering = config.define_enum_state(
     name = 'jax_traceback_filtering',
-    enum_values=["off", "tracebackhide", "remove_frames", "auto"],
+    enum_values=["off", "tracebackhide", "remove_frames", "quiet_remove_frames",
+                 "auto"],
     default="auto",
     help="Controls how JAX filters internal frames out of tracebacks.\n\n"
          "Valid values are:\n"
          " * \"off\": disables traceback filtering.\n"
-         " * \"auto\": use \"tracebackhide\" if running under a sufficiently "
-         "new IPython, or \"remove_frames\" otherwise.\n"
-         " * \"tracebackhide\": adds \"__tracebackhide__\" annotations to "
+         " * \"auto\": use \"tracebackhide\" if running under a sufficiently"
+         " new IPython, or \"remove_frames\" otherwise.\n"
+         " * \"tracebackhide\": adds \"__tracebackhide__\" annotations to"
          " hidden stack frames, which some traceback printers support.\n"
-         " * \"remove_frames\": removes hidden frames from tracebacks, and adds "
-         " the unfiltered traceback as a __cause__ of the exception.\n")
+         " * \"remove_frames\": removes hidden frames from tracebacks, and adds"
+         " the unfiltered traceback as a __cause__ of the exception.\n"
+         " * \"quiet_remove_frames\": removes hidden frames from tracebacks, and adds"
+         " a brief message (to the __cause__ of the exception) describing that this has"
+         " happened.\n")
 
 # This flag is for internal use.
 # TODO(tianjianlu): Removes once we always enable cusparse lowering.
@@ -1085,16 +1096,6 @@ config.define_bool_state(
     default=(lib.version >= (0, 3, 6)),
     help=('Enables using optimization-barrier op for lowering remat.'))
 
-# TODO(b/205307544): Remove flag once coordination service has rolled out.
-config.define_bool_state(
-    name='jax_coordination_service',
-    default=True,
-    help=(
-         'Use coordination service (experimental) instead of the default PjRT '
-         'distributed runtime.'
-    )
-)
-
 # TODO(sharadmv,mattjj): set default to True, then remove
 config.define_bool_state(
     name='jax_eager_pmap',
@@ -1112,6 +1113,14 @@ config.define_bool_state(
           'work under pmap/pjit.')
 )
 
+jax_xla_profile_version = config.define_int_state(
+    name='jax_xla_profile_version',
+    default=0,
+    help=('Optional profile version for XLA compilation. This is meaningful '
+          'only when XLA is configured to support the remote compilation '
+          'profile feature.')
+)
+
 @contextlib.contextmanager
 def explicit_device_put_scope() -> Iterator[None]:
   """Indicates that the current context is an explicit device_put*() call."""
@@ -1235,3 +1244,21 @@ def transfer_guard(new_val: str) -> Iterator[None]:
     stack.enter_context(transfer_guard_device_to_host(new_val))
     stack.enter_context(_transfer_guard(new_val))
     yield
+
+
+def _update_debug_log_modules(module_names_str: Optional[str]):
+  logging_config.disable_all_debug_logging()
+  if not module_names_str:
+    return
+  module_names = module_names_str.split(',')
+  for module_name in module_names:
+    logging_config.enable_debug_logging(module_name)
+
+# Don't define a context manager since this isn't threadsafe.
+config.define_string_state(
+    name='jax_debug_log_modules',
+    default='',
+    help=('Comma-separated list of module names (e.g. "jax" or '
+          '"jax._src.xla_bridge,jax._src.dispatch") to enable debug logging '
+          'for.'),
+    update_global_hook=_update_debug_log_modules)
