@@ -12,11 +12,11 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from collections.abc import Sequence
 import dataclasses
 from functools import update_wrapper, reduce, partial
 import inspect
-from typing import (
-  Any, Callable, Generic, List, Optional, Sequence, Tuple, TypeVar)
+from typing import Any, Callable, Generic, Optional, TypeVar
 
 from jax._src import core
 from jax._src import custom_api_util
@@ -49,8 +49,6 @@ traceback_util.register_exclusion(__file__)
 map = safe_map
 zip = safe_zip
 
-allowed_effects: effects.EffectTypeSet = (
-    effects.custom_derivatives_allowed_effects)
 
 ### util
 
@@ -137,13 +135,13 @@ class custom_jvp(Generic[ReturnValue]):
   .. _tutorial: https://jax.readthedocs.io/en/latest/notebooks/Custom_derivative_rules_for_Python_code.html
   """
   fun: Callable[..., ReturnValue]
-  nondiff_argnums: Tuple[int, ...]
-  jvp: Optional[Callable[..., Tuple[ReturnValue, ReturnValue]]] = None
+  nondiff_argnums: tuple[int, ...]
+  jvp: Optional[Callable[..., tuple[ReturnValue, ReturnValue]]] = None
   symbolic_zeros: bool = False
 
   def __init__(self,
                fun: Callable[..., ReturnValue],
-               nondiff_argnums: Tuple[int, ...] = (),
+               nondiff_argnums: tuple[int, ...] = (),
                ):
     update_wrapper(self, fun)
     self.fun = fun
@@ -152,9 +150,9 @@ class custom_jvp(Generic[ReturnValue]):
   __getattr__ = custom_api_util.forward_attr
 
   def defjvp(self,
-             jvp: Callable[..., Tuple[ReturnValue, ReturnValue]],
+             jvp: Callable[..., tuple[ReturnValue, ReturnValue]],
              symbolic_zeros: bool = False,
-             ) -> Callable[..., Tuple[ReturnValue, ReturnValue]]:
+             ) -> Callable[..., tuple[ReturnValue, ReturnValue]]:
     """Define a custom JVP rule for the function represented by this instance.
 
     Args:
@@ -416,7 +414,7 @@ def process_env_traces(primitive, level: int, jvp_was_run: bool, *args):
   yield outs, tuple(todo)  # Ensure the aux output is immutable
 
 
-allowed_effects.add_type(lax.InOutFeedEffect)
+effects.custom_derivatives_allowed_effects.add_type(lax.InOutFeedEffect)
 
 custom_jvp_call_p = CustomJVPCallPrimitive('custom_jvp_call')
 
@@ -424,7 +422,7 @@ def _custom_jvp_call_typecheck(_, *in_avals, call_jaxpr, jvp_jaxpr_thunk,
                                num_consts, symbolic_zeros):
   # TODO(mattjj): could do more checking here...
   del in_avals, jvp_jaxpr_thunk, num_consts
-  disallowed_effects = allowed_effects.filter_not_in(call_jaxpr.effects)
+  disallowed_effects = effects.custom_derivatives_allowed_effects.filter_not_in(call_jaxpr.effects)
   if disallowed_effects:
     raise NotImplementedError(
         f'Effects not supported in `custom_jvp`: {disallowed_effects}')
@@ -491,19 +489,19 @@ class custom_vjp(Generic[ReturnValue]):
 
   def __init__(self,
                fun: Callable[..., ReturnValue],
-               nondiff_argnums: Tuple[int, ...] = ()):
+               nondiff_argnums: tuple[int, ...] = ()):
     update_wrapper(self, fun)
     self.fun = fun
     self.nondiff_argnums = nondiff_argnums
-    self.fwd: Optional[Callable[..., Tuple[ReturnValue, Any]]] = None
-    self.bwd: Optional[Callable[..., Tuple[Any, ...]]] = None
+    self.fwd: Optional[Callable[..., tuple[ReturnValue, Any]]] = None
+    self.bwd: Optional[Callable[..., tuple[Any, ...]]] = None
     self.symbolic_zeros = False
 
   __getattr__ = custom_api_util.forward_attr
 
   def defvjp(self,
-             fwd: Callable[..., Tuple[ReturnValue, Any]],
-             bwd: Callable[..., Tuple[Any, ...]],
+             fwd: Callable[..., tuple[ReturnValue, Any]],
+             bwd: Callable[..., tuple[Any, ...]],
              symbolic_zeros: bool = False,
              ) -> None:
     """Define a custom VJP rule for the function represented by this instance.
@@ -531,13 +529,16 @@ class custom_vjp(Generic[ReturnValue]):
         derivative rules to detect when certain inputs, and when certain
         output cotangents, are not involved in differentiation. If ``True``:
 
-        * ``fwd`` must accept, in place of each leaf value ``x`` in the pytree
-          comprising an argument to the original function, an object with two
-          attributes instead: ``value`` and ``perturbed``. The ``value`` field
-          is the original primal argument, and ``perturbed`` is a boolean.
-          The ``perturbed`` bit indicates whether the argument is involved in
-          differentiation (i.e., if it is ``False``, then the corresponding
-          Jacobian "column" is zero).
+        * ``fwd`` must accept, in place of each leaf value ``x`` in
+          the pytree comprising an argument to the original function,
+          an object (of type
+          ``jax.custom_derivatives.CustomVJPPrimal``) with two
+          attributes instead: ``value`` and ``perturbed``. The
+          ``value`` field is the original primal argument, and
+          ``perturbed`` is a boolean.  The ``perturbed`` bit indicates
+          whether the argument is involved in differentiation (i.e.,
+          if it is ``False``, then the corresponding Jacobian "column"
+          is zero).
 
         * ``bwd`` will be passed objects representing static symbolic zeros in
           its cotangent argument in correspondence with unperturbed values;
@@ -622,6 +623,7 @@ class custom_vjp(Generic[ReturnValue]):
 
 @dataclasses.dataclass
 class CustomVJPPrimal:
+  """Primal to a ``custom_vjp``'s forward rule when ``symbolic_zeros`` is set"""
   value: Any
   perturbed: bool
 
@@ -813,7 +815,7 @@ def _custom_vjp_call_jaxpr_impl(*args, fun_jaxpr, **_):
   return core.jaxpr_as_fun(fun_jaxpr)(*args)
 
 def _custom_vjp_call_jaxpr_abstract_eval(*_, fun_jaxpr, **__):
-  disallowed_effects = allowed_effects.filter_not_in(fun_jaxpr.effects)
+  disallowed_effects = effects.custom_derivatives_allowed_effects.filter_not_in(fun_jaxpr.effects)
   if disallowed_effects:
     raise NotImplementedError(
         f'Effects not supported in `custom_vjp`: {disallowed_effects}')
@@ -828,10 +830,9 @@ CustomVJPCallPrimitive.initial_style = custom_vjp_call_jaxpr_p
 mlir.register_lowering(custom_vjp_call_jaxpr_p, mlir.lower_fun(
     _custom_vjp_call_jaxpr_impl, multiple_results=True))
 
-
 def _custom_vjp_call_jaxpr_jvp(
     primals, tangents, *, fun_jaxpr: core.ClosedJaxpr,
-    fwd_jaxpr_thunk: Callable[..., Tuple[core.Jaxpr, Sequence[Any]]],
+    fwd_jaxpr_thunk: Callable[..., tuple[core.Jaxpr, Sequence[Any]]],
     num_consts: int, bwd: Callable, out_trees: Callable, symbolic_zeros: bool):
   _, args = split_list(primals, [num_consts])
   consts_dot, args_dot = split_list(tangents, [num_consts])
@@ -850,6 +851,7 @@ def _custom_vjp_call_jaxpr_jvp(
   tangents_out = ad.custom_lin_p.bind(
       *res, *args_dot, num_res=res_tree.num_leaves, bwd=bwd,
       out_avals=avals_out, symbolic_zeros=symbolic_zeros)
+  tangents_out = map(lax.tie_p.bind, primals_out, tangents_out)
   tangents_out = map(ad.recast_to_float0, primals_out, tangents_out)
   return primals_out, tangents_out
 ad.primitive_jvps[custom_vjp_call_jaxpr_p] = _custom_vjp_call_jaxpr_jvp
@@ -857,7 +859,7 @@ ad.primitive_jvps[custom_vjp_call_jaxpr_p] = _custom_vjp_call_jaxpr_jvp
 def _custom_vjp_call_jaxpr_vmap(
     spmd_axis_name, axis_size, axis_name, main_type, args, in_dims, *,
     fun_jaxpr: core.ClosedJaxpr,
-    fwd_jaxpr_thunk: Callable[..., Tuple[core.Jaxpr, Sequence[Any]]],
+    fwd_jaxpr_thunk: Callable[..., tuple[core.Jaxpr, Sequence[Any]]],
     num_consts: int, bwd: Callable, out_trees: Callable, symbolic_zeros: bool):
   args = [batching.moveaxis(x, d, 0) if d is not not_mapped and d != 0
           else x for x, d in zip(args, in_dims)]
@@ -1006,7 +1008,7 @@ class Residuals:
     return cls(jaxpr, in_tree, out_tree, consts)
 
 
-def closure_convert(fun: Callable, *example_args) -> Tuple[Callable, List[Any]]:
+def closure_convert(fun: Callable, *example_args) -> tuple[Callable, list[Any]]:
   """Closure conversion utility, for use with higher-order custom derivatives.
 
   To define custom derivatives such as with ``jax.custom_vjp(f)``, the target
@@ -1196,7 +1198,7 @@ def linear_call(fun: Callable, fun_transpose: Callable, residual_args,
   Args:
     fun: a Python callable specifying a linear function. It should
       take two arguments: one of "residual" inputs (type ``r``),
-      i.e. inputs in which the function is not necessarly linear, and
+      i.e. inputs in which the function is not necessarily linear, and
       one of "linear" inputs (type ``a``).  It should return output
       whose components are linear in the linear input (type ``b``).
     fun_transpose: a Python callable specifying a structurally linear
