@@ -13,10 +13,11 @@
 # limitations under the License.
 """Module for JAX debugging primitives and related functionality."""
 
+from collections.abc import Sequence
 import functools
 import string
 import sys
-from typing import Any, Dict, Callable, Optional, Sequence, Set, Tuple, Union
+from typing import Any, Callable, Optional, Union
 import weakref
 
 import numpy as np
@@ -205,14 +206,14 @@ def debug_callback(callback: Callable[..., Any], *args: Any,
 
   For more explanation, see `External Callbacks`_.
 
-  `debug_callback` enables you to pass in a Python function that can be called
-  inside of a staged JAX program. A `debug_callback` follows existing JAX
+  ``jax.debug.callback`` enables you to pass in a Python function that can be called
+  inside of a staged JAX program. A ``jax.debug.callback`` follows existing JAX
   transformation *pure* operational semantics, which are therefore unaware of
   side-effects. This means the effect could be dropped, duplicated, or
   potentially reordered in the presence of higher-order primitives and
   transformations.
 
-  We want this behavior because we'd like `debug_callback` to be "innocuous",
+  We want this behavior because we'd like ``jax.debug.callback`` to be "innocuous",
   i.e. we want these primitives to change the JAX computation as little as
   possible while revealing as much about them as possible, such as which parts
   of the computation are duplicated or dropped.
@@ -273,8 +274,8 @@ def debug_print(fmt: str, *args, ordered: bool = False, **kwargs) -> None:
       input arguments.
     *args: A list of positional arguments to be formatted.
     ordered: A keyword only argument used to indicate whether or not the
-      staged out computation will enforce ordering of this ``debug_print``
-      w.r.t. other ordered ``debug_print`` calls.
+      staged out computation will enforce ordering of this ``jax.debug.print``
+      w.r.t. other ordered ``jax.debug.print`` calls.
     **kwargs: Additional keyword arguments to be formatted.
   """
   # Check that we provide the correct arguments to be formatted
@@ -413,8 +414,8 @@ def _raise_to_slice(slc: Union[slice, int]):
     return slice(slc, slc + 1)
   return slc
 
-Color = Union[Tuple[float, float, float], str]
-ColorMap = Callable[[float], Tuple[float, float, float, float]]
+Color = Union[tuple[float, float, float], str]
+ColorMap = Callable[[float], tuple[float, float, float, float]]
 
 def _canonicalize_color(color: Color) -> str:
   if isinstance(color, str):
@@ -464,9 +465,9 @@ def visualize_sharding(shape: Sequence[int], sharding: Sharding, *,
   device_kind = next(iter(sharding.device_set)).platform.upper()
 
   device_indices_map = sharding.devices_indices_map(tuple(shape))
-  slices: Dict[Tuple[int, ...], Set[int]] = {}
-  heights: Dict[Tuple[int, ...], Optional[float]] = {}
-  widths: Dict[Tuple[int, ...], float] = {}
+  slices: dict[tuple[int, ...], set[int]] = {}
+  heights: dict[tuple[int, ...], Optional[float]] = {}
+  widths: dict[tuple[int, ...], float] = {}
 
   for i, (dev, slcs) in enumerate(device_indices_map.items()):
     assert slcs is not None
