@@ -19,46 +19,137 @@
 # b) the set of supported types (e.g., bfloat16),
 # so we need our own implementation that deviates from NumPy in places.
 
+from __future__ import annotations
 
+import abc
 import builtins
 import functools
-from typing import (cast, overload, Any, Dict, List, Literal, Optional, Set,
-                    Tuple, Type, Union)
+from typing import cast, overload, Any, Literal, Optional, Union
 import warnings
 
 import ml_dtypes
 import numpy as np
 
-from jax._src.config import flags, config
-from jax._src.typing import DType, DTypeLike, OpaqueDType
+from jax._src.config import config
+from jax._src.typing import DType, DTypeLike
 
 from jax._src import traceback_util
 traceback_util.register_exclusion(__file__)
 
-FLAGS = flags.FLAGS
+try:
+  _ml_dtypes_version = tuple(map(int, ml_dtypes.__version__.split('.')[:3]))
+except:
+  pass
+else:
+  if _ml_dtypes_version < (0, 2, 0):
+    raise ValueError("JAX requires ml_dtypes version 0.2.0 or newer; "
+                     f"installed version is {ml_dtypes.__version__}.")
+
+
+class extended(np.generic):
+  """Scalar class for extended dtypes.
+
+  This is an abstract class that should never be instantiated, but rather
+  exists for the sake of `jnp.issubdtype`.
+
+  Examples:
+    >>> from jax import random
+    >>> from jax import dtypes
+    >>> key = random.key(0)
+    >>> jnp.issubdtype(key.dtype, dtypes.extended)
+    True
+  """
+  pass
+
+
+class prng_key(extended):
+  """Scalar class for PRNG Key dtypes.
+
+  This is an abstract class that should never be instantiated, but rather
+  exists for the sake of `jnp.issubdtype`.
+
+  Examples:
+    >>> from jax import random
+    >>> from jax import dtypes
+    >>> key = random.key(0)
+    >>> jnp.issubdtype(key.dtype, dtypes.prng_key)
+    True
+  """
+  pass
+
+
+class ExtendedDType(metaclass=abc.ABCMeta):
+  """Abstract Base Class for extended dtypes"""
+  @property
+  @abc.abstractmethod
+  def type(self) -> type: ...
+
 
 # fp8 support
-float8_e4m3fn: Type[np.generic] = ml_dtypes.float8_e4m3fn
-float8_e5m2: Type[np.generic] = ml_dtypes.float8_e5m2
+float8_e4m3b11fnuz: type[np.generic] = ml_dtypes.float8_e4m3b11fnuz
+float8_e4m3fn: type[np.generic] = ml_dtypes.float8_e4m3fn
+float8_e4m3fnuz: type[np.generic] = ml_dtypes.float8_e4m3fnuz
+float8_e5m2: type[np.generic] = ml_dtypes.float8_e5m2
+float8_e5m2fnuz: type[np.generic] = ml_dtypes.float8_e5m2fnuz
+
+_float8_e4m3b11fnuz_dtype: np.dtype = np.dtype(float8_e4m3b11fnuz)
 _float8_e4m3fn_dtype: np.dtype = np.dtype(float8_e4m3fn)
+_float8_e4m3fnuz_dtype: np.dtype = np.dtype(float8_e4m3fnuz)
 _float8_e5m2_dtype: np.dtype = np.dtype(float8_e5m2)
+_float8_e5m2fnuz_dtype: np.dtype = np.dtype(float8_e5m2fnuz)
 
 # bfloat16 support
-bfloat16: Type[np.generic] = ml_dtypes.bfloat16
+bfloat16: type[np.generic] = ml_dtypes.bfloat16
 _bfloat16_dtype: np.dtype = np.dtype(bfloat16)
 
+_custom_float_scalar_types = [
+    float8_e4m3b11fnuz,
+    float8_e4m3fn,
+    float8_e4m3fnuz,
+    float8_e5m2,
+    float8_e5m2fnuz,
+    bfloat16,
+]
+_custom_float_dtypes = [
+    _float8_e4m3b11fnuz_dtype,
+    _float8_e4m3fn_dtype,
+    _float8_e4m3fnuz_dtype,
+    _float8_e5m2_dtype,
+    _float8_e5m2fnuz_dtype,
+    _bfloat16_dtype,
+]
+_float8_dtypes = [
+    _float8_e4m3b11fnuz_dtype,
+    _float8_e4m3fn_dtype,
+    _float8_e4m3fnuz_dtype,
+    _float8_e5m2_dtype,
+    _float8_e5m2fnuz_dtype,
+]
+
+# 4-bit integer support
+int4: type[np.generic] = ml_dtypes.int4
+uint4: type[np.generic] = ml_dtypes.uint4
+
+_int4_dtype: np.dtype = np.dtype(int4)
+_uint4_dtype: np.dtype = np.dtype(uint4)
+
+_int4_dtypes = [
+    _int4_dtype,
+    _uint4_dtype,
+]
+
 # Default types.
 bool_: type = np.bool_
 int_: type = np.int32 if config.jax_default_dtype_bits == '32' else np.int64
 uint: type = np.uint32 if config.jax_default_dtype_bits == '32' else np.uint64
 float_: type = np.float32 if config.jax_default_dtype_bits == '32' else np.float64
 complex_: type = np.complex64 if config.jax_default_dtype_bits == '32' else np.complex128
-_default_types: Dict[str, type] = {'b': bool_, 'i': int_, 'u': uint, 'f': float_, 'c': complex_}
+_default_types: dict[str, type] = {'b': bool_, 'i': int_, 'u': uint, 'f': float_, 'c': complex_}
 
 # Trivial vectorspace datatype needed for tangent values of int/bool primals
 float0: np.dtype = np.dtype([('float0', np.void, 0)])
 
-_dtype_to_32bit_dtype: Dict[DType, DType] = {
+_dtype_to_32bit_dtype: dict[DType, DType] = {
     np.dtype('int64'): np.dtype('int32'),
     np.dtype('uint64'): np.dtype('uint32'),
     np.dtype('float64'): np.dtype('float32'),
@@ -68,7 +159,7 @@ _dtype_to_32bit_dtype: Dict[DType, DType] = {
 # Note: we promote narrow types to float32 here for backward compatibility
 # with earlier approaches. We might consider revisiting this, or perhaps
 # tying the logic more closely to the type promotion lattice.
-_dtype_to_inexact: Dict[DType, DType] = {
+_dtype_to_inexact: dict[DType, DType] = {
     np.dtype(k): np.dtype(v) for k, v in [
         ('bool', 'float32'),
         ('uint8', 'float32'), ('int8', 'float32'),
@@ -97,13 +188,12 @@ def to_complex_dtype(dtype: DTypeLike) -> DType:
   return np.dtype('complex64')
 
 
-@functools.lru_cache(maxsize=None)
-def _canonicalize_dtype(x64_enabled: bool, allow_opaque_dtype: bool, dtype: Any) -> Union[DType, OpaqueDType]:
-  from jax._src import core     # TODO(frostig): break this cycle
-  if core.is_opaque_dtype(dtype):
-    if not allow_opaque_dtype:
-      raise ValueError(f"Internal: canonicalize_dtype called onopaque dtype {dtype} "
-                       "with allow_opaque_dtype=False")
+@functools.cache
+def _canonicalize_dtype(x64_enabled: bool, allow_extended_dtype: bool, dtype: Any) -> Union[DType, ExtendedDType]:
+  if issubdtype(dtype, extended):
+    if not allow_extended_dtype:
+      raise ValueError(f"Internal: canonicalize_dtype called on extended dtype {dtype} "
+                       "with allow_extended_dtype=False")
     return dtype
   try:
     dtype_ = np.dtype(dtype)
@@ -116,17 +206,23 @@ def _canonicalize_dtype(x64_enabled: bool, allow_opaque_dtype: bool, dtype: Any)
     return _dtype_to_32bit_dtype.get(dtype_, dtype_)
 
 @overload
-def canonicalize_dtype(dtype: Any, allow_opaque_dtype: Literal[False] = False) -> DType: ...
+def canonicalize_dtype(dtype: Any, allow_extended_dtype: Literal[False] = False, allow_opaque_dtype: Any = None) -> DType: ...
 
 @overload
-def canonicalize_dtype(dtype: Any, allow_opaque_dtype: bool = False) -> Union[DType, OpaqueDType]: ...
+def canonicalize_dtype(dtype: Any, allow_extended_dtype: bool = False, allow_opaque_dtype: Any = None) -> Union[DType, ExtendedDType]: ...
 
-def canonicalize_dtype(dtype: Any, allow_opaque_dtype: bool = False) -> Union[DType, OpaqueDType]:
+def canonicalize_dtype(dtype: Any, allow_extended_dtype: bool = False, allow_opaque_dtype: Any = None) -> Union[DType, ExtendedDType]:
   """Convert from a dtype to a canonical dtype based on config.x64_enabled."""
-  return _canonicalize_dtype(config.x64_enabled, allow_opaque_dtype, dtype)
+  if allow_opaque_dtype is not None:
+    # TODO(jakevdp): complete the deprecation cycle (Deprecated July 24 2023).
+    warnings.warn(
+      "allow_opaque_dtype argument is deprecated; use allow_extended_dtype.",
+      DeprecationWarning)
+    allow_extended_dtype = allow_opaque_dtype
+  return _canonicalize_dtype(config.x64_enabled, allow_extended_dtype, dtype)  # type: ignore[bad-return-type]
 
 # Default dtypes corresponding to Python scalars.
-python_scalar_dtypes : Dict[type, DType] = {
+python_scalar_dtypes : dict[type, DType] = {
   bool: np.dtype('bool'),
   int: np.dtype('int64'),
   float: np.dtype('float64'),
@@ -136,12 +232,10 @@ python_scalar_dtypes : Dict[type, DType] = {
 def scalar_type_of(x: Any) -> type:
   """Return the scalar type associated with a JAX value."""
   typ = dtype(x)
-  if typ == bfloat16:
-    return float
-  elif typ == float8_e4m3fn:
-    return float
-  elif typ == float8_e5m2:
+  if typ in _custom_float_dtypes:
     return float
+  elif typ in _int4_dtypes:
+    return int
   elif np.issubdtype(typ, np.bool_):
     return bool
   elif np.issubdtype(typ, np.integer):
@@ -194,172 +288,8 @@ def coerce_to_array(x: Any, dtype: Optional[DTypeLike] = None) -> np.ndarray:
     dtype = _scalar_type_to_dtype(type(x), x)
   return np.asarray(x, dtype)
 
-iinfo = np.iinfo
-
-class _Bfloat16MachArLike:
-  def __init__(self):
-    smallest_normal = float.fromhex("0x1p-126")
-    self.smallest_normal = bfloat16(smallest_normal)
-    smallest_subnormal = float.fromhex("0x1p-133")
-    self.smallest_subnormal = bfloat16(smallest_subnormal)
-
-class _Float8E4m3FnMachArLike:
-  def __init__(self):
-    smallest_normal = float.fromhex("0x1p-6")
-    self.smallest_normal = float8_e4m3fn(smallest_normal)
-    smallest_subnormal = float.fromhex("0x1p-9")
-    self.smallest_subnormal = float8_e4m3fn(smallest_subnormal)
-
-class _Float8E5m2MachArLike:
-  def __init__(self):
-    smallest_normal = float.fromhex("0x1p-14")
-    self.smallest_normal = float8_e5m2(smallest_normal)
-    smallest_subnormal = float.fromhex("0x1p-16")
-    self.smallest_subnormal = float8_e5m2(smallest_subnormal)
-
-class finfo(np.finfo):
-  __doc__ = np.finfo.__doc__
-  _finfo_cache: Dict[np.dtype, np.finfo] = {}
-  @staticmethod
-  def _bfloat16_finfo():
-    def float_to_str(f):
-      return "%12.4e" % float(f)
-
-    bfloat16 = _bfloat16_dtype.type
-    tiny = float.fromhex("0x1p-126")
-    resolution = 0.01
-    eps = float.fromhex("0x1p-7")
-    epsneg = float.fromhex("0x1p-8")
-    max = float.fromhex("0x1.FEp127")
-
-    obj = object.__new__(np.finfo)
-    obj.dtype = _bfloat16_dtype
-    obj.bits = 16
-    obj.eps = bfloat16(eps)
-    obj.epsneg = bfloat16(epsneg)
-    obj.machep = -7
-    obj.negep = -8
-    obj.max = bfloat16(max)
-    obj.min = bfloat16(-max)
-    obj.nexp = 8
-    obj.nmant = 7
-    obj.iexp = obj.nexp
-    obj.maxexp = 128
-    obj.precision = 2
-    obj.resolution = bfloat16(resolution)
-    obj._machar = _Bfloat16MachArLike()
-    if not hasattr(obj, "tiny"):
-      obj.tiny = bfloat16(tiny)
-    if not hasattr(obj, "smallest_normal"):
-      obj.smallest_normal = obj._machar.smallest_normal
-    obj.smallest_subnormal = obj._machar.smallest_subnormal
-
-    obj._str_tiny = float_to_str(tiny)
-    obj._str_smallest_normal = float_to_str(tiny)
-    obj._str_max = float_to_str(max)
-    obj._str_epsneg = float_to_str(epsneg)
-    obj._str_eps = float_to_str(eps)
-    obj._str_resolution = float_to_str(resolution)
-    return obj
-
-  @staticmethod
-  def _float8_e4m3fn_finfo():
-    def float_to_str(f):
-      return "%6.2e" % float(f)
-
-    float8_e4m3fn = _float8_e4m3fn_dtype.type
-    tiny = float.fromhex("0x1p-6")
-    resolution = 0.1
-    eps = float.fromhex("0x1p-3")
-    epsneg = float.fromhex("0x1p-4")
-    max = float.fromhex("0x1.Cp8")
-
-    obj = object.__new__(np.finfo)
-    obj.dtype = _float8_e4m3fn_dtype
-    obj.bits = 8
-    obj.eps = float8_e4m3fn(eps)
-    obj.epsneg = float8_e4m3fn(epsneg)
-    obj.machep = -3
-    obj.negep = -4
-    obj.max = float8_e4m3fn(max)
-    obj.min = float8_e4m3fn(-max)
-    obj.nexp = 4
-    obj.nmant = 3
-    obj.iexp = obj.nexp
-    obj.maxexp = 9
-    obj.precision = 1
-    obj.resolution = float8_e4m3fn(resolution)
-    obj._machar = _Float8E4m3FnMachArLike()
-    if not hasattr(obj, "tiny"):
-      obj.tiny = float8_e4m3fn(tiny)
-    if not hasattr(obj, "smallest_normal"):
-      obj.smallest_normal = obj._machar.smallest_normal
-    obj.smallest_subnormal = obj._machar.smallest_subnormal
-
-    obj._str_tiny = float_to_str(tiny)
-    obj._str_smallest_normal = float_to_str(tiny)
-    obj._str_max = float_to_str(max)
-    obj._str_epsneg = float_to_str(epsneg)
-    obj._str_eps = float_to_str(eps)
-    obj._str_resolution = float_to_str(resolution)
-    return obj
-
-  @staticmethod
-  def _float8_e5m2_finfo():
-    def float_to_str(f):
-      return "%6.2e" % float(f)
-
-    float8_e5m2 = _float8_e5m2_dtype.type
-    tiny = float.fromhex("0x1p-14")
-    resolution = 0.1
-    eps = float.fromhex("0x1p-2")
-    epsneg = float.fromhex("0x1p-3")
-    max = float.fromhex("0x1.Cp15")
-
-    obj = object.__new__(np.finfo)
-    obj.dtype = _float8_e5m2_dtype
-    obj.bits = 8
-    obj.eps = float8_e5m2(eps)
-    obj.epsneg = float8_e5m2(epsneg)
-    obj.machep = -2
-    obj.negep = -3
-    obj.max = float8_e5m2(max)
-    obj.min = float8_e5m2(-max)
-    obj.nexp = 5
-    obj.nmant = 2
-    obj.iexp = obj.nexp
-    obj.maxexp = 16
-    obj.precision = 1
-    obj.resolution = float8_e5m2(resolution)
-    obj._machar = _Float8E5m2MachArLike()
-    if not hasattr(obj, "tiny"):
-      obj.tiny = float8_e5m2(tiny)
-    if not hasattr(obj, "smallest_normal"):
-      obj.smallest_normal = obj._machar.smallest_normal
-    obj.smallest_subnormal = obj._machar.smallest_subnormal
-
-    obj._str_tiny = float_to_str(tiny)
-    obj._str_smallest_normal = float_to_str(tiny)
-    obj._str_max = float_to_str(max)
-    obj._str_epsneg = float_to_str(epsneg)
-    obj._str_eps = float_to_str(eps)
-    obj._str_resolution = float_to_str(resolution)
-    return obj
-
-  def __new__(cls, dtype):
-    if isinstance(dtype, str) and dtype == 'bfloat16' or dtype == _bfloat16_dtype:
-      if _bfloat16_dtype not in cls._finfo_cache:
-        cls._finfo_cache[_bfloat16_dtype] = cls._bfloat16_finfo()
-      return cls._finfo_cache[_bfloat16_dtype]
-    if isinstance(dtype, str) and dtype == 'float8_e4m3fn' or dtype == _float8_e4m3fn_dtype:
-      if _float8_e4m3fn_dtype not in cls._finfo_cache:
-        cls._finfo_cache[_float8_e4m3fn_dtype] = cls._float8_e4m3fn_finfo()
-      return cls._finfo_cache[_float8_e4m3fn_dtype]
-    if isinstance(dtype, str) and dtype == 'float8_e5m2' or dtype == _float8_e5m2_dtype:
-      if _float8_e5m2_dtype not in cls._finfo_cache:
-        cls._finfo_cache[_float8_e5m2_dtype] = cls._float8_e5m2_finfo()
-      return cls._finfo_cache[_float8_e5m2_dtype]
-    return super().__new__(cls, dtype)
+iinfo = ml_dtypes.iinfo
+finfo = ml_dtypes.finfo
 
 def _issubclass(a: Any, b: Any) -> bool:
   """Determines if ``a`` is a subclass of ``b``.
@@ -372,72 +302,90 @@ def _issubclass(a: Any, b: Any) -> bool:
   except TypeError:
     return False
 
+
 def issubdtype(a: DTypeLike, b: DTypeLike) -> bool:
   """Returns True if first argument is a typecode lower/equal in type hierarchy.
 
   This is like :func:`numpy.issubdtype`, but can handle dtype extensions such as
-  :obj:`jax.dtypes.bfloat16`.
-  '"""
-  if a == "float8_e4m3fn":
-    a = float8_e4m3fn
-  if a == float8_e4m3fn:
-    if isinstance(b, np.dtype):
-      return b == _float8_e4m3fn_dtype
-    else:
-      return b in [float8_e4m3fn, np.floating, np.inexact, np.number]
-  if a == "float8_e5m2":
-    a = float8_e5m2
-  if a == float8_e5m2:
-    if isinstance(b, np.dtype):
-      return b == _float8_e5m2_dtype
-    else:
-      return b in [float8_e5m2, np.floating, np.inexact, np.number]
-  if a == "bfloat16":
-    a = bfloat16
-  if a == bfloat16:
-    if isinstance(b, np.dtype):
-      return b == _bfloat16_dtype
-    else:
-      return b in [bfloat16, np.floating, np.inexact, np.number]
-  if not _issubclass(b, np.generic):
-    # Workaround for JAX scalar types. NumPy's issubdtype has a backward
-    # compatibility behavior for the second argument of issubdtype that
-    # interacts badly with JAX's custom scalar types. As a workaround,
-    # explicitly cast the second argument to a NumPy type object.
-    b = np.dtype(b).type
-  try:
-    return np.issubdtype(a, b)
-  except TypeError:  # e.g. if 'a' is not a np.dtype
-    return False
+  :obj:`jax.dtypes.bfloat16` and `jax.dtypes.prng_key`.
+  """
+  # Main departures from np.issubdtype are:
+  # - "extended" dtypes (like prng key types) are not normal numpy dtypes, so we
+  #   need to handle them specifically. However, their scalar types do conform to
+  #   the numpy scalar type hierarchy.
+  # - custom dtypes (like bfloat16, int4, etc.) are normal numpy dtypes, but they
+  #   don't conform to the standard numpy type hierarchy (e.g. the bfloat16 scalar
+  #   type is not a subclass of np.floating) so we must also handle these specially.
+
+  # First handle extended dtypes. This is important for performance because
+  # isinstance(x, extended) is called frequently within JAX internals.
+  if _issubclass(b, extended):
+    if isinstance(a, ExtendedDType):
+      return _issubclass(a.type, b)
+    if _issubclass(a, np.generic):
+      return _issubclass(a, b)
+    return _issubclass(np.dtype(a).type, b)
+  if isinstance(b, ExtendedDType):
+    return isinstance(a, ExtendedDType) and a == b
+  if isinstance(a, ExtendedDType):
+    a = a.type
+
+  # For all others, normalize inputs to scalar types.
+  a_sctype = a if _issubclass(a, np.generic) else np.dtype(a).type
+  b_sctype = b if _issubclass(b, np.generic) else np.dtype(b).type
+
+  # Now do special handling of custom float and int types, as they don't conform
+  # to the normal scalar type hierarchy.
+  if a_sctype in _custom_float_scalar_types:
+    return b_sctype in {a_sctype, np.floating, np.inexact, np.number, np.generic}
+  if a_sctype == int4:
+    return b_sctype in {a_sctype, np.signedinteger, np.integer, np.number, np.generic}
+  if a_sctype == uint4:
+    return b_sctype in {a_sctype, np.unsignedinteger, np.integer, np.number, np.generic}
+
+  # Otherwise, fall back to numpy.issubdtype
+  return np.issubdtype(a_sctype, b_sctype)
 
 can_cast = np.can_cast
-issubsctype = np.issubsctype
 
 JAXType = Union[type, DType]
 
 # Enumeration of all valid JAX types in order.
-_weak_types: List[JAXType] = [int, float, complex]
-_bool_types: List[JAXType] = [np.dtype(bool)]
-_int_types: List[JAXType] = [
-    np.dtype('uint8'),
-    np.dtype('uint16'),
-    np.dtype('uint32'),
-    np.dtype('uint64'),
-    np.dtype('int8'),
-    np.dtype('int16'),
-    np.dtype('int32'),
-    np.dtype('int64'),
+_weak_types: list[JAXType] = [int, float, complex]
+_bool_types: list[JAXType] = [np.dtype(bool)]
+_int_types: list[JAXType]
+if int4 is not None:
+  _int_types = [
+      np.dtype(uint4),
+      np.dtype('uint8'),
+      np.dtype('uint16'),
+      np.dtype('uint32'),
+      np.dtype('uint64'),
+      np.dtype(int4),
+      np.dtype('int8'),
+      np.dtype('int16'),
+      np.dtype('int32'),
+      np.dtype('int64'),
+  ]
+else:
+  _int_types = [
+      np.dtype('uint8'),
+      np.dtype('uint16'),
+      np.dtype('uint32'),
+      np.dtype('uint64'),
+      np.dtype('int8'),
+      np.dtype('int16'),
+      np.dtype('int32'),
+      np.dtype('int64'),
+  ]
+
+_float_types: list[JAXType] = [
+    *_custom_float_dtypes,
+    np.dtype('float16'),
+    np.dtype('float32'),
+    np.dtype('float64'),
 ]
-_float_types: List[JAXType]
-_float_types = [
-  np.dtype(float8_e4m3fn),
-  np.dtype(float8_e5m2),
-  np.dtype(bfloat16),
-  np.dtype('float16'),
-  np.dtype('float32'),
-  np.dtype('float64'),
-]
-_complex_types: List[JAXType] = [
+_complex_types: list[JAXType] = [
     np.dtype('complex64'),
     np.dtype('complex128'),
 ]
@@ -449,33 +397,45 @@ def _jax_type(dtype: DType, weak_type: bool) -> JAXType:
   if weak_type:
     if dtype == bool:
       return dtype
-    if dtype in [_float8_e4m3fn_dtype, _float8_e5m2_dtype, _bfloat16_dtype]:
+    if dtype in _custom_float_dtypes:
       return float
     return type(dtype.type(0).item())
   return dtype
 
-def _dtype_and_weaktype(value: Any) -> Tuple[DType, bool]:
+def _dtype_and_weaktype(value: Any) -> tuple[DType, bool]:
   """Return a (dtype, weak_type) tuple for the given input."""
   return dtype(value), any(value is typ for typ in _weak_types) or is_weakly_typed(value)
 
-def _type_promotion_lattice(jax_numpy_dtype_promotion: str) -> Dict[JAXType, List[JAXType]]:
+def _type_promotion_lattice(jax_numpy_dtype_promotion: str) -> dict[JAXType, list[JAXType]]:
   """
   Return the type promotion lattice in the form of a DAG.
   This DAG maps each type to its immediately higher type on the lattice.
   """
   b1, = _bool_types
-  u1, u2, u4, u8, i1, i2, i4, i8 = _int_types
-  f1_e4m3fn, f1_e5m2, bf, f2, f4, f8 = _float_types
+  if int4 is not None:
+    _uint4, u1, u2, u4, u8, _int4, i1, i2, i4, i8 = _int_types  # pytype: disable=bad-unpacking
+  else:
+    u1, u2, u4, u8, i1, i2, i4, i8 = _int_types  # pytype: disable=bad-unpacking
+  *f1_types, bf, f2, f4, f8 = _float_types
   c4, c8 = _complex_types
   i_, f_, c_ = _weak_types
   if jax_numpy_dtype_promotion == 'standard':
-    return {
+    out: dict[JAXType, list[JAXType]]
+    out = {
       b1: [i_],
       u1: [i2, u2], u2: [i4, u4], u4: [i8, u8], u8: [f_],
       i_: [u1, i1], i1: [i2], i2: [i4], i4: [i8], i8: [f_],
-      f_: [f1_e4m3fn, f1_e5m2, bf, f2, c_], f1_e4m3fn: [], f1_e5m2: [], bf: [f4], f2: [f4], f4: [f8, c4], f8: [c8],
+      f_: [*f1_types, bf, f2, c_],
+      **{t: [] for t in f1_types}, bf: [f4], f2: [f4], f4: [f8, c4], f8: [c8],
       c_: [c4], c4: [c8], c8: [],
     }
+    if _int4_dtype is not None:
+      out[i_].append(_int4_dtype)
+      out[_int4_dtype] = []
+    if _uint4_dtype is not None:
+      out[i_].append(_uint4_dtype)
+      out[_uint4_dtype] = []
+    return out
   elif jax_numpy_dtype_promotion == 'strict':
     return {
       i_: [f_] + _int_types,
@@ -487,7 +447,7 @@ def _type_promotion_lattice(jax_numpy_dtype_promotion: str) -> Dict[JAXType, Lis
     raise ValueError(
       f"Unexpected value of jax_numpy_dtype_promotion={jax_numpy_dtype_promotion!r}")
 
-def _make_lattice_upper_bounds(jax_numpy_dtype_promotion: str) -> Dict[JAXType, Set[JAXType]]:
+def _make_lattice_upper_bounds(jax_numpy_dtype_promotion: str) -> dict[JAXType, set[JAXType]]:
   lattice = _type_promotion_lattice(jax_numpy_dtype_promotion)
   upper_bounds = {node: {node} for node in lattice}
   for n in lattice:
@@ -500,7 +460,7 @@ def _make_lattice_upper_bounds(jax_numpy_dtype_promotion: str) -> Dict[JAXType,
       upper_bounds[n] |= new_upper_bounds
   return upper_bounds
 
-_lattice_upper_bounds: Dict[str, Dict[JAXType, Set[JAXType]]] = {
+_lattice_upper_bounds: dict[str, dict[JAXType, set[JAXType]]] = {
   'standard': _make_lattice_upper_bounds('standard'),
   'strict': _make_lattice_upper_bounds('strict'),
 }
@@ -537,7 +497,12 @@ def _least_upper_bound(jax_numpy_dtype_promotion: str, *nodes: JAXType) -> JAXTy
   # So if N ∩ CUB(N) is nonempty, if follows that LUB(N) = N ∩ CUB(N).
   N = set(nodes)
   UB = _lattice_upper_bounds[jax_numpy_dtype_promotion]
-  CUB = set.intersection(*(UB[n] for n in N))
+  try:
+    bounds = [UB[n] for n in N]
+  except KeyError:
+    dtype = next(n for n in N if n not in UB)
+    raise ValueError(f"{dtype=} is not a valid dtype for JAX type promotion.")
+  CUB = set.intersection(*bounds)
   LUB = (CUB & N) or {c for c in CUB if CUB.issubset(UB[c])}
   if len(LUB) == 1:
     return LUB.pop()
@@ -547,6 +512,18 @@ def _least_upper_bound(jax_numpy_dtype_promotion: str, *nodes: JAXType) -> JAXTy
         f"Input dtypes {tuple(str(n) for n in nodes)} have no available implicit dtype "
         "promotion path when jax_numpy_dtype_promotion=strict. Try explicitly casting "
         "inputs to the desired output type, or set jax_numpy_dtype_promotion=standard.")
+    elif any(n in _float8_dtypes for n in nodes):
+      msg = (
+        f"Input dtypes {tuple(str(n) for n in nodes)} have no available implicit dtype "
+        "promotion path. To avoid unintended promotion, 8-bit floats do not support "
+        "implicit promotion. If you'd like your inputs to be promoted to another type, "
+        "you can do so explicitly using e.g. x.astype('float32')")
+    elif any(n in _int4_dtypes for n in nodes):
+      msg = (
+        f"Input dtypes {tuple(str(n) for n in nodes)} have no available implicit dtype "
+        "promotion path. To avoid unintended promotion, 4-bit integers do not support "
+        "implicit promotion. If you'd like your inputs to be promoted to another type, "
+        "you can do so explicitly using e.g. x.astype('int32')")
     else:
       msg = (
         f"Input dtypes {tuple(str(n) for n in nodes)} have no available implicit dtype "
@@ -597,30 +574,36 @@ def check_valid_dtype(dtype: DType) -> None:
 
 def dtype(x: Any, *, canonicalize: bool = False) -> DType:
   """Return the dtype object for a value or type, optionally canonicalized based on X64 mode."""
-  from jax._src import core     # TODO(frostig): break this cycle
   if x is None:
     raise ValueError(f"Invalid argument to dtype: {x}.")
   elif isinstance(x, type) and x in python_scalar_dtypes:
     dt = python_scalar_dtypes[x]
   elif type(x) in python_scalar_dtypes:
     dt = python_scalar_dtypes[type(x)]
-  elif core.is_opaque_dtype(getattr(x, 'dtype', None)):
+  elif _issubclass(x, np.generic):
+    return np.dtype(x)
+  elif issubdtype(getattr(x, 'dtype', None), extended):
     dt = x.dtype
   else:
     try:
       dt = np.result_type(x)
     except TypeError as err:
       raise TypeError(f"Cannot determine dtype of {x}") from err
-  if dt not in _jax_dtype_set:
+  if dt not in _jax_dtype_set and not issubdtype(dt, extended):
     raise TypeError(f"Value '{x}' with dtype {dt} is not a valid JAX array "
                     "type. Only arrays of numeric types are supported by JAX.")
-  return canonicalize_dtype(dt) if canonicalize else dt
+  # TODO(jakevdp): fix return type annotation and remove this ignore.
+  return canonicalize_dtype(dt, allow_extended_dtype=True) if canonicalize else dt  # type: ignore[return-value]
 
-def _lattice_result_type(*args: Any) -> Tuple[DType, bool]:
+def _lattice_result_type(*args: Any) -> tuple[DType, bool]:
   dtypes, weak_types = zip(*(_dtype_and_weaktype(arg) for arg in args))
   if len(dtypes) == 1:
     out_dtype = dtypes[0]
     out_weak_type = weak_types[0]
+  elif len(set(dtypes)) == 1 and not all(weak_types):
+    # Trivial promotion case. This allows extended dtypes through.
+    out_dtype = dtypes[0]
+    out_weak_type = False
   elif all(weak_types) and config.jax_numpy_dtype_promotion != 'strict':
     # If all inputs are weakly typed, we compute the bound of the strongly-typed
     # counterparts and apply the weak type at the end. This avoids returning the
@@ -638,15 +621,15 @@ def _lattice_result_type(*args: Any) -> Tuple[DType, bool]:
   return out_dtype, (out_dtype != bool_) and out_weak_type
 
 @overload
-def result_type(*args: Any, return_weak_type_flag: Literal[True]) -> Tuple[DType, bool]: ...
+def result_type(*args: Any, return_weak_type_flag: Literal[True]) -> tuple[DType, bool]: ...
 
 @overload
 def result_type(*args: Any, return_weak_type_flag: Literal[False] = False) -> DType: ...
 
 @overload
-def result_type(*args: Any, return_weak_type_flag: bool = False) -> Union[DType, Tuple[DType, bool]]: ...
+def result_type(*args: Any, return_weak_type_flag: bool = False) -> Union[DType, tuple[DType, bool]]: ...
 
-def result_type(*args: Any, return_weak_type_flag: bool = False) -> Union[DType, Tuple[DType, bool]]:
+def result_type(*args: Any, return_weak_type_flag: bool = False) -> Union[DType, tuple[DType, bool]]:
   """Convenience function to apply JAX argument dtype promotion.
 
   Args:
@@ -658,20 +641,27 @@ def result_type(*args: Any, return_weak_type_flag: bool = False) -> Union[DType,
   """
   if len(args) == 0:
     raise ValueError("at least one array or dtype is required")
+  dtype: DType | ExtendedDType
   dtype, weak_type = _lattice_result_type(*(float_ if arg is None else arg for arg in args))
   if weak_type:
     dtype = canonicalize_dtype(
-      _default_types['f' if dtype in [_float8_e4m3fn_dtype, _float8_e5m2_dtype, _bfloat16_dtype] else dtype.kind])
+      _default_types['f' if dtype in _custom_float_dtypes else dtype.kind])
   else:
-    dtype = canonicalize_dtype(dtype)
-  return (dtype, weak_type) if return_weak_type_flag else dtype
+    dtype = canonicalize_dtype(dtype, allow_extended_dtype=True)
+  # TODO(jakevdp): fix return type annotation and remove this ignore.
+  return (dtype, weak_type) if return_weak_type_flag else dtype  # type: ignore[return-value]
 
 def check_user_dtype_supported(dtype, fun_name=None):
+  if issubdtype(dtype, extended):
+    return
   # Avoid using `dtype in [...]` because of numpy dtype equality overloading.
   if isinstance(dtype, type) and dtype in {bool, int, float, builtins.complex}:
     return
   np_dtype = np.dtype(dtype)
-  is_custom_dtype = np_dtype.type in [float8_e4m3fn, float8_e5m2, bfloat16]
+  if int4 is not None:
+    is_custom_dtype = np_dtype.type in [*_custom_float_scalar_types, int4, uint4]
+  else:
+    is_custom_dtype = np_dtype.type in _custom_float_scalar_types
   if np_dtype.kind not in "biufc" and not is_custom_dtype:
     msg = f"JAX only supports number and bool dtypes, got dtype {dtype}"
     msg += f" in {fun_name}" if fun_name else ""
@@ -684,4 +674,41 @@ def check_user_dtype_supported(dtype, fun_name=None):
            "See https://github.com/google/jax#current-gotchas for more.")
     fun_name = f"requested in {fun_name}" if fun_name else ""
     truncated_dtype = canonicalize_dtype(dtype).name
-    warnings.warn(msg.format(dtype, fun_name , truncated_dtype), stacklevel=3)
+    warnings.warn(msg.format(dtype, fun_name, truncated_dtype), stacklevel=3)
+
+def safe_to_cast(input_dtype_or_value: Any,
+                 output_dtype_or_value: Any) -> bool:
+  """Check if a dtype/value is safe to cast to another dtype/value
+
+  Args:
+    input_dtype_or_value: a dtype or value (to be passed to result_type)
+      representing the source dtype.
+    output_dtype_or_value: a dtype or value (to be passed to result_type)
+      representing the target dtype.
+
+  Returns:
+    boolean representing whether the values are safe to cast according to
+    default type promotion semantics.
+
+  Raises:
+    TypePromotionError: if the inputs have differing types and no type promotion
+    path under the current jax_numpy_dtype_promotion setting.
+
+  Examples:
+
+    >>> safe_to_cast('int32', 'float64')
+    True
+    >>> safe_to_cast('float64', 'int32')
+    False
+    >>> safe_to_cast('float32', 'complex64')
+    True
+    >>> safe_to_cast('complex64', 'float64')
+    False
+  """
+  input_dtype = dtype(input_dtype_or_value, canonicalize=True)
+  output_dtype = dtype(output_dtype_or_value, canonicalize=True)
+  if input_dtype == output_dtype:
+    return True
+  # We deliberately use output_dtype rather than output_dtype_or_value here:
+  # this effectively treats the output dtype as always strongly-typed.
+  return result_type(input_dtype_or_value, output_dtype) == output_dtype
