@@ -28,7 +28,7 @@ from jax._src.util import safe_map, safe_zip
 
 import numpy as np
 
-from jax.config import config
+from jax import config
 config.parse_flags_with_absl()
 
 map, unsafe_map = safe_map, map
@@ -167,6 +167,7 @@ def lax_ops():
       ),
       op_record("is_finite", 1, float_dtypes, test_util.rand_small),
       op_record("exp", 1, float_dtypes + complex_dtypes, test_util.rand_small),
+      op_record("exp2", 1, float_dtypes + complex_dtypes, test_util.rand_small),
       # TODO(b/142975473): on CPU, expm1 for float64 is only accurate to ~float32
       # precision.
       op_record(
@@ -262,9 +263,7 @@ def lax_ops():
           float_dtypes,
           test_util.rand_positive,
           {
-              np.float32: (
-                  1e-3 if test_util.device_under_test() == "tpu" else 1e-5
-              ),
+              np.float32: 1e-5,
               np.float64: 1e-14,
           },
       ),
