@@ -12,14 +12,15 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 """Module for conditional control flow primitives."""
+
 import collections
+from collections.abc import Sequence
 import functools
 from functools import partial
 import inspect
 import itertools
 import operator
-
-from typing import Callable, Sequence, List, Tuple
+from typing import Callable
 
 from jax import config
 from jax.tree_util import tree_flatten, tree_unflatten
@@ -54,7 +55,6 @@ from jax._src.lax.control_flow.common import (
     _make_closed_jaxpr,
     _prune_zeros,
     _typecheck_param,
-    allowed_effects,
     )
 
 map, unsafe_map = safe_map, map
@@ -78,6 +78,11 @@ def switch(index, branches: Sequence[Callable], *operands,
       index = clamp(0, index, len(branches) - 1)
       return branches[index](*operands)
 
+  Internally this wraps XLA's `Conditional
+  <https://www.tensorflow.org/xla/operation_semantics#conditional>`_
+  operator. However, when transformed with :func:`~jax.vmap` to operate over a
+  batch of predicates, ``cond`` is converted to :func:`~jax.lax.select`.
+
   Args:
     index: Integer scalar type, indicating which branch function to apply.
     branches: Sequence of functions (A -> B) to be applied based on ``index``.
@@ -138,7 +143,7 @@ def switch(index, branches: Sequence[Callable], *operands,
                           out_trees[0], jaxprs[0].out_avals,
                           out_tree, jaxpr.out_avals)
   joined_effects = core.join_effects(*(jaxpr.effects for jaxpr in jaxprs))
-  disallowed_effects = allowed_effects.filter_not_in(joined_effects)
+  disallowed_effects = effects.control_flow_allowed_effects.filter_not_in(joined_effects)
   if disallowed_effects:
     raise NotImplementedError(
         f'Effects not supported in `switch`: {disallowed_effects}')
@@ -242,12 +247,15 @@ def _cond(pred, true_fun: Callable, false_fun: Callable, *operands,
     raise ValueError("Cannot pass `Ref`s into `cond`.")
   true_jaxpr, false_jaxpr = jaxprs
   out_tree, false_out_tree = out_trees
+  if any(isinstance(out_aval, AbstractRef) for out_aval in
+         true_jaxpr.out_avals + false_jaxpr.out_avals):
+    raise ValueError("Cannot return `Ref`s from `cond`.")
 
   _check_tree_and_avals("true_fun and false_fun output",
                         out_tree, true_jaxpr.out_avals,
                         false_out_tree, false_jaxpr.out_avals)
   joined_effects = core.join_effects(true_jaxpr.effects, false_jaxpr.effects)
-  disallowed_effects = allowed_effects.filter_not_in(joined_effects)
+  disallowed_effects = effects.control_flow_allowed_effects.filter_not_in(joined_effects)
   if disallowed_effects:
     raise NotImplementedError(
         f'Effects not supported in `cond`: {disallowed_effects}')
@@ -276,8 +284,11 @@ def cond(*args, **kwargs):
     pass
   else:
     assert not ba.kwargs  # no catch-all **kwargs in _cond_with_per_branch
-    _, _, maybe_true_fun, _, maybe_false_fun = ba.args
-    if callable(maybe_true_fun) and callable(maybe_false_fun):
+    _, true_operand, true_fun, false_operand, false_fun = ba.args
+    if callable(true_operand) and callable(true_fun):
+      # treat this as modern cond (with two operands)
+      return _cond(*args, **kwargs)
+    if callable(true_fun) and callable(false_fun):
       return _cond_with_per_branch_args(*ba.args)
 
   return _cond(*args, **kwargs)
@@ -316,7 +327,7 @@ def _join_cond_effects(branches: Sequence[core.Jaxpr]) -> effects.Effects:
 
 def _cond_abstract_eval(*avals, branches, **_):
   joined_effects = _join_cond_effects(branches)
-  disallowed_effects = allowed_effects.filter_not_in(joined_effects)
+  disallowed_effects = effects.control_flow_allowed_effects.filter_not_in(joined_effects)
   if disallowed_effects:
     raise NotImplementedError(
         f'Effects not supported in `cond`: {disallowed_effects}')
@@ -506,7 +517,7 @@ def _cond_partial_eval_custom(saveable, unks_in, inst_in, eqn):
 
   # First, compute output unknowns (unks_out), where an output of the cond is
   # unknown if it would be unknown on any of the branches.
-  unks_out: List[bool] = [False] * len(eqn.outvars)
+  unks_out: list[bool] = [False] * len(eqn.outvars)
   for jaxpr in branches:
     _, _, unks_out_, _, _ = pe.partial_eval_jaxpr_custom(
         jaxpr.jaxpr, in_unknowns=ops_uk, in_inst=True,
@@ -515,9 +526,9 @@ def _cond_partial_eval_custom(saveable, unks_in, inst_in, eqn):
 
   # Next, use the computed output unknowns to build a known jaxpr and a staged
   # jaxpr for each branch.
-  branches_known_ : List[core.ClosedJaxpr] = []
-  branches_staged_: List[core.ClosedJaxpr] = []
-  branch_res_avals: List[core.AbstractValue] = []
+  branches_known_ : list[core.ClosedJaxpr] = []
+  branches_staged_: list[core.ClosedJaxpr] = []
+  branch_res_avals: list[core.AbstractValue] = []
   for jaxpr in branches:
     jaxpr_known, jaxpr_staged, _, inst_out, num_res = \
         pe.partial_eval_jaxpr_custom(
@@ -648,13 +659,13 @@ def _ordered_unique(xs):
   d = collections.OrderedDict((x, None) for x in xs)
   return list(d.keys())
 
-def _cond_dce_rule(used_outputs: List[bool], eqn: core.JaxprEqn,
-                   ) -> Tuple[List[bool], core.JaxprEqn]:
+def _cond_dce_rule(used_outputs: list[bool], eqn: core.JaxprEqn,
+                   ) -> tuple[list[bool], core.JaxprEqn]:
   closed_branches = eqn.params['branches']
   branches = [closed_jaxpr.jaxpr for closed_jaxpr in closed_branches]
 
   # First, compute which inputs are used in any branch (not including `pred`).
-  used_inputs: List[bool] = [False] * (len(eqn.invars) - 1)  # -1 for pred
+  used_inputs: list[bool] = [False] * (len(eqn.invars) - 1)  # -1 for pred
   for jaxpr in branches:
     _, used_inputs_ = pe.dce_jaxpr(jaxpr, used_outputs, instantiate=False)
     used_inputs = map(operator.or_, used_inputs, used_inputs_)
@@ -756,7 +767,7 @@ def _cond_typecheck(bind_time, *in_atoms, branches, linear):
   jaxpr0_in_avals_str = _avals_short(jaxpr0.in_avals)
   jaxpr0_out_avals_str = _avals_short(jaxpr0.out_avals)
   joined_effects = _join_cond_effects(branches)
-  disallowed_effects = allowed_effects.filter_not_in(joined_effects)
+  disallowed_effects = effects.control_flow_allowed_effects.filter_not_in(joined_effects)
   if disallowed_effects:
     raise NotImplementedError(
         f'Effects not supported in `cond`: {disallowed_effects}')
@@ -841,10 +852,10 @@ def _cond_lowering(ctx, index, *args, branches, linear):
     with ir.InsertionPoint(branch):
       sub_ctx = ctx.module_context.replace(
           name_stack=name_stack.extend(f'branch_{i}_fun'))
+      consts = [mlir.ir_constants(xla.canonicalize_dtype(x)) for x in jaxpr.consts]
       out_vals, tokens_out = mlir.jaxpr_subcomp(
           sub_ctx, jaxpr.jaxpr, tokens_in,
-          map(mlir.ir_constants, jaxpr.consts),
-          *map(mlir.wrap_singleton_ir_values, args),
+          consts, *map(mlir.wrap_singleton_ir_values, args),
           dim_var_values=ctx.dim_var_values)
       out_tokens = [tokens_out.get(eff) for eff in ordered_effects]
       out_vals = [*out_tokens, *out_vals]
@@ -863,8 +874,8 @@ def _cond_state_discharge_rule(in_avals, out_avals, *args, branches, linear):
       core.ClosedJaxpr(discharge_state(branch.jaxpr, ())[0], ())
       for branch in branches)
   out_vals = cond_p.bind(*args, branches=discharged_branches, linear=linear)
-  out_ref_vals, out_vals = util.split_list(
-      out_vals, [len(out_vals) - len(out_avals)])
+  out_vals, out_ref_vals = util.split_list(
+      out_vals, [len(out_avals)])
   ref_val_iter = iter(out_ref_vals)
   new_invals = []
   for aval in in_avals:
