@@ -12,10 +12,11 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 """Module for the `for_loop` primitive."""
+
+from collections.abc import Sequence
 import functools
 import operator
-
-from typing import Any, Callable, Generic, List, Optional, Sequence, Set, Tuple, TypeVar, Union
+from typing import Any, Callable, Generic, Optional, TypeVar, Union
 
 import jax.numpy as jnp
 from jax import lax
@@ -33,7 +34,12 @@ from jax._src import dispatch
 from jax._src import dtypes
 from jax._src import linear_util as lu
 from jax._src import source_info_util
-from jax._src import state
+from jax._src.state.types import (ReadEffect, AbstractRef, StateEffect)
+from jax._src.state import discharge as state_discharge
+from jax._src.state import primitives as state_primitives
+from jax._src.state import utils as state_utils
+from jax._src.state import types as state_types
+from jax._src.typing import Array
 from jax._src.util import (partition_list, merge_lists, safe_map, safe_zip,
                            split_list, split_dict)
 from jax._src.lax.control_flow import loops
@@ -48,17 +54,11 @@ zip, unsafe_zip = safe_zip, zip
 S = TypeVar('S')
 T = TypeVar('T')
 class Ref(Generic[T]): pass
-Array = Any
 
-ReadEffect = state.ReadEffect
-WriteEffect = state.WriteEffect
-AccumEffect = state.AccumEffect
-StateEffect = state.StateEffect
-AbstractRef = state.AbstractRef
-ref_set = state.ref_set
-ref_get = state.ref_get
-ref_addupdate = state.ref_addupdate
-discharge_state = state.discharge_state
+ref_set = state_primitives.ref_set
+ref_get = state_primitives.ref_get
+ref_addupdate = state_primitives.ref_addupdate
+discharge_state = state_discharge.discharge_state
 
 
 ## `for_loop` implementation
@@ -75,7 +75,7 @@ def _hoist_consts_to_refs(jaxpr: core.Jaxpr) -> core.Jaxpr:
   const_avals, const_ref_avals = partition_list(is_const_ref, all_const_avals)
   const_avals = map(AbstractRef, const_avals)
   merged_const_avals = merge_lists(is_const_ref, const_avals, const_ref_avals)
-  i_aval, *arg_avals = [var.aval for var in jaxpr.invars]
+  i_aval, *arg_avals = (var.aval for var in jaxpr.invars)
   in_avals = [i_aval, *merged_const_avals, *arg_avals]
   num_consts = len(merged_const_avals)
 
@@ -93,19 +93,13 @@ def _hoist_consts_to_refs(jaxpr: core.Jaxpr) -> core.Jaxpr:
 
 def _trace_to_jaxpr_with_refs(f, state_tree: PyTreeDef,
                               state_avals: Sequence[core.AbstractValue]
-                              ) -> Tuple[core.Jaxpr, List[Any], PyTreeDef]:
+                              ) -> tuple[core.Jaxpr, list[Any], PyTreeDef]:
   f, out_tree_thunk = flatten_fun_nokwargs(
       lu.wrap_init(f), treedef_tuple((tree_structure(0), state_tree)))
   jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(
       f, state_avals)
   return jaxpr, consts, out_tree_thunk()
 
-def val_to_ref_aval(x) -> AbstractRef:
-  aval = core.raise_to_shaped(core.get_aval(x))
-  if type(aval) is not core.ShapedArray:
-    raise Exception(f"can't make ref from {x}")
-  return AbstractRef(aval)
-
 def for_loop(nsteps: Union[int, Sequence[int]],
              body: Callable[[Array, Ref[S]], None], init_state: S,
              *, reverse: bool = False, unroll: int = 1) -> S:
@@ -151,14 +145,14 @@ def for_loop(nsteps: Union[int, Sequence[int]],
   if len(nsteps) > 1:
     outer_step, *rest_steps = nsteps
     def wrapped_body(i, refs):
-      vals = tree_map(lambda ref: state.ref_get(ref, ()), refs)
+      vals = tree_map(lambda ref: ref_get(ref, ()), refs)
       vals = for_loop(
           rest_steps, functools.partial(body, i), vals, unroll=unroll)
-      tree_map(lambda ref, val: state.ref_set(ref, (), val), refs, vals)
+      tree_map(lambda ref, val: ref_set(ref, (), val), refs, vals)
     return for_loop(outer_step, wrapped_body, init_state, unroll=unroll)
   nsteps, = nsteps
   flat_state, state_tree = tree_flatten(init_state)
-  state_avals = map(val_to_ref_aval, flat_state)
+  state_avals = map(state_utils.val_to_ref_aval, flat_state)
   idx_aval = core.ShapedArray((), jnp.dtype("int32"))
   jaxpr, consts, out_tree = _trace_to_jaxpr_with_refs(
       body, state_tree, [idx_aval, *state_avals])
@@ -179,12 +173,12 @@ Carry = TypeVar('Carry')
 X = TypeVar('X')
 Y = TypeVar('Y')
 
-def scan(f: Callable[[Carry, X], Tuple[Carry, Y]],
+def scan(f: Callable[[Carry, X], tuple[Carry, Y]],
          init: Carry,
          xs: X,
          length: Optional[int] = None,
          reverse: bool = False,
-         unroll: int = 1) -> Tuple[Carry, Y]:
+         unroll: int = 1) -> tuple[Carry, Y]:
   if not callable(f):
     raise TypeError("scan: f argument should be a callable.")
   if unroll < 1:
@@ -247,20 +241,20 @@ def scan(f: Callable[[Carry, X], Tuple[Carry, Y]],
 @for_p.def_effectful_abstract_eval
 def _for_abstract_eval(*avals, jaxpr, **__):
   # Find out for each of the `Ref`s in our jaxpr what effects they have.
-  jaxpr_aval_effects = state.get_ref_state_effects(
+  jaxpr_aval_effects = state_types.get_ref_state_effects(
       [v.aval for v in jaxpr.invars], jaxpr.effects)[1:]
-  aval_effects = [set(eff.replace(input_index=eff.input_index - 1)
-                      for eff in effs) for aval, effs
+  aval_effects = [{eff.replace(input_index=eff.input_index - 1)
+                      for eff in effs} for aval, effs
                   in zip(avals, jaxpr_aval_effects)
                   if isinstance(aval, AbstractRef)]
   nonlocal_state_effects = core.join_effects(*aval_effects)
   return list(avals), nonlocal_state_effects
 
-@state.register_discharge_rule(for_p)
+@state_discharge.register_discharge_rule(for_p)
 def _for_discharge_rule(in_avals, _, *args: Any, jaxpr: core.Jaxpr,
                         reverse: bool, which_linear: Sequence[bool],
                         nsteps: int, unroll: int
-                        ) -> Tuple[Sequence[Optional[Any]], Sequence[Any]]:
+                        ) -> tuple[Sequence[Optional[Any]], Sequence[Any]]:
   out_vals = for_p.bind(*args, jaxpr=jaxpr, reverse=reverse,
                         which_linear=which_linear, nsteps=nsteps,
                         unroll=unroll)
@@ -378,7 +372,7 @@ def _partial_eval_jaxpr_custom(jaxpr, in_unknowns, policy):
 
 _save_everything = lambda *_, **__: True
 
-def _is_read_only(ref_effects: Set[StateEffect]) -> bool:
+def _is_read_only(ref_effects: set[StateEffect]) -> bool:
   assert len(ref_effects) > 0
   if len(ref_effects) > 1:
     # Means we must have a write or accum effect so not read-only
@@ -386,9 +380,9 @@ def _is_read_only(ref_effects: Set[StateEffect]) -> bool:
   eff, = ref_effects
   return isinstance(eff, ReadEffect)
 
-def _loop_invariant_outputs(jaxpr: core.Jaxpr) -> List[bool]:
+def _loop_invariant_outputs(jaxpr: core.Jaxpr) -> list[bool]:
   # Get effects for each of the jaxpr inputs and remove the loop index.
-  ref_effects = state.get_ref_state_effects(
+  ref_effects = state_types.get_ref_state_effects(
       [v.aval for v in jaxpr.invars], jaxpr.effects)[1:]
   # We first assume that *read-only `Ref`s* are loop-invariant. We can safely do
   # this because the only way something can be loop-varying is if we write to it
@@ -413,8 +407,8 @@ def _loop_invariant_outputs(jaxpr: core.Jaxpr) -> List[bool]:
 
 def _for_partial_eval(trace: pe.JaxprTrace, *tracers: pe.JaxprTracer,
                       jaxpr: core.Jaxpr, nsteps: int, reverse: bool,
-                      which_linear: Tuple[bool, ...],
-                      unroll: int) -> List[pe.JaxprTracer]:
+                      which_linear: tuple[bool, ...],
+                      unroll: int) -> list[pe.JaxprTracer]:
   num_inputs = len(tracers)
   assert num_inputs == len(jaxpr.invars) - 1
   in_unknowns = [not t.pval.is_known() for t in tracers]
@@ -643,7 +637,7 @@ pe.partial_eval_jaxpr_custom_rules[for_p] = _for_partial_eval_custom
 
 def _convert_outputs_to_writes(
     nsteps: int, jaxpr: core.Jaxpr, loop_invar_res: Sequence[bool]
-    ) -> Tuple[core.Jaxpr, List[core.ShapedArray]]:
+    ) -> tuple[core.Jaxpr, list[core.ShapedArray]]:
   assert not jaxpr.constvars, "Jaxpr shouldn't have constvars."
 
   in_avals = [v.aval for v in jaxpr.invars]  # [i, *orig_ref_avals]
@@ -661,7 +655,7 @@ def _convert_outputs_to_writes(
         res_ref[i] = res_val
     return []
   # TODO(mattjj, sharadmv): better handling of tokens, which don't have shape/dtype
-  res_ref_avals: List[core.AbstractValue] = [
+  res_ref_avals: list[core.AbstractValue] = [
       AbstractRef(v.aval) if loop_invar else  # pytype: disable=attribute-error
       AbstractRef(core.ShapedArray((nsteps, *v.aval.shape),  # pytype: disable=attribute-error
                   v.aval.dtype))  # pytype: disable=attribute-error
@@ -686,7 +680,7 @@ def _convert_inputs_to_reads(
 
   res_val_avals, (i_aval,), orig_ref_avals = \
       split_list([v.aval for v in jaxpr.invars], [num_res, 1])
-  res_ref_avals: List[core.AbstractValue] = [
+  res_ref_avals: list[core.AbstractValue] = [
       AbstractRef(aval) if loop_invar else  # pytype: disable=attribute-error
       AbstractRef(core.ShapedArray((nsteps, *aval.shape),  # pytype: disable=attribute-error
                   aval.dtype))  # pytype: disable=attribute-error
@@ -696,7 +690,7 @@ def _convert_inputs_to_reads(
       eval_jaxpr, [i_aval, *res_ref_avals, *orig_ref_avals])
   return jaxpr
 
-def transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: List[bool]) -> core.Jaxpr:
+def transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: list[bool]) -> core.Jaxpr:
   def trans(i, *args):
     # First we want to run the computation to read all the residual refs. We can
     # do that by using partial evaluation with all linear inputs unknown.
@@ -725,7 +719,7 @@ def transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: List[bool]) -> core.Jaxpr:
 
 def _for_transpose(in_cts, *args, jaxpr, nsteps, reverse, which_linear, unroll):
   # if any in_ct is nonzero, we definitely want it in args_ (and the
-  # corresponding x in args could be an undefined primal, but doesnt have to be)
+  # corresponding x in args could be an undefined primal, but doesn't have to be)
   # for non-res stuff:
   #   getting and setting => (nonzero ct, UndefinedPrimal arg)
   #   just setting =>        (nonzero ct, not UndefinedPrimal, dummy value)
@@ -771,7 +765,7 @@ def discharged_for_loop(nsteps, body, init_state, *, reverse: bool = False):
   Potentially useful for testing and benchmarking.
   """
   flat_state, state_tree = tree_flatten(init_state)
-  state_avals = map(val_to_ref_aval, flat_state)
+  state_avals = map(state_utils.val_to_ref_aval, flat_state)
   idx_aval = core.ShapedArray((), jnp.dtype("int32"))
   jaxpr, consts, out_tree = _trace_to_jaxpr_with_refs(
       body, state_tree, [idx_aval, *state_avals])
