@@ -16,7 +16,7 @@ import inspect
 import functools
 from functools import partial
 import math
-from typing import cast, Any, Callable, List, Literal, Optional, Tuple, TypeVar, Union, overload
+from typing import cast, Any, Callable, Literal, Optional, TypeVar, Union, overload
 import warnings
 
 import numpy as np
@@ -137,7 +137,7 @@ def cholesky(x: Array, *, symmetrize_input: bool = True) -> Array:
 
 @_warn_on_positional_kwargs
 def eig(x: ArrayLike, *, compute_left_eigenvectors: bool = True,
-        compute_right_eigenvectors: bool = True) -> List[Array]:
+        compute_right_eigenvectors: bool = True) -> list[Array]:
   """Eigendecomposition of a general matrix.
 
   Nonsymmetric eigendecomposition is at present only implemented on CPU.
@@ -147,7 +147,7 @@ def eig(x: ArrayLike, *, compute_left_eigenvectors: bool = True,
 
 @_warn_on_positional_kwargs
 def eigh(x: Array, *, lower: bool = True, symmetrize_input: bool = True,
-         sort_eigenvalues: bool = True) -> Tuple[Array, Array]:
+         sort_eigenvalues: bool = True) -> tuple[Array, Array]:
   r"""Eigendecomposition of a Hermitian matrix.
 
   Computes the eigenvectors and eigenvalues of a complex Hermitian or real
@@ -200,7 +200,7 @@ def lu_pivots_to_permutation(pivots: ArrayLike, permutation_size: int) -> Array:
   return permutation
 
 
-def lu(x: ArrayLike) -> Tuple[Array, Array, Array]:
+def lu(x: ArrayLike) -> tuple[Array, Array, Array]:
   """LU decomposition with partial pivoting.
 
   Computes the matrix decomposition:
@@ -234,7 +234,7 @@ def lu(x: ArrayLike) -> Tuple[Array, Array, Array]:
   return lu, pivots, permutation
 
 @_warn_on_positional_kwargs
-def qr(x: ArrayLike, *, full_matrices: bool = True) -> Tuple[Array, Array]:
+def qr(x: ArrayLike, *, full_matrices: bool = True) -> tuple[Array, Array]:
   """QR decomposition.
 
   Computes the QR decomposition
@@ -265,17 +265,17 @@ def qr(x: ArrayLike, *, full_matrices: bool = True) -> Tuple[Array, Array]:
   return q, r
 
 @overload
-def svd(x: ArrayLike, *, full_matrices: bool = True, compute_uv: Literal[True]) -> Tuple[Array, Array, Array]: ...
+def svd(x: ArrayLike, *, full_matrices: bool = True, compute_uv: Literal[True]) -> tuple[Array, Array, Array]: ...
 
 @overload
 def svd(x: ArrayLike, *, full_matrices: bool = True, compute_uv: Literal[False]) -> Array: ...
 
 @overload
-def svd(x: ArrayLike, *, full_matrices: bool = True, compute_uv: bool = True) -> Union[Array, Tuple[Array, Array, Array]]: ...
+def svd(x: ArrayLike, *, full_matrices: bool = True, compute_uv: bool = True) -> Union[Array, tuple[Array, Array, Array]]: ...
 
 # TODO: Add `max_qdwh_iterations` to the function signature for TPU SVD.
 @_warn_on_positional_kwargs
-def svd(x: ArrayLike, *, full_matrices: bool = True, compute_uv: bool = True) -> Union[Array, Tuple[Array, Array, Array]]:
+def svd(x: ArrayLike, *, full_matrices: bool = True, compute_uv: bool = True) -> Union[Array, tuple[Array, Array, Array]]:
   """Singular value decomposition.
 
   Returns the singular values if compute_uv is False, otherwise returns a triple
@@ -422,13 +422,21 @@ def _cholesky_lowering(ctx, x):
 
 mlir.register_lowering(cholesky_p, _cholesky_lowering)
 
-def _cholesky_cpu_gpu_lowering(potrf_impl, ctx, operand):
-  if any(not is_constant_shape(a.shape) for a in (ctx.avals_in + ctx.avals_out)):
-    raise NotImplementedError("Shape polymorphism for custom call is not implemented (cholesky); b/261671778")
+def _cholesky_cpu_lowering(ctx, operand):
   operand_aval, = ctx.avals_in
   out_aval, = ctx.avals_out
   batch_dims = operand_aval.shape[:-2]
-  result, info = potrf_impl(operand_aval.dtype, operand, lower=True)
+  if jaxlib_version < (0, 4, 13):
+    if not is_constant_shape(operand_aval.shape):
+      raise NotImplementedError(
+          "Shape polymorphism for native serialization for cholesky on CPU is "
+          f"not implemented; b/261671778; {operand_aval.shape}")
+    result, info = lapack.potrf_hlo(operand_aval.dtype, operand, lower=True)  # type: ignore
+  else:
+    op_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, operand_aval.shape)
+    result, info = lapack.potrf_hlo(operand_aval.dtype, operand, lower=True,
+                                    a_shape_vals=op_shape_vals)
+
   ok = mlir.compare_hlo(
       info, mlir.full_like_aval(ctx, 0, ShapedArray(batch_dims, np.dtype(np.int32))),
       "EQ", "SIGNED")
@@ -442,9 +450,7 @@ def _cholesky_cpu_gpu_lowering(potrf_impl, ctx, operand):
       result, out_aval, _nan_like_hlo(ctx, out_aval), out_aval)]
 
 mlir.register_lowering(
-    cholesky_p,
-    partial(_cholesky_cpu_gpu_lowering, lapack.potrf_hlo),
-    platform='cpu')
+    cholesky_p, _cholesky_cpu_lowering, platform='cpu')
 
 # Asymmetric eigendecomposition
 
@@ -487,15 +493,26 @@ def eig_abstract_eval(operand, *, compute_left_eigenvectors,
 
 def _eig_cpu_lowering(ctx, operand, *, compute_left_eigenvectors,
                       compute_right_eigenvectors):
-  if any(not is_constant_shape(a.shape) for a in (ctx.avals_in + ctx.avals_out)):
-    raise NotImplementedError("Shape polymorphism for custom call is not implemented (eig); b/261671778")
   operand_aval, = ctx.avals_in
   out_aval = ctx.avals_out[0]
   batch_dims = operand_aval.shape[:-2]
-
-  w, vl, vr, info = lapack.geev_hlo(operand_aval.dtype, operand,
-                                    jobvl=compute_left_eigenvectors,
-                                    jobvr=compute_right_eigenvectors)
+  if jaxlib_version < (0, 4, 13):
+    if any(not is_constant_shape(a.shape) for a in ctx.avals_in):
+      raise NotImplementedError(
+          "Shape polymorphism for eig is not implemented. "
+          "Try upgrading jaxlib")
+    w, vl, vr, info = lapack.geev_hlo(operand_aval.dtype, operand,  # type: ignore
+                                      jobvl=compute_left_eigenvectors,
+                                      jobvr=compute_right_eigenvectors)
+  else:
+    if jaxlib_version < (0, 4, 14):
+      op_shape_vals = mlir.eval_dynamic_shape_as_vals(ctx, operand_aval.shape)
+    else:
+      op_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, operand_aval.shape)
+    w, vl, vr, info = lapack.geev_hlo(operand_aval.dtype, operand,
+                                      input_shape_vals=op_shape_vals,
+                                      jobvl=compute_left_eigenvectors,
+                                      jobvr=compute_right_eigenvectors)
 
   ok = mlir.compare_hlo(
       info, mlir.full_like_aval(ctx, 0, ShapedArray(batch_dims, np.dtype(np.int32))),
@@ -572,7 +589,7 @@ ad.primitive_jvps[eig_p] = eig_jvp_rule
 
 
 def eigh_jacobi(x: ArrayLike, *, lower: bool = True,
-                sort_eigenvalues: bool = True) -> Tuple[Array, Array]:
+                sort_eigenvalues: bool = True) -> tuple[Array, Array]:
   """Helper Jacobi eigendecomposition implemented by XLA.
 
   Used as a subroutine of QDWH-eig on TPU."""
@@ -618,11 +635,8 @@ def _eigh_jacobi_lowering_rule(ctx, operand, lower, sort_eigenvalues):
 
   if any(not is_constant_shape(aval_out.shape)
          for aval_out in ctx.avals_out):
-    if jaxlib_version < (0, 4, 9):
-      raise ValueError("shape polymorphism with native lowering for eigh on "
-                       "TPU requires jaxlib version 0.4.9.")
     result_shapes = [
-        mlir.shape_tensor(mlir.eval_dynamic_shape(ctx, aval_out.shape))
+        mlir.eval_dynamic_shape_as_tensor(ctx, aval_out.shape)
         # The custom call returns the results swapped
         for aval_out in list(reversed(ctx.avals_out))
     ]
@@ -630,8 +644,8 @@ def _eigh_jacobi_lowering_rule(ctx, operand, lower, sort_eigenvalues):
     result_shapes = None
   op = mlir.custom_call(
       "Eigh",
-      result_types,
-      [operand],
+      result_types=result_types,
+      operands=[operand],
       backend_config=backend_config,
       api_version=1,
       result_shapes=result_shapes,
@@ -674,31 +688,29 @@ def _eigh_cpu_gpu_lowering(syevd_impl, ctx, operand, *, lower,
 
   batch_dims = operand_aval.shape[:-2]
 
-  if jaxlib_version < (0, 4, 9):
-    if not is_constant_shape(operand_aval.shape):
-      raise NotImplementedError("Shape polymorphism for native lowering for "
-                                "eigh requires "
-                                "jaxlib version 0.4.9; b/261671778")
-    v, w, info = syevd_impl(operand_aval.dtype, operand, lower=lower)
-  else:
-    # The eigh implementation on CPU and GPU uses lapack helper routines to
-    # find the size of the workspace based on the non-batch dimensions.
-    # Therefore, we cannot yet support dynamic non-batch dimensions.
-    if not is_constant_shape(operand_aval.shape[-2:]):
-      raise NotImplementedError(
-          "Shape polymorphism for for native lowering for eigh is implemented "
-          f"only for the batch dimensions: {operand_aval.shape}")
+  # The eigh implementation on CPU and GPU uses lapack helper routines to
+  # find the size of the workspace based on the non-batch dimensions.
+  # Therefore, we cannot yet support dynamic non-batch dimensions.
+  if not is_constant_shape(operand_aval.shape[-2:]):
+    raise NotImplementedError(
+        "Shape polymorphism for for native lowering for eigh is implemented "
+        f"only for the batch dimensions: {operand_aval.shape}")
 
+  if jaxlib_version < (0, 4, 14):
     batch_size_num = math.prod(batch_dims) if batch_dims else 1
     batch_size = mlir.eval_dynamic_shape(ctx, (batch_size_num,))[0]
     if isinstance(batch_size, int):
       batch_size = mlir.ir_constant(np.int32(batch_size))
-    v_shape: ir.Value = mlir.shape_tensor(mlir.eval_dynamic_shape(ctx, v_aval.shape))
-    w_shape: ir.Value = mlir.shape_tensor(mlir.eval_dynamic_shape(ctx, w_aval.shape))
-    info_shape: ir.Value = mlir.shape_tensor(mlir.eval_dynamic_shape(ctx, batch_dims))
+    v_shape: ir.Value = mlir.eval_dynamic_shape_as_tensor(ctx, v_aval.shape)
+    w_shape: ir.Value = mlir.eval_dynamic_shape_as_tensor(ctx, w_aval.shape)
+    info_shape: ir.Value = mlir.eval_dynamic_shape_as_tensor(ctx, batch_dims)
     v, w, info = syevd_impl(operand_aval.dtype, operand, batch_size,
                             v_shape, w_shape, info_shape,
                             lower=lower)
+  else:
+    op_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, operand_aval.shape)
+    v, w, info = syevd_impl(operand_aval.dtype, operand,
+                            a_shape_vals=op_shape_vals, lower=lower)
 
   zeros = mlir.full_like_aval(ctx, 0, ShapedArray(batch_dims, np.dtype(np.int32)))
   ok = mlir.compare_hlo(info, zeros, "EQ", "SIGNED")
@@ -951,16 +963,23 @@ mlir.register_lowering(triangular_solve_p, _triangular_solve_lowering)
 def _triangular_solve_cpu_lower(
     ctx, a, b, *, left_side, lower, transpose_a,
     conjugate_a, unit_diagonal):
-  a_aval, _ = ctx.avals_in
+  a_aval, b_aval = ctx.avals_in
 
   if conjugate_a and not transpose_a:
     a = chlo.ConjOp(a).result
     conjugate_a = False
   if len(a_aval.shape) == 2 and np.dtype(a_aval.dtype) in _cpu_lapack_types:
     alpha = mlir.ir_constant(np.array(1, dtype=a_aval.dtype))
-    return [lapack.trsm_hlo(
-      a_aval.dtype, alpha,
-      a, b, left_side, lower, transpose_a, conjugate_a, unit_diagonal)]
+    if jaxlib_version < (0, 4, 14):
+      return [lapack.trsm_hlo(
+        a_aval.dtype, alpha,
+        a, b, left_side, lower, transpose_a, conjugate_a, unit_diagonal)]  # type: ignore
+    else:
+      b_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, b_aval.shape)
+      return [lapack.trsm_hlo(
+        a_aval.dtype, alpha,
+        a, b, left_side, lower, transpose_a, conjugate_a, unit_diagonal,
+        b_shape_vals=b_shape_vals)]
   else:
     # Fall back to the HLO implementation for unsupported types or batching.
     # TODO: Consider swapping XLA for LAPACK in batched case
@@ -1225,14 +1244,31 @@ def _lu_batching_rule(batched_args, batch_dims):
   x = batching.moveaxis(x, bd, 0)
   return lu_p.bind(x), (0, 0, 0)
 
-def _lu_cpu_gpu_lowering(getrf_impl, ctx, operand):
-  if any(not is_constant_shape(a.shape) for a in (ctx.avals_in + ctx.avals_out)):
-    raise NotImplementedError("Shape polymorphism for custom call is not implemented (lu); b/261671778")
+def _lu_cpu_gpu_lowering(getrf_impl, ctx, operand, *,
+                         platform: str):
   operand_aval, = ctx.avals_in
+  # It should be possible to support fully-dynamic shapes, but since
+  # the last two dimensions (m, n) are used in more involved ways, we only
+  # support dynamic dimensions for the batch size for now.
+  if not is_constant_shape(operand_aval.shape[-2:]):
+    raise NotImplementedError(
+      "Shape polymorphism for native lowering for lu on CPU and GPU is "
+      f"implemented only for the batch dimensions: {operand_aval.shape}")
+
   out_aval, pivot_aval, perm_aval = ctx.avals_out
   batch_dims = operand_aval.shape[:-2]
   m = operand_aval.shape[-2]
-  lu, pivot, info = getrf_impl(operand_aval.dtype, operand)
+  if platform in ["cuda", "rocm"] or jaxlib_version < (0, 4, 14):
+    # TODO(necula): remove the platform kwarg when we implement GPU support.
+    if not is_constant_shape(operand_aval.shape):
+      raise NotImplementedError(
+          "Shape polymorphism for native serialization for lu on GPU is not "
+          f"implemented; b/261671778; {operand_aval.shape}")
+    lu, pivot, info = getrf_impl(operand_aval.dtype, operand)
+  else:
+    op_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, operand_aval.shape)
+    lu, pivot, info = getrf_impl(
+        operand_aval.dtype, operand, a_shape_vals=op_shape_vals)
   # Subtract 1 from the pivot to get 0-based indices.
   pivot = hlo.SubtractOp(pivot, mlir.full_like_aval(ctx, 1, pivot_aval)).result
   ok = mlir.compare_hlo(
@@ -1253,19 +1289,21 @@ def _lu_cpu_gpu_lowering(getrf_impl, ctx, operand):
 
 
 def _lu_tpu_lowering_rule(ctx, operand):
-  if any(not is_constant_shape(a.shape) for a in (ctx.avals_in + ctx.avals_out)):
-    raise NotImplementedError(f"Shape polymorphism for custom call is not implemented (lu); b/261671778; {ctx.avals_in + ctx.avals_out}")
   result_types = [
     mlir.aval_to_ir_type(ctx.avals_out[0]),
     mlir.aval_to_ir_type(ctx.avals_out[1]),
-    mlir.aval_to_ir_type(ctx.avals_out[2])
-  ]
-  op = hlo.CustomCallOp(
-    result_types,
-    [operand],
-    call_target_name=ir.StringAttr.get("LuDecomposition"),
-    has_side_effect=ir.BoolAttr.get(False),
-  )
+    mlir.aval_to_ir_type(ctx.avals_out[2])]
+  if any(not is_constant_shape(a.shape) for a in ctx.avals_out):
+    result_shapes = [
+      mlir.eval_dynamic_shape_as_tensor(ctx, a.shape)
+      for a in ctx.avals_out]
+  else:
+    result_shapes = None
+  op = mlir.custom_call(
+    "LuDecomposition",
+    result_types=result_types,
+    operands=[operand],
+    result_shapes=result_shapes)
   return op.results
 
 
@@ -1278,14 +1316,17 @@ ad.primitive_jvps[lu_p] = _lu_jvp_rule
 batching.primitive_batchers[lu_p] = _lu_batching_rule
 
 mlir.register_lowering(lu_p,
-                        partial(_lu_cpu_gpu_lowering, lapack.getrf_hlo),
+                        partial(_lu_cpu_gpu_lowering, lapack.getrf_hlo,
+                                platform='cpu'),
                         platform='cpu')
 
 mlir.register_lowering(
-    lu_p, partial(_lu_cpu_gpu_lowering, gpu_solver.cuda_getrf),
+    lu_p, partial(_lu_cpu_gpu_lowering, gpu_solver.cuda_getrf,
+                  platform='cuda'),
     platform='cuda')
 mlir.register_lowering(
-    lu_p, partial(_lu_cpu_gpu_lowering, gpu_solver.rocm_getrf),
+    lu_p, partial(_lu_cpu_gpu_lowering, gpu_solver.rocm_getrf,
+                  platform='rocm'),
     platform='rocm')
 
 mlir.register_lowering(lu_p, _lu_tpu_lowering_rule, platform='tpu')
@@ -1353,7 +1394,7 @@ def lu_solve(lu: ArrayLike, permutation: ArrayLike, b: ArrayLike,
 # geqrf and orgqr. The names, while cryptic Fortran alphabet soup, are LAPACK's
 # names for the primitives, and we stick with them for consistency.
 
-def geqrf(a: ArrayLike) -> Tuple[Array, Array]:
+def geqrf(a: ArrayLike) -> tuple[Array, Array]:
   """Computes the QR decomposition of a matrix.
 
   Args:
@@ -1388,35 +1429,53 @@ def _geqrf_lowering_rule(ctx, operand):
   if any(not is_constant_shape(aval_out.shape)
          for aval_out in ctx.avals_out):
     result_shapes = [
-        mlir.shape_tensor(mlir.eval_dynamic_shape(ctx, aval_out.shape))
+        mlir.eval_dynamic_shape_as_tensor(ctx, aval_out.shape)
         for aval_out in ctx.avals_out
     ]
   else:
     result_shapes = None
   op = mlir.custom_call(
       "Qr",
-      result_types,
-      [operand],
+      result_types=result_types,
+      operands=[operand],
       api_version=1,
       result_shapes=result_shapes
   )
   return op.results
 
-def _geqrf_cpu_gpu_lowering(geqrf_impl, batched_geqrf_impl, ctx, a):
-  if any(not is_constant_shape(a.shape) for a in (ctx.avals_in + ctx.avals_out)):
-    raise NotImplementedError("Shape polymorphism for custom call is not implemented (geqrf); b/261671778")
+def _geqrf_cpu_gpu_lowering(geqrf_impl, batched_geqrf_impl, ctx, a, *,
+                            platform: str):
   a_aval, taus_aval = ctx.avals_out
   *batch_dims, m, n = a_aval.shape
+  # It should be possible to support fully-dynamic shapes, but since
+  # the last two dimensions (m, n) are used in more involved ways, we only
+  # support dynamic dimensions for the batch size for now.
+  if not is_constant_shape([m, n]):
+    raise NotImplementedError(
+      "Shape polymorphism for native serialization for qr on CPU and GPU is "
+      f"implemented only for the batch dimensions: {a_aval.shape}")
   batch = math.prod(batch_dims)
 
   if batch == 0 or m == 0 or n == 0:
     return mlir.full_like_aval(ctx, 0, a_aval), mlir.full_like_aval(ctx, 0, taus_aval)
 
+  if not is_constant_shape(a_aval.shape):
+    if platform in ["cuda", "rocm"] or jaxlib_version < (0, 4, 13):
+      # TODO(necula): remove the platform kwarg when we implement GPU support.
+      raise NotImplementedError(
+          "Shape polymorphism for native serialization for QR is not "
+          f"implemented, try to upgrade jaxlib; b/261671778; {a_aval.shape}")
+
   if (batched_geqrf_impl is not None and batch > 1 and m // batch <= 128 and
       n // batch <= 128):
     a_out, taus = batched_geqrf_impl(a_aval.dtype, a)
   else:
-    a_out, taus, info_geqrf = geqrf_impl(a_aval.dtype, a)
+    if platform in ["cuda", "rocm"] or jaxlib_version < (0, 4, 13):
+      a_out, taus, info_geqrf = geqrf_impl(a_aval.dtype, a)  # type: ignore
+    else:
+      a_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, a_aval.shape)
+      a_out, taus, info_geqrf = geqrf_impl(a_aval.dtype, a,
+                                           a_shape_vals=a_shape_vals)
     zeros = mlir.full_like_aval(ctx, 0, ShapedArray(batch_dims, np.dtype(np.int32)))
     ok = mlir.compare_hlo(info_geqrf, zeros, "EQ", "SIGNED")
     select_ok_a_aval = ShapedArray(batch_dims + [1, 1], np.dtype(np.bool_))
@@ -1437,17 +1496,20 @@ batching.primitive_batchers[geqrf_p] = _geqrf_batching_rule
 mlir.register_lowering(geqrf_p, _geqrf_lowering_rule)
 
 mlir.register_lowering(
-    geqrf_p, partial(_geqrf_cpu_gpu_lowering, lapack.geqrf_hlo, None),
+    geqrf_p, partial(_geqrf_cpu_gpu_lowering, lapack.geqrf_hlo, None,
+                     platform='cpu'),
     platform='cpu')
 mlir.register_lowering(
     geqrf_p,
     partial(_geqrf_cpu_gpu_lowering, gpu_solver.cuda_geqrf,
-            gpu_solver.cuda_geqrf_batched),
+            gpu_solver.cuda_geqrf_batched,
+            platform='cuda'),
     platform='cuda')
 mlir.register_lowering(
     geqrf_p,
     partial(_geqrf_cpu_gpu_lowering, gpu_solver.rocm_geqrf,
-            gpu_solver.rocm_geqrf_batched),
+            gpu_solver.rocm_geqrf_batched,
+            platform='rocm'),
     platform='rocm')
 
 
@@ -1494,27 +1556,42 @@ def _householder_product_lowering_rule(ctx, a, taus):
   aval_out, = ctx.avals_out
   if not is_constant_shape(aval_out.shape):
     result_shapes = [
-        mlir.shape_tensor(mlir.eval_dynamic_shape(ctx, aval_out.shape))]
+        mlir.eval_dynamic_shape_as_tensor(ctx, aval_out.shape)]
   else:
     result_shapes = None
   op = mlir.custom_call(
       "ProductOfElementaryHouseholderReflectors",
-      [mlir.aval_to_ir_type(aval_out)],
-      [a, taus],
+      result_types=[mlir.aval_to_ir_type(aval_out)],
+      operands=[a, taus],
       api_version=1,
       result_shapes=result_shapes)
   return [op.result]
 
-def _householder_product_cpu_gpu_lowering(orgqr_impl, ctx, a, taus):
-  if any(not is_constant_shape(a.shape) for a in (ctx.avals_in + ctx.avals_out)):
-    raise NotImplementedError("Shape polymorphism for custom call is not implemented (householder product); b/261671778")
-  a_aval, _ = ctx.avals_in
+def _householder_product_cpu_gpu_lowering(orgqr_impl, ctx, a, taus, *,
+                                          platform: str):
+  a_aval, taus_aval = ctx.avals_in
   *batch_dims, m, n = a_aval.shape
+  if not is_constant_shape([m, n]):
+    raise NotImplementedError(
+      "Shape polymorphism for native serialization for householder_product on "
+      f"CPU and GPU is implemented only for the batch dimensions: {a_aval.shape}")
 
   if m == 0 or n == 0:
     return [mlir.full_like_aval(ctx, 0, a_aval)]
 
-  a, info_orgqr = orgqr_impl(a_aval.dtype, a, taus)
+  if platform in ["rocm", "cuda"] or jaxlib_version < (0, 4, 13):
+    # TODO(necula): remove the platform kwarg when we implement GPU support.
+    if not is_constant_shape(a_aval.shape):
+      raise NotImplementedError(
+          "Shape polymorphism for native serialization for householder_product "
+          f"on GPU is not implemented; b/261671778; {a_aval.shape}")
+    a, info_orgqr = orgqr_impl(a_aval.dtype, a, taus)  # type: ignore
+  else:
+    a_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, a_aval.shape)
+    tau_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, taus_aval.shape)
+    a, info_orgqr = orgqr_impl(a_aval.dtype, a, taus,
+                               a_shape_vals=a_shape_vals,
+                               tau_shape_vals=tau_shape_vals)
   zeros = mlir.full_like_aval(ctx, 0, ShapedArray(batch_dims, np.dtype(np.int32)))
   ok = mlir.compare_hlo(info_orgqr, zeros, "EQ", "SIGNED")
   select_a_aval = ShapedArray(batch_dims + [1, 1], np.dtype(np.bool_))
@@ -1532,15 +1609,18 @@ mlir.register_lowering(householder_product_p, _householder_product_lowering_rule
 
 mlir.register_lowering(
     householder_product_p,
-    partial(_householder_product_cpu_gpu_lowering, lapack.orgqr_hlo),
+    partial(_householder_product_cpu_gpu_lowering, lapack.orgqr_hlo,
+            platform='cpu'),
     platform='cpu')
 mlir.register_lowering(
     householder_product_p,
-    partial(_householder_product_cpu_gpu_lowering, gpu_solver.cuda_orgqr),
+    partial(_householder_product_cpu_gpu_lowering, gpu_solver.cuda_orgqr,
+            platform='cuda'),
     platform='cuda')
 mlir.register_lowering(
     householder_product_p,
-    partial(_householder_product_cpu_gpu_lowering, gpu_solver.rocm_orgqr),
+    partial(_householder_product_cpu_gpu_lowering, gpu_solver.rocm_orgqr,
+            platform='rocm'),
     platform='rocm')
 
 
@@ -1617,7 +1697,7 @@ qr_p.def_abstract_eval(_qr_abstract_eval)
 ad.primitive_jvps[qr_p] = qr_jvp_rule
 batching.primitive_batchers[qr_p] = _qr_batching_rule
 
-mlir.register_lowering(qr_p, mlir.lower_fun(_qr_lowering));
+mlir.register_lowering(qr_p, mlir.lower_fun(_qr_lowering))
 
 
 # Singular value decomposition
@@ -1705,21 +1785,37 @@ def _empty_svd(a, *, full_matrices, compute_uv):
   return s, u, v
 
 def _svd_cpu_gpu_lowering(gesvd_impl, ctx, operand, *, full_matrices,
-                          compute_uv):
-  if any(not is_constant_shape(a.shape) for a in (ctx.avals_in + ctx.avals_out)):
-    raise NotImplementedError("Shape polymorphism for custom call is not implemented (svd); b/261671778")
+                          compute_uv, platform: str):
   operand_aval, = ctx.avals_in
   s_aval = ctx.avals_out[0]
   m, n = operand_aval.shape[-2:]
+  # Since the last two dimensions (m, n) are used to compute the workspace
+  # size, we support dynamic dimensions only for the batch size for now.
+  if not is_constant_shape([m, n]):
+    raise NotImplementedError(
+      "Shape polymorphism for native serialization for svd on CPU and GPU is "
+      f"implemented only for the batch dimensions: {operand_aval.shape}")
   batch_dims = operand_aval.shape[:-2]
 
   if m == 0 or n == 0:
     return mlir.lower_fun(_empty_svd, multiple_results=True)(
       ctx, operand, full_matrices=full_matrices, compute_uv=compute_uv)
 
-  s, u, vt, info = gesvd_impl(operand_aval.dtype, operand,
-                              full_matrices=full_matrices,
-                              compute_uv=compute_uv)
+  if platform in ["cuda", "rocm"] or jaxlib_version < (0, 4, 13):
+    if not is_constant_shape(operand_aval.shape):
+      # TODO(necula): remove the platform kwarg when we implement GPU support.
+      raise NotImplementedError(
+          "Shape polymorphism for native serialization for SVD is not "
+          f"implemented, try to upgrade jaxlib; b/261671778; {operand_aval.shape}")
+    s, u, vt, info = gesvd_impl(operand_aval.dtype, operand,
+                                full_matrices=full_matrices,
+                                compute_uv=compute_uv)
+  else:
+    a_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, operand_aval.shape)
+    s, u, vt, info = gesvd_impl(operand_aval.dtype, operand,
+                                full_matrices=full_matrices,
+                                compute_uv=compute_uv,
+                                a_shape_vals=a_shape_vals)
   zeros = mlir.full_like_aval(ctx, 0, ShapedArray(batch_dims, np.dtype(np.int32)))
   ok = mlir.compare_hlo(info, zeros, "EQ", "SIGNED")
   select_s_aval = ShapedArray(batch_dims + (1,), np.dtype(np.bool_))
@@ -1795,28 +1891,80 @@ ad.primitive_jvps[svd_p] = _svd_jvp_rule
 batching.primitive_batchers[svd_p] = _svd_batching_rule
 
 mlir.register_lowering(
-    svd_p, partial(_svd_cpu_gpu_lowering, lapack.gesdd_hlo),
+    svd_p, partial(_svd_cpu_gpu_lowering, lapack.gesdd_hlo,
+                   platform='cpu'),
     platform='cpu')
 mlir.register_lowering(
-  svd_p, partial(_svd_cpu_gpu_lowering, gpu_solver.cuda_gesvd),
+  svd_p, partial(_svd_cpu_gpu_lowering, gpu_solver.cuda_gesvd,
+                 platform='cuda'),
   platform='cuda')
 mlir.register_lowering(
-  svd_p, partial(_svd_cpu_gpu_lowering, gpu_solver.rocm_gesvd),
+  svd_p, partial(_svd_cpu_gpu_lowering, gpu_solver.rocm_gesvd,
+                 platform='rocm'),
   platform='rocm')
 
 mlir.register_lowering(svd_p, _svd_tpu_lowering_rule)
 
+
 def _tridiagonal_solve_gpu_lowering(lowering, ctx, dl, d, du, b, *, m, n, ldb, t):
-  return [lowering(dl, d, du, b, m=m, n=n, ldb=ldb,
-                   t=dtypes.canonicalize_dtype(t))]
+  _, _, _, b_aval = ctx.avals_in
+  b_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, b_aval.shape)
+  if jaxlib_version >= (0, 4, 15):
+    return [lowering(
+        dl, d, du, b, m=m, n=n, ldb=ldb, t=dtypes.canonicalize_dtype(t),
+        b_shape_vals=b_shape_vals)]
+  else:
+    return [lowering(
+        dl, d, du, b, m=m, n=n, ldb=ldb, t=dtypes.canonicalize_dtype(t))]
+
+
+def _tridiagonal_solve_transpose_rule(cotangent, dl, d, du, b, *, m, n, ldb, t):
+  del m, n, ldb, t
+  # Tridiagonal solve is nonlinear in the tridiagonal arguments and linear
+  # otherwise.
+  assert not (ad.is_undefined_primal(dl) or ad.is_undefined_primal(d) or
+              ad.is_undefined_primal(du)) and ad.is_undefined_primal(b)
+  if type(cotangent) is ad_util.Zero:
+    cotangent_b = ad_util.Zero(b.aval)
+  else:
+    cotangent_b = tridiagonal_solve(dl, d, du, cotangent)
+  return [None, None, None, cotangent_b]
+
+
+def _tridiagonal_solve_batching_rule(
+    batched_args, batch_dims, *, m, n, ldb, t):
+  del m, n, ldb, t
+  dl, d, du, b = batched_args
+  bdl, bd, bdu, bb = batch_dims
+  if (bdl is batching.not_mapped and
+      bd is batching.not_mapped and
+      bdu is batching.not_mapped):
+
+    b = batching.moveaxis(b, bb, -2)
+    b_flat = b.reshape(b.shape[:-3]  + (b.shape[-3], b.shape[-2] * b.shape[-1]))
+    bdim_out = b.ndim - 2
+    out_flat = tridiagonal_solve(dl, d, du, b_flat)
+    return out_flat.reshape(b.shape), bdim_out
+  else:
+    size = next(t.shape[i] for t, i in zip(batched_args, batch_dims)
+                if i is not None)
+    dl = batching.bdim_at_front(dl, bdl, size)
+    d = batching.bdim_at_front(d, bd, size)
+    du = batching.bdim_at_front(du, bdu, size)
+    b = batching.bdim_at_front(b, bb, size)
+    return tridiagonal_solve(dl, d, du, b), 0
+
 
 tridiagonal_solve_p = Primitive('tridiagonal_solve')
 tridiagonal_solve_p.multiple_results = False
 tridiagonal_solve_p.def_impl(
     functools.partial(dispatch.apply_primitive, tridiagonal_solve_p))
 tridiagonal_solve_p.def_abstract_eval(lambda dl, d, du, b, *, m, n, ldb, t: b)
+ad.primitive_transposes[tridiagonal_solve_p] = _tridiagonal_solve_transpose_rule
+batching.primitive_batchers[tridiagonal_solve_p] = _tridiagonal_solve_batching_rule
 # TODO(tomhennigan): Consider AD rules using lax.custom_linear_solve?
 
+
 mlir.register_lowering(
     tridiagonal_solve_p,
     partial(_tridiagonal_solve_gpu_lowering, gpu_sparse.cuda_gtsv2),
@@ -1829,12 +1977,26 @@ mlir.register_lowering(
 
 def _tridiagonal_solve_jax(dl, d, du, b, **kw):
   """Pure JAX implementation of `tridiagonal_solve`."""
-  prepend_zero = lambda x: jnp.append(jnp.zeros([1], dtype=x.dtype), x[:-1])
+  def prepend_zero(x):
+    return jnp.append(
+        jnp.zeros((1,) + x.shape[1:], dtype=x.dtype),
+        x[:-1], axis=0)
   fwd1 = lambda tu_, x: x[1] / (x[0] - x[2] * tu_)
-  fwd2 = lambda b_, x: (x[0] - x[3] * b_) / (x[1] - x[3] * x[2])
-  bwd1 = lambda x_, x: x[0] - x[1] * x_
+
+  def fwd2(b_, x):
+    return (x[0] - x[3][jnp.newaxis, ...] * b_) / (
+        x[1] - x[3] * x[2])[jnp.newaxis, ...]
+
+  bwd1 = lambda x_, x: x[0] - x[1][jnp.newaxis, ...] * x_
   double = lambda f, args: (f(*args), f(*args))
 
+  # Move relevant dimensions to the front for the scan.
+  dl = jnp.moveaxis(dl, -1, 0)
+  d = jnp.moveaxis(d, -1, 0)
+  du = jnp.moveaxis(du, -1, 0)
+  b = jnp.moveaxis(b, -1, 0)
+  b = jnp.moveaxis(b, -1, 0)
+
   # Forward pass.
   _, tu_ = lax.scan(lambda tu_, x: double(fwd1, (tu_, x)),
                     du[0] / d[0],
@@ -1842,7 +2004,7 @@ def _tridiagonal_solve_jax(dl, d, du, b, **kw):
                     unroll=32)
 
   _, b_ = lax.scan(lambda b_, x: double(fwd2, (b_, x)),
-                   b[0] / d[0],
+                   b[0] / d[0:1],
                    (b, d, prepend_zero(tu_), dl),
                    unroll=32)
 
@@ -1852,7 +2014,10 @@ def _tridiagonal_solve_jax(dl, d, du, b, **kw):
                    (b_[::-1], tu_[::-1]),
                    unroll=32)
 
-  return x_[::-1]
+  result = x_[::-1]
+  result = jnp.moveaxis(result, 0, -1)
+  result = jnp.moveaxis(result, 0, -1)
+  return result
 
 
 mlir.register_lowering(tridiagonal_solve_p, mlir.lower_fun(
@@ -1868,31 +2033,30 @@ def tridiagonal_solve(dl: Array, d: Array, du: Array, b: Array) -> Array:
     A . X = B
 
   Args:
-    dl: The lower diagonal of A: ``dl[i] := A[i, i-1]`` for i in ``[0,m)``.
+
+    dl: A batch of vectors with shape ``[..., m]``.
+      The lower diagonal of A: ``dl[i] := A[i, i-1]`` for i in ``[0,m)``.
       Note that ``dl[0] = 0``.
-    d: The middle diagnoal of A: ``d[i]  := A[i, i]`` for i in ``[0,m)``.
-    du: The upper diagonal of A: ``du[i] := A[i, i+1]`` for i in ``[0,m)``.
+    d: A batch of vectors with shape ``[..., m]``.
+      The middle diagonal of A: ``d[i]  := A[i, i]`` for i in ``[0,m)``.
+    du: A batch of vectors with shape ``[..., m]``.
+      The upper diagonal of A: ``du[i] := A[i, i+1]`` for i in ``[0,m)``.
       Note that ``dl[m - 1] = 0``.
     b: Right hand side matrix.
 
   Returns:
     Solution ``X`` of tridiagonal system.
   """
-  if dl.ndim != 1 or d.ndim != 1 or du.ndim != 1:
-    raise ValueError('dl, d and du must be vectors')
-
   if dl.shape != d.shape or d.shape != du.shape:
     raise ValueError(
         f'dl={dl.shape}, d={d.shape} and du={du.shape} must all be `[m]`')
 
-  if b.ndim != 2:
-    raise ValueError(f'b={b.shape} must be a matrix')
-
-  m, = dl.shape
+  m = dl.shape[-1]
   if m < 3:
     raise ValueError(f'm ({m}) must be >= 3')
 
-  ldb, n = b.shape
+  ldb = b.shape[-2]
+  n = b.shape[-1]
   if ldb < max(1, m):
     raise ValueError(f'Leading dimension of b={ldb} must be ≥ max(1, {m})')
 
@@ -1914,7 +2078,7 @@ def tridiagonal_solve(dl: Array, d: Array, du: Array, b: Array) -> Array:
 def schur(x: ArrayLike, *,
           compute_schur_vectors: bool = True,
           sort_eig_vals: bool = False,
-          select_callable: Optional[Callable[..., Any]] = None) -> Tuple[Array, Array]:
+          select_callable: Optional[Callable[..., Any]] = None) -> tuple[Array, Array]:
   return schur_p.bind(
       x,
       compute_schur_vectors=compute_schur_vectors,
@@ -1956,10 +2120,18 @@ def _schur_cpu_lowering(ctx, operand, *, compute_schur_vectors, sort_eig_vals,
   operand_aval, = ctx.avals_in
   batch_dims = operand_aval.shape[:-2]
 
-  gees_result = lapack.gees_hlo(operand_aval.dtype, operand,
-                                jobvs=compute_schur_vectors,
-                                sort=sort_eig_vals,
-                                select=select_callable)
+  if jaxlib_version < (0, 4, 14):
+    gees_result = lapack.gees_hlo(operand_aval.dtype, operand,
+                                  jobvs=compute_schur_vectors,
+                                  sort=sort_eig_vals,
+                                  select=select_callable)  # type: ignore
+  else:
+    a_shape_vals = mlir.eval_dynamic_shape_as_ivals(ctx, operand_aval.shape)
+    gees_result = lapack.gees_hlo(operand_aval.dtype, operand,
+                                  jobvs=compute_schur_vectors,
+                                  sort=sort_eig_vals,
+                                  select=select_callable,
+                                  a_shape_vals=a_shape_vals)
 
   # Number of return values depends on value of sort_eig_vals.
   T, vs, *_, info = gees_result
@@ -2003,7 +2175,7 @@ def _schur_batching_rule(batched_args, batch_dims, *, compute_schur_vectors,
       select_callable=select_callable), (0,) * (1 + compute_schur_vectors)
 
 
-def _schur_jvp_rule(primals, tangents, *, compute_schur_vectors, sort_eig_vals):
+def _schur_jvp_rule(primals, tangents, **kwds):
   raise NotImplementedError(
       'The differentiation rules for the Schur factorization have not been implemented.'
   )
@@ -2021,7 +2193,7 @@ ad.primitive_jvps[schur_p] = _schur_jvp_rule
 
 # hessenberg: Upper Hessenberg reduction
 
-def hessenberg(a: ArrayLike) -> Tuple[Array, Array]:
+def hessenberg(a: ArrayLike) -> tuple[Array, Array]:
   """Reduces a square matrix to upper Hessenberg form.
 
   Currently implemented on CPU only.
@@ -2093,7 +2265,7 @@ mlir.register_lowering(hessenberg_p, _hessenberg_cpu_hlo, platform='cpu')
 # tridiagonal: Upper Hessenberg reduction
 
 def tridiagonal(a: ArrayLike, *, lower=True
-               ) -> Tuple[Array, Array, Array, Array]:
+               ) -> tuple[Array, Array, Array, Array]:
   """Reduces a symmetric/Hermitian matrix to tridiagonal form.
 
   Currently implemented on CPU and GPU only.
