@@ -12,43 +12,56 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import contextlib
-import numpy as np
-import itertools as it
 from collections import OrderedDict, abc
-from typing import (Callable, Iterable, Tuple, Optional, Dict, Any, Set,
-                    NamedTuple, Union, Sequence)
+from collections.abc import Iterable, Sequence, Mapping
+import contextlib
 from functools import wraps, partial, partialmethod, lru_cache
+import itertools as it
+import math
+from typing import (Callable, Optional, Any,
+                    NamedTuple, Union)
 
+import numpy as np
+
+from jax import lax
 from jax import numpy as jnp
+
 from jax._src import core
-from jax._src import mesh
-from jax._src import linear_util as lu
-from jax import stages
 from jax._src import dispatch
 from jax._src import effects
-from jax.tree_util import (tree_flatten, tree_unflatten, all_leaves, tree_map,
-                           treedef_tuple)
+from jax._src import mesh as mesh_lib
+from jax._src import linear_util as lu
+from jax._src import op_shardings
+from jax._src import sharding_impls
+from jax._src import source_info_util
+from jax._src import stages
+from jax._src import traceback_util
 from jax._src.api_util import (flatten_fun_nokwargs, flatten_axes,
                                _ensure_index_tuple, donation_vector,
                                shaped_abstractify, check_callable)
-from jax._src import source_info_util
-from jax._src import traceback_util
-from jax._src.config import config
-from jax.errors import JAXTypeError
 from jax._src.array import ArrayImpl
-from jax._src.sharding_impls import NamedSharding
+from jax._src.config import config
+from jax._src.errors import JAXTypeError
+from jax._src.interpreters import ad
+from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
-from jax._src.interpreters import partial_eval as pe
-from jax._src.interpreters import pxla
 from jax._src.interpreters import xla
-from jax.interpreters import batching
-from jax.interpreters import ad
+from jax._src.interpreters import partial_eval as pe
+from jax._src.interpreters.partial_eval import (
+  trace_to_subjaxpr_dynamic, DynamicJaxprTracer,
+  convert_constvars_jaxpr, new_jaxpr_eqn)
+from jax._src.interpreters import pxla
+from jax._src.pjit import (sharding_constraint_p, get_unconstrained_dims,
+                           GSPMDSharding)
+from jax._src.sharding_impls import (
+    ArrayMapping, NamedSharding, ParsedPartitionSpec,
+    array_mapping_to_axis_resources)
+from jax._src.tree_util import (tree_flatten, tree_unflatten, all_leaves,
+                                tree_map, treedef_tuple)
 from jax._src.util import (safe_map, safe_zip, HashableFunction, unzip2, unzip3,
                            as_hashable_function, distributed_debug_log,
                            tuple_insert, moveaxis, split_list, wrap_name,
                            merge_lists, partition_list)
-from jax import lax
 
 source_info_util.register_exclusion(__file__)
 traceback_util.register_exclusion(__file__)
@@ -82,12 +95,12 @@ class FrozenDict(abc.Mapping):
 # Multi-dimensional generalized map
 
 AxisName = core.AxisName
-ResourceAxisName = mesh.ResourceAxisName  # Different name just for documentation purposes
-Mesh = mesh.Mesh
-MeshAxisName = mesh.MeshAxisName
-ResourceEnv = mesh.ResourceEnv
-EMPTY_ENV = mesh.EMPTY_ENV
-thread_resources = mesh.thread_resources
+ResourceAxisName = mesh_lib.ResourceAxisName  # Different name just for documentation purposes
+Mesh = mesh_lib.Mesh
+MeshAxisName = mesh_lib.MeshAxisName
+ResourceEnv = mesh_lib.ResourceEnv
+EMPTY_ENV = mesh_lib.EMPTY_ENV
+thread_resources = mesh_lib.thread_resources
 
 
 class SerialLoop:
@@ -149,15 +162,19 @@ def serial_loop(name: ResourceAxisName, length: int):
 
   Example::
 
-    with loop('l', 4):
-      out = xmap(
-        lambda x: jnp.sin(x) * 5,  # This will be called 4 times with different
-                                   # slices of x.
-        in_axes=['i'], out_axes=['i'],
-        axis_resources={'i': 'l'})(x)
+    >>> x = jnp.linspace(0, jnp.pi, 4)
+    ...
+    >>> with serial_loop('l', len(x)):
+    ...   out = xmap(
+    ...     lambda x: jnp.sin(x) * 5,  # This will be called 4 times with different
+    ...                                # slices of x.
+    ...     in_axes=['i'], out_axes=['i'],
+    ...     axis_resources={'i': 'l'})(x)
+    >>> out.shape
+    (4,)
   """
   old_env: ResourceEnv = getattr(thread_resources, "env", EMPTY_ENV)
-  thread_resources.env = old_env.with_extra_loop(mesh.Loop(name, length))
+  thread_resources.env = old_env.with_extra_loop(mesh_lib.Loop(name, length))
   try:
     yield
   finally:
@@ -253,7 +270,7 @@ def _prepare_axes(axes, arg_name):
   return tree_unflatten(treedef, entries), entries, treedef
 
 Resource = Union[ResourceAxisName, SerialLoop]
-ResourceSet = Union[Resource, Tuple[Resource, ...]]
+ResourceSet = Union[Resource, tuple[Resource, ...]]
 
 # TODO: Some syntactic sugar to make the API more usable in a single-axis case?
 # TODO: Are the resource axes scoped lexically or dynamically? Dynamically for now!
@@ -261,8 +278,8 @@ def xmap(fun: Callable,
          in_axes,
          out_axes,
          *,
-         axis_sizes: Dict[AxisName, int] = {},
-         axis_resources: Dict[AxisName, ResourceSet] = {},
+         axis_sizes: Optional[Mapping[AxisName, int]] = None,
+         axis_resources: Optional[Mapping[AxisName, ResourceSet]] = None,
          donate_argnums: Union[int, Sequence[int]] = (),
          backend: Optional[str] = None) -> stages.Wrapped:
   """Assign a positional signature to a program that uses named array axes.
@@ -450,6 +467,9 @@ def xmap(fun: Callable,
     out_axes, out_axes_entries, out_axes_treedef = _prepare_axes(out_axes, "out_axes")
     out_axes_entries = tuple(out_axes_entries)  # Make entries hashable
 
+  axis_sizes = {} if axis_sizes is None else axis_sizes
+  axis_resources = {} if axis_resources is None else axis_resources
+
   axis_sizes_names = set(axis_sizes.keys())
   in_axes_names = set(it.chain(*(spec.keys() for spec in in_axes_entries)))
   defined_names = axis_sizes_names | in_axes_names
@@ -473,7 +493,7 @@ def xmap(fun: Callable,
                      f"in_axes or axis_sizes, but the following are missing: "
                      f"{out_axes_names - defined_names}")
 
-  normalized_axis_resources: Dict[AxisName, Tuple[ResourceAxisName, ...]] = {}
+  normalized_axis_resources: dict[AxisName, tuple[ResourceAxisName, ...]] = {}
   for axis in defined_names:
     resources = axis_resources.get(axis, ())
     if not isinstance(resources, tuple):
@@ -506,7 +526,7 @@ def xmap(fun: Callable,
     args_flat, in_tree = tree_flatten(args)
     fun_flat, out_tree = flatten_fun_nokwargs(lu.wrap_init(fun), in_tree)
     if donate_argnums:
-      donated_invars = donation_vector(donate_argnums, args, ())
+      donated_invars = donation_vector(donate_argnums, (), args, {})
     else:
       donated_invars = (False,) * len(args_flat)
     in_axes_flat = _flatten_axes("xmap in_axes", in_tree, in_axes, tupled_args=True)
@@ -584,7 +604,9 @@ def xmap(fun: Callable,
     return verify_outputs(out_flat, out_tree, params)
 
   @decorate_serial
-  def lower(*args, _experimental_lowering_platform: Optional[str] = None):
+  def lower(*args, **kwargs):
+    lowering_parameters = kwargs.pop(
+        '_experimental_lowering_platform', mlir.LoweringParameters())
     fun_flat, args_flat, params, in_tree, out_tree = infer_params(*args)
     avals_flat = [shaped_abstractify(arg) for arg in args_flat]
     computation = make_xmap_callable(
@@ -592,7 +614,7 @@ def xmap(fun: Callable,
         params['donated_invars'], params['global_axis_sizes'], params['axis_resources'],
         params['resource_env'], params['backend'], params['spmd_in_axes'],
         params['spmd_out_axes_thunk'],
-        _experimental_lowering_platform, *avals_flat)
+        lowering_parameters, *avals_flat)
 
     in_tree = treedef_tuple([in_tree, tree_flatten({})[1]])
     in_avals = in_tree.unflatten(avals_flat)
@@ -611,7 +633,7 @@ def xmap_impl(fun: lu.WrappedFun, *args, name, in_axes, out_axes_thunk, donated_
       fun, name, in_axes, out_axes_thunk, donated_invars, global_axis_sizes,
       axis_resources, resource_env, backend,
       spmd_in_axes, spmd_out_axes_thunk,
-      None, *in_avals).compile().unsafe_call
+      mlir.LoweringParameters(), *in_avals).compile().unsafe_call
   distributed_debug_log(("Running xmapped function", name),
                         ("python function", fun.f),
                         ("mesh", resource_env.physical_mesh),
@@ -624,7 +646,7 @@ def make_xmap_callable(fun: lu.WrappedFun,
                        in_axes, out_axes_thunk, donated_invars,
                        global_axis_sizes, axis_resources, resource_env, backend,
                        spmd_in_axes, spmd_out_axes_thunk,
-                       lowering_platform: Optional[str],
+                       lowering_parameters: mlir.LoweringParameters,
                        *in_avals):
   plan = EvaluationPlan.from_axis_resources(
       axis_resources, resource_env, global_axis_sizes)
@@ -634,9 +656,9 @@ def make_xmap_callable(fun: lu.WrappedFun,
   mapped_in_avals = [_delete_aval_axes(aval, in_axes, global_axis_sizes)
                      for aval, in_axes in zip(in_avals, in_axes)]
   with core.extend_axis_env_nd(global_axis_sizes.items()):
-    with dispatch.log_elapsed_time(f"Finished tracing + transforming {fun.__name__} "
-                                   "for xmap in {elapsed_time} sec",
-                                    event=dispatch.JAXPR_TRACE_EVENT):
+    with dispatch.log_elapsed_time(
+        "Finished tracing + transforming {fun_name} for xmap in {elapsed_time} sec",
+        fun_name=fun.__name__, event=dispatch.JAXPR_TRACE_EVENT):
       jaxpr, out_avals, consts = pe.trace_to_jaxpr_final(fun, mapped_in_avals)
   out_axes = out_axes_thunk()
   _check_out_avals_vs_out_axes(out_avals, out_axes, global_axis_sizes)
@@ -669,29 +691,29 @@ def make_xmap_callable(fun: lu.WrappedFun,
       tiling_method = pxla.TileManual(manual_mesh_axes)
     else:
       tiling_method = pxla.TileVectorize()
-    in_shardings = [NamedSharding(mesh, pxla.array_mapping_to_axis_resources(i))
+    in_shardings = [NamedSharding(mesh, array_mapping_to_axis_resources(i))
                     for i in mesh_in_axes]
-    out_shardings = [NamedSharding(mesh, pxla.array_mapping_to_axis_resources(o))
+    out_shardings = [NamedSharding(mesh, array_mapping_to_axis_resources(o))
                      for o in mesh_out_axes]
     return pxla.lower_mesh_computation(
         f, 'xmap', name, mesh,
         in_shardings, out_shardings, donated_invars,
         use_spmd_lowering, in_avals,
         tiling_method=tiling_method,
-        lowering_platform=lowering_platform)
+        lowering_parameters=lowering_parameters)
   else:
     return dispatch.sharded_lowering(
-        f, name, donated_invars, True, *[(a, None) for a in in_avals],
-        lowering_platform=lowering_platform)
+        f, name, donated_invars, True, False, in_avals, (None,) * len(in_avals),
+        lowering_parameters=lowering_parameters)
 
 
 class EvaluationPlan(NamedTuple):
   """Encapsulates preprocessing common to top-level xmap invocations and its translation rule."""
   resource_env: ResourceEnv
-  physical_axis_resources: Dict[AxisName, Tuple[ResourceAxisName, ...]]
-  loop_axis_resources: Dict[AxisName, Tuple[ResourceAxisName, ...]]
-  axis_subst_dict: Dict[AxisName, Tuple[ResourceAxisName, ...]]
-  axis_vmap_size: Dict[AxisName, Optional[int]]
+  physical_axis_resources: dict[AxisName, tuple[ResourceAxisName, ...]]
+  loop_axis_resources: dict[AxisName, tuple[ResourceAxisName, ...]]
+  axis_subst_dict: dict[AxisName, tuple[ResourceAxisName, ...]]
+  axis_vmap_size: dict[AxisName, Optional[int]]
 
   @property
   def axis_subst(self) -> core.AxisSubst:
@@ -709,15 +731,15 @@ class EvaluationPlan(NamedTuple):
 
   @classmethod
   def from_axis_resources(cls,
-                          axis_resources: Dict[AxisName, Tuple[ResourceAxisName, ...]],
+                          axis_resources: dict[AxisName, tuple[ResourceAxisName, ...]],
                           resource_env: ResourceEnv,
-                          global_axis_sizes: Dict[AxisName, int]):
+                          global_axis_sizes: dict[AxisName, int]):
     physical_axis_resources, loop_axis_resources = _unzip_axis_resources(
             axis_resources, resource_env)
     axis_resource_count = _get_axis_resource_count(
         axis_resources, resource_env)
     axis_subst_dict = dict(axis_resources)
-    axis_vmap_size: Dict[AxisName, Optional[int]] = {}
+    axis_vmap_size: dict[AxisName, Optional[int]] = {}
     for naxis, raxes in sorted(axis_resources.items(), key=lambda x: str(x[0])):
       num_resources = axis_resource_count[naxis]
       assert global_axis_sizes[naxis] % num_resources.nglobal == 0
@@ -846,7 +868,7 @@ core.axis_substitution_rules[xmap_p] = _xmap_axis_subst
 # NOTE: We don't have to handle spmd_{in|out}_axes here, because
 # SPMD batching always gets involved as the last transform before XLA translation
 ad.JVPTrace.process_xmap = ad.JVPTrace.process_call  # type: ignore
-ad.call_param_updaters[xmap_p] = xla.xla_call_jvp_update_params
+ad.call_param_updaters[xmap_p] = pxla.xla_call_jvp_update_params
 
 def _xmap_transpose(params, call_jaxpr, args, cts_in, cts_in_avals, reduce_axes):
   all_args, in_tree_def = tree_flatten(((), args, cts_in))  # empty consts
@@ -929,7 +951,7 @@ def _resource_typing_xmap(avals,
     raise JAXTypeError(
         f"Detected disallowed xmap axis name shadowing at "
         f"{source_info_util.summarize(source_info)} "
-        f"(shadowed axes: {mesh.show_axes(overlap)})")
+        f"(shadowed axes: {mesh_lib.show_axes(overlap)})")
 
   if resource_env.physical_mesh != params['resource_env'].physical_mesh:
     raise RuntimeError("Changing the physical mesh is not allowed inside xmap.")
@@ -957,17 +979,14 @@ def _resource_typing_xmap(avals,
         raise JAXTypeError(
             f"One of xmapped function ({params['name']}) outputs is broadcast "
             f"along axis `{baxis}` which is assigned to resources "
-            f"{mesh.show_axes(baxis_resources)}, but the output is already "
-            f"partitioned along {mesh.show_axes(overlap)}, because its "
-            f"named shape contains {mesh.show_axes(partitioning_axes)}")
+            f"{mesh_lib.show_axes(baxis_resources)}, but the output is already "
+            f"partitioned along {mesh_lib.show_axes(overlap)}, because its "
+            f"named shape contains {mesh_lib.show_axes(partitioning_axes)}")
 pxla.custom_resource_typing_rules[xmap_p] = _resource_typing_xmap
 
 
 # This is DynamicJaxprTrace.process_map with some very minor modifications
 def _dynamic_jaxpr_process_xmap(self, primitive, f, tracers, params):
-  from jax._src.interpreters.partial_eval import (
-    trace_to_subjaxpr_dynamic, DynamicJaxprTracer,
-    convert_constvars_jaxpr, new_jaxpr_eqn)
   assert primitive is xmap_p
   in_avals = [t.aval for t in tracers]
   global_axis_sizes = params['global_axis_sizes']
@@ -1021,7 +1040,7 @@ def _xmap_partial_eval_custom_params_updater(
     unks_in: Sequence[bool], inst_in: Sequence[bool],
     kept_outs_known: Sequence[bool], kept_outs_staged: Sequence[bool],
     num_res: int, params_known: dict, params_staged: dict
-  ) -> Tuple[dict, dict]:
+  ) -> tuple[dict, dict]:
   assert params_known['spmd_in_axes'] is None is params_known['spmd_out_axes']
   assert params_staged['spmd_in_axes'] is None is params_staged['spmd_out_axes']
 
@@ -1264,7 +1283,7 @@ batching.BatchTrace.post_process_xmap = _batch_trace_post_process_xmap
 # -------- nested xmap handling --------
 
 def _xmap_lowering_rule(ctx, *args, **kwargs):
-  if isinstance(ctx.module_context.axis_context, mlir.SPMDAxisContext):
+  if isinstance(ctx.module_context.axis_context, sharding_impls.SPMDAxisContext):
     if config.experimental_xmap_spmd_lowering_manual:
       return _xmap_lowering_rule_spmd_manual(ctx, *args, **kwargs)
     else:
@@ -1272,9 +1291,9 @@ def _xmap_lowering_rule(ctx, *args, **kwargs):
   # Here ShardingContext is used in place of ReplicaAxisContext because when
   # axis_resources and mesh is not used with xmap, `make_xmap_callable` will
   # go via `dispatch.sharded_lowering` path which sets the context to
-  # ShardingContext. mlir.ShardingContext is not used for SPMD.
+  # ShardingContext. sharding_impls.ShardingContext is not used for SPMD.
   elif isinstance(ctx.module_context.axis_context,
-                  (mlir.ReplicaAxisContext, mlir.ShardingContext)):
+                  (sharding_impls.ReplicaAxisContext, sharding_impls.ShardingContext)):
     return _xmap_lowering_rule_replica(ctx, *args, **kwargs)
   else:
     raise AssertionError("Unrecognized axis context type!")
@@ -1286,7 +1305,7 @@ def _xmap_lowering_rule_replica(ctx, *in_nodes,
                                 global_axis_sizes,
                                 spmd_in_axes, spmd_out_axes,
                                 axis_resources, resource_env, backend):
-  xla.check_backend_matches(backend, ctx.module_context.platform)
+  mlir.check_backend_matches(backend, ctx.module_context.platform)
   # The only way for any of those two assertions to be violated is when xmap
   # is using the SPMD lowering, but then this rule shouldn't even trigger.
   assert spmd_in_axes is None and spmd_out_axes is None
@@ -1317,7 +1336,7 @@ def _xmap_lowering_rule_replica(ctx, *in_nodes,
   #       them!
   vectorized_jaxpr, out_avals, consts = pe.trace_to_jaxpr_dynamic(f, local_avals)
   _check_out_avals_vs_out_axes(out_avals, out_axes, global_axis_sizes)
-  const_nodes = map(mlir.ir_constants, consts)
+  const_nodes = [mlir.ir_constants(xla.canonicalize_dtype(x)) for x in consts]
 
   local_mesh_shape = mesh.local_mesh.shape
   tiled_ins = (
@@ -1363,7 +1382,7 @@ def _xmap_lowering_rule_spmd(ctx, *global_in_nodes,
                              donated_invars, global_axis_sizes, spmd_in_axes,
                              spmd_out_axes, axis_resources,
                              resource_env, backend):
-  xla.check_backend_matches(backend, ctx.module_context.platform)
+  mlir.check_backend_matches(backend, ctx.module_context.platform)
   plan = EvaluationPlan.from_axis_resources(
       axis_resources, resource_env, global_axis_sizes)
 
@@ -1377,7 +1396,7 @@ def _xmap_lowering_rule_spmd(ctx, *global_in_nodes,
 
   # XXX: We modify mesh_in_axes and mesh_out_axes here
   def add_spmd_axes(
-      flat_mesh_axes: Sequence[pxla.ArrayMapping],
+      flat_mesh_axes: Sequence[ArrayMapping],
       flat_extra_axes: Optional[Sequence[Sequence[Sequence[MeshAxisName]]]]):
     if flat_extra_axes is None:
       return
@@ -1394,11 +1413,13 @@ def _xmap_lowering_rule_spmd(ctx, *global_in_nodes,
 
   global_sharding_spec = pxla.mesh_sharding_specs(mesh.shape, mesh.axis_names)
   sharded_global_in_nodes = [
-    [mlir.wrap_with_sharding_op(node, global_sharding_spec(aval, aval_axes).sharding_proto())]
+    [mlir.wrap_with_sharding_op(
+        ctx, node, aval,
+        global_sharding_spec(aval, aval_axes).sharding_proto().to_proto())]
     if aval_axes else [node]
     for node, aval, aval_axes in zip(global_in_nodes, global_in_avals, mesh_in_axes)
   ]
-  const_nodes = map(mlir.ir_constants, consts)
+  const_nodes = [mlir.ir_constants(xla.canonicalize_dtype(x)) for x in consts]
 
   # We in-line here rather than generating a Call HLO as in the xla_call
   # translation rule just because the extra tuple stuff is a pain.
@@ -1412,7 +1433,9 @@ def _xmap_lowering_rule_spmd(ctx, *global_in_nodes,
       dim_var_values=ctx.dim_var_values)
 
   sharded_global_out_nodes = [
-    mlir.wrap_with_sharding_op(node, global_sharding_spec(aval, aval_axes).sharding_proto())
+    mlir.wrap_with_sharding_op(
+        ctx, node, aval,
+        global_sharding_spec(aval, aval_axes).sharding_proto().to_proto())
     if aval_axes else node
     for (node,), aval, aval_axes in zip(global_out_nodes, global_out_avals, mesh_out_axes)
   ]
@@ -1427,7 +1450,7 @@ def _xmap_lowering_rule_spmd_manual(ctx, *global_in_nodes,
                                     resource_env, backend):
   assert spmd_in_axes is None and spmd_out_axes is None
   # This first part (up to vtile_manual) is shared with non-MANUAL SPMD rule.
-  xla.check_backend_matches(backend, ctx.module_context.platform)
+  mlir.check_backend_matches(backend, ctx.module_context.platform)
   plan = EvaluationPlan.from_axis_resources(
       axis_resources, resource_env, global_axis_sizes)
   manual_mesh_axes = frozenset(it.chain.from_iterable(plan.physical_axis_resources.values()))
@@ -1447,11 +1470,12 @@ def _xmap_lowering_rule_spmd_manual(ctx, *global_in_nodes,
   #       them!
   global_in_avals = ctx.avals_in
   vectorized_jaxpr, global_out_avals, consts = pe.trace_to_jaxpr_dynamic(f, global_in_avals)
-  const_nodes = map(mlir.ir_constants, consts)
+  const_nodes = [mlir.ir_constants(xla.canonicalize_dtype(x)) for x in consts]
 
   # We in-line here rather than generating a Call HLO as in the xla_call
   # translation rule just because the extra tuple stuff is a pain.
-  assert isinstance(ctx.module_context.axis_context, mlir.SPMDAxisContext)
+  assert isinstance(ctx.module_context.axis_context,
+                    sharding_impls.SPMDAxisContext)
   sub_ctx = ctx.module_context.replace(
       name_stack=ctx.module_context.name_stack.extend(wrap_name(name, 'xmap')),
       axis_context=ctx.module_context.axis_context.extend_manual(manual_mesh_axes))
@@ -1551,7 +1575,7 @@ class ResourceCount(NamedTuple):
 
 
 def _get_axis_resource_count(
-    axis_resources, resource_env) -> Dict[ResourceAxisName, ResourceCount]:
+    axis_resources, resource_env) -> dict[ResourceAxisName, ResourceCount]:
   global_res_shape = resource_env.shape
   local_res_shape = None
 
@@ -1562,17 +1586,17 @@ def _get_axis_resource_count(
     if local_res_shape is None:
       nlocal = None
     else:
-      nlocal = int(np.prod(map(local_res_shape.get, resources), dtype=np.int64))
+      nlocal = math.prod(map(local_res_shape.get, resources))
     resource_count_map[axis] = ResourceCount(
-        int(np.prod(map(global_res_shape.get, resources), dtype=np.int64)),
+        math.prod(map(global_res_shape.get, resources)),
         nlocal, distributed)
   return resource_count_map
 
 
 def _get_axis_sizes(args_flat: Iterable[Any],
                     in_axes_flat: Iterable[AxisNamePos],
-                    global_axis_sizes: Dict[AxisName, int],
-                    axis_resource_count: Dict[AxisName, ResourceCount]):
+                    global_axis_sizes: dict[AxisName, int],
+                    axis_resource_count: dict[AxisName, ResourceCount]):
   global_axis_sizes = dict(global_axis_sizes)
   for arg, in_axes in zip(args_flat, in_axes_flat):
     for name, dim in in_axes.items():
@@ -1596,16 +1620,6 @@ def _get_axis_sizes(args_flat: Iterable[Any],
   return FrozenDict(global_axis_sizes)
 
 
-def lookup_exactly_one_of(d: AxisNamePos, names: Set[AxisName]) -> Optional[int]:
-  res = None
-  for name in names:
-    if name in d:
-      if res is not None:
-        raise ValueError("An input was mapped to the same resource twice")
-      res = d[name]
-  return res
-
-
 @lu.transformation
 def hide_mapped_axes(flat_in_axes, flat_out_axes, *flat_args):
   def _squeeze_mapped_axes(arg, axes: AxisNamePos):
@@ -1631,7 +1645,7 @@ def hide_mapped_axes(flat_in_axes, flat_out_axes, *flat_args):
   yield map(_unsqueeze_mapped_axes, flat_outputs, flat_out_axes)
 
 
-def _jaxpr_resources(jaxpr, resource_env) -> Set[ResourceAxisName]:
+def _jaxpr_resources(jaxpr, resource_env) -> set[ResourceAxisName]:
   if isinstance(jaxpr, core.ClosedJaxpr):
     jaxpr = jaxpr.jaxpr
   assert isinstance(jaxpr, core.Jaxpr)
@@ -1649,7 +1663,7 @@ def _jaxpr_resources(jaxpr, resource_env) -> Set[ResourceAxisName]:
 
 
 def _to_resource_axes(axes_specs: Sequence[AxisNamePos],
-                      axis_resources: Dict[AxisName, Tuple[ResourceAxisName, ...]]):
+                      axis_resources: dict[AxisName, tuple[ResourceAxisName, ...]]):
   """
   Convert in/out_axes parameters ranging over logical dimensions to
   ones that range over resource dimensions.
@@ -1683,7 +1697,7 @@ def _slice_tile(x, dim: Optional[int], i, n: int):
   return lax.dynamic_slice_in_dim(x, i * tile_size, slice_size=tile_size, axis=dim)
 
 
-def _unzip_axis_resources(axis_resources: Dict[AxisName, Tuple[ResourceAxisName, ...]],
+def _unzip_axis_resources(axis_resources: dict[AxisName, tuple[ResourceAxisName, ...]],
                           resource_env: ResourceEnv):
   """Splits axis_resources into separate dicts for physical and loop resources."""
   physical_axis_resources = {}
@@ -1706,7 +1720,7 @@ def _unzip_axis_resources(axis_resources: Dict[AxisName, Tuple[ResourceAxisName,
 
 def _check_out_avals_vs_out_axes(out_avals: Sequence[core.AbstractValue],
                                  out_axes: Sequence[AxisNamePos],
-                                 global_axis_sizes: Dict[AxisName, int]):
+                                 global_axis_sizes: dict[AxisName, int]):
   defined_axes = set(global_axis_sizes)
   for aval, axes in zip(out_avals, out_axes):
     if not isinstance(aval, core.ShapedArray):
@@ -1725,11 +1739,12 @@ def _check_out_avals_vs_out_axes(out_avals: Sequence[core.AbstractValue],
 def _check_gda_or_array_xmap_partitioning(axis_resources, resource_env,
                                           global_axis_sizes, in_axes_flat,
                                           args_flat):
-  @lru_cache()
+  @lru_cache
   def _check_sharding(in_sharding, xmap_sharding, ndim, arr_flavor):
-    if not pxla.are_op_shardings_equal(
-        in_sharding._to_xla_op_sharding(ndim),
-        xmap_sharding._to_xla_op_sharding(ndim)):
+    if (not op_shardings.are_op_shardings_equal(
+          in_sharding._to_xla_hlo_sharding(ndim),
+          xmap_sharding._to_xla_hlo_sharding(ndim)) or
+        in_sharding.memory_kind != xmap_sharding.memory_kind):
       raise ValueError(
           f"Got an input {arr_flavor} to xmap with different partitioning than "
           "specified in xmap. The partitioning must match. "
@@ -1750,7 +1765,7 @@ def _check_gda_or_array_xmap_partitioning(axis_resources, resource_env,
 
       s = arg.sharding
       xmap_sharding = pxla.create_mesh_pspec_sharding(
-          mesh, pxla.array_mapping_to_axis_resources(xmap_array_mapping))
+          mesh, array_mapping_to_axis_resources(xmap_array_mapping))
       # This check is cached because comparing OpSharding is expensive during
       # dispatch and if the shardings are the same, then there is no need to
       # compare twice.
@@ -1775,10 +1790,6 @@ def _check_no_loop_collectives(jaxpr, loop_axis_resources):
 
 
 def _fix_inferred_spmd_sharding(jaxpr, resource_env, gen_fresh_name = None):
-  from jax._src.pjit import (
-      sharding_constraint_p, ParsedPartitionSpec, get_unconstrained_dims,
-      GSPMDSharding)
-
   rec = lambda jaxpr: _fix_inferred_spmd_sharding(jaxpr, resource_env, gen_fresh_name)
   if isinstance(jaxpr, core.ClosedJaxpr):
     return jaxpr.map_jaxpr(rec)
