@@ -18,7 +18,7 @@ from functools import partial
 import operator
 import warnings
 import numpy as np
-from typing import Any, Optional, Tuple, Union
+from typing import Any, Optional, Union
 
 import jax
 import jax.numpy as jnp
@@ -28,15 +28,16 @@ from jax._src import core
 from jax._src import dtypes
 from jax._src import util
 from jax._src.core import AxisName
+from jax._src.numpy import util as numpy_util
+from jax._src.typing import Array, ArrayLike
 from jax._src.ops.special import logsumexp as _logsumexp
 
-Array = Any
 
 # activations
 
 @custom_jvp
 @jax.jit
-def relu(x: Array) -> Array:
+def relu(x: ArrayLike) -> Array:
   r"""Rectified linear unit activation function.
 
   Computes the element-wise function:
@@ -55,13 +56,24 @@ def relu(x: Array) -> Array:
 
   Args:
     x : input array
+
+  Returns:
+    An array.
+
+  Example:
+    >>> jax.nn.relu(jax.numpy.array([-2., -1., -0.5, 0, 0.5, 1., 2.]))
+    Array([0. , 0. , 0. , 0. , 0.5, 1. , 2. ], dtype=float32)
+
+  See also:
+    :func:`relu6`
+
   """
   return jnp.maximum(x, 0)
 # For behavior at 0, see https://openreview.net/forum?id=urrcVI-_jRm
 relu.defjvps(lambda g, ans, x: lax.select(x > 0, g, lax.full_like(g, 0)))
 
 @jax.jit
-def softplus(x: Array) -> Array:
+def softplus(x: ArrayLike) -> Array:
   r"""Softplus activation function.
 
   Computes the element-wise function
@@ -75,7 +87,7 @@ def softplus(x: Array) -> Array:
   return jnp.logaddexp(x, 0)
 
 @jax.jit
-def soft_sign(x: Array) -> Array:
+def soft_sign(x: ArrayLike) -> Array:
   r"""Soft-sign activation function.
 
   Computes the element-wise function
@@ -86,10 +98,12 @@ def soft_sign(x: Array) -> Array:
   Args:
     x : input array
   """
-  return x / (jnp.abs(x) + 1)
+  numpy_util.check_arraylike("soft_sign", x)
+  x_arr = jnp.asarray(x)
+  return x_arr / (jnp.abs(x_arr) + 1)
 
 @jax.jit
-def sigmoid(x: Array) -> Array:
+def sigmoid(x: ArrayLike) -> Array:
   r"""Sigmoid activation function.
 
   Computes the element-wise function:
@@ -99,27 +113,44 @@ def sigmoid(x: Array) -> Array:
 
   Args:
     x : input array
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`log_sigmoid`
+
   """
   return lax.logistic(x)
 
 @jax.jit
-def silu(x: Array) -> Array:
-  r"""SiLU activation function.
+def silu(x: ArrayLike) -> Array:
+  r"""SiLU (a.k.a. swish) activation function.
 
   Computes the element-wise function:
 
   .. math::
     \mathrm{silu}(x) = x \cdot \mathrm{sigmoid}(x) = \frac{x}{1 + e^{-x}}
 
+  :func:`swish` and :func:`silu` are both aliases for the same function.
+
   Args:
     x : input array
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`sigmoid`
   """
-  return x * sigmoid(x)
+  numpy_util.check_arraylike("silu", x)
+  x_arr = jnp.asarray(x)
+  return x_arr * sigmoid(x_arr)
 
 swish = silu
 
 @jax.jit
-def log_sigmoid(x: Array) -> Array:
+def log_sigmoid(x: ArrayLike) -> Array:
   r"""Log-sigmoid activation function.
 
   Computes the element-wise function:
@@ -129,11 +160,19 @@ def log_sigmoid(x: Array) -> Array:
 
   Args:
     x : input array
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`sigmoid`
   """
-  return -softplus(-x)
+  numpy_util.check_arraylike("log_sigmoid", x)
+  x_arr = jnp.asarray(x)
+  return -softplus(-x_arr)
 
 @jax.jit
-def elu(x: Array, alpha: Array = 1.0) -> Array:
+def elu(x: ArrayLike, alpha: ArrayLike = 1.0) -> Array:
   r"""Exponential linear unit activation function.
 
   Computes the element-wise function:
@@ -147,12 +186,21 @@ def elu(x: Array, alpha: Array = 1.0) -> Array:
   Args:
     x : input array
     alpha : scalar or array of alpha values (default: 1.0)
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`selu`
   """
-  safe_x = jnp.where(x > 0, 0., x)
-  return jnp.where(x > 0, x, alpha * jnp.expm1(safe_x))
+  numpy_util.check_arraylike("elu", x)
+  x_arr = jnp.asarray(x)
+  return jnp.where(x_arr > 0,
+                   x_arr,
+                   alpha * jnp.expm1(jnp.where(x_arr > 0, 0., x_arr)))
 
 @jax.jit
-def leaky_relu(x: Array, negative_slope: Array = 1e-2) -> Array:
+def leaky_relu(x: ArrayLike, negative_slope: ArrayLike = 1e-2) -> Array:
   r"""Leaky rectified linear unit activation function.
 
   Computes the element-wise function:
@@ -168,11 +216,19 @@ def leaky_relu(x: Array, negative_slope: Array = 1e-2) -> Array:
   Args:
     x : input array
     negative_slope : array or scalar specifying the negative slope (default: 0.01)
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`relu`
   """
-  return jnp.where(x >= 0, x, negative_slope * x)
+  numpy_util.check_arraylike("leaky_relu", x)
+  x_arr = jnp.asarray(x)
+  return jnp.where(x_arr >= 0, x_arr, negative_slope * x_arr)
 
 @jax.jit
-def hard_tanh(x: Array) -> Array:
+def hard_tanh(x: ArrayLike) -> Array:
   r"""Hard :math:`\mathrm{tanh}` activation function.
 
   Computes the element-wise function:
@@ -186,11 +242,16 @@ def hard_tanh(x: Array) -> Array:
 
   Args:
     x : input array
+
+  Returns:
+    An array.
   """
-  return jnp.where(x > 1, 1, jnp.where(x < -1, -1, x))
+  numpy_util.check_arraylike("hard_tanh", x)
+  x_arr = jnp.asarray(x)
+  return jnp.where(x_arr > 1, 1, jnp.where(x_arr < -1, -1, x_arr))
 
 @jax.jit
-def celu(x: Array, alpha: Array = 1.0) -> Array:
+def celu(x: ArrayLike, alpha: ArrayLike = 1.0) -> Array:
   r"""Continuously-differentiable exponential linear unit activation.
 
   Computes the element-wise function:
@@ -208,11 +269,14 @@ def celu(x: Array, alpha: Array = 1.0) -> Array:
   Args:
     x : input array
     alpha : array or scalar (default: 1.0)
+
+  Returns:
+    An array.
   """
   return jnp.maximum(x, 0.0) + alpha * jnp.expm1(jnp.minimum(x, 0.0) / alpha)
 
 @jax.jit
-def selu(x: Array) -> Array:
+def selu(x: ArrayLike) -> Array:
   r"""Scaled exponential linear unit activation.
 
   Computes the element-wise function:
@@ -232,6 +296,12 @@ def selu(x: Array) -> Array:
 
   Args:
     x : input array
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`elu`
   """
   alpha = 1.6732632423543772848170429916717
   scale = 1.0507009873554804934193349852946
@@ -239,7 +309,7 @@ def selu(x: Array) -> Array:
 
 # TODO(phawkins): this jit was found to change numerics in a test. Debug this.
 # @partial(jax.jit, static_argnames=("approximate",))
-def gelu(x: Array, approximate: bool = True) -> Array:
+def gelu(x: ArrayLike, approximate: bool = True) -> Array:
   r"""Gaussian error linear unit activation function.
 
   If ``approximate=False``, computes the element-wise function:
@@ -261,29 +331,45 @@ def gelu(x: Array, approximate: bool = True) -> Array:
     x : input array
     approximate: whether to use the approximate or exact formulation.
   """
-
-  # Promote to nearest float-like dtype.
-  x = x.astype(dtypes.to_inexact_dtype(x.dtype))
+  [x_arr] = numpy_util.promote_args_inexact("gelu", x)
 
   if approximate:
-    sqrt_2_over_pi = np.sqrt(2 / np.pi).astype(x.dtype)
-    cdf = 0.5 * (1.0 + jnp.tanh(sqrt_2_over_pi * (x + 0.044715 * (x ** 3))))
-    return x * cdf
+    sqrt_2_over_pi = np.sqrt(2 / np.pi).astype(x_arr.dtype)
+    cdf = 0.5 * (1.0 + jnp.tanh(sqrt_2_over_pi * (x_arr + 0.044715 * (x_arr ** 3))))
+    return x_arr * cdf
   else:
-    sqrt_2 = np.sqrt(2).astype(x.dtype)
-    return jnp.array(x * (lax.erf(x / sqrt_2) + 1) / 2, dtype=x.dtype)
+    sqrt_2 = np.sqrt(2).astype(x_arr.dtype)
+    return jnp.array(x_arr * (lax.erf(x_arr / sqrt_2) + 1) / 2, dtype=x_arr.dtype)
 
 @partial(jax.jit, static_argnames=("axis",))
-def glu(x: Array, axis: int = -1) -> Array:
-  """Gated linear unit activation function.
+def glu(x: ArrayLike, axis: int = -1) -> Array:
+  r"""Gated linear unit activation function.
+
+  Computes the function:
+
+  .. math::
+    \mathrm{glu}(x) =  x\left[\ldots, 0:\frac{n}{2}, \ldots\right] \cdot
+      \mathrm{sigmoid} \left( x\left[\ldots, \frac{n}{2}:n, \ldots\right]
+        \right)
+
+  where the array is split into two along ``axis``. The size of the ``axis``
+  dimension must be divisible by two.
 
   Args:
     x : input array
     axis: the axis along which the split should be computed (default: -1)
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`sigmoid`
   """
-  size = x.shape[axis]
+  numpy_util.check_arraylike("glu", x)
+  x_arr = jnp.asarray(x)
+  size = x_arr.shape[axis]
   assert size % 2 == 0, "axis size must be divisible by 2"
-  x1, x2 = jnp.split(x, 2, axis)
+  x1, x2 = jnp.split(x_arr, 2, axis)
   return x1 * sigmoid(x2)
 
 # other functions
@@ -292,10 +378,10 @@ logsumexp = _logsumexp
 
 
 @partial(jax.jit, static_argnames=("axis",))
-def log_softmax(x: Array,
-                axis: Optional[Union[int, Tuple[int, ...]]] = -1,
-                where: Optional[Array] = None,
-                initial: Optional[Array] = None) -> Array:
+def log_softmax(x: ArrayLike,
+                axis: Optional[Union[int, tuple[int, ...]]] = -1,
+                where: Optional[ArrayLike] = None,
+                initial: Optional[ArrayLike] = None) -> Array:
   r"""Log-Softmax function.
 
   Computes the logarithm of the :code:`softmax` function, which rescales
@@ -312,9 +398,17 @@ def log_softmax(x: Array,
     where: Elements to include in the :code:`log_softmax`.
     initial: The minimum value used to shift the input array. Must be present
       when :code:`where` is not None.
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`softmax`
   """
-  x_max = jnp.max(x, axis, where=where, initial=initial, keepdims=True)
-  shifted = x - lax.stop_gradient(x_max)
+  numpy_util.check_arraylike("log_softmax", x)
+  x_arr = jnp.asarray(x)
+  x_max = jnp.max(x_arr, axis, where=where, initial=initial, keepdims=True)
+  shifted = x_arr - lax.stop_gradient(x_max)
   shifted_logsumexp = jnp.log(
       jnp.sum(jnp.exp(shifted), axis, where=where, keepdims=True))
   result = shifted - shifted_logsumexp
@@ -325,10 +419,10 @@ def log_softmax(x: Array,
 
 # TODO(phawkins): this jit was found to change numerics in a test. Debug this.
 #@partial(jax.jit, static_argnames=("axis",))
-def softmax(x: Array,
-            axis: Optional[Union[int, Tuple[int, ...]]] = -1,
-            where: Optional[Array] = None,
-            initial: Optional[Array] = None) -> Array:
+def softmax(x: ArrayLike,
+            axis: Optional[Union[int, tuple[int, ...]]] = -1,
+            where: Optional[ArrayLike] = None,
+            initial: Optional[ArrayLike] = None) -> Array:
   r"""Softmax function.
 
   Computes the function which rescales elements to the range :math:`[0, 1]`
@@ -345,19 +439,28 @@ def softmax(x: Array,
     where: Elements to include in the :code:`softmax`.
     initial: The minimum value used to shift the input array. Must be present
       when :code:`where` is not None.
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`log_softmax`
   """
   if jax.config.jax_softmax_custom_jvp:
-    return _softmax(x, axis, where, initial)
+    # mypy is confused by the `functools.partial` application in the definition
+    # of `_softmax` and incorrectly concludes that `_softmax` returns
+    # `ReturnValue` -- the unsubstituted type parameter of `custom_jvp`.
+    return _softmax(x, axis, where, initial)  # type: ignore[return-value]
   else:
     return _softmax_deprecated(x, axis, where, initial)
 
 # TODO(mattjj): replace softmax with _softmax when deprecation flag is removed
 @partial(jax.custom_jvp, nondiff_argnums=(1,))
 def _softmax(
-    x,
-    axis: Optional[Union[int, Tuple[int, ...]]] = -1,
-    where: Optional[Array] = None,
-    initial: Optional[Array] = None) -> Array:
+    x: ArrayLike,
+    axis: Optional[Union[int, tuple[int, ...]]] = -1,
+    where: Optional[ArrayLike] = None,
+    initial: Optional[ArrayLike] = None) -> Array:
   x_max = jnp.max(x, axis, where=where, initial=initial, keepdims=True)
   unnormalized = jnp.exp(x - x_max)
   result = unnormalized / jnp.sum(unnormalized, axis, where=where, keepdims=True)
@@ -371,7 +474,11 @@ def _softmax_jvp(axis, primals, tangents):
   y = _softmax(x, axis, where, initial)
   return y, y * (x_dot - (y * x_dot).sum(axis, where=where, keepdims=True))
 
-def _softmax_deprecated(x, axis, where, initial):
+def _softmax_deprecated(
+    x: ArrayLike,
+    axis: Optional[Union[int, tuple[int, ...]]] = -1,
+    where: Optional[ArrayLike] = None,
+    initial: Optional[ArrayLike] = None) -> Array:
   x_max = jnp.max(x, axis, where=where, initial=initial, keepdims=True)
   unnormalized = jnp.exp(x - lax.stop_gradient(x_max))
   result = unnormalized / jnp.sum(unnormalized, axis, where=where, keepdims=True)
@@ -381,13 +488,15 @@ def _softmax_deprecated(x, axis, where, initial):
 
 
 @partial(jax.jit, static_argnames=("axis",))
-def standardize(x: Array,
-              axis: Optional[Union[int, Tuple[int, ...]]] = -1,
-              mean: Optional[Array] = None,
-              variance: Optional[Array] = None,
-              epsilon: Array = 1e-5,
-              where: Optional[Array] = None) -> Array:
+def standardize(x: ArrayLike,
+              axis: Optional[Union[int, tuple[int, ...]]] = -1,
+              mean: Optional[ArrayLike] = None,
+              variance: Optional[ArrayLike] = None,
+              epsilon: ArrayLike = 1e-5,
+              where: Optional[ArrayLike] = None) -> Array:
   r"""Normalizes an array by subtracting ``mean`` and dividing by :math:`\sqrt{\mathrm{variance}}`."""
+  numpy_util.check_arraylike("standardize", x)
+  numpy_util.check_arraylike_or_none("standardize", mean, variance, where)
   if mean is None:
     mean = jnp.mean(x, axis, keepdims=True, where=where)
   if variance is None:
@@ -397,44 +506,45 @@ def standardize(x: Array,
     # when used in neural network normalization layers
     variance = jnp.mean(
         jnp.square(x), axis, keepdims=True, where=where) - jnp.square(mean)
-  return (x - mean) * lax.rsqrt(variance + epsilon)
+  return jnp.subtract(x, jnp.asarray(mean)) * lax.rsqrt(jnp.asarray(variance) + epsilon)
 
-def normalize(x: Array,
-              axis: Optional[Union[int, Tuple[int, ...]]] = -1,
-              mean: Optional[Array] = None,
-              variance: Optional[Array] = None,
-              epsilon: Array = 1e-5,
-              where: Optional[Array] = None) -> Array:
+def normalize(x: ArrayLike,
+            axis: Optional[Union[int, tuple[int, ...]]] = -1,
+            mean: Optional[ArrayLike] = None,
+            variance: Optional[ArrayLike] = None,
+            epsilon: ArrayLike = 1e-5,
+            where: Optional[ArrayLike] = None) -> Array:
   r"""Normalizes an array by subtracting ``mean`` and dividing by :math:`\sqrt{\mathrm{variance}}`."""
   warnings.warn("jax.nn.normalize will be deprecated. Use jax.nn.standardize instead.", DeprecationWarning)
   return standardize(x, axis, mean, variance, epsilon, where)
 
+# TODO(slebedev): Change the type of `x` to `ArrayLike`.
 @partial(jax.jit, static_argnames=("num_classes", "dtype", "axis"))
-def _one_hot(x: Array, num_classes: int, *,
+def _one_hot(x: Any, num_classes: int, *,
              dtype: Any, axis: Union[int, AxisName]) -> Array:
-  if not core.is_special_dim_size(num_classes):
-    num_classes = core.concrete_or_error(
-        int, num_classes,
-        "The error arose in jax.nn.one_hot argument `num_classes`.")
+  num_classes = core.concrete_dim_or_error(
+      num_classes,
+      "The error arose in jax.nn.one_hot argument `num_classes`.")
   dtype = dtypes.canonicalize_dtype(dtype)
-  x = jnp.asarray(x)
+  x_arr = jnp.asarray(x)
   try:
-    output_pos_axis = util.canonicalize_axis(axis, x.ndim + 1)
+    output_pos_axis = util.canonicalize_axis(axis, x_arr.ndim + 1)
   except TypeError:
     axis_size = lax.psum(1, axis)
     if num_classes != axis_size:
       raise ValueError(f"Expected num_classes to match the size of axis {axis}, "
                        f"but {num_classes} != {axis_size}") from None
     axis_idx = lax.axis_index(axis)
-    return jnp.asarray(x == axis_idx, dtype=dtype)
+    return jnp.asarray(x_arr == axis_idx, dtype=dtype)
   axis = operator.index(axis)  # type: ignore[arg-type]
-  lhs = lax.expand_dims(x, (axis,))
-  rhs_shape = [1] * x.ndim
+  lhs = lax.expand_dims(x_arr, (axis,))
+  rhs_shape = [1] * x_arr.ndim
   rhs_shape.insert(output_pos_axis, num_classes)
-  rhs = lax.broadcasted_iota(x.dtype, rhs_shape, output_pos_axis)
+  rhs = lax.broadcasted_iota(x_arr.dtype, rhs_shape, output_pos_axis)
   return jnp.asarray(lhs == rhs, dtype=dtype)
 
-def one_hot(x: Array, num_classes: int, *,
+# TODO(slebedev): Change the type of `x` to `ArrayLike`.
+def one_hot(x: Any, num_classes: int, *,
             dtype: Any = jnp.float_, axis: Union[int, AxisName] = -1) -> Array:
   """One-hot encodes the given indices.
 
@@ -459,16 +569,15 @@ def one_hot(x: Array, num_classes: int, *,
     axis: the axis or axes along which the function should be
       computed.
   """
-  if not core.is_special_dim_size(num_classes):
-    num_classes = core.concrete_or_error(
-        int, num_classes,
-        "The error arose in jax.nn.one_hot argument `num_classes`.")
+  num_classes = core.concrete_dim_or_error(
+      num_classes,
+      "The error arose in jax.nn.one_hot argument `num_classes`.")
   return _one_hot(x, num_classes, dtype=dtype, axis=axis)
 
 
 @jax.custom_jvp
 @jax.jit
-def relu6(x: Array) -> Array:
+def relu6(x: ArrayLike) -> Array:
   r"""Rectified Linear Unit 6 activation function.
 
   Computes the element-wise function
@@ -486,16 +595,21 @@ def relu6(x: Array) -> Array:
   .. math::
     \nabla \mathrm{relu}(6) = 0
 
-
   Args:
     x : input array
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`relu`
   """
   return jnp.minimum(jnp.maximum(x, 0), 6.)
 relu6.defjvps(lambda g, ans, x:
               lax.select((x > 0) & (x < 6), g, lax.full_like(g, 0)))
 
 @jax.jit
-def hard_sigmoid(x: Array) -> Array:
+def hard_sigmoid(x: ArrayLike) -> Array:
   r"""Hard Sigmoid activation function.
 
   Computes the element-wise function
@@ -505,21 +619,38 @@ def hard_sigmoid(x: Array) -> Array:
 
   Args:
     x : input array
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`relu6`
   """
   return relu6(x + 3.) / 6.
 
 @jax.jit
-def hard_silu(x: Array) -> Array:
-  r"""Hard SiLU activation function
+def hard_silu(x: ArrayLike) -> Array:
+  r"""Hard SiLU (swish) activation function
 
   Computes the element-wise function
 
   .. math::
     \mathrm{hard\_silu}(x) = x \cdot \mathrm{hard\_sigmoid}(x)
 
+  Both :func:`hard_silu` and :func:`hard_swish` are aliases for the same
+  function.
+
   Args:
     x : input array
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`hard_sigmoid`
   """
-  return x * hard_sigmoid(x)
+  numpy_util.check_arraylike("hard_silu", x)
+  x_arr = jnp.asarray(x)
+  return x_arr * hard_sigmoid(x_arr)
 
 hard_swish = hard_silu
