@@ -21,17 +21,15 @@ This is done dynamically in order to avoid circular imports.
 
 __all__ = ['register_jax_array_methods']
 
+import abc
 from functools import partial, wraps
-import inspect
-from typing import Any, List, Optional, Tuple, Union
-import warnings
+from typing import Any, Optional, Union
 
 import numpy as np
 import jax
 from jax import lax
 from jax._src import core
 from jax._src import dtypes
-from jax._src import device_array
 from jax._src.api_util import _ensure_index_tuple
 from jax._src.array import ArrayImpl
 from jax._src.lax import lax as lax_internal
@@ -41,12 +39,15 @@ from jax._src.numpy import ufuncs
 from jax._src.numpy import util
 from jax._src.ops import scatter
 from jax._src.typing import Array, ArrayLike, DimSize, DTypeLike, Shape
-from jax._src.util import safe_zip
+from jax._src.util import safe_zip, safe_map
+
+map, unsafe_map = safe_map, map
+zip, unsafe_zip = safe_zip, zip
 
 
 ### add method and operator overloads to arraylike classes
 
-# We add operator overloads to DeviceArray and ShapedArray. These method and
+# We add operator overloads to Array and ShapedArray. These method and
 # operator overloads mainly just forward calls to the corresponding lax_numpy
 # functions, which can themselves handle instances from any of these classes.
 
@@ -54,7 +55,7 @@ from jax._src.util import safe_zip
 def _astype(arr: ArrayLike, dtype: DTypeLike) -> Array:
   """Copy the array and cast to a specified dtype.
 
-  This is implemeted via :func:`jax.lax.convert_element_type`, which may
+  This is implemented via :func:`jax.lax.convert_element_type`, which may
   have slightly different behavior than :meth:`numpy.ndarray.astype` in
   some cases. In particular, the details of float-to-int and int-to-float
   casts are implementation dependent.
@@ -119,12 +120,16 @@ def _compute_newshape(a: ArrayLike, newshape: Union[DimSize, Shape]) -> Shape:
   try:
     iter(newshape)  # type: ignore[arg-type]
   except:
-    iterable = False
-  else:
-    iterable = True
-  newshape = core.canonicalize_shape(newshape if iterable else [newshape])  # type: ignore[arg-type]
-  return tuple(- core.divide_shape_sizes(np.shape(a), newshape)
-               if core.symbolic_equal_dim(d, -1) else d
+    newshape = [newshape]
+  newshape = core.canonicalize_shape(newshape)  # type: ignore[arg-type]
+  neg1s = [i for i, d in enumerate(newshape) if type(d) is int and d == -1]
+  if len(neg1s) == 1:
+    i, = neg1s
+    sz = core.cancel_divide_tracers(np.shape(a), (*newshape[:i], *newshape[i+1:]))
+    if sz is not None:
+      return (*newshape[:i], sz, *newshape[i+1:])
+  return tuple(-core.divide_shape_sizes(np.shape(a), newshape)
+               if core.definitely_equal(d, -1) else d
                for d in newshape)
 
 
@@ -235,7 +240,7 @@ def _view(arr: Array, dtype: DTypeLike = None, type: None = None) -> Array:
 
 
 def _notimplemented_flat(self):
-  raise NotImplementedError("JAX DeviceArrays do not implement the arr.flat property: "
+  raise NotImplementedError("JAX Arrays do not implement the arr.flat property: "
                             "consider arr.flatten() instead.")
 
 _accepted_binop_types = (int, float, complex, np.generic, np.ndarray, Array)
@@ -249,7 +254,9 @@ def _defer_to_unrecognized_arg(opchar, binary_op, swap=False):
     args = (other, self) if swap else (self, other)
     if isinstance(other, _accepted_binop_types):
       return binary_op(*args)
-    if isinstance(other, _rejected_binop_types):
+    # Note: don't use isinstance here, because we don't want to raise for
+    # subclasses, e.g. NamedTuple objects that may override operators.
+    if type(other) in _rejected_binop_types:
       raise TypeError(f"unsupported operand type(s) for {opchar}: "
                       f"{type(args[0]).__name__!r} and {type(args[1]).__name__!r}")
     return NotImplemented
@@ -277,7 +284,7 @@ def _deepcopy(self: Array, memo: Any) -> Array:
 
 # Experimental support for NumPy's module dispatch with NEP-37.
 # Currently requires https://github.com/seberg/numpy-dispatch
-_JAX_ARRAY_TYPES = (device_array.DeviceArray, core.Tracer, ArrayImpl)
+_JAX_ARRAY_TYPES = (core.Tracer, ArrayImpl)
 _HANDLED_ARRAY_TYPES = _JAX_ARRAY_TYPES + (np.ndarray,)
 
 def __array_module__(self, types):
@@ -295,49 +302,19 @@ def _compress_method(a: ArrayLike, condition: ArrayLike,
   return lax_numpy.jaxcompress(condition, a, axis, out)
 
 
-@util._wraps(lax.broadcast, lax_description="""
-Deprecated. Use :func:`jax.lax.broadcast` instead.
-""")
-def _deprecated_broadcast(*args, **kwargs):
-  warnings.warn(
-    "The arr.broadcast() method is deprecated. Use jax.lax.broadcast instead.",
-    category=FutureWarning)
-  return lax.broadcast(*args, **kwargs)
-
-
-@util._wraps(lax.broadcast, lax_description="""
-Deprecated. Use :func:`jax.lax.broadcast_in_dim` instead.
-""")
-def _deprecated_broadcast_in_dim(*args, **kwargs):
-  warnings.warn(
-    "The arr.broadcast_in_dim() method is deprecated. Use jax.lax.broadcast_in_dim instead.",
-    category=FutureWarning)
-  return lax.broadcast_in_dim(*args, **kwargs)
-
-
-@util._wraps(lax.broadcast, lax_description="""
-Deprecated. Use :func:`jax.numpy.split` instead.
-""")
-def _deprecated_split(*args, **kwargs):
-  warnings.warn(
-    "The arr.split() method is deprecated. Use jax.numpy.split instead.",
-    category=FutureWarning)
-  return lax_numpy.split(*args, **kwargs)
-
-
 @core.stash_axis_env()
 @partial(jax.jit, static_argnums=(1,2,3))
 def _multi_slice(arr: ArrayLike,
-                 start_indices: Tuple[Tuple[int, ...]],
-                 limit_indices: Tuple[Tuple[int, ...]],
-                 removed_dims: Tuple[Tuple[int, ...]]) -> List[Array]:
+                 start_indices: tuple[tuple[int, ...]],
+                 limit_indices: tuple[tuple[int, ...]],
+                 removed_dims: tuple[tuple[int, ...]]) -> list[Array]:
   """Extracts multiple slices from `arr`.
 
-  This is used to shard DeviceArray arguments to pmap. It's implemented as a
-  DeviceArray method here to avoid circular imports.
+  This is used to shard Array arguments to pmap. It's implemented as a
+  Array method here to avoid circular imports.
   """
-  results: List[Array] = []
-  for starts, limits, removed in safe_zip(start_indices, limit_indices, removed_dims):
+  results: list[Array] = []
+  for starts, limits, removed in zip(start_indices, limit_indices, removed_dims):
     sliced = lax.slice(arr, starts, limits)
     if removed:
       sliced = lax.squeeze(sliced, removed)
@@ -347,7 +324,7 @@ def _multi_slice(arr: ArrayLike,
 # The next two functions are related to iter(device_array), implemented here to
 # avoid circular imports.
 @jax.jit
-def _unstack(x: Array) -> List[Array]:
+def _unstack(x: Array) -> list[Array]:
   return [lax.index_in_dim(x, i, keepdims=False) for i in range(x.shape[0])]
 
 def _chunk_iter(x, size):
@@ -360,13 +337,16 @@ def _chunk_iter(x, size):
     if tail:
       yield lax.dynamic_slice_in_dim(x, num_chunks * size, tail)
 
+def _getitem(self, item):
+  return lax_numpy._rewriting_take(self, item)
+
 # Syntactic sugar for scatter operations.
 class _IndexUpdateHelper:
   # Note: this docstring will appear as the docstring for the `at` property.
   """Helper property for index update functionality.
 
   The ``at`` property provides a functionally pure equivalent of in-place
-  array modificatons.
+  array modifications.
 
   In particular:
 
@@ -458,27 +438,6 @@ class _IndexUpdateHelper:
     return f"_IndexUpdateHelper({repr(self.array)})"
 
 
-# TODO(jakevdp): remove these deprecation warnings after June 2023
-def allow_pass_by_position_with_warning(f):
-  @wraps(f)
-  def wrapped(*args, **kwargs):
-    sig = inspect.signature(f)
-    try:
-      sig.bind(*args, **kwargs)
-    except TypeError:
-      argspec = inspect.getfullargspec(f)
-      n_positional = len(argspec.args)
-      keywords = argspec.kwonlyargs[:len(args) - n_positional]
-      warnings.warn(
-          f"jnp.ndarray.at[...].{f.__name__}: Passing '{keywords[0]}' by position is deprecated. "
-          f"Pass by keyword instead", category=FutureWarning, stacklevel=2)
-      converted_kwargs = dict(zip(keywords, args[n_positional:]))
-      return f(*args[:n_positional], **converted_kwargs, **kwargs)
-    else:
-      return f(*args, **kwargs)
-  return wrapped
-
-
 class _IndexUpdateRef:
   """Helper object to call indexed update functions for an (advanced) index.
 
@@ -495,7 +454,6 @@ class _IndexUpdateRef:
   def __repr__(self):
     return f"_IndexUpdateRef({repr(self.array)}, {repr(self.index)})"
 
-  @allow_pass_by_position_with_warning
   def get(self, *, indices_are_sorted=False, unique_indices=False,
           mode=None, fill_value=None):
     """Equivalent to ``x[idx]``.
@@ -512,7 +470,6 @@ class _IndexUpdateRef:
                                      unique_indices=unique_indices, mode=mode,
                                      fill_value=fill_value)
 
-  @allow_pass_by_position_with_warning
   def set(self, values, *, indices_are_sorted=False, unique_indices=False,
           mode=None):
     """Pure equivalent of ``x[idx] = y``.
@@ -526,7 +483,6 @@ class _IndexUpdateRef:
                                    indices_are_sorted=indices_are_sorted,
                                    unique_indices=unique_indices, mode=mode)
 
-  @allow_pass_by_position_with_warning
   def apply(self, func, *, indices_are_sorted=False, unique_indices=False,
             mode=None):
     """Pure equivalent of ``func.at(x, idx)`` for a unary ufunc ``func``.
@@ -542,15 +498,14 @@ class _IndexUpdateRef:
 
     See :mod:`jax.ops` for details.
     """
-    def _scatter_apply(x, indices, _, dims, **kwargs):
-      return lax.scatter_apply(x, indices, func, dims, **kwargs)
+    def _scatter_apply(x, indices, y, dims, **kwargs):
+      return lax.scatter_apply(x, indices, func, dims, update_shape=y.shape, **kwargs)
     return scatter._scatter_update(self.array, self.index,
                                    lax_internal._zero(self.array.dtype),
                                    _scatter_apply,
                                    indices_are_sorted=indices_are_sorted,
                                    unique_indices=unique_indices, mode=mode)
 
-  @allow_pass_by_position_with_warning
   def add(self, values, *, indices_are_sorted=False, unique_indices=False,
           mode=None):
     """Pure equivalent of ``x[idx] += y``.
@@ -565,7 +520,6 @@ class _IndexUpdateRef:
                                    indices_are_sorted=indices_are_sorted,
                                    unique_indices=unique_indices, mode=mode)
 
-  @allow_pass_by_position_with_warning
   def multiply(self, values, *, indices_are_sorted=False, unique_indices=False,
                mode=None):
     """Pure equivalent of ``x[idx] *= y``.
@@ -582,7 +536,6 @@ class _IndexUpdateRef:
                                    mode=mode)
   mul = multiply
 
-  @allow_pass_by_position_with_warning
   def divide(self, values, *, indices_are_sorted=False, unique_indices=False,
              mode=None):
     """Pure equivalent of ``x[idx] /= y``.
@@ -599,7 +552,6 @@ class _IndexUpdateRef:
                               indices_are_sorted=indices_are_sorted,
                               unique_indices=unique_indices, mode=mode))
 
-  @allow_pass_by_position_with_warning
   def power(self, values, *, indices_are_sorted=False, unique_indices=False,
             mode=None):
     """Pure equivalent of ``x[idx] **= y``.
@@ -616,7 +568,6 @@ class _IndexUpdateRef:
                               indices_are_sorted=indices_are_sorted,
                               unique_indices=unique_indices, mode=mode))
 
-  @allow_pass_by_position_with_warning
   def min(self, values, *, indices_are_sorted=False, unique_indices=False,
           mode=None):
     """Pure equivalent of ``x[idx] = minimum(x[idx], y)``.
@@ -632,7 +583,6 @@ class _IndexUpdateRef:
                                    indices_are_sorted=indices_are_sorted,
                                    unique_indices=unique_indices, mode=mode)
 
-  @allow_pass_by_position_with_warning
   def max(self, values, *, indices_are_sorted=False, unique_indices=False,
           mode=None):
     """Pure equivalent of ``x[idx] = maximum(x[idx], y)``.
@@ -649,7 +599,7 @@ class _IndexUpdateRef:
                                    unique_indices=unique_indices, mode=mode)
 
 _array_operators = {
-  "getitem": lax_numpy._rewriting_take,
+  "getitem": _getitem,
   "setitem": _unimplemented_setitem,
   "copy": _copy,
   "deepcopy": _deepcopy,
@@ -741,12 +691,6 @@ _array_methods = {
   # Methods exposed in order to avoid circular imports
   "_split": lax_numpy.split,  # used in jacfwd/jacrev
   "_multi_slice": _multi_slice,  # used in pxla for sharding
-
-  # Deprecated methods.
-  # TODO(jakevdp): remove these after June 2023
-  "broadcast": _deprecated_broadcast,
-  "broadcast_in_dim": _deprecated_broadcast_in_dim,
-  "split": _deprecated_split,
 }
 
 _impl_only_array_methods = {
@@ -757,6 +701,7 @@ _impl_only_array_methods = {
 _array_properties = {
   "flat": _notimplemented_flat,
   "T": lax_numpy.transpose,
+  "mT": lax_numpy.matrix_transpose,
   "real": ufuncs.real,
   "imag": ufuncs.imag,
   "nbytes": _nbytes,
@@ -778,9 +723,35 @@ def _set_shaped_array_attributes(shaped_array):
     setattr(shaped_array, prop_name, core.aval_property(prop))
   setattr(shaped_array, "_array_module", staticmethod(__array_module__))
 
+def _forward_operator_to_aval(name):
+  def op(self, *args):
+    return getattr(self.aval, f"_{name}")(self, *args)
+  return op
 
-def _set_device_array_base_attributes(device_array, include=None, exclude=None):
-  # Forward operators, methods, and properties on DeviceArray to lax_numpy
+def _forward_method_to_aval(name):
+  def meth(self, *args, **kwargs):
+    return getattr(self.aval, name).fun(self, *args, **kwargs)
+  return meth
+
+def _forward_property_to_aval(name):
+  @property
+  def prop(self):
+    return getattr(self.aval, name).fget(self)
+  return prop
+
+def _set_tracer_aval_forwarding(tracer, exclude=()):
+  for operator_name in _array_operators:
+    if operator_name not in exclude:
+      setattr(tracer, f"__{operator_name}__", _forward_operator_to_aval(operator_name))
+  for method_name in _array_methods:
+    if method_name not in exclude:
+      setattr(tracer, method_name, _forward_method_to_aval(method_name))
+  for prop_name in _array_properties:
+    if prop_name not in exclude:
+      setattr(tracer, prop_name, _forward_property_to_aval(prop_name))
+
+def _set_array_base_attributes(device_array, include=None, exclude=None):
+  # Forward operators, methods, and properties on Array to lax_numpy
   # functions (with no Tracers involved; this forwarding is direct)
   def maybe_setattr(attr_name, target):
     if exclude is not None and attr_name in exclude:
@@ -798,20 +769,36 @@ def _set_device_array_base_attributes(device_array, include=None, exclude=None):
   for name, func in _impl_only_array_methods.items():
     setattr(device_array, name, func)
 
-def _set_device_array_attributes(device_array):
+def _set_array_attributes(device_array):
   setattr(device_array, "__array_module__", __array_module__)
 
+def _make_abstract_method(name, func):
+  @abc.abstractmethod
+  @wraps(func)
+  def method(*args, **kwargs):
+    raise NotImplementedError(f"Cannot call abstract method {name}")
+  return method
+
+def _set_array_abstract_methods(basearray):
+  for operator_name, function in _array_operators.items():
+    setattr(basearray, f"__{operator_name}__",
+            _make_abstract_method(f"__{operator_name}__", function))
+  for method_name, method in _array_methods.items():
+    setattr(basearray, method_name,
+            _make_abstract_method(method_name, method))
+  for prop_name, prop in _array_properties.items():
+    setattr(basearray, prop_name,
+            property(_make_abstract_method(prop_name, prop)))
 
 def register_jax_array_methods():
   """Call this function once to register methods of JAX arrays"""
   _set_shaped_array_attributes(core.ShapedArray)
   _set_shaped_array_attributes(core.DShapedArray)
 
-  _set_device_array_base_attributes(device_array.DeviceArray)
-  _set_device_array_base_attributes(ArrayImpl, exclude={'__getitem__'})
+  _set_array_base_attributes(ArrayImpl, exclude={'__getitem__'})
+  _set_tracer_aval_forwarding(core.Tracer, exclude={*_impl_only_array_methods, "at"})
+  _set_array_attributes(ArrayImpl)
 
-  for t in device_array.device_array_types:
-    _set_device_array_attributes(t)
-    _set_device_array_attributes(ArrayImpl)
+  _set_array_abstract_methods(Array)
 
   Array.at.__doc__ = _IndexUpdateHelper.__doc__
