@@ -18,7 +18,7 @@ from functools import partial
 import numpy as np
 import textwrap
 import operator
-from typing import Literal, Optional, Tuple, Union, cast, overload
+from typing import Literal, Optional, Union, cast, overload
 
 import jax
 from jax import jit, custom_jvp
@@ -33,12 +33,8 @@ from jax._src.util import canonicalize_axis
 from jax._src.typing import ArrayLike, Array
 
 
-def _T(x: ArrayLike) -> Array:
-  return jnp.swapaxes(x, -1, -2)
-
-
 def _H(x: ArrayLike) -> Array:
-  return ufuncs.conjugate(jnp.swapaxes(x, -1, -2))
+  return ufuncs.conjugate(jnp.matrix_transpose(x))
 
 
 def _symmetrize(x: Array) -> Array: return (x + _H(x)) / 2
@@ -53,10 +49,10 @@ def cholesky(a: ArrayLike) -> Array:
 
 @overload
 def svd(a: ArrayLike, full_matrices: bool = True, *, compute_uv: Literal[True],
-        hermitian: bool = False) -> Tuple[Array, Array, Array]: ...
+        hermitian: bool = False) -> tuple[Array, Array, Array]: ...
 @overload
 def svd(a: ArrayLike, full_matrices: bool, compute_uv: Literal[True],
-        hermitian: bool = False) -> Tuple[Array, Array, Array]: ...
+        hermitian: bool = False) -> tuple[Array, Array, Array]: ...
 @overload
 def svd(a: ArrayLike, full_matrices: bool = True, *, compute_uv: Literal[False],
         hermitian: bool = False) -> Array: ...
@@ -65,12 +61,12 @@ def svd(a: ArrayLike, full_matrices: bool, compute_uv: Literal[False],
         hermitian: bool = False) -> Array: ...
 @overload
 def svd(a: ArrayLike, full_matrices: bool = True, compute_uv: bool = True,
-        hermitian: bool = False) -> Union[Array, Tuple[Array, Array, Array]]: ...
+        hermitian: bool = False) -> Union[Array, tuple[Array, Array, Array]]: ...
 
 @_wraps(np.linalg.svd)
 @partial(jit, static_argnames=('full_matrices', 'compute_uv', 'hermitian'))
 def svd(a: ArrayLike, full_matrices: bool = True, compute_uv: bool = True,
-        hermitian: bool = False) -> Union[Array, Tuple[Array, Array, Array]]:
+        hermitian: bool = False) -> Union[Array, tuple[Array, Array, Array]]:
   check_arraylike("jnp.linalg.svd", a)
   a, = promote_dtypes_inexact(jnp.asarray(a))
   if hermitian:
@@ -146,7 +142,7 @@ def matrix_rank(M: ArrayLike, tol: Optional[ArrayLike] = None) -> Array:
 
 
 @custom_jvp
-def _slogdet_lu(a: Array) -> Tuple[Array, Array]:
+def _slogdet_lu(a: Array) -> tuple[Array, Array]:
   dtype = lax.dtype(a)
   lu, pivot, _ = lax_linalg.lu(a)
   diag = jnp.diagonal(lu, axis1=-2, axis2=-1)
@@ -168,7 +164,7 @@ def _slogdet_lu(a: Array) -> Tuple[Array, Array]:
   return sign, ufuncs.real(logdet)
 
 @custom_jvp
-def _slogdet_qr(a: Array) -> Tuple[Array, Array]:
+def _slogdet_qr(a: Array) -> tuple[Array, Array]:
   # Implementation of slogdet using QR decomposition. One reason we might prefer
   # QR decomposition is that it is more amenable to a fast batched
   # implementation on TPU because of the lack of row pivoting.
@@ -196,7 +192,7 @@ def _slogdet_qr(a: Array) -> Tuple[Array, Array]:
         LU decomposition if ``None``.
     """))
 @partial(jit, static_argnames=('method',))
-def slogdet(a: ArrayLike, *, method: Optional[str] = None) -> Tuple[Array, Array]:
+def slogdet(a: ArrayLike, *, method: Optional[str] = None) -> tuple[Array, Array]:
   check_arraylike("jnp.linalg.slogdet", a)
   a, = promote_dtypes_inexact(jnp.asarray(a))
   a_shape = jnp.shape(a)
@@ -227,7 +223,7 @@ def _slogdet_jvp(primals, tangents):
 _slogdet_lu.defjvp(_slogdet_jvp)
 _slogdet_qr.defjvp(_slogdet_jvp)
 
-def _cofactor_solve(a: ArrayLike, b: ArrayLike) -> Tuple[Array, Array]:
+def _cofactor_solve(a: ArrayLike, b: ArrayLike) -> tuple[Array, Array]:
   """Equivalent to det(a)*solve(a, b) for nonsingular mat.
 
   Intermediate function used for jvp and vjp of det.
@@ -368,7 +364,7 @@ At present, non-symmetric eigendecomposition is only implemented on the CPU
 backend. However eigendecomposition for symmetric/Hermitian matrices is
 implemented more widely (see :func:`jax.numpy.linalg.eigh`).
 """)
-def eig(a: ArrayLike) -> Tuple[Array, Array]:
+def eig(a: ArrayLike) -> tuple[Array, Array]:
   check_arraylike("jnp.linalg.eig", a)
   a, = promote_dtypes_inexact(jnp.asarray(a))
   w, v = lax_linalg.eig(a, compute_left_eigenvectors=False)
@@ -386,7 +382,7 @@ def eigvals(a: ArrayLike) -> Array:
 @_wraps(np.linalg.eigh)
 @partial(jit, static_argnames=('UPLO', 'symmetrize_input'))
 def eigh(a: ArrayLike, UPLO: Optional[str] = None,
-         symmetrize_input: bool = True) -> Tuple[Array, Array]:
+         symmetrize_input: bool = True) -> tuple[Array, Array]:
   check_arraylike("jnp.linalg.eigh", a)
   if UPLO is None or UPLO == "L":
     lower = True
@@ -436,7 +432,7 @@ def pinv(a: ArrayLike, rcond: Optional[ArrayLike] = None,
   rcond = lax.expand_dims(rcond[..., jnp.newaxis], range(s.ndim - rcond.ndim - 1))
   cutoff = rcond * s[..., 0:1]
   s = jnp.where(s > cutoff, s, jnp.inf).astype(u.dtype)
-  res = jnp.matmul(_T(vh), ufuncs.divide(_T(u), s[..., jnp.newaxis]),
+  res = jnp.matmul(vh.mT, ufuncs.divide(u.mT, s[..., jnp.newaxis]),
                    precision=lax.Precision.HIGHEST)
   return lax.convert_element_type(res, arr.dtype)
 
@@ -485,7 +481,7 @@ def inv(a: ArrayLike) -> Array:
 @_wraps(np.linalg.norm)
 @partial(jit, static_argnames=('ord', 'axis', 'keepdims'))
 def norm(x: ArrayLike, ord: Union[int, str, None] = None,
-         axis: Union[None, Tuple[int, ...], int] = None,
+         axis: Union[None, tuple[int, ...], int] = None,
          keepdims: bool = False) -> Array:
   check_arraylike("jnp.linalg.norm", x)
   x, = promote_dtypes_inexact(jnp.asarray(x))
@@ -536,7 +532,7 @@ def norm(x: ArrayLike, ord: Union[int, str, None] = None,
       return ufuncs.power(out, ord_inv)
 
   elif num_axes == 2:
-    row_axis, col_axis = cast(Tuple[int, ...], axis)
+    row_axis, col_axis = cast(tuple[int, ...], axis)
     if ord is None or ord in ('f', 'fro'):
       return ufuncs.sqrt(reductions.sum(ufuncs.real(x * ufuncs.conj(x)), axis=axis,
                                         keepdims=keepdims))
@@ -582,16 +578,16 @@ def norm(x: ArrayLike, ord: Union[int, str, None] = None,
 @overload
 def qr(a: ArrayLike, mode: Literal["r"]) -> Array: ...
 @overload
-def qr(a: ArrayLike, mode: str = "reduced") -> Union[Array, Tuple[Array, Array]]: ...
+def qr(a: ArrayLike, mode: str = "reduced") -> Union[Array, tuple[Array, Array]]: ...
 
 @_wraps(np.linalg.qr)
 @partial(jit, static_argnames=('mode',))
-def qr(a: ArrayLike, mode: str = "reduced") -> Union[Array, Tuple[Array, Array]]:
+def qr(a: ArrayLike, mode: str = "reduced") -> Union[Array, tuple[Array, Array]]:
   check_arraylike("jnp.linalg.qr", a)
   a, = promote_dtypes_inexact(jnp.asarray(a))
   if mode == "raw":
     a, taus = lax_linalg.geqrf(a)
-    return _T(a), taus
+    return a.mT, taus
   if mode in ("reduced", "r", "full"):
     full_matrices = False
   elif mode == "complete":
@@ -613,7 +609,7 @@ def solve(a: ArrayLike, b: ArrayLike) -> Array:
 
 
 def _lstsq(a: ArrayLike, b: ArrayLike, rcond: Optional[float], *,
-           numpy_resid: bool = False) -> Tuple[Array, Array, Array, Array]:
+           numpy_resid: bool = False) -> tuple[Array, Array, Array, Array]:
   # TODO: add lstsq to lax_linalg and implement this function via those wrappers.
   # TODO: add custom jvp rule for more robust lstsq differentiation
   a, b = promote_dtypes_inexact(a, b)
@@ -673,7 +669,7 @@ _jit_lstsq = jit(partial(_lstsq, numpy_resid=False))
     poorly behaved for some inputs, particularly for low-rank `a`.
     """))
 def lstsq(a: ArrayLike, b: ArrayLike, rcond: Optional[float] = None, *,
-          numpy_resid: bool = False) -> Tuple[Array, Array, Array, Array]:
+          numpy_resid: bool = False) -> tuple[Array, Array, Array, Array]:
   check_arraylike("jnp.linalg.lstsq", a, b)
   if numpy_resid:
     return _lstsq(a, b, rcond, numpy_resid=True)
