@@ -13,10 +13,11 @@
 # limitations under the License.
 
 import builtins
+from collections.abc import Sequence
 from functools import partial
 import math
 import operator
-from typing import overload, Any, Callable, Literal, Optional, Sequence, Tuple, Union
+from typing import overload, Any, Callable, Literal, Optional, Protocol, Union
 import warnings
 
 import numpy as np
@@ -25,13 +26,15 @@ from jax import lax
 from jax._src import api
 from jax._src import core
 from jax._src import dtypes
+from jax._src.numpy import ufuncs
 from jax._src.numpy.util import (
-    _broadcast_to, _check_arraylike, _complex_elem_type,
-    _promote_dtypes_inexact, _promote_dtypes_numeric, _where, _wraps)
+    _broadcast_to, check_arraylike, _complex_elem_type,
+    promote_dtypes_inexact, promote_dtypes_numeric, _where, _wraps)
 from jax._src.lax import lax as lax_internal
 from jax._src.typing import Array, ArrayLike, DType, DTypeLike
 from jax._src.util import (
-    canonicalize_axis as _canonicalize_axis, maybe_named_axis)
+    canonicalize_axis as _canonicalize_axis, maybe_named_axis,
+    NumpyComplexWarning)
 
 
 _all = builtins.all
@@ -40,11 +43,6 @@ _lax_const = lax_internal._const
 
 Axis = Union[None, int, Sequence[int]]
 
-
-def _asarray(a: ArrayLike) -> Array:
-  # simplified version of jnp.asarray() for local use.
-  return a if isinstance(a, Array) else api.device_put(a)
-
 def _isscalar(element: Any) -> bool:
   if hasattr(element, '__jax_array__'):
     element = element.__jax_array__()
@@ -52,8 +50,8 @@ def _isscalar(element: Any) -> bool:
 
 def _moveaxis(a: ArrayLike, source: int, destination: int) -> Array:
   # simplified version of jnp.moveaxis() for local use.
-  _check_arraylike("moveaxis", a)
-  a = _asarray(a)
+  check_arraylike("moveaxis", a)
+  a = lax_internal.asarray(a)
   source = _canonicalize_axis(source, np.ndim(a))
   destination = _canonicalize_axis(destination, np.ndim(a))
   perm = [i for i in range(np.ndim(a)) if i != source]
@@ -83,7 +81,7 @@ def _reduction(a: ArrayLike, name: str, np_fun: Any, op: ReductionOp, init_val:
   # exists, passing along all its arguments.
   if out is not None:
     raise NotImplementedError(f"The 'out' argument to jnp.{name} is not supported.")
-  _check_arraylike(name, a)
+  check_arraylike(name, a)
   dtypes.check_user_dtype_supported(dtype, name)
   axis = core.concrete_or_error(None, axis, f"axis argument to jnp.{name}().")
 
@@ -91,13 +89,13 @@ def _reduction(a: ArrayLike, name: str, np_fun: Any, op: ReductionOp, init_val:
     raise ValueError(f"reduction operation {name} does not have an identity, so to use a "
                      f"where mask one has to specify 'initial'")
 
-  a = a if isinstance(a, Array) else _asarray(a)
+  a = a if isinstance(a, Array) else lax_internal.asarray(a)
   a = preproc(a) if preproc else a
   pos_dims, dims = _reduction_dims(a, axis)
 
   if initial is None and not has_identity:
     shape = np.shape(a)
-    if not _all(core.greater_equal_dim(shape[d], 1) for d in pos_dims):
+    if not _all(shape[d] >= 1 for d in pos_dims):
       raise ValueError(f"zero-size array to reduction operation {name} which has no identity")
 
   result_dtype = dtype or dtypes.dtype(a)
@@ -134,7 +132,7 @@ def _reduction(a: ArrayLike, name: str, np_fun: Any, op: ReductionOp, init_val:
   else:
     result = lax.reduce(a, init_val, op, dims)
   if initial is not None:
-    initial_arr = lax.convert_element_type(initial, _asarray(a).dtype)
+    initial_arr = lax.convert_element_type(initial, lax_internal.asarray(a).dtype)
     if initial_arr.shape != ():
       raise ValueError("initial value must be a scalar. "
                        f"Got array of shape {initial_arr.shape}")
@@ -176,11 +174,11 @@ def _reduction_init_val(a: ArrayLike, init_val: Any) -> np.ndarray:
 
 def _cast_to_bool(operand: ArrayLike) -> Array:
   with warnings.catch_warnings():
-    warnings.filterwarnings("ignore", category=np.ComplexWarning)
+    warnings.filterwarnings("ignore", category=NumpyComplexWarning)
     return lax.convert_element_type(operand, np.bool_)
 
 def _cast_to_numeric(operand: ArrayLike) -> Array:
-  return _promote_dtypes_numeric(operand)[0]
+  return promote_dtypes_numeric(operand)[0]
 
 
 def _ensure_optional_axes(x: Axis) -> Axis:
@@ -298,11 +296,8 @@ def any(a: ArrayLike, axis: Axis = None, out: None = None,
   return _reduce_any(a, axis=_ensure_optional_axes(axis), out=out,
                      keepdims=keepdims, where=where)
 
-product = prod
 amin = min
 amax = max
-alltrue = all
-sometrue = any
 
 def _axis_size(a: ArrayLike, axis: Union[int, Sequence[int]]):
   if not isinstance(axis, (tuple, list)):
@@ -325,12 +320,23 @@ def mean(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None,
 @partial(api.jit, static_argnames=('axis', 'dtype', 'keepdims'), inline=True)
 def _mean(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None,
           out: None = None, keepdims: bool = False, *,
+          upcast_f16_for_computation: bool = True,
           where: Optional[ArrayLike] = None) -> Array:
-  _check_arraylike("mean", a)
-  dtypes.check_user_dtype_supported(dtype, "mean")
+  check_arraylike("mean", a)
   if out is not None:
     raise NotImplementedError("The 'out' argument to jnp.mean is not supported.")
 
+  if dtype is None:
+    result_dtype = dtypes.to_inexact_dtype(dtypes.dtype(a, canonicalize=True))
+  else:
+    dtypes.check_user_dtype_supported(dtype, "mean")
+    result_dtype = dtypes.canonicalize_dtype(dtype)
+
+  if upcast_f16_for_computation and dtypes.issubdtype(result_dtype, np.inexact):
+    computation_dtype = _upcast_f16(result_dtype)
+  else:
+    computation_dtype = result_dtype
+
   if where is None:
     if axis is None:
       normalizer = core.dimension_as_value(np.size(a))
@@ -339,13 +345,10 @@ def _mean(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None,
   else:
     normalizer = sum(_broadcast_to(where, np.shape(a)), axis, dtype=dtype, keepdims=keepdims)
 
-  if dtype is None:
-    dtype = dtypes.to_inexact_dtype(dtypes.dtype(a))
-  dtype = dtypes.canonicalize_dtype(dtype)
-
   return lax.div(
-      sum(a, axis, dtype=dtype, keepdims=keepdims, where=where),
-      lax.convert_element_type(normalizer, dtype))
+      sum(a, axis, dtype=computation_dtype, keepdims=keepdims, where=where),
+      lax.convert_element_type(normalizer, computation_dtype)
+  ).astype(result_dtype)
 
 @overload
 def average(a: ArrayLike, axis: Axis = None, weights: Optional[ArrayLike] = None,
@@ -355,18 +358,18 @@ def average(a: ArrayLike, axis: Axis = None, weights: Optional[ArrayLike] = None
             returned: Literal[True], keepdims: bool = False) -> Array: ...
 @overload
 def average(a: ArrayLike, axis: Axis = None, weights: Optional[ArrayLike] = None,
-            returned: bool = False, keepdims: bool = False) -> Union[Array, Tuple[Array, Array]]: ...
+            returned: bool = False, keepdims: bool = False) -> Union[Array, tuple[Array, Array]]: ...
 @_wraps(np.average)
 def average(a: ArrayLike, axis: Axis = None, weights: Optional[ArrayLike] = None,
-            returned: bool = False, keepdims: bool = False) -> Union[Array, Tuple[Array, Array]]:
+            returned: bool = False, keepdims: bool = False) -> Union[Array, tuple[Array, Array]]:
   return _average(a, _ensure_optional_axes(axis), weights, returned, keepdims)
 
 @partial(api.jit, static_argnames=('axis', 'returned', 'keepdims'), inline=True)
 def _average(a: ArrayLike, axis: Axis = None, weights: Optional[ArrayLike] = None,
-             returned: bool = False, keepdims: bool = False) -> Union[Array, Tuple[Array, Array]]:
+             returned: bool = False, keepdims: bool = False) -> Union[Array, tuple[Array, Array]]:
   if weights is None: # Treat all weights as 1
-    _check_arraylike("average", a)
-    a, = _promote_dtypes_inexact(a)
+    check_arraylike("average", a)
+    a, = promote_dtypes_inexact(a)
     avg = mean(a, axis=axis, keepdims=keepdims)
     if axis is None:
       weights_sum = lax.full((), core.dimension_as_value(a.size), dtype=avg.dtype)
@@ -375,8 +378,8 @@ def _average(a: ArrayLike, axis: Axis = None, weights: Optional[ArrayLike] = Non
     else:
       weights_sum = lax.full_like(avg, core.dimension_as_value(a.shape[axis]))  # type: ignore[index]
   else:
-    _check_arraylike("average", a, weights)
-    a, weights = _promote_dtypes_inexact(a, weights)
+    check_arraylike("average", a, weights)
+    a, weights = promote_dtypes_inexact(a, weights)
 
     a_shape = np.shape(a)
     a_ndim = len(a_shape)
@@ -399,7 +402,7 @@ def _average(a: ArrayLike, axis: Axis = None, weights: Optional[ArrayLike] = Non
                          "weights differ.")
       elif isinstance(axis, tuple):
         raise ValueError("Single axis expected when shapes of a and weights differ")
-      elif not core.symbolic_equal_dim(weights_shape[0], a_shape[axis]):
+      elif not core.definitely_equal(weights_shape[0], a_shape[axis]):
         raise ValueError("Length of weights not "
                          "compatible with specified axis.")
 
@@ -427,17 +430,18 @@ def var(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None,
 def _var(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None,
          out: None = None, ddof: int = 0, keepdims: bool = False, *,
          where: Optional[ArrayLike] = None) -> Array:
-  _check_arraylike("var", a)
+  check_arraylike("var", a)
   dtypes.check_user_dtype_supported(dtype, "var")
   if out is not None:
     raise NotImplementedError("The 'out' argument to jnp.var is not supported.")
 
   computation_dtype, dtype = _var_promote_types(dtypes.dtype(a), dtype)
-  a = _asarray(a).astype(computation_dtype)
+  a = lax_internal.asarray(a).astype(computation_dtype)
   a_mean = mean(a, axis, dtype=computation_dtype, keepdims=True, where=where)
   centered = lax.sub(a, a_mean)
-  if dtypes.issubdtype(centered.dtype, np.complexfloating):
+  if dtypes.issubdtype(computation_dtype, np.complexfloating):
     centered = lax.real(lax.mul(centered, lax.conj(centered)))
+    computation_dtype = centered.dtype  # avoid casting to complex below.
   else:
     centered = lax.square(centered)
 
@@ -446,16 +450,16 @@ def _var(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None,
       normalizer = core.dimension_as_value(np.size(a))
     else:
       normalizer = core.dimension_as_value(_axis_size(a, axis))
+    normalizer = lax.convert_element_type(normalizer, computation_dtype)
   else:
-    normalizer = sum(_broadcast_to(where, np.shape(a)), axis, dtype=dtype, keepdims=keepdims)
-  normalizer = normalizer - ddof
-
-  result = sum(centered, axis, keepdims=keepdims, where=where)
-  result = lax.div(result, lax.convert_element_type(normalizer, result.dtype))
-  return lax.convert_element_type(result, dtype)
+    normalizer = sum(_broadcast_to(where, np.shape(a)), axis,
+                     dtype=computation_dtype, keepdims=keepdims)
+  normalizer = lax.sub(normalizer, lax.convert_element_type(ddof, computation_dtype))
+  result = sum(centered, axis, dtype=computation_dtype, keepdims=keepdims, where=where)
+  return lax.div(result, normalizer).astype(dtype)
 
 
-def _var_promote_types(a_dtype: DTypeLike, dtype: DTypeLike) -> Tuple[DType, DType]:
+def _var_promote_types(a_dtype: DTypeLike, dtype: DTypeLike) -> tuple[DType, DType]:
   if dtype:
     if (not dtypes.issubdtype(dtype, np.complexfloating) and
         dtypes.issubdtype(a_dtype, np.complexfloating)):
@@ -487,8 +491,10 @@ def std(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None,
 def _std(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None,
          out: None = None, ddof: int = 0, keepdims: bool = False, *,
          where: Optional[ArrayLike] = None) -> Array:
-  _check_arraylike("std", a)
+  check_arraylike("std", a)
   dtypes.check_user_dtype_supported(dtype, "std")
+  if dtype is not None and not dtypes.issubdtype(dtype, np.inexact):
+    raise ValueError(f"dtype argument to jnp.std must be inexact; got {dtype}")
   if out is not None:
     raise NotImplementedError("The 'out' argument to jnp.std is not supported.")
   return lax.sqrt(var(a, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims, where=where))
@@ -502,7 +508,7 @@ def ptp(a: ArrayLike, axis: Axis = None, out: None = None,
 @partial(api.jit, static_argnames=('axis', 'keepdims'))
 def _ptp(a: ArrayLike, axis: Axis = None, out: None = None,
          keepdims: bool = False) -> Array:
-  _check_arraylike("ptp", a)
+  check_arraylike("ptp", a)
   if out is not None:
     raise NotImplementedError("The 'out' argument to jnp.ptp is not supported.")
   x = amax(a, axis=axis, keepdims=keepdims)
@@ -514,15 +520,15 @@ def _ptp(a: ArrayLike, axis: Axis = None, out: None = None,
 @partial(api.jit, static_argnames=('axis', 'keepdims'))
 def count_nonzero(a: ArrayLike, axis: Axis = None,
                   keepdims: bool = False) -> Array:
-  _check_arraylike("count_nonzero", a)
+  check_arraylike("count_nonzero", a)
   return sum(lax.ne(a, _lax_const(a, 0)), axis=axis,
-             dtype=dtypes.canonicalize_dtype(np.int_), keepdims=keepdims)
+             dtype=dtypes.canonicalize_dtype(int), keepdims=keepdims)
 
 
 def _nan_reduction(a: ArrayLike, name: str, jnp_reduction: Callable[..., Array],
                    init_val: ArrayLike, nan_if_all_nan: bool,
                    axis: Axis = None, keepdims: bool = False, **kwargs) -> Array:
-  _check_arraylike(name, a)
+  check_arraylike(name, a)
   if not dtypes.issubdtype(dtypes.dtype(a), np.inexact):
     return jnp_reduction(a, axis=axis, keepdims=keepdims, **kwargs)
 
@@ -580,7 +586,7 @@ def nanprod(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None, out: None
 @partial(api.jit, static_argnames=('axis', 'dtype', 'keepdims'))
 def nanmean(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None, out: None = None,
             keepdims: bool = False, where: Optional[ArrayLike] = None) -> Array:
-  _check_arraylike("nanmean", a)
+  check_arraylike("nanmean", a)
   dtypes.check_user_dtype_supported(dtype, "nanmean")
   if out is not None:
     raise NotImplementedError("The 'out' argument to jnp.nanmean is not supported.")
@@ -600,13 +606,13 @@ def nanmean(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None, out: None
 def nanvar(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None, out: None = None,
            ddof: int = 0, keepdims: bool = False,
            where: Optional[ArrayLike] = None) -> Array:
-  _check_arraylike("nanvar", a)
+  check_arraylike("nanvar", a)
   dtypes.check_user_dtype_supported(dtype, "nanvar")
   if out is not None:
     raise NotImplementedError("The 'out' argument to jnp.nanvar is not supported.")
 
   computation_dtype, dtype = _var_promote_types(dtypes.dtype(a), dtype)
-  a = _asarray(a).astype(computation_dtype)
+  a = lax_internal.asarray(a).astype(computation_dtype)
   a_mean = nanmean(a, axis, dtype=computation_dtype, keepdims=True, where=where)
 
   centered = _where(lax_internal._isnan(a), 0, lax.sub(a, a_mean))  # double-where trick for gradients.
@@ -631,16 +637,20 @@ def nanvar(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None, out: None =
 def nanstd(a: ArrayLike, axis: Axis = None, dtype: DTypeLike = None, out: None = None,
            ddof: int = 0, keepdims: bool = False,
            where: Optional[ArrayLike] = None) -> Array:
-  _check_arraylike("nanstd", a)
+  check_arraylike("nanstd", a)
   dtypes.check_user_dtype_supported(dtype, "nanstd")
   if out is not None:
     raise NotImplementedError("The 'out' argument to jnp.nanstd is not supported.")
   return lax.sqrt(nanvar(a, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims, where=where))
 
 
-# TODO(jakevdp): use a protocol here for better typing?
+class CumulativeReduction(Protocol):
+  def __call__(self, a: ArrayLike, axis: Axis = None,
+               dtype: DTypeLike = None, out: None = None) -> Array: ...
+
+
 def _make_cumulative_reduction(np_reduction: Any, reduction: Callable[..., Array],
-                               fill_nan: bool = False, fill_value: ArrayLike = 0) -> Callable[..., Array]:
+                               fill_nan: bool = False, fill_value: ArrayLike = 0) -> CumulativeReduction:
   @_wraps(np_reduction, skip_params=['out'])
   def cumulative_reduction(a: ArrayLike, axis: Axis = None,
                            dtype: DTypeLike = None, out: None = None) -> Array:
@@ -649,7 +659,7 @@ def _make_cumulative_reduction(np_reduction: Any, reduction: Callable[..., Array
   @partial(api.jit, static_argnames=('axis', 'dtype'))
   def _cumulative_reduction(a: ArrayLike, axis: Axis = None,
                             dtype: DTypeLike = None, out: None = None) -> Array:
-    _check_arraylike(np_reduction.__name__, a)
+    check_arraylike(np_reduction.__name__, a)
     if out is not None:
       raise NotImplementedError(f"The 'out' argument to jnp.{np_reduction.__name__} "
                                 f"is not supported.")
@@ -657,6 +667,7 @@ def _make_cumulative_reduction(np_reduction: Any, reduction: Callable[..., Array
 
     if axis is None or _isscalar(a):
       a = lax.reshape(a, (np.size(a),))
+    if axis is None:
       axis = 0
 
     a_shape = list(np.shape(a))
@@ -678,8 +689,205 @@ def _make_cumulative_reduction(np_reduction: Any, reduction: Callable[..., Array
 
 cumsum = _make_cumulative_reduction(np.cumsum, lax.cumsum, fill_nan=False)
 cumprod = _make_cumulative_reduction(np.cumprod, lax.cumprod, fill_nan=False)
-cumproduct = cumprod
 nancumsum = _make_cumulative_reduction(np.nancumsum, lax.cumsum,
                                        fill_nan=True, fill_value=0)
 nancumprod = _make_cumulative_reduction(np.nancumprod, lax.cumprod,
                                         fill_nan=True, fill_value=1)
+
+# Quantiles
+@_wraps(np.quantile, skip_params=['out', 'overwrite_input'])
+@partial(api.jit, static_argnames=('axis', 'overwrite_input', 'interpolation',
+                               'keepdims', 'method'))
+def quantile(a: ArrayLike, q: ArrayLike, axis: Optional[Union[int, tuple[int, ...]]] = None,
+             out: None = None, overwrite_input: bool = False, method: str = "linear",
+             keepdims: bool = False, interpolation: None = None) -> Array:
+  check_arraylike("quantile", a, q)
+  if overwrite_input or out is not None:
+    msg = ("jax.numpy.quantile does not support overwrite_input=True or "
+           "out != None")
+    raise ValueError(msg)
+  if interpolation is not None:
+    warnings.warn("The interpolation= argument to 'quantile' is deprecated. "
+                  "Use 'method=' instead.", DeprecationWarning)
+  return _quantile(lax_internal.asarray(a), lax_internal.asarray(q), axis, interpolation or method, keepdims, False)
+
+@_wraps(np.nanquantile, skip_params=['out', 'overwrite_input'])
+@partial(api.jit, static_argnames=('axis', 'overwrite_input', 'interpolation',
+                               'keepdims', 'method'))
+def nanquantile(a: ArrayLike, q: ArrayLike, axis: Optional[Union[int, tuple[int, ...]]] = None,
+                out: None = None, overwrite_input: bool = False, method: str = "linear",
+                keepdims: bool = False, interpolation: None = None) -> Array:
+  check_arraylike("nanquantile", a, q)
+  if overwrite_input or out is not None:
+    msg = ("jax.numpy.nanquantile does not support overwrite_input=True or "
+           "out != None")
+    raise ValueError(msg)
+  if interpolation is not None:
+    warnings.warn("The interpolation= argument to 'nanquantile' is deprecated. "
+                  "Use 'method=' instead.", DeprecationWarning)
+  return _quantile(lax_internal.asarray(a), lax_internal.asarray(q), axis, interpolation or method, keepdims, True)
+
+def _quantile(a: Array, q: Array, axis: Optional[Union[int, tuple[int, ...]]],
+              interpolation: str, keepdims: bool, squash_nans: bool) -> Array:
+  if interpolation not in ["linear", "lower", "higher", "midpoint", "nearest"]:
+    raise ValueError("interpolation can only be 'linear', 'lower', 'higher', "
+                     "'midpoint', or 'nearest'")
+  a, = promote_dtypes_inexact(a)
+  keepdim = []
+  if dtypes.issubdtype(a.dtype, np.complexfloating):
+    raise ValueError("quantile does not support complex input, as the operation is poorly defined.")
+  if axis is None:
+    a = a.ravel()
+    axis = 0
+  elif isinstance(axis, tuple):
+    keepdim = list(a.shape)
+    nd = a.ndim
+    axis = tuple(_canonicalize_axis(ax, nd) for ax in axis)
+    if len(set(axis)) != len(axis):
+      raise ValueError('repeated axis')
+    for ax in axis:
+      keepdim[ax] = 1
+
+    keep = set(range(nd)) - set(axis)
+    # prepare permutation
+    dimensions = list(range(nd))
+    for i, s in enumerate(sorted(keep)):
+      dimensions[i], dimensions[s] = dimensions[s], dimensions[i]
+    do_not_touch_shape = tuple(x for idx,x in enumerate(a.shape) if idx not in axis)
+    touch_shape = tuple(x for idx,x in enumerate(a.shape) if idx in axis)
+    a = lax.reshape(a, do_not_touch_shape + (math.prod(touch_shape),), dimensions)
+    axis = _canonicalize_axis(-1, a.ndim)
+  else:
+    axis = _canonicalize_axis(axis, a.ndim)
+
+  q_shape = q.shape
+  q_ndim = q.ndim
+  if q_ndim > 1:
+    raise ValueError(f"q must be have rank <= 1, got shape {q.shape}")
+
+  a_shape = a.shape
+
+  if squash_nans:
+    a = _where(ufuncs.isnan(a), np.nan, a) # Ensure nans are positive so they sort to the end.
+    a = lax.sort(a, dimension=axis)
+    counts = sum(ufuncs.logical_not(ufuncs.isnan(a)), axis=axis, dtype=q.dtype, keepdims=keepdims)
+    shape_after_reduction = counts.shape
+    q = lax.expand_dims(
+      q, tuple(range(q_ndim, len(shape_after_reduction) + q_ndim)))
+    counts = lax.expand_dims(counts, tuple(range(q_ndim)))
+    q = lax.mul(q, lax.sub(counts, _lax_const(q, 1)))
+    low = lax.floor(q)
+    high = lax.ceil(q)
+    high_weight = lax.sub(q, low)
+    low_weight = lax.sub(_lax_const(high_weight, 1), high_weight)
+
+    low = lax.max(_lax_const(low, 0), lax.min(low, counts - 1))
+    high = lax.max(_lax_const(high, 0), lax.min(high, counts - 1))
+    low = lax.convert_element_type(low, int)
+    high = lax.convert_element_type(high, int)
+    out_shape = q_shape + shape_after_reduction
+    index = [lax.broadcasted_iota(int, out_shape, dim + q_ndim)
+             for dim in range(len(shape_after_reduction))]
+    if keepdims:
+      index[axis] = low
+    else:
+      index.insert(axis, low)
+    low_value = a[tuple(index)]
+    index[axis] = high
+    high_value = a[tuple(index)]
+  else:
+    a = _where(any(ufuncs.isnan(a), axis=axis, keepdims=True), np.nan, a)
+    a = lax.sort(a, dimension=axis)
+    n = lax.convert_element_type(a_shape[axis], lax_internal._dtype(q))
+    q = lax.mul(q, n - 1)
+    low = lax.floor(q)
+    high = lax.ceil(q)
+    high_weight = lax.sub(q, low)
+    low_weight = lax.sub(_lax_const(high_weight, 1), high_weight)
+
+    low = lax.clamp(_lax_const(low, 0), low, n - 1)
+    high = lax.clamp(_lax_const(high, 0), high, n - 1)
+    low = lax.convert_element_type(low, int)
+    high = lax.convert_element_type(high, int)
+
+    slice_sizes = list(a_shape)
+    slice_sizes[axis] = 1
+    dnums = lax.GatherDimensionNumbers(
+      offset_dims=tuple(range(
+        q_ndim,
+        len(a_shape) + q_ndim if keepdims else len(a_shape) + q_ndim - 1)),
+      collapsed_slice_dims=() if keepdims else (axis,),
+      start_index_map=(axis,))
+    low_value = lax.gather(a, low[..., None], dimension_numbers=dnums,
+                           slice_sizes=slice_sizes)
+    high_value = lax.gather(a, high[..., None], dimension_numbers=dnums,
+                            slice_sizes=slice_sizes)
+    if q_ndim == 1:
+      low_weight = lax.broadcast_in_dim(low_weight, low_value.shape,
+                                        broadcast_dimensions=(0,))
+      high_weight = lax.broadcast_in_dim(high_weight, high_value.shape,
+                                        broadcast_dimensions=(0,))
+
+  if interpolation == "linear":
+    result = lax.add(lax.mul(low_value.astype(q.dtype), low_weight),
+                     lax.mul(high_value.astype(q.dtype), high_weight))
+  elif interpolation == "lower":
+    result = low_value
+  elif interpolation == "higher":
+    result = high_value
+  elif interpolation == "nearest":
+    pred = lax.le(high_weight, _lax_const(high_weight, 0.5))
+    result = lax.select(pred, low_value, high_value)
+  elif interpolation == "midpoint":
+    result = lax.mul(lax.add(low_value, high_value), _lax_const(low_value, 0.5))
+  else:
+    raise ValueError(f"interpolation={interpolation!r} not recognized")
+  if keepdims and keepdim:
+    if q_ndim > 0:
+      keepdim = [np.shape(q)[0], *keepdim]
+    result = result.reshape(keepdim)
+  return lax.convert_element_type(result, a.dtype)
+
+@_wraps(np.percentile, skip_params=['out', 'overwrite_input'])
+@partial(api.jit, static_argnames=('axis', 'overwrite_input', 'interpolation',
+                                   'keepdims', 'method'))
+def percentile(a: ArrayLike, q: ArrayLike,
+               axis: Optional[Union[int, tuple[int, ...]]] = None,
+               out: None = None, overwrite_input: bool = False, method: str = "linear",
+               keepdims: bool = False, interpolation: None = None) -> Array:
+  check_arraylike("percentile", a, q)
+  q, = promote_dtypes_inexact(q)
+  return quantile(a, q / 100, axis=axis, out=out, overwrite_input=overwrite_input,
+                  interpolation=interpolation, method=method, keepdims=keepdims)
+
+@_wraps(np.nanpercentile, skip_params=['out', 'overwrite_input'])
+@partial(api.jit, static_argnames=('axis', 'overwrite_input', 'interpolation',
+                               'keepdims', 'method'))
+def nanpercentile(a: ArrayLike, q: ArrayLike,
+                  axis: Optional[Union[int, tuple[int, ...]]] = None,
+                  out: None = None, overwrite_input: bool = False, method: str = "linear",
+                  keepdims: bool = False, interpolation: None = None) -> Array:
+  check_arraylike("nanpercentile", a, q)
+  q = ufuncs.true_divide(q, 100.0)
+  return nanquantile(a, q, axis=axis, out=out, overwrite_input=overwrite_input,
+                     interpolation=interpolation, method=method,
+                     keepdims=keepdims)
+
+@_wraps(np.median, skip_params=['out', 'overwrite_input'])
+@partial(api.jit, static_argnames=('axis', 'overwrite_input', 'keepdims'))
+def median(a: ArrayLike, axis: Optional[Union[int, tuple[int, ...]]] = None,
+           out: None = None, overwrite_input: bool = False,
+           keepdims: bool = False) -> Array:
+  check_arraylike("median", a)
+  return quantile(a, 0.5, axis=axis, out=out, overwrite_input=overwrite_input,
+                  keepdims=keepdims, method='midpoint')
+
+@_wraps(np.nanmedian, skip_params=['out', 'overwrite_input'])
+@partial(api.jit, static_argnames=('axis', 'overwrite_input', 'keepdims'))
+def nanmedian(a: ArrayLike, axis: Optional[Union[int, tuple[int, ...]]] = None,
+              out: None = None, overwrite_input: bool = False,
+              keepdims: bool = False) -> Array:
+  check_arraylike("nanmedian", a)
+  return nanquantile(a, 0.5, axis=axis, out=out,
+                     overwrite_input=overwrite_input, keepdims=keepdims,
+                     method='midpoint')
