@@ -16,7 +16,7 @@ from functools import partial
 import math
 import operator
 from textwrap import dedent as _dedent
-from typing import Optional, Tuple, Union
+from typing import Optional, Union, cast
 
 import numpy as np
 
@@ -90,15 +90,15 @@ def setdiff1d(ar1: ArrayLike, ar2: ArrayLike, assume_unique: bool = False,
   if arr1.size == 0:
     return full_like(arr1, fill_value, shape=size or 0)
   if not assume_unique:
-    arr1 = unique(arr1, size=size and arr1.size)
+    arr1 = cast(Array, unique(arr1, size=size and arr1.size))
   mask = in1d(arr1, ar2, invert=True)
   if size is None:
     return arr1[mask]
   else:
     if not (assume_unique or size is None):
       # Set mask to zero at locations corresponding to unique() padding.
-      n_unique = arr1.size + 1 - (arr1 == arr1[0]).sum()  # pytype: disable=attribute-error  # always-use-return-annotations
-      mask = where(arange(arr1.size) < n_unique, mask, False)  # pytype: disable=attribute-error  # always-use-return-annotations
+      n_unique = arr1.size + 1 - (arr1 == arr1[0]).sum()
+      mask = where(arange(arr1.size) < n_unique, mask, False)
     return where(arange(size) < mask.sum(), arr1[where(mask, size=size)], fill_value)
 
 
@@ -124,7 +124,9 @@ def union1d(ar1: ArrayLike, ar2: ArrayLike,
     ar2 = core.concrete_or_error(None, ar2, "The error arose in union1d()")
   else:
     size = core.concrete_or_error(operator.index, size, "The error arose in union1d()")
-  return unique(concatenate((ar1, ar2), axis=None), size=size, fill_value=fill_value)  # pytype: disable=bad-return-type  # always-use-return-annotations
+  out = unique(concatenate((ar1, ar2), axis=None), size=size,
+               fill_value=fill_value)
+  return cast(Array, out)
 
 
 @_wraps(np.setxor1d, lax_description="""
@@ -153,7 +155,7 @@ def setxor1d(ar1: ArrayLike, ar2: ArrayLike, assume_unique: bool = False) -> Arr
 
 
 @partial(jit, static_argnames=['return_indices'])
-def _intersect1d_sorted_mask(ar1: ArrayLike, ar2: ArrayLike, return_indices: bool = False) -> Tuple[Array, ...]:
+def _intersect1d_sorted_mask(ar1: ArrayLike, ar2: ArrayLike, return_indices: bool = False) -> tuple[Array, ...]:
   """
     Helper function for intersect1d which is jit-able
     """
@@ -173,7 +175,7 @@ def _intersect1d_sorted_mask(ar1: ArrayLike, ar2: ArrayLike, return_indices: boo
 
 @_wraps(np.intersect1d)
 def intersect1d(ar1: ArrayLike, ar2: ArrayLike, assume_unique: bool = False,
-                return_indices: bool = False) -> Union[Array, Tuple[Array, Array, Array]]:
+                return_indices: bool = False) -> Union[Array, tuple[Array, Array, Array]]:
   check_arraylike("intersect1d", ar1, ar2)
   ar1 = core.concrete_or_error(None, ar1, "The error arose in intersect1d()")
   ar2 = core.concrete_or_error(None, ar2, "The error arose in intersect1d()")
@@ -224,7 +226,7 @@ UNIQUE_SIZE_HINT = (
   "a concrete value for the size argument, which will determine the output size.")
 
 @partial(jit, static_argnums=1)
-def _unique_sorted_mask(ar: Array, axis: int) -> Tuple[Array, Array, Array]:
+def _unique_sorted_mask(ar: Array, axis: int) -> tuple[Array, Array, Array]:
   aux = moveaxis(ar, axis, 0)
   if np.issubdtype(aux.dtype, np.complexfloating):
     # Work around issue in sorting of complex numbers with Nan only in the
@@ -253,7 +255,7 @@ def _unique_sorted_mask(ar: Array, axis: int) -> Tuple[Array, Array, Array]:
 def _unique(ar: Array, axis: int, return_index: bool = False, return_inverse: bool = False,
             return_counts: bool = False, size: Optional[int] = None,
             fill_value: Optional[ArrayLike] = None, return_true_size: bool = False
-            ) -> Union[Array, Tuple[Array, ...]]:
+            ) -> Union[Array, tuple[Array, ...]]:
   """
   Find the unique elements of an array along a particular axis.
   """
@@ -278,7 +280,7 @@ def _unique(ar: Array, axis: int, return_index: bool = False, return_inverse: bo
       result = full_like(result, fill_value, shape=(size, *result.shape[1:]))
   result = moveaxis(result, 0, axis)
 
-  ret: Tuple[Array, ...] = (result,)
+  ret: tuple[Array, ...] = (result,)
   if return_index:
     if aux.size:
       ret += (perm[ind],)
