@@ -12,12 +12,12 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from collections.abc import Sequence
 from functools import partial
 import re
 import textwrap
-from typing import (
-    Any, Callable, Dict, List, NamedTuple, Optional, Sequence, Set, Type, TypeVar
-)
+from typing import Any, Callable, NamedTuple, Optional, TypeVar
+
 import warnings
 
 from jax._src import dtypes
@@ -53,7 +53,7 @@ class ParsedDoc(NamedTuple):
   signature: str = ""
   summary: str = ""
   front_matter: str = ""
-  sections: Dict[str, str] = {}
+  sections: dict[str, str] = {}
 
 
 def _parse_numpydoc(docstr: Optional[str]) -> ParsedDoc:
@@ -101,7 +101,7 @@ def _parse_numpydoc(docstr: Optional[str]) -> ParsedDoc:
                    front_matter=front_matter, sections=sections)
 
 
-def _parse_parameters(body: str) -> Dict[str, str]:
+def _parse_parameters(body: str) -> dict[str, str]:
   """Parse the Parameters section of a docstring."""
   title, underline, content = body.split('\n', 2)
   assert title == 'Parameters'
@@ -110,7 +110,7 @@ def _parse_parameters(body: str) -> Dict[str, str]:
   return {p.partition(' : ')[0].partition(', ')[0]: p for p in parameters}
 
 
-def _parse_extra_params(extra_params: str) -> Dict[str, str]:
+def _parse_extra_params(extra_params: str) -> dict[str, str]:
   """Parse the extra parameters passed to _wraps()"""
   parameters = _parameter_break.split(extra_params.strip('\n'))
   return {p.partition(' : ')[0].partition(', ')[0]: p for p in parameters}
@@ -223,19 +223,10 @@ def _wraps(
 
 _dtype = partial(dtypes.dtype, canonicalize=True)
 
-def _asarray(arr: ArrayLike) -> Array:
-  """
-  Pared-down utility to convert object to a DeviceArray.
-  Note this will not correctly handle lists or tuples.
-  """
-  check_arraylike("_asarray", arr)
-  dtype, weak_type = dtypes._lattice_result_type(arr)
-  return lax._convert_element_type(arr, dtype, weak_type)
-
-def promote_shapes(fun_name: str, *args: ArrayLike) -> List[Array]:
+def promote_shapes(fun_name: str, *args: ArrayLike) -> list[Array]:
   """Apply NumPy-style broadcasting, making args shape-compatible for lax.py."""
   if len(args) < 2:
-    return [_asarray(arg) for arg in args]
+    return [lax.asarray(arg) for arg in args]
   else:
     shapes = [np.shape(arg) for arg in args]
     if config.jax_dynamic_shapes:
@@ -246,10 +237,10 @@ def promote_shapes(fun_name: str, *args: ArrayLike) -> List[Array]:
       return [_broadcast_to(arg, res_shape) for arg, shp in zip(args, shapes)]
     else:
       if all(len(shapes[0]) == len(s) for s in shapes[1:]):
-        return [_asarray(arg) for arg in args]  # no need for rank promotion, so rely on lax promotion
+        return [lax.asarray(arg) for arg in args]  # no need for rank promotion, so rely on lax promotion
       nonscalar_ranks = {len(shp) for shp in shapes if shp}
       if len(nonscalar_ranks) < 2:
-        return [_asarray(arg) for arg in args]  # rely on lax scalar promotion
+        return [lax.asarray(arg) for arg in args]  # rely on lax scalar promotion
       else:
         if config.jax_numpy_rank_promotion != "allow":
           _rank_promotion_warning_or_error(fun_name, shapes)
@@ -273,29 +264,29 @@ def _rank_promotion_warning_or_error(fun_name: str, shapes: Sequence[Shape]):
     raise ValueError(msg.format(fun_name, ' '.join(map(str, shapes))))
 
 
-def promote_dtypes(*args: ArrayLike) -> List[Array]:
+def promote_dtypes(*args: ArrayLike) -> list[Array]:
   """Convenience function to apply Numpy argument dtype promotion."""
   # TODO(dougalm,mattjj): This is a performance bottleneck. Consider memoizing.
   if len(args) < 2:
-    return [_asarray(arg) for arg in args]
+    return [lax.asarray(arg) for arg in args]
   else:
     to_dtype, weak_type = dtypes._lattice_result_type(*args)
-    to_dtype = dtypes.canonicalize_dtype(to_dtype)
+    to_dtype = dtypes.canonicalize_dtype(to_dtype, allow_extended_dtype=True)  # type: ignore[assignment]
     return [lax._convert_element_type(x, to_dtype, weak_type) for x in args]
 
 
-def promote_dtypes_inexact(*args: ArrayLike) -> List[Array]:
+def promote_dtypes_inexact(*args: ArrayLike) -> list[Array]:
   """Convenience function to apply Numpy argument dtype promotion.
 
   Promotes arguments to an inexact type."""
   to_dtype, weak_type = dtypes._lattice_result_type(*args)
-  to_dtype = dtypes.canonicalize_dtype(to_dtype)
+  to_dtype = dtypes.canonicalize_dtype(to_dtype, allow_extended_dtype=True)  # type: ignore[assignment]
   to_dtype_inexact = dtypes.to_inexact_dtype(to_dtype)
   return [lax._convert_element_type(x, to_dtype_inexact, weak_type)
           for x in args]
 
 
-def promote_dtypes_numeric(*args: ArrayLike) -> List[Array]:
+def promote_dtypes_numeric(*args: ArrayLike) -> list[Array]:
   """Convenience function to apply Numpy argument dtype promotion.
 
   Promotes arguments to a numeric (non-bool) type."""
@@ -306,7 +297,7 @@ def promote_dtypes_numeric(*args: ArrayLike) -> List[Array]:
           for x in args]
 
 
-def promote_dtypes_complex(*args: ArrayLike) -> List[Array]:
+def promote_dtypes_complex(*args: ArrayLike) -> list[Array]:
   """Convenience function to apply Numpy argument dtype promotion.
 
   Promotes arguments to a complex type."""
@@ -327,20 +318,18 @@ def _arraylike(x: ArrayLike) -> bool:
           hasattr(x, '__jax_array__') or np.isscalar(x))
 
 
-def _stackable(*args: Any) -> bool:
-  return all(type(arg) in stackables for arg in args)
-stackables: Set[Type] = set()
-_register_stackable: Callable[[Type], None] = stackables.add
-
-
-def check_arraylike(fun_name: str, *args: Any):
+def check_arraylike(fun_name: str, *args: Any, emit_warning=False, stacklevel=3):
   """Check if all args fit JAX's definition of arraylike."""
   assert isinstance(fun_name, str), f"fun_name must be a string. Got {fun_name}"
   if any(not _arraylike(arg) for arg in args):
     pos, arg = next((i, arg) for i, arg in enumerate(args)
                     if not _arraylike(arg))
-    msg = "{} requires ndarray or scalar arguments, got {} at position {}."
-    raise TypeError(msg.format(fun_name, type(arg), pos))
+    msg = f"{fun_name} requires ndarray or scalar arguments, got {type(arg)} at position {pos}."
+    if emit_warning:
+      warnings.warn(msg + "In a future JAX release this will be an error.",
+                    category=DeprecationWarning, stacklevel=stacklevel)
+    else:
+      raise TypeError(msg.format(fun_name, type(arg), pos))
 
 
 def check_arraylike_or_none(fun_name: str, *args: Any):
@@ -352,7 +341,7 @@ def check_arraylike_or_none(fun_name: str, *args: Any):
     raise TypeError(msg.format(fun_name, type(arg), pos))
 
 
-def _check_no_float0s(fun_name: str, *args: Any):
+def check_no_float0s(fun_name: str, *args: Any):
   """Check if none of the args have dtype float0."""
   if any(dtypes.dtype(arg) == dtypes.float0 for arg in args):
     raise TypeError(
@@ -363,22 +352,23 @@ def _check_no_float0s(fun_name: str, *args: Any):
         "to cast a float0 array to a regular zeros array. \n"
         "If you didn't expect to get a float0 you might have accidentally "
         "taken a gradient with respect to an integer argument.")
+_check_no_float0s = check_no_float0s
 
 
-def promote_args(fun_name: str, *args: ArrayLike) -> List[Array]:
+def promote_args(fun_name: str, *args: ArrayLike) -> list[Array]:
   """Convenience function to apply Numpy argument shape and dtype promotion."""
   check_arraylike(fun_name, *args)
   _check_no_float0s(fun_name, *args)
   return promote_shapes(fun_name, *promote_dtypes(*args))
 
 
-def promote_args_numeric(fun_name: str, *args: ArrayLike) -> List[Array]:
+def promote_args_numeric(fun_name: str, *args: ArrayLike) -> list[Array]:
   check_arraylike(fun_name, *args)
   _check_no_float0s(fun_name, *args)
   return promote_shapes(fun_name, *promote_dtypes_numeric(*args))
 
 
-def promote_args_inexact(fun_name: str, *args: ArrayLike) -> List[Array]:
+def promote_args_inexact(fun_name: str, *args: ArrayLike) -> list[Array]:
   """Convenience function to apply Numpy argument shape and dtype promotion.
 
   Promotes non-inexact types to an inexact type."""
@@ -388,35 +378,33 @@ def promote_args_inexact(fun_name: str, *args: ArrayLike) -> List[Array]:
 
 
 @partial(api.jit, inline=True)
-def _broadcast_arrays(*args: ArrayLike) -> List[Array]:
+def _broadcast_arrays(*args: ArrayLike) -> list[Array]:
   """Like Numpy's broadcast_arrays but doesn't return views."""
   shapes = [np.shape(arg) for arg in args]
-  if not shapes or all(core.symbolic_equal_shape(shapes[0], s) for s in shapes):
-    return [_asarray(arg) for arg in args]
+  if not shapes or all(core.definitely_equal_shape(shapes[0], s) for s in shapes):
+    return [lax.asarray(arg) for arg in args]
   result_shape = lax.broadcast_shapes(*shapes)
   return [_broadcast_to(arg, result_shape) for arg in args]
 
 
 def _broadcast_to(arr: ArrayLike, shape: Shape) -> Array:
-  if hasattr(arr, "broadcast_to"):
-    return arr.broadcast_to(shape)  # type: ignore[union-attr]
   check_arraylike("broadcast_to", arr)
-  arr = arr if isinstance(arr, Array) else _asarray(arr)
+  arr = arr if isinstance(arr, Array) else lax.asarray(arr)
   if not isinstance(shape, tuple) and np.ndim(shape) == 0:
     shape = (shape,)
   shape = core.canonicalize_shape(shape)  # check that shape is concrete
   arr_shape = np.shape(arr)
-  if core.symbolic_equal_shape(arr_shape, shape):
+  if core.definitely_equal_shape(arr_shape, shape):
     return arr
   else:
     nlead = len(shape) - len(arr_shape)
     shape_tail = shape[nlead:]
-    compatible = all(core.symbolic_equal_one_of_dim(arr_d, [1, shape_d])
+    compatible = all(core.definitely_equal_one_of_dim(arr_d, [1, shape_d])
                      for arr_d, shape_d in safe_zip(arr_shape, shape_tail))
     if nlead < 0 or not compatible:
       msg = "Incompatible shapes for broadcasting: {} and requested shape {}"
       raise ValueError(msg.format(arr_shape, shape))
-    diff, = np.where(tuple(not core.symbolic_equal_dim(arr_d, shape_d)
+    diff, = np.where(tuple(not core.definitely_equal(arr_d, shape_d)
                            for arr_d, shape_d in safe_zip(arr_shape, shape_tail)))
     new_dims = tuple(range(nlead)) + tuple(nlead + diff)
     kept_dims = tuple(np.delete(np.arange(len(shape)), new_dims))
