@@ -14,12 +14,14 @@
 
 # Helpers for indexed updates.
 
+from collections.abc import Sequence
 import sys
-from typing import Any, Callable, Optional, Sequence, Tuple, Union
+from typing import Callable, Optional, Union
 import warnings
 
 import numpy as np
 
+from jax import config
 from jax import lax
 
 from jax._src import core
@@ -29,15 +31,15 @@ from jax._src.lax import lax as lax_internal
 from jax._src.numpy import lax_numpy as jnp
 from jax._src.numpy import reductions
 from jax._src.numpy.util import check_arraylike, promote_dtypes
+from jax._src.typing import Array, ArrayLike
 
 
-Array = Any
 if sys.version_info >= (3, 10):
     from types import EllipsisType
     SingleIndex = Union[None, int, slice, Sequence[int], Array, EllipsisType]
 else:
     SingleIndex = Union[None, int, slice, Sequence[int], Array]
-Index = Union[SingleIndex, Tuple[SingleIndex, ...]]
+Index = Union[SingleIndex, tuple[SingleIndex, ...]]
 Scalar = Union[complex, float, int, np.number]
 
 
@@ -87,10 +89,11 @@ def _scatter_impl(x, y, scatter_op, treedef, static_idx, dynamic_idx,
   dtype = lax.dtype(x)
   weak_type = dtypes.is_weakly_typed(x)
 
-  if dtype != lax.dtype(y) and dtype != dtypes.result_type(x, y):
+  if not dtypes.safe_to_cast(y, x):
     # TODO(jakevdp): change this to an error after the deprecation period.
     warnings.warn("scatter inputs have incompatible types: cannot safely cast "
-                  f"value from dtype={lax.dtype(y)} to dtype={lax.dtype(x)}. "
+                  f"value from dtype={lax.dtype(y)} to dtype={lax.dtype(x)} "
+                  f"with jax_numpy_dtype_promotion={config.jax_numpy_dtype_promotion!r}. "
                   "In future JAX releases this will result in an error.",
                   FutureWarning)
 
@@ -151,8 +154,8 @@ def _get_identity(op, dtype):
 
 
 def _segment_update(name: str,
-                    data: Array,
-                    segment_ids: Array,
+                    data: ArrayLike,
+                    segment_ids: ArrayLike,
                     scatter_op: Callable,
                     num_segments: Optional[int] = None,
                     indices_are_sorted: bool = False,
@@ -192,8 +195,8 @@ def _segment_update(name: str,
   return reducer(out, axis=0).astype(dtype)
 
 
-def segment_sum(data: Array,
-                segment_ids: Array,
+def segment_sum(data: ArrayLike,
+                segment_ids: ArrayLike,
                 num_segments: Optional[int] = None,
                 indices_are_sorted: bool = False,
                 unique_indices: bool = False,
@@ -247,8 +250,8 @@ def segment_sum(data: Array,
       indices_are_sorted, unique_indices, bucket_size, reductions.sum, mode=mode)
 
 
-def segment_prod(data: Array,
-                 segment_ids: Array,
+def segment_prod(data: ArrayLike,
+                 segment_ids: ArrayLike,
                  num_segments: Optional[int] = None,
                  indices_are_sorted: bool = False,
                  unique_indices: bool = False,
@@ -303,8 +306,8 @@ def segment_prod(data: Array,
       indices_are_sorted, unique_indices, bucket_size, reductions.prod, mode=mode)
 
 
-def segment_max(data: Array,
-                segment_ids: Array,
+def segment_max(data: ArrayLike,
+                segment_ids: ArrayLike,
                 num_segments: Optional[int] = None,
                 indices_are_sorted: bool = False,
                 unique_indices: bool = False,
@@ -358,8 +361,8 @@ def segment_max(data: Array,
       indices_are_sorted, unique_indices, bucket_size, reductions.max, mode=mode)
 
 
-def segment_min(data: Array,
-                segment_ids: Array,
+def segment_min(data: ArrayLike,
+                segment_ids: ArrayLike,
                 num_segments: Optional[int] = None,
                 indices_are_sorted: bool = False,
                 unique_indices: bool = False,
