@@ -12,18 +12,19 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from collections.abc import Sequence, Iterable
 import dataclasses
+from functools import partial, lru_cache
 import inspect
+import itertools as it
 import logging
 import weakref
-import numpy as np
-from typing import (Callable, Sequence, Tuple, Union, cast, List, Optional,
-                    Iterable, NamedTuple, Any)
-import itertools as it
-from functools import partial, lru_cache
+from typing import Callable, Union, cast, Optional, NamedTuple, Any
 import threading
 import warnings
 
+import numpy as np
+
 from jax._src import core
 from jax._src import stages
 from jax._src import dispatch
@@ -32,6 +33,7 @@ from jax._src import linear_util as lu
 from jax._src import op_shardings
 from jax._src import sharding_impls
 from jax._src import source_info_util
+from jax._src import tree_util
 from jax._src import traceback_util
 from jax._src import api
 from jax._src import xla_bridge as xb
@@ -52,6 +54,7 @@ from jax._src.interpreters import pxla
 from jax._src.lib.mlir import ir
 from jax._src.lib.mlir.dialects import func as func_dialect
 from jax._src.lib import xla_client as xc
+from jax._src.lib import xla_extension_version
 from jax._src.sharding_impls import (
     NamedSharding, XLACompatibleSharding, GSPMDSharding,
     XLADeviceAssignment, SingleDeviceSharding, PmapSharding,
@@ -96,16 +99,16 @@ def _find_arg_mismatch(arg_list, fails, fun_name):
     if first_err.m_type == pxla.MismatchType.ARG_SHARDING:
       if first_err.da == inp_da:
         mismatched_args_msg.append(
-            (f"argument {name} of {fun_name} with shape {aval.str_short()} and "
-             f"{first_err._dev_ids_plat_str}"))
+            f"argument {name} of {fun_name} with shape {aval.str_short()} and "
+             f"{first_err._dev_ids_plat_str}")
         break
 
   for name, inp_da, aval in arg_list:
     if second_err.m_type == pxla.MismatchType.ARG_SHARDING:
       if second_err.da == inp_da:
         mismatched_args_msg.append(
-            (f"argument {name} of {fun_name} with shape {aval.str_short()} and "
-             f"{second_err._dev_ids_plat_str}"))
+            f"argument {name} of {fun_name} with shape {aval.str_short()} and "
+             f"{second_err._dev_ids_plat_str}")
         break
   return mismatched_args_msg
 
@@ -155,7 +158,7 @@ def _device_assignment_mismatch_error(fun_name, fails, args_flat, api_name,
 
 
 def _python_pjit_helper(fun, infer_params_fn, *args, **kwargs):
-  args_flat, _, params, in_tree, out_tree, _, jaxpr = infer_params_fn(
+  args_flat, _, params, in_tree, out_tree, _ = infer_params_fn(
       *args, **kwargs)
   for arg in args_flat:
     dispatch.check_arg(arg)
@@ -170,7 +173,7 @@ def _python_pjit_helper(fun, infer_params_fn, *args, **kwargs):
         fun_name, fails, args_flat, api_name, arg_names)
     raise ValueError(msg) from None
   outs = tree_unflatten(out_tree, out_flat)
-  return outs, out_flat, out_tree, args_flat, jaxpr
+  return outs, out_flat, out_tree, args_flat, params['jaxpr']
 
 
 def _python_pjit(fun: Callable, infer_params_fn):
@@ -189,6 +192,8 @@ def _python_pjit(fun: Callable, infer_params_fn):
 
 
 def _get_fastpath_data(executable, out_tree, args_flat, out_flat):
+  out_flat, out_tree = pxla.reflatten_outputs_for_dispatch(out_tree, out_flat)
+
   use_fastpath = (
       executable is not None and
       isinstance(executable, pxla.MeshExecutable) and
@@ -253,10 +258,11 @@ def _cpp_pjit(fun: Callable, infer_params_fn, static_argnums, static_argnames,
     fastpath_data = _get_fastpath_data(executable, out_tree, args_flat, out_flat)
     return outs, fastpath_data
 
-  cpp_pjit_f = xc._xla.pjit(  # type: ignore
-      getattr(fun, "__name__", "<unnamed function>"),  # type: ignore
-      fun, cache_miss, static_argnums, static_argnames,  # type: ignore
-      donate_argnums, _get_cpp_global_cache(pjit_has_explicit_sharding))  # type: ignore
+  cpp_pjit_f = xc._xla.pjit(
+    getattr(fun, "__name__", "<unnamed function>"),
+    fun, cache_miss, static_argnums, static_argnames,
+    donate_argnums, tree_util.dispatch_registry,
+    _get_cpp_global_cache(pjit_has_explicit_sharding))
 
   cpp_pjitted_f = wraps(fun)(cpp_pjit_f)
   cpp_pjitted_f._fun = fun
@@ -264,37 +270,9 @@ def _cpp_pjit(fun: Callable, infer_params_fn, static_argnums, static_argnames,
   return cpp_pjitted_f
 
 
-def _resolve_axis_resources_and_shardings_arg(
-    in_shardings, out_shardings, in_axis_resources, out_axis_resources):
-  if not is_unspecified(in_shardings) and not is_unspecified(in_axis_resources):
-    raise ValueError(
-        'Setting both in_shardings and in_axis_resources is not '
-        'allowed. in_axis_resources is deprecated. Please use in_shardings.')
-  if not is_unspecified(out_shardings) and not is_unspecified(out_axis_resources):
-    raise ValueError(
-        'Setting both out_shardings and out_axis_resources is not '
-        'allowed. out_axis_resources is deprecated. Please use out_shardings.')
-  if (not is_unspecified(in_axis_resources) or
-      not is_unspecified(out_axis_resources)):
-    warnings.warn(
-        'in_axis_resources and out_axis_resources are deprecated. Please use '
-        'in_shardings and out_shardings as their replacement.',
-        DeprecationWarning)
-
-  if not is_unspecified(in_axis_resources):
-    final_in_shardings = in_axis_resources
-  else:
-    final_in_shardings = in_shardings
-
-  if not is_unspecified(out_axis_resources):
-    final_out_shardings = out_axis_resources
-  else:
-    final_out_shardings = out_shardings
-  return final_in_shardings, final_out_shardings
-
-
 def pre_infer_params(fun, in_shardings, out_shardings,
-                     donate_argnums, static_argnums, static_argnames, device,
+                     donate_argnums, donate_argnames,
+                     static_argnums, static_argnames, device,
                      backend, abstracted_axes):
   if abstracted_axes and not config.jax_dynamic_shapes:
     raise ValueError("abstracted_axes must be used with --jax_dynamic_shapes")
@@ -311,10 +289,10 @@ def pre_infer_params(fun, in_shardings, out_shardings,
     if device is not None and backend is not None:
       raise ValueError("can't specify both a device and a backend for jit, "
                        f"got {device=} and {backend=}")
-    if not is_unspecified(in_shardings):
+    if in_shardings is not None and not is_unspecified(in_shardings):
       raise ValueError('If backend or device is specified on jit, then '
                        'in_shardings should not be specified.')
-    if not is_unspecified(out_shardings):
+    if out_shardings is not None and not is_unspecified(out_shardings):
       raise ValueError('If backend or device is specified on jit, then '
                        'out_shardings should not be specified.')
 
@@ -329,11 +307,11 @@ def pre_infer_params(fun, in_shardings, out_shardings,
   in_shardings, _, _ = prepare_axis_resources(in_shardings, 'in_shardings')
   out_shardings, _, _ = prepare_axis_resources(out_shardings, 'out_shardings')
 
-  donate_argnums, static_argnums, static_argnames = resolve_argnums(
-      fun, donate_argnums, static_argnums, static_argnames)
+  donate_argnums, donate_argnames, static_argnums, static_argnames = resolve_argnums(
+      fun, donate_argnums, donate_argnames, static_argnums, static_argnames)
 
-  return (in_shardings, out_shardings, donate_argnums, static_argnums,
-          static_argnames)
+  return (in_shardings, out_shardings, donate_argnums, donate_argnames,
+          static_argnums, static_argnames)
 
 
 def post_infer_params(fun, infer_params_fn, static_argnums, static_argnames,
@@ -347,10 +325,10 @@ def post_infer_params(fun, infer_params_fn, static_argnums, static_argnames,
 
   @api_boundary
   def lower(*args, **kwargs):
-    _experimental_lowering_platform = kwargs.pop(
-        '_experimental_lowering_platform', None)
+    lowering_parameters = kwargs.pop(
+        '_experimental_lowering_parameters', mlir.LoweringParameters())
     (args_flat, flat_global_in_avals, params, in_tree, out_tree,
-     donate_argnums, _) = infer_params_fn(*args, **kwargs)
+     donated_invars) = infer_params_fn(*args, **kwargs)
     resource_env = params['resource_env']
     mesh = None if resource_env is None else resource_env.physical_mesh
     try:
@@ -359,8 +337,8 @@ def post_infer_params(fun, infer_params_fn, static_argnums, static_argnames,
       lowering = _pjit_lower(
           params['jaxpr'], in_shardings, params['out_shardings'],
           params['resource_env'], params['donated_invars'], params['name'],
-          params['keep_unused'], params['inline'], always_lower=True,
-          lowering_platform=_experimental_lowering_platform)
+          params['keep_unused'], params['inline'],
+          lowering_parameters=lowering_parameters)
     except pxla.DeviceAssignmentMismatchError as e:
       fails, = e.args
       api_name = 'jit' if params['resource_env'] is None else 'pjit'
@@ -375,6 +353,7 @@ def post_infer_params(fun, infer_params_fn, static_argnums, static_argnames,
     else:
       args_kwargs_in_tree = treedef_tuple([in_tree, tree_flatten({})[1]])
 
+    donate_argnums = tuple(i for i, d in enumerate(donated_invars) if d)
     return stages.Lowered.from_flat_info(
         lowering, args_kwargs_in_tree, flat_global_in_avals, donate_argnums,
         out_tree)
@@ -397,9 +376,10 @@ class PjitInfo(NamedTuple):
   fun: Callable
   in_shardings: Any
   out_shardings: Any
-  static_argnums: Tuple[int, ...]
-  static_argnames: Tuple[str, ...]
-  donate_argnums: Tuple[int, ...]
+  static_argnums: tuple[int, ...]
+  static_argnames: tuple[str, ...]
+  donate_argnums: tuple[int, ...]
+  donate_argnames: tuple[str, ...]
   device: Optional[xc.Device]
   backend: Optional[str]
   keep_unused: bool
@@ -410,10 +390,11 @@ class PjitInfo(NamedTuple):
 
 def common_infer_params(pjit_info_args, *args, **kwargs):
   (fun, user_in_shardings, user_out_shardings, static_argnums, static_argnames,
-   donate_argnums, device, backend, keep_unused, inline,
+   donate_argnums, donate_argnames, device, backend, keep_unused, inline,
    resource_env, abstracted_axes) = pjit_info_args
 
-  if kwargs and not is_unspecified(user_in_shardings):
+  if (kwargs and user_in_shardings is not None and
+      not is_unspecified(user_in_shardings)):
     raise ValueError(
         "pjit does not support kwargs when in_shardings is specified.")
 
@@ -447,13 +428,15 @@ def common_infer_params(pjit_info_args, *args, **kwargs):
   else:
     explicit_args, in_tree = tree_flatten(dyn_args)
     flat_fun, out_tree = flatten_fun_nokwargs(f, in_tree)
-    dyn_kwargs = ()
+    dyn_kwargs = {}
   del kwargs
 
-  if donate_argnums and not config.jax_debug_nans:
-    donated_invars = donation_vector(donate_argnums, dyn_args, dyn_kwargs)
+  if (donate_argnums or donate_argnames) and not config.jax_debug_nans:
+    donated_invars = donation_vector(
+        donate_argnums, donate_argnames, dyn_args, dyn_kwargs)
   else:
     donated_invars = (False,) * len(explicit_args)
+  del donate_argnums, donate_argnames
 
   # If backend or device is set as an arg on jit, then resolve them to
   # in_shardings and out_shardings as if user passed in in_shardings
@@ -467,14 +450,17 @@ def common_infer_params(pjit_info_args, *args, **kwargs):
     in_shardings = tree_map(
         lambda x: _create_sharding_for_array(pjit_mesh, x, 'in_shardings',
                                              jit_name),
-        user_in_shardings)
+        user_in_shardings, is_leaf=lambda x: x is None)
     out_shardings = tree_map(
         lambda x: _create_sharding_for_array(pjit_mesh, x, 'out_shardings',
                                              jit_name),
-        user_out_shardings)
+        user_out_shardings, is_leaf=lambda x: x is None)
 
   del user_in_shardings, user_out_shardings
 
+  assert in_shardings is not None or all(i is not None for i in in_shardings)
+  assert out_shardings is not None or all(o is not None for o in out_shardings)
+
   if config.jax_dynamic_shapes:
     in_type = pe.infer_lambda_input_type(axes_specs, explicit_args)
     in_avals = tuple(a for a, e in in_type if e)
@@ -528,10 +514,10 @@ def common_infer_params(pjit_info_args, *args, **kwargs):
       inline=inline,
   )
   return (consts + args_flat, in_type, params, in_tree, out_tree(),
-          donate_argnums, jaxpr)
+          donated_invars)
 
 def _extract_implicit_args(
-  in_type: Sequence[Tuple[core.AbstractValue, bool]],
+  in_type: Sequence[tuple[core.AbstractValue, bool]],
   explicit_args: Sequence[Any]
 ) -> Sequence[core.Tracer]:
   """
@@ -561,7 +547,7 @@ def _extract_implicit_args(
   return [x for x, (_, e) in zip(args, in_type) if not e]  # type: ignore
 
 def _flat_axes_specs(abstracted_axes, *args, **kwargs
-                     ) -> Optional[List[pe.AbstractedAxesSpec]]:
+                     ) -> Optional[list[pe.AbstractedAxesSpec]]:
   if abstracted_axes is None: return None
   if kwargs: raise NotImplementedError
   def ax_leaf(l):
@@ -576,11 +562,10 @@ def pjit(
     fun: Callable,
     in_shardings=UNSPECIFIED,
     out_shardings=UNSPECIFIED,
-    in_axis_resources=UNSPECIFIED,
-    out_axis_resources=UNSPECIFIED,
     static_argnums: Union[int, Sequence[int], None] = None,
     static_argnames: Union[str, Iterable[str], None] = None,
-    donate_argnums: Union[int, Sequence[int]] = (),
+    donate_argnums: Union[int, Sequence[int], None] = None,
+    donate_argnames: Union[str, Iterable[str], None] = None,
     keep_unused: bool = False,
     device: Optional[xc.Device] = None,
     backend: Optional[str] = None,
@@ -661,11 +646,19 @@ def pjit(
       - :py:class:`XLACompatibleSharding`, which will decide how the value
         will be partitioned. With this, using a mesh context manager is not
         required.
+      - :py:obj:`None` is a special case whose semantics are:
+          - if the mesh context manager is *not* provided, JAX has the freedom to
+            choose whatever sharding it wants.
+            For in_shardings, JAX will mark is as replicated but this behavior
+            can change in the future.
+            For out_shardings, we will rely on the XLA GSPMD partitioner to
+            determine the output shardings.
+          - If the mesh context manager is provided, None will imply that the
+            value will be replicated on all devices of the mesh.
       - For backwards compatibility, in_shardings still supports ingesting
-        :py:class:`PartitionSpec` and :py:obj:`None`. These 2 options can
-        *only* be used with the mesh context manager.
+        :py:class:`PartitionSpec`. This option can *only* be used with the
+        mesh context manager.
 
-        - :py:obj:`None`, in which case the value will be replicated on all devices
         - :py:class:`PartitionSpec`, a tuple of length at most equal to the rank
           of the partitioned value. Each element can be a :py:obj:`None`, a mesh
           axis or a tuple of mesh axes, and specifies the set of resources assigned
@@ -677,8 +670,6 @@ def pjit(
       assignment for function outputs.
       The ``out_shardings`` argument is optional. If not specified, :py:func:`jax.jit`
       will use GSPMD's sharding propagation to determine how to shard the outputs.
-    in_axis_resources: (Deprecated) Please use in_shardings.
-    out_axis_resources: (Deprecated) Please use out_shardings.
     static_argnums: An optional int or collection of ints that specify which
       positional arguments to treat as static (compile-time constant).
       Operations that only depend on static arguments will be constant-folded in
@@ -697,14 +688,32 @@ def pjit(
       comment on ``static_argnums`` for details. If not
       provided but ``static_argnums`` is set, the default is based on calling
       ``inspect.signature(fun)`` to find corresponding named arguments.
-    donate_argnums: Specify which argument buffers are "donated" to the computation.
-      It is safe to donate argument buffers if you no longer need them once the
-      computation has finished. In some cases XLA can make use of donated
-      buffers to reduce the amount of memory needed to perform a computation,
-      for example recycling one of your input buffers to store a result. You
-      should not reuse buffers that you donate to a computation, JAX will raise
-      an error if you try to.
-      For more details on buffer donation see the `FAQ <https://jax.readthedocs.io/en/latest/faq.html#buffer-donation>`_.
+    donate_argnums: Specify which positional argument buffers are "donated" to
+      the computation. It is safe to donate argument buffers if you no longer
+      need them once the computation has finished. In some cases XLA can make
+      use of donated buffers to reduce the amount of memory needed to perform a
+      computation, for example recycling one of your input buffers to store a
+      result. You should not reuse buffers that you donate to a computation, JAX
+      will raise an error if you try to. By default, no argument buffers are
+      donated.
+
+      If neither ``donate_argnums`` nor ``donate_argnames`` is provided, no
+      arguments are donated. If ``donate_argnums`` is not provided but
+      ``donate_argnames`` is, or vice versa, JAX uses
+      :code:`inspect.signature(fun)` to find any positional arguments that
+      correspond to ``donate_argnames``
+      (or vice versa). If both ``donate_argnums`` and ``donate_argnames`` are
+      provided, ``inspect.signature`` is not used, and only actual
+      parameters listed in either ``donate_argnums`` or ``donate_argnames`` will
+      be donated.
+
+      For more details on buffer donation see the
+      `FAQ <https://jax.readthedocs.io/en/latest/faq.html#buffer-donation>`_.
+    donate_argnames: An optional string or collection of strings specifying
+      which named arguments are donated to the computation. See the
+      comment on ``donate_argnums`` for details. If not
+      provided but ``donate_argnums`` is set, the default is based on calling
+      ``inspect.signature(fun)`` to find corresponding named arguments.
     keep_unused: If `False` (the default), arguments that JAX determines to be
       unused by `fun` *may* be dropped from resulting compiled XLA executables.
       Such arguments will not be transferred to the device nor provided to the
@@ -740,12 +749,9 @@ def pjit(
   ...   print(f(x))  # doctest: +SKIP
   [ 0.5  2.   4.   6.   8.  10.  12.  10. ]
   """
-  in_shardings, out_shardings = _resolve_axis_resources_and_shardings_arg(
-      in_shardings, out_shardings, in_axis_resources, out_axis_resources)
-
-  (in_shardings, out_shardings, donate_argnums, static_argnums,
+  (in_shardings, out_shardings, donate_argnums, donate_argnames, static_argnums,
    static_argnames) = pre_infer_params(
-       fun, in_shardings, out_shardings, donate_argnums,
+       fun, in_shardings, out_shardings, donate_argnums, donate_argnames,
        static_argnums, static_argnames, device, backend, abstracted_axes)
 
   def infer_params(*args, **kwargs):
@@ -755,8 +761,8 @@ def pjit(
           fun=fun, in_shardings=in_shardings,
           out_shardings=out_shardings, static_argnums=static_argnums,
           static_argnames=static_argnames, donate_argnums=donate_argnums,
-          device=device, backend=backend, keep_unused=keep_unused,
-          inline=inline, resource_env=resource_env,
+          donate_argnames=donate_argnames, device=device, backend=backend,
+          keep_unused=keep_unused, inline=inline, resource_env=resource_env,
           abstracted_axes=abstracted_axes)
     return common_infer_params(pjit_info_args, *args, **kwargs)
 
@@ -774,14 +780,9 @@ def hashable_pytree(pytree):
                           closure=(treedef, vals))
 
 
-@lru_cache(maxsize=4096)
-def _create_mesh_pspec_sharding_from_parsed_pspec(mesh, x):
-  if is_unspecified_or_auto(x):
-    return x
-  return pxla.create_mesh_pspec_sharding(mesh, x.user_spec, x)
-
-
 def _create_sharding_for_array(mesh, x, name, api_name):
+  if x is None and (mesh is None or mesh.empty):
+    return UNSPECIFIED
   if isinstance(x, XLACompatibleSharding) or is_unspecified_or_auto(x):
     return x
   if mesh is None:
@@ -804,8 +805,9 @@ def _create_sharding_for_array(mesh, x, name, api_name):
         f' site? Alternatively, provide `XLACompatibleSharding`s to {name} and'
         ' then the mesh context manager is not required.')
   # A nice user error is raised in prepare_axis_resources.
-  assert isinstance(x, ParsedPartitionSpec), x
-  return _create_mesh_pspec_sharding_from_parsed_pspec(mesh, x)
+  assert x is None or isinstance(x, ParsedPartitionSpec), x
+  return (pxla.create_mesh_pspec_sharding(mesh, x)
+          if x is None else pxla.create_mesh_pspec_sharding(mesh, x.user_spec, x))
 
 
 def _create_sharding_with_device_backend(device, backend):
@@ -814,8 +816,7 @@ def _create_sharding_with_device_backend(device, backend):
     out = SingleDeviceSharding(device)
   elif backend is not None:
     assert device is None
-    out = SingleDeviceSharding(
-        xb.get_backend(backend).get_default_device_assignment(1)[0])
+    out = SingleDeviceSharding(xb.get_backend(backend).local_devices()[0])
   return out
 
 
@@ -968,12 +969,13 @@ def _pjit_jaxpr(fun, out_shardings_thunk, in_type, debug_info,
 
 
 def pjit_check_aval_sharding(
-    shardings, flat_avals, names: Optional[Tuple[str, ...]],
+    shardings, flat_avals, names: Optional[tuple[str, ...]],
     what_aval: str, allow_uneven_sharding: bool):
   new_names = [''] * len(shardings) if names is None else names
   for aval, s, name in zip(flat_avals, shardings, new_names):
     if is_unspecified_or_auto(s):
       continue
+    s = getattr(s, '_original_sharding', s)
     name_str = f' with pytree key path {name}' if name else ''
     shape = aval.shape
     try:
@@ -1024,6 +1026,10 @@ def _resolve_in_shardings(
   for a in args:
     if hasattr(a, 'sharding'):
       arg_s = a.sharding
+      # arg sharding can be None in case of ShapeDtypeStruct. jax.Array does
+      # not allow None as the sharding.
+      if arg_s is None:
+        continue
       if not isinstance(arg_s, XLACompatibleSharding):
         raise ValueError(f'One of the argument to pjit got sharding {arg_s} '
                          'which is not a subclass of XLACompatibleSharding.')
@@ -1045,8 +1051,11 @@ def _resolve_in_shardings(
 
   resolved_in_shardings = []
   for arg, pjit_in_s in zip(args, pjit_in_shardings):
+    # arg sharding can be None in case of ShapeDtypeStruct. jax.Array does
+    # not allow None as the sharding.
     arg_s, committed = ((arg.sharding, getattr(arg, '_committed', True))
-                        if hasattr(arg, 'sharding') else (UNSPECIFIED, False))
+                        if hasattr(arg, 'sharding') and arg.sharding is not None
+                        else (UNSPECIFIED, False))
     if is_unspecified(pjit_in_s):
       if is_unspecified(arg_s):
         resolved_in_shardings.append(arg_s)
@@ -1082,6 +1091,15 @@ def _resolve_in_shardings(
             'https://jax.readthedocs.io/en/latest/jax_array_migration.html#handling-of-host-local-inputs-to-pjit-like-batch-etc. '
             f'Got arg shape: {arg.shape}, arg value: {arg}')
       if not is_unspecified(arg_s):
+        # jax.jit does not allow resharding across different memory kinds even
+        # if the argument is uncommitted. Use jax.device_put for those cases,
+        # either outside or inside jax.jit.
+        if pjit_in_s.memory_kind != arg_s.memory_kind:  # type: ignore
+          raise ValueError(
+              'Memory kinds passed to jax.jit does not match memory kind on the'
+              f' respective arg. Got pjit memory kind: {pjit_in_s.memory_kind}, '  # type: ignore
+              f'arg memory kind: {arg_s.memory_kind} for '  # type: ignore
+              f'arg shape: {shaped_abstractify(arg).str_short()}')
         if (committed and
             not isinstance(arg_s, PmapSharding) and
             not op_shardings.are_op_shardings_equal(
@@ -1091,8 +1109,8 @@ def _resolve_in_shardings(
           raise ValueError('Sharding passed to pjit does not match the sharding '
                            'on the respective arg. '
                            f'Got pjit sharding: {op},\n'
-                           f'arg sharding: {arg_s} for arg shape: {arg.shape}, '
-                           f'arg value: {arg}')
+                           f'arg sharding: {arg_s} for '
+                           f'arg shape: {shaped_abstractify(arg).str_short()}')
       resolved_in_shardings.append(pjit_in_s)
 
   return tuple(resolved_in_shardings)
@@ -1110,7 +1128,7 @@ def _pjit_call_impl_python(
   compiled = _pjit_lower(
       jaxpr, in_shardings, out_shardings, resource_env,
       donated_invars, name, keep_unused, inline,
-      always_lower=False, lowering_platform=None).compile()
+      lowering_parameters=mlir.LoweringParameters()).compile()
   _most_recent_pjit_call_executable.weak_key_dict[jaxpr] = compiled
   # This check is expensive so only do it if enable_checks is on.
   if compiled._auto_spmd_lowering and config.jax_enable_checks:
@@ -1191,6 +1209,7 @@ def _pjit_call_impl(*args, jaxpr,
   has_explicit_sharding = _pjit_explicit_sharding(
       in_shardings, out_shardings, None, None)
   return xc._xla.pjit(name, f, call_impl_cache_miss, [], [], donated_argnums,
+                      tree_util.dispatch_registry,
                       _get_cpp_global_cache(has_explicit_sharding))(*args)
 
 pjit_p.def_impl(_pjit_call_impl)
@@ -1198,7 +1217,7 @@ pjit_p.def_impl(_pjit_call_impl)
 
 @dataclasses.dataclass(frozen=True)
 class SameDeviceAssignmentTuple:
-  shardings: Tuple[PjitSharding, ...]
+  shardings: tuple[PjitSharding, ...]
   # device_assignment is Optional because shardings can contain `AUTO` and in
   # that case `mesh` is compulsory to be used. So in that case
   # `_pjit_lower_cached` cache, resource_env will check against the devices.
@@ -1221,8 +1240,9 @@ class SameDeviceAssignmentTuple:
       s = getattr(s, "_original_sharding", s)
       o = getattr(o, "_original_sharding", o)
       if isinstance(s, GSPMDSharding) and isinstance(o, GSPMDSharding):
-        eq.append(op_shardings.are_op_shardings_equal(
-            s._hlo_sharding, o._hlo_sharding))
+        eq.append(
+            op_shardings.are_op_shardings_equal(s._hlo_sharding, o._hlo_sharding)
+            and s.memory_kind == o.memory_kind)
       else:
         eq.append(s == o)
     return all(eq) and self.device_assignment == other.device_assignment
@@ -1249,12 +1269,11 @@ def _pjit_lower_cached(
     name: str,
     keep_unused: bool,
     inline: bool,
-    always_lower: bool,
     *,
-    lowering_platform: Optional[str]):
-  in_shardings: Tuple[PjitShardingMinusUnspecified, ...] = cast(
-      Tuple[PjitShardingMinusUnspecified, ...], sdat_in_shardings.shardings)
-  out_shardings: Tuple[PjitSharding, ...] = sdat_out_shardings.shardings
+    lowering_parameters: mlir.LoweringParameters):
+  in_shardings: tuple[PjitShardingMinusUnspecified, ...] = cast(
+      tuple[PjitShardingMinusUnspecified, ...], sdat_in_shardings.shardings)
+  out_shardings: tuple[PjitSharding, ...] = sdat_out_shardings.shardings
 
   if resource_env is not None:
     pxla.resource_typecheck(jaxpr, resource_env, {}, lambda: "pjit")
@@ -1274,15 +1293,16 @@ def _pjit_lower_cached(
       jaxpr, api_name, name, mesh,
       in_shardings, out_shardings, donated_invars,
       True, jaxpr.in_avals, tiling_method=None,
-      lowering_platform=lowering_platform)
+      lowering_parameters=lowering_parameters)
   else:
     return pxla.lower_sharding_computation(
         jaxpr, api_name, name, in_shardings, out_shardings,
         tuple(donated_invars), tuple(jaxpr.in_avals),
-        keep_unused=keep_unused, inline=inline, always_lower=always_lower,
+        keep_unused=keep_unused, inline=inline,
         devices_from_context=(
             None if mesh is None or mesh.empty else list(mesh.devices.flat)),
-        lowering_platform=lowering_platform)
+        lowering_parameters=lowering_parameters,
+)
 
 
 def pjit_staging_rule(trace, *args, **params):
@@ -1313,7 +1333,7 @@ pe.custom_staging_rules[pjit_p] = pjit_staging_rule
 
 # TODO(mattjj): remove/trivialize this when jaxprs have type annotation on them,
 # since it's actually not possible in general to infer the type from the term
-def _out_type(jaxpr: core.ClosedJaxpr) -> List[core.AbstractValue]:
+def _out_type(jaxpr: core.ClosedJaxpr) -> list[core.AbstractValue]:
   out = []
   in_idx = {v: i for i, v in enumerate(jaxpr.jaxpr.invars)}
   out_idx = {x: i for i, x in enumerate(jaxpr.jaxpr.invars)
@@ -1381,6 +1401,7 @@ def _pjit_batcher(insert_axis, spmd_axis_name,
                   vals_in, dims_in,
                   jaxpr, in_shardings, out_shardings,
                   resource_env, donated_invars, name, keep_unused, inline):
+  segment_lens, dims_in = batching.indirectify_ragged_axes(dims_in)
   new_jaxpr, axes_out = batching.batch_jaxpr2(
       jaxpr, axis_size, dims_in, axis_name=axis_name,
       spmd_axis_name=spmd_axis_name, main_type=main_type)
@@ -1394,6 +1415,7 @@ def _pjit_batcher(insert_axis, spmd_axis_name,
   else:
     mesh = None
 
+  # TODO(axch): prepend with Nones (?) to account for new segment_lens inputs
   in_shardings = tuple(
       _pjit_batcher_for_sharding(i, axis_in, new_parts, mesh, aval.ndim)
       if axis_in is not None else i
@@ -1412,7 +1434,9 @@ def _pjit_batcher(insert_axis, spmd_axis_name,
     name=name,
     keep_unused=keep_unused,
     inline=inline)
-  return vals_out, axes_out
+  resolved_axes_out = batching.resolve_ragged_axes_against_inputs_outputs(
+      vals_in, vals_out, axes_out)
+  return vals_out, resolved_axes_out
 
 batching.spmd_axis_primitive_batchers[pjit_p] = partial(_pjit_batcher, False)
 batching.axis_primitive_batchers[pjit_p] = partial(_pjit_batcher, False, None)
@@ -1420,7 +1444,7 @@ pxla.spmd_primitive_batchers[pjit_p] = partial(_pjit_batcher, True, None)
 
 def _pjit_batcher_for_sharding(
     s: Union[GSPMDSharding, UnspecifiedValue],
-    dim: int, val: Tuple[str, ...], mesh, ndim: int):
+    dim: int, val: tuple[str, ...], mesh, ndim: int):
   if is_unspecified(s):
     return s
   if not val:
@@ -1441,7 +1465,16 @@ def _pjit_batcher_for_sharding(
     assert isinstance(s, GSPMDSharding)
     if isinstance(getattr(s, '_original_sharding', None), NamedSharding):
       mesh = s._original_sharding.mesh  # type: ignore
-    assert mesh is not None and not mesh.empty
+    if mesh is None or mesh.empty:
+      s_type = (f', got: {repr(s._original_sharding)}'
+                if hasattr(s, '_original_sharding') else '')
+      raise ValueError(
+          'If you are using xmap or spmd_axis_name parameter of jax.vmap,'
+          ' please make sure to run your jitted function inside the mesh'
+          ' context manager. Only `jax.lax.with_sharding_constraint` with'
+          ' `jax.sharding.NamedSharding` as an input can be transformed with'
+          ' spmd_axis_name batching rules outside of an explicit mesh context'
+          f' manager scope{s_type}')
     parsed_pspec = parse_flatten_op_sharding(s._hlo_sharding, mesh)[0]  # type: ignore
     parsed_pspec = parsed_pspec.insert_axis_partitions(dim, val)
     mps = NamedSharding._from_parsed_pspec(mesh, parsed_pspec)
@@ -1480,7 +1513,7 @@ ad.primitive_jvps[pjit_p] = _pjit_jvp
 
 @weakref_lru_cache
 def _known_jaxpr_fwd(known_jaxpr: core.ClosedJaxpr,
-                     fwds_known: Tuple[Optional[int]]) -> core.ClosedJaxpr:
+                     fwds_known: tuple[Optional[int]]) -> core.ClosedJaxpr:
   updated_jaxpr = known_jaxpr.jaxpr.replace(
       outvars=[x for x, i in zip(known_jaxpr.jaxpr.outvars, fwds_known)
                if i is None])
@@ -1593,7 +1626,7 @@ def _pjit_partial_eval_custom_params_updater(
     unks_in: Sequence[bool], inst_in: Sequence[bool],
     kept_outs_known: Sequence[bool], kept_outs_staged: Sequence[bool],
     num_res: int, params_known: dict, params_staged: dict
-  ) -> Tuple[dict, dict]:
+  ) -> tuple[dict, dict]:
   # prune inputs to jaxpr_known according to unks_in
   donated_invars_known, _ = pe.partition_list(unks_in, params_known['donated_invars'])
   in_shardings_known, _ = pe.partition_list(unks_in, params_known['in_shardings'])
@@ -1678,14 +1711,14 @@ ad.reducing_transposes[pjit_p] = _pjit_transpose
 
 @weakref_lru_cache
 def _dce_jaxpr_pjit(
-    jaxpr: core.ClosedJaxpr, used_outputs: Tuple[bool]
-) -> Tuple[core.ClosedJaxpr, List[bool]]:
+    jaxpr: core.ClosedJaxpr, used_outputs: tuple[bool]
+) -> tuple[core.ClosedJaxpr, list[bool]]:
   new_jaxpr, used_inputs = pe.dce_jaxpr(jaxpr.jaxpr, used_outputs)
   return core.ClosedJaxpr(new_jaxpr, jaxpr.consts), used_inputs
 
 
-def dce_jaxpr_pjit_rule(used_outputs: List[bool], eqn: core.JaxprEqn
-                        ) -> Tuple[List[bool], Optional[core.JaxprEqn]]:
+def dce_jaxpr_pjit_rule(used_outputs: list[bool], eqn: core.JaxprEqn
+                        ) -> tuple[list[bool], Optional[core.JaxprEqn]]:
   dced_jaxpr, used_inputs = _dce_jaxpr_pjit(
       eqn.params['jaxpr'], tuple(used_outputs))
 
@@ -1801,7 +1834,7 @@ def with_sharding_constraint(x, shardings):
   of how to use this function, see `Distributed arrays and automatic parallelization`_.
 
   Args:
-    x: PyTree of jax.Arrays which will have their shardings constrainted
+    x: PyTree of jax.Arrays which will have their shardings constrained
     shardings: PyTree of sharding specifications. Valid values are the same as for
       the ``in_shardings`` argument of :func:`jax.experimental.pjit`.
   Returns:
@@ -1860,14 +1893,17 @@ def _sharding_constraint_hlo_lowering(ctx, x_node, *, sharding,
   out_aval, = ctx.avals_out
   axis_ctx = ctx.module_context.axis_context
   # axis_ctx and manual_axes is *only used with xmap* and xmap only works with
-  # NamedSharding. So convert the GSPMDSharding to NamedSharding
-  # and then convert it back with the added special axes.
+  # NamedSharding. So update the NamedSharding to have the manual axes.
   if isinstance(axis_ctx, sharding_impls.SPMDAxisContext):
     mesh = resource_env.physical_mesh
     parsed_pspec = parse_flatten_op_sharding(sharding._hlo_sharding, mesh)[0]
-    mps = NamedSharding._from_parsed_pspec(mesh, parsed_pspec)
-    sharding = GSPMDSharding(
-        mps._device_assignment, mps._to_xla_hlo_sharding(aval.ndim, axis_ctx=axis_ctx))
+    if xla_extension_version >= 188:
+      sharding = NamedSharding._from_parsed_pspec(
+          mesh, parsed_pspec, _manual_axes=axis_ctx.manual_axes)
+    else:
+      mps = NamedSharding._from_parsed_pspec(mesh, parsed_pspec)
+      sharding = GSPMDSharding(
+          mps._device_assignment, mps._to_xla_hlo_sharding(aval.ndim, axis_ctx=axis_ctx))
   return [
       mlir.wrap_with_sharding_op(ctx,
           x_node, out_aval,
@@ -1925,7 +1961,8 @@ def to_gspmd_sharding(s: XLACompatibleSharding, ndim: int,
                       device_or_backend_set: bool = False) -> GSPMDSharding:
   if isinstance(s, GSPMDSharding):
     return s
-  gs = GSPMDSharding(s._device_assignment, s._to_xla_hlo_sharding(ndim))
+  gs = GSPMDSharding(s._device_assignment, s._to_xla_hlo_sharding(ndim),
+                      memory_kind=s.memory_kind)
   gs._original_sharding = s
   if device_or_backend_set:
     gs._original_sharding._device_backend = device_or_backend_set
@@ -1954,14 +1991,14 @@ def _get_partition_spec(ppspec: Sequence[ParsedPartitionSpec]) -> Sequence[Parti
   return [get_single_pspec(p) for p in ppspec]
 
 
-def _get_op_sharding_from_executable(
-    executable) -> Tuple[Sequence[xc.OpSharding], Sequence[xc.OpSharding]]:
-  in_op_shardings: List[xc.OpSharding] = []
+def get_op_sharding_from_executable(
+    executable) -> tuple[Sequence[xc.OpSharding], Sequence[xc.OpSharding]]:
+  in_op_shardings: list[xc.OpSharding] = []
   parameter_shardings_from_xla = executable.get_parameter_shardings()
   if parameter_shardings_from_xla is not None:
     in_op_shardings = parameter_shardings_from_xla
 
-  out_op_shardings: List[xc.OpSharding] = []
+  out_op_shardings: list[xc.OpSharding] = []
   output_shardings_from_xla = executable.get_output_shardings()
   if output_shardings_from_xla is not None:
     out_op_shardings = output_shardings_from_xla
@@ -1969,19 +2006,23 @@ def _get_op_sharding_from_executable(
   return in_op_shardings, out_op_shardings
 
 
-def _get_ppspec_from_executable(executable, mesh) -> Tuple[Sequence[ParsedPartitionSpec], Sequence[ParsedPartitionSpec]]:
-  input_op_shardings: Sequence[xc.OpSharding] = executable.hlo_modules()[0].spmd_parameters_shardings
-  output_op_sharding: xc.OpSharding = executable.hlo_modules()[0].spmd_output_sharding
-  in_ppspec: List[ParsedPartitionSpec] = []
+def _get_ppspec_from_executable(executable, mesh) -> tuple[Sequence[ParsedPartitionSpec], Sequence[ParsedPartitionSpec]]:
+  input_op_shardings, output_op_sharding = get_op_sharding_from_executable(
+      executable
+  )
+  in_ppspec: list[ParsedPartitionSpec] = []
   for s in input_op_shardings:
     in_ppspec.extend(parse_flatten_op_sharding(s, mesh))
-  out_ppspec = parse_flatten_op_sharding(output_op_sharding, mesh)
+
+  out_ppspec: list[ParsedPartitionSpec] = []
+  for s in output_op_sharding:
+    out_ppspec.extend(parse_flatten_op_sharding(s, mesh))
   return in_ppspec, out_ppspec
 
 
-def _get_pspec_from_executable(
+def get_pspec_from_executable(
     executable, mesh: pxla.Mesh
-) -> Tuple[Tuple[PartitionSpec, ...], Tuple[PartitionSpec, ...]]:
+) -> tuple[tuple[PartitionSpec, ...], tuple[PartitionSpec, ...]]:
   in_ppspec, out_ppspec = _get_ppspec_from_executable(executable, mesh)
   out_partition_spec = _get_partition_spec(out_ppspec)
   in_partition_spec = _get_partition_spec(in_ppspec)
