@@ -26,17 +26,26 @@
 #
 
 import abc
+from collections.abc import Sequence
 import enum
-import sys
 from functools import partial
-from typing import List, NamedTuple, Optional, Sequence, Tuple, Union
-from jax._src.config import config
+import sys
+from typing import NamedTuple, Optional, Union
+
+from jax._src import config
 
 try:
   import colorama  # pytype: disable=import-error
 except ImportError:
   colorama = None
 
+
+_PPRINT_USE_COLOR = config.DEFINE_bool(
+    'jax_pprint_use_color',
+    config.bool_env('JAX_PPRINT_USE_COLOR', True),
+    help='Enable jaxpr pretty-printing with colorful syntax highlighting.'
+)
+
 def _can_use_color() -> bool:
   try:
     # Check if we're in IPython or Colab
@@ -61,7 +70,7 @@ class Doc(abc.ABC):
   def format(self, width: int = 80, use_color: Optional[bool] = None,
              annotation_prefix=" # ") -> str:
     if use_color is None:
-      use_color = CAN_USE_COLOR and config.FLAGS.jax_pprint_use_color
+      use_color = CAN_USE_COLOR and _PPRINT_USE_COLOR.value
     return _format(self, width, use_color=use_color,
                    annotation_prefix=annotation_prefix)
 
@@ -95,7 +104,7 @@ class _TextDoc(Doc):
 
 class _ConcatDoc(Doc):
   __slots__ = ("children",)
-  children: List[Doc]
+  children: list[Doc]
 
   def __init__(self, children: Sequence[Doc]):
     self.children = list(children)
@@ -164,7 +173,7 @@ _BreakMode = enum.Enum("_BreakMode", ["FLAT", "BREAK"])
 # non-recursive formulation using an explicit stack, necessary because Python
 # doesn't have a tail recursion optimization.
 
-def _fits(doc: Doc, width: int, agenda: List[Tuple[int, _BreakMode, Doc]]
+def _fits(doc: Doc, width: int, agenda: list[tuple[int, _BreakMode, Doc]]
          ) -> bool:
   while width >= 0 and len(agenda) > 0:
     i, m, doc = agenda.pop()
@@ -234,11 +243,11 @@ class _State(NamedTuple):
 class _Line(NamedTuple):
   text: str
   width: int
-  annotations: Union[Optional[str], List[str]]
+  annotations: Union[Optional[str], list[str]]
 
 
 def _update_color(use_color: bool, state: _ColorState, update: _ColorState
-                 ) -> Tuple[_ColorState, str]:
+                 ) -> tuple[_ColorState, str]:
   if not use_color or colorama is None:
     return update, ""
   color_str = ""
@@ -247,7 +256,7 @@ def _update_color(use_color: bool, state: _ColorState, update: _ColorState
   if state.background != update.background:
     color_str += getattr(colorama.Back, str(update.background.name))
   if state.intensity != update.intensity:
-    color_str += colorama.Style.NORMAL
+    color_str += colorama.Style.NORMAL  # pytype: disable=unsupported-operands
     color_str += getattr(colorama.Style, str(update.intensity.name))
   return update, color_str
 
