@@ -104,7 +104,7 @@ def start_trace(log_dir, create_perfetto_link: bool = False,
       Perfetto trace viewer UI (https://ui.perfetto.dev). The file will also be
       generated if ``create_perfetto_link`` is true. This could be useful if you
       want to generate a Perfetto-compatible trace without blocking the
-      processs.
+      process.
   """
   with _profile_state.lock:
     if _profile_state.profile_session is not None:
@@ -228,7 +228,7 @@ def trace(log_dir, create_perfetto_link=False, create_perfetto_trace=False):
       Perfetto trace viewer UI (https://ui.perfetto.dev). The file will also be
       generated if ``create_perfetto_link`` is true. This could be useful if you
       want to generate a Perfetto-compatible trace without blocking the
-      processs.
+      process.
   """
   start_trace(log_dir, create_perfetto_link, create_perfetto_trace)
   try:
@@ -321,7 +321,7 @@ def device_memory_profile(backend: Optional[str] = None) -> bytes:
   """Captures a JAX device memory profile as ``pprof``-format protocol buffer.
 
   A device memory profile is a snapshot of the state of memory, that describes the JAX
-  :class:`jax.DeviceArray` and executable objects present in memory and their
+  :class:`~jax.Array` and executable objects present in memory and their
   allocation sites.
 
   For more information how to use the device memory profiler, see
