@@ -15,13 +15,11 @@
 from functools import partial
 import operator
 
-from jax import config
-from jax.tree_util import tree_map, tree_reduce
 from jax._src import api
 from jax._src import dtypes as _dtypes
-from jax._src import xla_bridge
-from jax._src.config import flags
-from jax._src.lib import xla_client
+from jax._src.config import config
+from jax._src.tree_util import tree_map, tree_reduce
+
 import numpy as np
 
 
@@ -31,10 +29,7 @@ import numpy as np
 __all__ = ['check_grads', 'check_jvp', 'check_vjp']
 
 
-FLAGS = flags.FLAGS
-
 EPS = 1e-4
-_fp8_enabled = xla_client._version >= 117
 
 
 def _dtype(x):
@@ -49,14 +44,21 @@ def _dtype(x):
 _default_tolerance = {
   _dtypes.float0: 0,
   np.dtype(np.bool_): 0,
+  np.dtype(_dtypes.int4): 0,
   np.dtype(np.int8): 0,
   np.dtype(np.int16): 0,
   np.dtype(np.int32): 0,
   np.dtype(np.int64): 0,
+  np.dtype(_dtypes.uint4): 0,
   np.dtype(np.uint8): 0,
   np.dtype(np.uint16): 0,
   np.dtype(np.uint32): 0,
   np.dtype(np.uint64): 0,
+  np.dtype(_dtypes.float8_e4m3b11fnuz): 1e-1,
+  np.dtype(_dtypes.float8_e4m3fn): 1e-1,
+  np.dtype(_dtypes.float8_e4m3fnuz): 1e-1,
+  np.dtype(_dtypes.float8_e5m2): 1e-1,
+  np.dtype(_dtypes.float8_e5m2fnuz): 1e-1,
   np.dtype(_dtypes.bfloat16): 1e-2,
   np.dtype(np.float16): 1e-3,
   np.dtype(np.float32): 1e-6,
@@ -64,23 +66,17 @@ _default_tolerance = {
   np.dtype(np.complex64): 1e-6,
   np.dtype(np.complex128): 1e-15,
 }
-if _fp8_enabled:
-  _default_tolerance.update({
-    np.dtype(_dtypes.float8_e4m3fn): 1e-1,
-    np.dtype(_dtypes.float8_e5m2): 1e-1,
-  })
-
 
 def default_tolerance():
-  if device_under_test() != "tpu":
-    return _default_tolerance
-  tol = _default_tolerance.copy()
-  tol[np.dtype(np.float32)] = 1e-3
-  tol[np.dtype(np.complex64)] = 1e-3
-  return tol
+  return _default_tolerance
 
 
 default_gradient_tolerance = {
+  np.dtype(_dtypes.float8_e4m3b11fnuz): 1e-1,
+  np.dtype(_dtypes.float8_e4m3fn): 1e-1,
+  np.dtype(_dtypes.float8_e4m3fnuz): 1e-1,
+  np.dtype(_dtypes.float8_e5m2): 1e-1,
+  np.dtype(_dtypes.float8_e5m2fnuz): 1e-1,
   np.dtype(_dtypes.bfloat16): 1e-1,
   np.dtype(np.float16): 1e-2,
   np.dtype(np.float32): 2e-3,
@@ -88,11 +84,6 @@ default_gradient_tolerance = {
   np.dtype(np.complex64): 1e-3,
   np.dtype(np.complex128): 1e-5,
 }
-if _fp8_enabled:
-  default_gradient_tolerance.update({
-    np.dtype(_dtypes.float8_e4m3fn): 1e-1,
-    np.dtype(_dtypes.float8_e5m2): 1e-1,
-  })
 
 def is_python_scalar(val):
   return not isinstance(val, np.generic) and isinstance(val, (bool, int, float, complex))
@@ -101,10 +92,8 @@ def _assert_numpy_allclose(a, b, atol=None, rtol=None, err_msg=''):
   if a.dtype == b.dtype == _dtypes.float0:
     np.testing.assert_array_equal(a, b, err_msg=err_msg)
     return
-  if _fp8_enabled:
-    custom_dtypes = [_dtypes.float8_e4m3fn, _dtypes.float8_e5m2, _dtypes.bfloat16]
-  else:
-    custom_dtypes = [_dtypes.bfloat16]
+  custom_dtypes = [_dtypes.float8_e4m3b11fnuz, _dtypes.float8_e4m3fn,
+                   _dtypes.float8_e5m2, _dtypes.bfloat16]
   a = a.astype(np.float32) if a.dtype in custom_dtypes else a
   b = b.astype(np.float32) if b.dtype in custom_dtypes else b
   kw = {}
@@ -296,7 +285,3 @@ def check_grads(f, args, order,
         _check_grads(f_vjp, args, order - 1, rev_msg)
 
   _check_grads(f, args, order)
-
-
-def device_under_test():
-  return getattr(FLAGS, 'jax_test_dut', None) or xla_bridge.get_backend().platform
