@@ -11,7 +11,9 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-from typing import Any, Callable, Mapping, Optional, Tuple, Union
+
+from collections.abc import Mapping
+from typing import Any, Callable, Optional, Union
 
 import jax
 from jax._src.scipy.optimize.bfgs import minimize_bfgs
@@ -50,7 +52,7 @@ class OptimizeResults(NamedTuple):
 def minimize(
     fun: Callable,
     x0: jax.Array,
-    args: Tuple = (),
+    args: tuple = (),
     *,
     method: str,
     tol: Optional[float] = None,
@@ -116,7 +118,7 @@ def minimize(
 
   if method.lower() == 'l-bfgs-experimental-do-not-rely-on-this':
     results = _minimize_lbfgs(fun_with_args, x0, **options)
-    success = results.converged & (~results.failed)
+    success = results.converged & jnp.logical_not(results.failed)
     return OptimizeResults(x=results.x_k,
                            success=success,
                            status=results.status,
