@@ -12,18 +12,22 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from collections.abc import Sequence
 from functools import partial
+import math
 import operator
-from typing import Callable, Optional, Tuple, Union
+from typing import Callable, Optional, Union
 import warnings
 
 import numpy as np
 import scipy.signal as osp_signal
+from scipy.fft import next_fast_len as osp_fft_next_fast_len
 
 import jax
 import jax.numpy.fft
 import jax.numpy as jnp
 from jax import lax
+from jax._src.api_util import _ensure_index_tuple
 from jax._src import dtypes
 from jax._src.lax.lax import PrecisionLike
 from jax._src.numpy import linalg
@@ -34,6 +38,62 @@ from jax._src.typing import Array, ArrayLike
 from jax._src.util import canonicalize_axis, tuple_delete, tuple_insert
 
 
+@_wraps(osp_signal.fftconvolve)
+def fftconvolve(in1: ArrayLike, in2: ArrayLike, mode: str = "full",
+                axes: Optional[Sequence[int]] = None) -> Array:
+  check_arraylike('fftconvolve', in1, in2)
+  in1, in2 = promote_dtypes_inexact(in1, in2)
+  if in1.ndim != in2.ndim:
+    raise ValueError("in1 and in2 should have the same dimensionality")
+  if mode not in ["same", "full", "valid"]:
+    raise ValueError("mode must be one of ['same', 'full', 'valid']")
+  _fftconvolve = partial(_fftconvolve_unbatched, mode=mode)
+  if axes is None:
+    return _fftconvolve(in1, in2)
+  axes = _ensure_index_tuple(axes)
+  axes = tuple(canonicalize_axis(ax, in1.ndim) for ax in axes)
+  mapped_axes = set(range(in1.ndim)) - set(axes)
+  if any(in1.shape[i] != in2.shape[i] for i in mapped_axes):
+    raise ValueError(f"mapped axes must have same shape; got {in1.shape=} {in2.shape=} {axes=}")
+  for ax in sorted(mapped_axes):
+    _fftconvolve = jax.vmap(_fftconvolve, in_axes=ax, out_axes=ax)
+  return _fftconvolve(in1, in2)
+
+def _fftconvolve_unbatched(in1: Array, in2: Array, mode: str) -> Array:
+  full_shape = tuple(s1 + s2 - 1 for s1, s2 in zip(in1.shape, in2.shape))
+  fft_shape = tuple(osp_fft_next_fast_len(s) for s in full_shape)
+
+  if mode == 'valid':
+    no_swap = all(s1 >= s2 for s1, s2 in zip(in1.shape, in2.shape))
+    swap = all(s1 <= s2 for s1, s2 in zip(in1.shape, in2.shape))
+    if not (no_swap or swap):
+      raise ValueError("For 'valid' mode, One input must be at least as "
+                       "large as the other in every dimension.")
+    if swap:
+      in1, in2 = in2, in1
+
+  if jnp.iscomplexobj(in1):
+    fft, ifft = jnp.fft.fftn, jnp.fft.ifftn
+  else:
+    fft, ifft = jnp.fft.rfftn, jnp.fft.irfftn
+  sp1 = fft(in1, fft_shape)
+  sp2 = fft(in2, fft_shape)
+  conv = ifft(sp1 * sp2, fft_shape)
+
+  if mode == "full":
+    out_shape = full_shape
+  elif mode == "same":
+    out_shape = in1.shape
+  elif mode == "valid":
+    out_shape = tuple(s1 - s2 + 1 for s1, s2 in zip(in1.shape, in2.shape))
+  else:
+    raise ValueError(f"Unrecognized {mode=}")
+
+  start_indices = tuple((full_size - out_size) // 2
+                        for full_size, out_size in zip(full_shape, out_shape))
+  return lax.dynamic_slice(conv, start_indices, out_shape)
+
+
 # Note: we do not re-use the code from jax.numpy.convolve here, because the handling
 # of padding differs slightly between the two implementations (particularly for
 # mode='same').
@@ -74,9 +134,12 @@ def _convolve_nd(in1: Array, in2: Array, mode: str, *, precision: PrecisionLike)
 @_wraps(osp_signal.convolve)
 def convolve(in1: Array, in2: Array, mode: str = 'full', method: str = 'auto',
              precision: PrecisionLike = None) -> Array:
-  if method != 'auto':
-    warnings.warn("convolve() ignores method argument")
-  return _convolve_nd(in1, in2, mode, precision=precision)
+  if method == 'fft':
+    return fftconvolve(in1, in2, mode=mode)
+  elif method in ['direct', 'auto']:
+    return _convolve_nd(in1, in2, mode, precision=precision)
+  else:
+    raise ValueError(f"Got {method=}; expected 'auto', 'fft', or 'direct'.")
 
 
 @_wraps(osp_signal.convolve2d)
@@ -92,9 +155,7 @@ def convolve2d(in1: Array, in2: Array, mode: str = 'full', boundary: str = 'fill
 @_wraps(osp_signal.correlate)
 def correlate(in1: Array, in2: Array, mode: str = 'full', method: str = 'auto',
               precision: PrecisionLike = None) -> Array:
-  if method != 'auto':
-    warnings.warn("correlate() ignores method argument")
-  return _convolve_nd(in1, jnp.flip(in2.conj()), mode, precision=precision)
+  return convolve(in1, jnp.flip(in2.conj()), mode, precision=precision, method=method)
 
 
 @_wraps(osp_signal.correlate2d)
@@ -173,7 +234,7 @@ def _fft_helper(x: Array, win: Array, detrend_func: Callable[[Array], Array],
   else:
     step = nperseg - noverlap
     batch_shape = list(batch_shape)
-    x = x.reshape((int(np.prod(batch_shape)), signal_length))[..., np.newaxis]
+    x = x.reshape((math.prod(batch_shape), signal_length, 1))
     result = jax.lax.conv_general_dilated_patches(
         x, (nperseg,), (step,),
         'VALID',
@@ -229,7 +290,7 @@ def _spectral_helper(x: Array, y: Optional[ArrayLike], fs: ArrayLike = 1.0,
                      detrend_type: Union[bool, str, Callable[[Array], Array]] = 'constant',
                      return_onesided: bool = True, scaling: str = 'density',
                      axis: int = -1, mode: str = 'psd', boundary: Optional[str] = None,
-                     padded: bool = False) -> Tuple[Array, Array, Array]:
+                     padded: bool = False) -> tuple[Array, Array, Array]:
   """LAX-backend implementation of `scipy.signal._spectral_helper`.
 
   Unlike the original helper function, `y` can be None for explicitly
@@ -440,7 +501,7 @@ def _spectral_helper(x: Array, y: Optional[ArrayLike], fs: ArrayLike = 1.0,
 def stft(x: Array, fs: ArrayLike = 1.0, window: str = 'hann', nperseg: int = 256,
          noverlap: Optional[int] = None, nfft: Optional[int] = None,
          detrend: bool = False, return_onesided: bool = True, boundary: Optional[str] = 'zeros',
-         padded: bool = True, axis: int = -1) -> Tuple[Array, Array, Array]:
+         padded: bool = True, axis: int = -1) -> tuple[Array, Array, Array]:
   return _spectral_helper(x, None, fs, window, nperseg, noverlap,
                           nfft, detrend, return_onesided,
                           scaling='spectrum', axis=axis,
@@ -460,7 +521,7 @@ def csd(x: Array, y: Optional[ArrayLike], fs: ArrayLike = 1.0, window: str = 'ha
         nperseg: Optional[int] = None, noverlap: Optional[int] = None,
         nfft: Optional[int] = None, detrend: str = 'constant',
         return_onesided: bool = True, scaling: str = 'density',
-        axis: int = -1, average: str = 'mean') -> Tuple[Array, Array]:
+        axis: int = -1, average: str = 'mean') -> tuple[Array, Array]:
   freqs, _, Pxy = _spectral_helper(x, y, fs, window, nperseg, noverlap, nfft,
                                   detrend, return_onesided, scaling, axis,
                                   mode='psd')
@@ -493,7 +554,7 @@ def welch(x: Array, fs: ArrayLike = 1.0, window: str = 'hann',
           nperseg: Optional[int] = None, noverlap: Optional[int] = None,
           nfft: Optional[int] = None, detrend: str = 'constant',
           return_onesided: bool = True, scaling: str = 'density',
-          axis: int = -1, average: str = 'mean') -> Tuple[Array, Array]:
+          axis: int = -1, average: str = 'mean') -> tuple[Array, Array]:
   freqs, Pxx = csd(x, None, fs=fs, window=window, nperseg=nperseg,
                    noverlap=noverlap, nfft=nfft, detrend=detrend,
                    return_onesided=return_onesided, scaling=scaling,
@@ -520,7 +581,7 @@ def _overlap_and_add(x: Array, step_size: int) -> Array:
     raise ValueError('Input must have (..., frames, frame_length) shape.')
 
   *batch_shape, nframes, segment_len = x.shape
-  flat_batchsize = np.prod(batch_shape, dtype=np.int64)
+  flat_batchsize = math.prod(batch_shape)
   x = x.reshape((flat_batchsize, nframes, segment_len))
   output_size = step_size * (nframes - 1) + segment_len
   nstep_per_segment = 1 + (segment_len - 1) // step_size
@@ -555,7 +616,7 @@ def istft(Zxx: Array, fs: ArrayLike = 1.0, window: str = 'hann',
           nperseg: Optional[int] = None, noverlap: Optional[int] = None,
           nfft: Optional[int] = None, input_onesided: bool = True,
           boundary: bool = True, time_axis: int = -1,
-          freq_axis: int = -2) -> Tuple[Array, Array]:
+          freq_axis: int = -2) -> tuple[Array, Array]:
   # Input validation
   check_arraylike("istft", Zxx)
   if Zxx.ndim < 2:
