@@ -30,11 +30,9 @@ executable protocols described above.
 """
 from __future__ import annotations
 
-import warnings
-
+from collections.abc import Sequence
 from dataclasses import dataclass
-from typing import (
-    Any, Dict, List, NamedTuple, Optional, Protocol, Sequence, Tuple, Union)
+from typing import Any, NamedTuple, Protocol, Union
 
 import jax
 
@@ -56,7 +54,7 @@ xla_extension = xc._xla
 map, unsafe_map = util.safe_map, map
 zip, unsafe_zip = util.safe_zip, zip
 
-CompilerOptions = Dict[str, Union[str, bool]]
+CompilerOptions = dict[str, Union[str, bool]]
 
 
 # -- Internal protocols
@@ -149,11 +147,11 @@ class Lowering(Protocol):
   """Protocol for lowerings, which a user-facing ``Lowered`` encapsulates."""
 
   def compile(
-      self, compiler_options: Optional[CompilerOptions] = None) -> Executable:
+      self, compiler_options: CompilerOptions | None = None) -> Executable:
     """Compile and return a corresponding ``Executable``."""
     raise NotImplementedError
 
-  def as_text(self, dialect: Optional[str] = None) -> str:
+  def as_text(self, dialect: str | None = None) -> str:
     """A human-readable text representation of this lowering.
 
     Intended for visualization and debugging purposes. This need not be a valid
@@ -161,7 +159,7 @@ class Lowering(Protocol):
     """
     raise NotImplementedError
 
-  def compiler_ir(self, dialect: Optional[str] = None) -> Any:
+  def compiler_ir(self, dialect: str | None = None) -> Any:
     """An arbitrary object representation of this lowering.
 
     Intended for debugging purposes. This need not be a valid nor reliable
@@ -232,11 +230,21 @@ class XlaExecutable(Executable):
       else:
         raise
 
-  def cost_analysis(self) -> List[Dict[str, float]]:
+    # TODO(skyewm): this should return a single dict (I think returning a list
+    # was to support MPMD executables, which never fully landed)
+  def cost_analysis(self) -> list[dict[str, float]]:
     xla_ext_exe = self.xla_extension_executable()
-    err_msg = ("cost analysis unsupported on current XLA backend: "
-               f"{type(xla_ext_exe)}")
+
     # TODO(b/259255524): Unify/merge the two cost_analysis calls below.
+    if hasattr(xla_ext_exe, "cost_analysis"):
+      try:
+        return [xla_ext_exe.cost_analysis()]
+      except xla_extension.XlaRuntimeError as e:
+        msg, *_ = e.args
+        if not (type(msg) is str and msg.startswith("UNIMPLEMENTED")):
+          raise
+
+    # Try client method if executable cost_analysis method is unimplemented
     if hasattr(xla_ext_exe, "client"):
       try:
         return [
@@ -245,21 +253,20 @@ class XlaExecutable(Executable):
         ]
       except xla_extension.XlaRuntimeError as e:
         msg, *_ = e.args
-        if type(msg) is str and msg.startswith("UNIMPLEMENTED"):
-          raise NotImplementedError(err_msg) from e
-        else:
+        if not (type(msg) is str and msg.startswith("UNIMPLEMENTED")):
           raise
-    elif hasattr(xla_ext_exe, "cost_analysis"):
-      try:
-        return xla_ext_exe.cost_analysis()
-      except xla_extension.XlaRuntimeError as e:
-        msg, *_ = e.args
-        if type(msg) is str and msg.startswith("UNIMPLEMENTED"):
-          raise NotImplementedError(err_msg) from e
-        else:
-          raise
-    else:
-      raise NotImplementedError(err_msg)
+
+    if (
+        xla_ext_exe is None
+        and hasattr(self, "unsafe_call")
+        and hasattr(self.unsafe_call, "compiled")
+        and hasattr(self.unsafe_call.compiled, "cost_analysis")
+    ):
+      return [self.unsafe_call.compiled.cost_analysis()]
+
+    raise NotImplementedError(
+        f"cost analysis unsupported on current XLA backend: {type(xla_ext_exe)}"
+    )
 
   def memory_analysis(self) -> Any:
     xla_ext_exe = self.xla_extension_executable()
@@ -283,7 +290,7 @@ class XlaExecutable(Executable):
 class XlaLowering(Lowering):
   """Adapts our various internal XLA-backed computations into a ``Lowering``."""
 
-  compile_args: Dict[str, Any]
+  compile_args: dict[str, Any]
 
   def hlo(self) -> xc.XlaComputation:
     """Return an HLO representation of this computation."""
@@ -303,10 +310,10 @@ class XlaLowering(Lowering):
     raise NotImplementedError("must override")
 
   def compile(
-      self, compiler_options: Optional[CompilerOptions] = None) -> Executable:
+      self, compiler_options: CompilerOptions | None = None) -> Executable:
     raise NotImplementedError("must override")
 
-  def as_text(self, dialect: Optional[str] = None) -> str:
+  def as_text(self, dialect: str | None = None) -> str:
     if dialect is None:
       dialect = "stablehlo"
     if dialect == "mhlo":
@@ -318,7 +325,7 @@ class XlaLowering(Lowering):
     else:
       raise ValueError(f"unknown dialect: {dialect}")
 
-  def compiler_ir(self, dialect: Optional[str] = None) -> Any:
+  def compiler_ir(self, dialect: str | None = None) -> Any:
     if dialect is None:
       dialect = "stablehlo"
     if dialect == "mhlo":
@@ -330,7 +337,7 @@ class XlaLowering(Lowering):
     else:
       raise ValueError(f"unknown dialect: {dialect}")
 
-  def cost_analysis(self) -> Dict[str, float]:
+  def cost_analysis(self) -> dict[str, float]:
     raise NotImplementedError("must override")
 
 
@@ -401,22 +408,7 @@ class Compiled(Stage):
                                       self.in_tree, self.out_tree)
     self._call = None
 
-  def compiler_ir(self):
-    """Post-compilation IR.
-
-    Compilation typically involves code transformation and
-    optimization. This method exists to reflect the compiler's
-    representation of the program after such passes, whenever
-    possible.
-    """
-    # TODO(frostig): remove (deprecated)
-    warnings.warn(
-        "compiler_ir() is deprecated, consider runtime_executable() instead",
-        DeprecationWarning)
-    exe = self.runtime_executable()
-    return exe.hlo_modules() if exe is not None else None
-
-  def as_text(self) -> Optional[str]:
+  def as_text(self) -> str | None:
     """A human-readable text representation of this executable.
 
     Intended for visualization and debugging purposes. This is not a valid nor
@@ -430,7 +422,7 @@ class Compiled(Stage):
     except NotImplementedError:
       return None
 
-  def cost_analysis(self) -> Optional[Any]:
+  def cost_analysis(self) -> Any | None:
     """A summary of execution cost estimates.
 
     Intended for visualization and debugging purposes. The object output by
@@ -448,7 +440,7 @@ class Compiled(Stage):
     except NotImplementedError:
       return None
 
-  def memory_analysis(self) -> Optional[Any]:
+  def memory_analysis(self) -> Any | None:
     """A summary of estimated memory requirements.
 
     Intended for visualization and debugging purposes. The object output by
@@ -466,7 +458,7 @@ class Compiled(Stage):
     except NotImplementedError:
       return None
 
-  def runtime_executable(self) -> Optional[Any]:
+  def runtime_executable(self) -> Any | None:
     """An arbitrary object representation of this executable.
 
     Intended for debugging purposes. This is not valid nor reliable
@@ -493,7 +485,7 @@ class Compiled(Stage):
     # This is because `__call__` passes in `self._params` as the first argument.
     # Instead of making the call signature `call(params, *args, **kwargs)`
     # extract it from args because `params` can be passed as a kwarg by users
-    # which might confict here.
+    # which might conflict here.
     params = args[0]
     args = args[1:]
     if jax.config.jax_dynamic_shapes:
@@ -533,9 +525,8 @@ class Compiled(Stage):
 
   def __call__(self, *args, **kwargs):
     if self._call is None:
-      self._call = self._executable.create_cpp_call(self._no_kwargs,
-                                                        self.in_tree,
-                                                        self.out_tree)
+      self._call = self._executable.create_cpp_call(
+          self._no_kwargs, self.in_tree, self.out_tree)
       if self._call is None:
         params = self._params
         def cpp_call_fallback(*args, **kwargs):
@@ -577,7 +568,7 @@ class Lowered(Stage):
                      lowering: XlaLowering,
                      in_tree: tree_util.PyTreeDef,
                      in_avals,
-                     donate_argnums: Tuple[int, ...],
+                     donate_argnums: tuple[int, ...],
                      out_tree: tree_util.PyTreeDef,
                      no_kwargs: bool = False):
     """Initialize from flat info (``in_avals`` etc.) and an input PyTreeDef.
@@ -596,9 +587,9 @@ class Lowered(Stage):
         no_kwargs=no_kwargs)
 
   def compile(
-      self, compiler_options: Optional[CompilerOptions] = None) -> Compiled:
+      self, compiler_options: CompilerOptions | None = None) -> Compiled:
     """Compile, returning a corresponding ``Compiled`` instance."""
-    kw: Dict[str, Any] = {"compiler_options": compiler_options}
+    kw: dict[str, Any] = {"compiler_options": compiler_options}
     return Compiled(
         self._lowering.compile(**kw),  # pytype: disable=wrong-keyword-args
         self.args_info,
@@ -606,7 +597,7 @@ class Lowered(Stage):
         no_kwargs=self._no_kwargs,
     )
 
-  def as_text(self, dialect: Optional[str] = None) -> str:
+  def as_text(self, dialect: str | None = None) -> str:
     """A human-readable text representation of this lowering.
 
     Intended for visualization and debugging purposes. This need not be a valid
@@ -617,7 +608,7 @@ class Lowered(Stage):
     """
     return self._lowering.as_text(dialect)
 
-  def compiler_ir(self, dialect: Optional[str] = None) -> Optional[Any]:
+  def compiler_ir(self, dialect: str | None = None) -> Any | None:
     """An arbitrary object representation of this lowering.
 
     Intended for debugging purposes. This is not a valid nor reliable
@@ -635,7 +626,7 @@ class Lowered(Stage):
     except NotImplementedError:
       return None
 
-  def cost_analysis(self) -> Optional[Any]:
+  def cost_analysis(self) -> Any | None:
     """A summary of execution cost estimates.
 
     Intended for visualization and debugging purposes. The object output by
