@@ -12,10 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 """Module for state."""
-from jax._src.state.types import (ShapedArrayRef, ReadEffect, WriteEffect,
+from jax._src.state.types import (AbstractRef, ReadEffect, WriteEffect,
                                   AccumEffect, StateEffect, RefEffect,
-                                  get_ref_state_effects)
-from jax._src.state.primitives import (ref_get, ref_set, ref_swap,
-                                       ref_addupdate, get_p, swap_p,
-                                       addupdate_p)
-from jax._src.state.discharge import discharge_state, register_discharge_rule
+                                  get_ref_state_effects, shaped_array_ref,
+                                  RefView)
