@@ -13,24 +13,25 @@
 # limitations under the License.
 """Module for discharging state primitives."""
 from __future__ import annotations
+
+from collections.abc import Sequence
 import dataclasses
 from functools import partial
 import operator
-
-from typing import Any, Callable, Dict, List, Optional, Protocol, Sequence, Tuple, Union
+from typing import Any, Callable, Protocol
 
 import numpy as np
 
-
 from jax._src import api_util
 from jax._src import ad_util
 from jax._src import core
 from jax._src import linear_util as lu
-from jax._src.interpreters import partial_eval as pe
 from jax._src import source_info_util
 from jax._src import tree_util
 from jax._src.config import config
 from jax._src.interpreters import ad
+from jax._src.interpreters import mlir
+from jax._src.interpreters import partial_eval as pe
 from jax._src.lax import lax
 from jax._src.lax import slicing as lax_slicing
 from jax._src.state.types import AbstractRef, RefEffect
@@ -53,8 +54,8 @@ PyTreeDef = tree_util.PyTreeDef
 # `Read/Write/Accum` effects.
 
 def discharge_state(jaxpr: core.Jaxpr, consts: Sequence[Any], * ,
-                    should_discharge: Union[bool, Sequence[bool]] = True
-                    ) -> Tuple[core.Jaxpr, List[Any]]:
+                    should_discharge: bool | Sequence[bool] = True
+                    ) -> tuple[core.Jaxpr, list[Any]]:
   """Converts a jaxpr that takes in `Ref`s into one that doesn't."""
   if isinstance(should_discharge, bool):
     should_discharge = [should_discharge] * len(jaxpr.invars)
@@ -68,7 +69,7 @@ def discharge_state(jaxpr: core.Jaxpr, consts: Sequence[Any], * ,
 
 @dataclasses.dataclass
 class Environment:
-  env: Dict[core.Var, Any]
+  env: dict[core.Var, Any]
 
   def read(self, v: core.Atom) -> Any:
     if type(v) is core.Literal:
@@ -83,7 +84,7 @@ class DischargeRule(Protocol):
 
   def __call__(self, in_avals: Sequence[core.AbstractValue],
       out_avals: Sequence[core.AbstractValue], *args: Any,
-      **params: Any) -> Tuple[Sequence[Optional[Any]], Sequence[Any]]:
+      **params: Any) -> tuple[Sequence[Any | None], Sequence[Any]]:
     ...
 
 _discharge_rules: dict[core.Primitive, DischargeRule] = {}
@@ -106,9 +107,9 @@ def _eval_jaxpr_discharge_state(
   # regular values in this interpreter.
   map(env.write, jaxpr.invars, args)
 
-  refs_to_discharge = set(id(v.aval) for v, d
+  refs_to_discharge = {id(v.aval) for v, d
                           in zip(jaxpr.invars, should_discharge) if d
-                          and isinstance(v.aval, AbstractRef))
+                          and isinstance(v.aval, AbstractRef)}
 
   for eqn in jaxpr.eqns:
     if _has_refs(eqn) and any(id(v.aval) in refs_to_discharge
@@ -173,7 +174,7 @@ def _prepend_scatter(x, idx, indexed_dims, val, *, add=False):
 
 def _indexer(idx, indexed_dims):
   idx_ = iter(idx)
-  indexer = tuple([next(idx_) if b else slice(None) for b in indexed_dims])
+  indexer = tuple(next(idx_) if b else slice(None) for b in indexed_dims)
   assert next(idx_, None) is None
   return indexer
 
@@ -274,6 +275,7 @@ def _run_state_impl(*args: Any, jaxpr: core.Jaxpr,
   discharged_jaxpr, consts = discharge_state(jaxpr, ())
   return core.eval_jaxpr(discharged_jaxpr, consts, *args)
 run_state_p.def_impl(_run_state_impl)
+mlir.register_lowering(run_state_p, mlir.lower_fun(_run_state_impl))
 
 def _run_state_abstract_eval(*avals: core.AbstractValue, jaxpr: core.Jaxpr,
                              which_linear: tuple[bool, ...]):
@@ -325,7 +327,7 @@ ad.primitive_jvps[run_state_p] = _run_state_jvp
 _save_everything = lambda *_, **__: True
 
 def _convert_outputs_to_writes(
-    jaxpr: core.Jaxpr) -> Tuple[core.Jaxpr, List[core.ShapedArray]]:
+    jaxpr: core.Jaxpr) -> tuple[core.Jaxpr, list[core.ShapedArray]]:
   assert not jaxpr.constvars, "Jaxpr shouldn't have constvars."
 
   in_avals = [v.aval for v in jaxpr.invars]
@@ -465,7 +467,7 @@ def _run_state_partial_eval(trace: pe.JaxprTrace, *tracers: pe.JaxprTracer,
 pe.custom_partial_eval_rules[run_state_p] = _run_state_partial_eval
 
 def _run_state_partial_eval_custom(
-    saveable: Callable[..., bool],
+    saveable: Callable[..., pe.RematCases_],
     in_unknowns: Sequence[bool],
     in_inst: Sequence[bool],
     eqn: core.JaxprEqn):
@@ -627,7 +629,7 @@ def _transpose_jaxpr(jaxpr: core.Jaxpr, which_linear: Sequence[bool]
 def _run_state_transpose(in_cts, *args, jaxpr: core.Jaxpr,
                          which_linear: tuple[bool, ...]):
   # if any in_ct is nonzero, we definitely want it in args_ (and the
-  # corresponding x in args could be an undefined primal, but doesnt have to be)
+  # corresponding x in args could be an undefined primal, but doesn't have to be)
   # for non-res stuff:
   #   getting and setting => (nonzero ct, UndefinedPrimal arg)
   #   just setting =>        (nonzero ct, not UndefinedPrimal, dummy value)
@@ -675,7 +677,7 @@ def _run_state_discharge_rule(in_avals: Sequence[core.AbstractValue],
 
 def initial_style_jaxpr(
     fun: Callable, in_tree: PyTreeDef, in_avals: Sequence[core.AbstractValue]
-  ) -> Tuple[core.Jaxpr, List[Any], PyTreeDef]:
+  ) -> tuple[core.Jaxpr, list[Any], PyTreeDef]:
   return _initial_style_jaxpr(fun, in_tree, tuple(in_avals))
 
 @weakref_lru_cache
