@@ -14,26 +14,22 @@
 """Module for state types."""
 from __future__ import annotations
 
+from collections.abc import Sequence
+import dataclasses
 import math
-from typing import Any, Generic, List, Sequence, Set, Tuple, TypeVar, Union
+from typing import Any, Generic, TypeVar, Union
 
 from jax._src import core
 from jax._src import effects
 from jax._src import pretty_printer as pp
-from jax._src import xla_bridge
-from jax._src.lib import xla_client
 from jax._src.util import safe_map, safe_zip
-
-xc = xla_client
-xb = xla_bridge
+from jax._src.typing import Array
 
 ## JAX utilities
 
 map, unsafe_map = safe_map, map
 zip, unsafe_zip = safe_zip, zip
 
-Array = Any
-
 _ref_effect_color = pp.Color.GREEN
 
 class RefEffect(effects.JaxprInputEffect):
@@ -78,6 +74,25 @@ StateEffect = Union[ReadEffect, WriteEffect, AccumEffect]
 
 Aval = TypeVar("Aval", bound=core.AbstractValue)
 
+@dataclasses.dataclass
+class RefIndexer:
+  ref: Any
+
+  def __getitem__(self, slc):
+    if not isinstance(slc, tuple):
+      slc = (slc,)
+    return RefView(self.ref, slc)
+
+@dataclasses.dataclass
+class RefView:
+  ref: Any
+  indexer: Any
+
+  @property
+  def at(self):
+    raise NotImplementedError("Can't call `.at` multiple times.")
+
+
 # We need an aval for `Ref`s so we can represent `get` and `swap` in Jaxprs.
 class AbstractRef(core.AbstractValue, Generic[Aval]):
   __slots__ = ["inner_aval"]
@@ -104,28 +119,32 @@ class AbstractRef(core.AbstractValue, Generic[Aval]):
       raise ValueError(f"`Ref{{{self.inner_aval.str_short()}}} has no `dtype`.")
     return self.inner_aval.dtype
 
+  @core.aval_property
+  def at(self):
+    return RefIndexer(self)
+
   @core.aval_method
   @staticmethod
   def get(tracer, idx=()):
-    from jax._src.state.primitives import ref_get
+    from jax._src.state.primitives import ref_get  # pytype: disable=import-error
     return ref_get(tracer, idx)
 
   @core.aval_method
   @staticmethod
   def set(tracer, value, idx=()):
-    from jax._src.state.primitives import ref_set
+    from jax._src.state.primitives import ref_set  # pytype: disable=import-error
     return ref_set(tracer, idx, value)
 
   def _getitem(self, tracer, idx) -> Array:
     if not isinstance(idx, tuple):
       idx = idx,
-    from jax._src.state.primitives import ref_get
+    from jax._src.state.primitives import ref_get  # pytype: disable=import-error
     return ref_get(tracer, idx)
 
   def _setitem(self, tracer, idx, value) -> None:
     if not isinstance(idx, tuple):
       idx = idx,
-    from jax._src.state.primitives import ref_set
+    from jax._src.state.primitives import ref_set  # pytype: disable=import-error
     return ref_set(tracer, idx, value)
 
   def __repr__(self) -> str:
@@ -155,12 +174,12 @@ core.aval_mapping_handlers[AbstractRef] = (_map_ref, _unmap_ref)
 
 def get_ref_state_effects(
     avals: Sequence[core.AbstractValue],
-    effects: core.Effects) -> List[Set[StateEffect]]:
+    effects: core.Effects) -> list[set[StateEffect]]:
   return [{eff for eff in effects
            if isinstance(eff, (ReadEffect, WriteEffect, AccumEffect))
            and eff.input_index == i} for i, _ in enumerate(avals)]
 
-def shaped_array_ref(shape: Tuple[int, ...], dtype,
+def shaped_array_ref(shape: tuple[int, ...], dtype,
                      weak_type: bool = False,
                      named_shape = None) -> AbstractRef[core.AbstractValue]:
   return AbstractRef(core.ShapedArray(shape, dtype, weak_type=weak_type,
