@@ -11,7 +11,9 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from __future__ import annotations
 
+from collections.abc import Generator, Iterable, Sequence
 from contextlib import contextmanager, ExitStack
 import inspect
 import io
@@ -22,7 +24,7 @@ import re
 import os
 import tempfile
 import textwrap
-from typing import Callable, List, Generator, Optional, Sequence, Tuple, Union
+from typing import Any, Callable, Optional
 import unittest
 import warnings
 import zlib
@@ -40,18 +42,21 @@ from jax._src.interpreters import mlir
 from jax.tree_util import tree_map, tree_all, tree_flatten, tree_unflatten
 from jax._src import api
 from jax._src import pjit as pjit_lib
+from jax._src import config as jax_config
 from jax._src import core
 from jax._src import dispatch
 from jax._src import dtypes as _dtypes
+from jax._src import monitoring
+from jax._src import stages
 from jax._src.interpreters import pxla
-from jax._src.config import (flags, bool_env, config,
+from jax._src.config import (bool_env, config,
                              raise_persistent_cache_errors,
                              persistent_cache_min_compile_time_secs)
 from jax._src.numpy.util import promote_dtypes, promote_dtypes_inexact
 from jax._src.util import unzip2
 from jax._src.public_test_util import (  # noqa: F401
     _assert_numpy_allclose, _check_dtypes_match, _default_tolerance, _dtype, check_close, check_grads,
-    check_jvp, check_vjp, default_gradient_tolerance, default_tolerance, device_under_test, tolerance)
+    check_jvp, check_vjp, default_gradient_tolerance, default_tolerance, tolerance)
 from jax._src import xla_bridge
 
 
@@ -59,19 +64,18 @@ from jax._src import xla_bridge
 # jax.test_util. Functionality appearing here is for internal use only, and
 # may be changed or removed at any time and without any deprecation cycle.
 
-FLAGS = flags.FLAGS
-flags.DEFINE_string(
+_TEST_DUT = jax_config.DEFINE_string(
     'jax_test_dut', '',
     help=
     'Describes the device under test in case special consideration is required.'
 )
 
-flags.DEFINE_integer(
+_NUM_GENERATED_CASES = jax_config.DEFINE_integer(
   'jax_num_generated_cases',
   int(os.getenv('JAX_NUM_GENERATED_CASES', '10')),
   help='Number of generated cases to test')
 
-flags.DEFINE_integer(
+_MAX_CASES_SAMPLING_RETRIES = jax_config.DEFINE_integer(
   'max_cases_sampling_retries',
   int(os.getenv('JAX_MAX_CASES_SAMPLING_RETRIES', '100')),
   'Number of times a failed test sample should be retried. '
@@ -79,29 +83,35 @@ flags.DEFINE_integer(
   'sampling process is terminated.'
 )
 
-flags.DEFINE_bool(
+_SKIP_SLOW_TESTS = jax_config.DEFINE_bool(
     'jax_skip_slow_tests',
     bool_env('JAX_SKIP_SLOW_TESTS', False),
     help='Skip tests marked as slow (> 5 sec).'
 )
 
-flags.DEFINE_string(
+_TEST_TARGETS = jax_config.DEFINE_string(
   'test_targets', os.getenv('JAX_TEST_TARGETS', ''),
   'Regular expression specifying which tests to run, called via re.search on '
   'the test name. If empty or unspecified, run all tests.'
 )
-flags.DEFINE_string(
+_EXCLUDE_TEST_TARGETS = jax_config.DEFINE_string(
   'exclude_test_targets', os.getenv('JAX_EXCLUDE_TEST_TARGETS', ''),
   'Regular expression specifying which tests NOT to run, called via re.search '
   'on the test name. If empty or unspecified, run all tests.'
 )
-
-flags.DEFINE_bool(
+TEST_WITH_PERSISTENT_COMPILATION_CACHE = jax_config.DEFINE_bool(
     'jax_test_with_persistent_compilation_cache',
     bool_env('JAX_TEST_WITH_PERSISTENT_COMPILATION_CACHE', False),
     help='If enabled, the persistent compilation cache will be enabled for all '
     'test cases. This can be used to increase compilation cache coverage.')
 
+# We sanitize test names to ensure they work with "unitttest -k" and
+# "pytest -k" test filtering. pytest accepts '[' and ']' but unittest -k
+# does not. We replace sequences of problematic characters with a single '_'.
+kSanitizeNameRE = re.compile(r"[ \"'\[\](){}<>=,._]+")
+def sanitize_test_name(s: str) -> str:
+  return kSanitizeNameRE.sub("_", s)
+
 def num_float_bits(dtype):
   return _dtypes.finfo(_dtypes.canonicalize_dtype(dtype)).bits
 
@@ -231,6 +241,22 @@ def count_pjit_cpp_cache_miss():
     pjit_lib._pjit_lower = original_pjit_lower
 
 
+@contextmanager
+def count_aot_jit_cpp_cache_miss():
+  original_call = stages.Compiled.call
+  count = [0]
+
+  def compiled_call_count(*args, **kwargs):
+    count[0] += 1
+    return original_call(*args, **kwargs)
+
+  stages.Compiled.call = compiled_call_count
+  try:
+    yield count
+  finally:
+    stages.Compiled.call = original_call
+
+
 @contextmanager
 def count_jit_and_pmap_compiles():
   # No need to clear any caches since we generally jit and pmap fresh callables
@@ -257,14 +283,9 @@ def assert_num_jit_and_pmap_compilations(times):
     raise AssertionError(f"Expected exactly {times} XLA compilations, "
                          f"but executed {count[0]}")
 
-def if_device_under_test(device_type: Union[str, Sequence[str]],
-                         if_true, if_false):
-  """Chooses `if_true` of `if_false` based on device_under_test."""
-  if device_under_test() in ([device_type] if isinstance(device_type, str)
-                             else device_type):
-    return if_true
-  else:
-    return if_false
+
+def device_under_test():
+  return _TEST_DUT.value or xla_bridge.get_backend().platform
 
 def supported_dtypes():
   if device_under_test() == "tpu":
@@ -292,14 +313,6 @@ def is_cloud_tpu():
   return 'libtpu' in xla_bridge.get_backend().platform_version
 
 
-def is_se_tpu():
-  return (
-      is_cloud_tpu() and not xla_bridge.using_pjrt_c_api()
-  ) or xla_bridge.get_backend().platform_version.startswith(
-      'StreamExecutor TPU'
-  )
-
-
 def is_device_tpu_v4():
   return jax.devices()[0].device_kind == "TPU v4"
 
@@ -313,13 +326,25 @@ def _get_device_tags():
     device_tags = {device_under_test()}
   return device_tags
 
-def skip_on_devices(*disabled_devices):
-  """A decorator for test methods to skip the test on certain devices."""
+def test_device_matches(device_types: Iterable[str]) -> bool:
+  assert not isinstance(
+      device_types, str
+  ), 'device_types should be a list of strings'
+  tags = _get_device_tags()
+  for device_type in device_types:
+    assert isinstance(device_type, str), device_type
+    if device_type in tags:
+      return True
+  return False
+
+test_device_matches.__test__ = False  # This isn't a test case, pytest.
+
+def _device_filter(predicate):
   def skip(test_method):
     @functools.wraps(test_method)
     def test_method_wrapper(self, *args, **kwargs):
       device_tags = _get_device_tags()
-      if device_tags & set(disabled_devices):
+      if not predicate():
         test_name = getattr(test_method, '__name__', '[unknown test]')
         raise unittest.SkipTest(
           f"{test_name} not supported on device with tags {device_tags}.")
@@ -327,6 +352,22 @@ def skip_on_devices(*disabled_devices):
     return test_method_wrapper
   return skip
 
+def skip_on_devices(*disabled_devices):
+  """A decorator for test methods to skip the test on certain devices."""
+  return _device_filter(lambda: not test_device_matches(disabled_devices))
+
+def run_on_devices(*enabled_devices):
+  """A decorator for test methods to run the test only on certain devices."""
+  return _device_filter(lambda: test_device_matches(enabled_devices))
+
+def device_supports_buffer_donation():
+  """A decorator for test methods to run the test only on devices that support
+  buffer donation."""
+  return _device_filter(
+      lambda: test_device_matches(mlir._platforms_with_donation)
+  )
+
+
 def set_host_platform_device_count(nr_devices: int):
   """Returns a closure that undoes the operation."""
   prev_xla_flags = os.getenv("XLA_FLAGS")
@@ -346,19 +387,6 @@ def set_host_platform_device_count(nr_devices: int):
   return undo
 
 
-def skip_on_xla_cpu_mlir(test_method):
-  """A decorator to skip tests when MLIR lowering is enabled."""
-  @functools.wraps(test_method)
-  def test_method_wrapper(self, *args, **kwargs):
-    xla_flags = os.getenv('XLA_FLAGS') or ''
-    if '--xla_cpu_use_xla_runtime' in xla_flags:
-      test_name = getattr(test_method, '__name__', '[unknown test]')
-      raise unittest.SkipTest(
-          f'{test_name} not supported on XLA:CPU MLIR')
-    return test_method(self, *args, **kwargs)
-  return test_method_wrapper
-
-
 def skip_on_flag(flag_name, skip_value):
   """A decorator for test methods to skip the test when flags are set."""
   def skip(test_method):        # pylint: disable=missing-docstring
@@ -727,10 +755,17 @@ def assert_dot_precision(expected_precision, fun, *args):
     else:
       assert precision == expected_precision, msg
 
+def assert_dot_preferred_element_type(expected, fun, *args, **kwargs):
+  jaxpr = api.make_jaxpr(partial(fun, **kwargs))(*args)
+  pref_eltypes = [eqn.params['preferred_element_type'] for eqn in iter_eqns(jaxpr.jaxpr)
+                   if eqn.primitive == lax.dot_general_p]
+  for pref_eltype in pref_eltypes:
+    msg = f"Unexpected preferred_element_type: {expected} != {pref_eltype}"
+    assert expected == pref_eltype, msg
 
 def cases_from_gens(*gens):
   sizes = [1, 3, 10]
-  cases_per_size = int(FLAGS.jax_num_generated_cases / len(sizes)) + 1
+  cases_per_size = int(_NUM_GENERATED_CASES.value / len(sizes)) + 1
   for size in sizes:
     for i in range(cases_per_size):
       yield (f'_{size}_{i}',) + tuple(gen(size) for gen in gens)
@@ -743,8 +778,8 @@ def named_cases_from_sampler(gen):
     if not isinstance(x, (list, tuple)):
       x = list(x)
     return [x[rng.randint(len(x))]]
-  while (len(seen) < FLAGS.jax_num_generated_cases and
-         retries < FLAGS.max_cases_sampling_retries):
+  while (len(seen) < _NUM_GENERATED_CASES.value and
+         retries < _MAX_CASES_SAMPLING_RETRIES.value):
     retries += 1
     cases = list(gen(choose_one))
     if not cases:
@@ -761,7 +796,7 @@ def named_cases_from_sampler(gen):
 
 # Random sampling for every parameterized test is expensive. Do it once and
 # cache the result.
-@functools.lru_cache(maxsize=None)
+@functools.cache
 def _choice(n, m):
   rng = np.random.RandomState(42)
   return rng.choice(n, size=m, replace=False)
@@ -772,7 +807,7 @@ def sample_product_testcases(*args, **kw):
   kw = [(k, list(v)) for k, v in kw.items()]
   n = math.prod(len(a) for a in args) * math.prod(len(v) for _, v in kw)
   testcases = []
-  for i in _choice(n, min(n, FLAGS.jax_num_generated_cases)):
+  for i in _choice(n, min(n, _NUM_GENERATED_CASES.value)):
     testcase = {}
     for a in args:
       testcase.update(a[i % len(a)])
@@ -803,12 +838,12 @@ def sample_product(*args, **kw):
 class JaxTestLoader(absltest.TestLoader):
   def getTestCaseNames(self, testCaseClass):
     names = super().getTestCaseNames(testCaseClass)
-    if FLAGS.test_targets:
-      pattern = re.compile(FLAGS.test_targets)
+    if _TEST_TARGETS.value:
+      pattern = re.compile(_TEST_TARGETS.value)
       names = [name for name in names
                if pattern.search(f"{testCaseClass.__name__}.{name}")]
-    if FLAGS.exclude_test_targets:
-      pattern = re.compile(FLAGS.exclude_test_targets)
+    if _EXCLUDE_TEST_TARGETS.value:
+      pattern = re.compile(_EXCLUDE_TEST_TARGETS.value)
       names = [name for name in names
                if not pattern.search(f"{testCaseClass.__name__}.{name}")]
     return names
@@ -818,7 +853,10 @@ def with_config(**kwds):
   """Test case decorator for subclasses of JaxTestCase"""
   def decorator(cls):
     assert inspect.isclass(cls) and issubclass(cls, JaxTestCase), "@with_config can only wrap JaxTestCase class definitions."
-    cls._default_config = {**JaxTestCase._default_config, **kwds}
+    cls._default_config = {}
+    for b in cls.__bases__:
+      cls._default_config.update(b._default_config)
+    cls._default_config.update(kwds)
     return cls
   return decorator
 
@@ -845,6 +883,7 @@ class JaxTestCase(parameterized.TestCase):
     'jax_numpy_dtype_promotion': 'strict',
     'jax_numpy_rank_promotion': 'raise',
     'jax_traceback_filtering': 'off',
+    'jax_legacy_prng_key': 'error',
   }
 
   _compilation_cache_exit_stack: Optional[ExitStack] = None
@@ -873,7 +912,7 @@ class JaxTestCase(parameterized.TestCase):
 
   @classmethod
   def setUpClass(cls):
-    if FLAGS.jax_test_with_persistent_compilation_cache:
+    if TEST_WITH_PERSISTENT_COMPILATION_CACHE.value:
       cls._compilation_cache_exit_stack = ExitStack()
       stack = cls._compilation_cache_exit_stack
       stack.enter_context(raise_persistent_cache_errors(True))
@@ -886,7 +925,7 @@ class JaxTestCase(parameterized.TestCase):
 
   @classmethod
   def tearDownClass(cls):
-    if FLAGS.jax_test_with_persistent_compilation_cache:
+    if TEST_WITH_PERSISTENT_COMPILATION_CACHE.value:
       cls._compilation_cache_exit_stack.close()
 
   def rng(self):
@@ -914,8 +953,8 @@ class JaxTestCase(parameterized.TestCase):
 
   def assertDtypesMatch(self, x, y, *, canonicalize_dtypes=True):
     if not config.x64_enabled and canonicalize_dtypes:
-      self.assertEqual(_dtypes.canonicalize_dtype(_dtype(x), allow_opaque_dtype=True),
-                       _dtypes.canonicalize_dtype(_dtype(y), allow_opaque_dtype=True))
+      self.assertEqual(_dtypes.canonicalize_dtype(_dtype(x), allow_extended_dtype=True),
+                       _dtypes.canonicalize_dtype(_dtype(y), allow_extended_dtype=True))
     else:
       self.assertEqual(_dtype(x), _dtype(y))
 
@@ -1057,7 +1096,7 @@ def ignore_warning(**kw):
 
 # -------------------- Mesh parametrization helpers --------------------
 
-MeshSpec = List[Tuple[str, int]]
+MeshSpec = list[tuple[str, int]]
 
 @contextmanager
 def with_mesh(named_shape: MeshSpec) -> Generator[None, None, None]:
@@ -1136,6 +1175,13 @@ class _LazyDtypes:
     supported = supported_dtypes()
     return type(dtypes)(d for d in dtypes if d in supported)
 
+  @_cached_property
+  def custom_floats(self):
+    return [np.dtype(t) for t in [
+      _dtypes.bfloat16, _dtypes.float8_e4m3b11fnuz,
+      _dtypes.float8_e4m3fn, _dtypes.float8_e4m3fnuz,
+      _dtypes.float8_e5m2, _dtypes.float8_e5m2fnuz]]
+
   @_cached_property
   def floating(self):
     return self.supported([np.float32, np.float64])
@@ -1199,7 +1245,7 @@ def strict_promotion_if_dtypes_match(dtypes):
   return jax.numpy_dtype_promotion('standard')
 
 _version_regex = re.compile(r"([0-9]+(?:\.[0-9]+)*)(?:(rc|dev).*)?")
-def _parse_version(v: str) -> Tuple[int, ...]:
+def _parse_version(v: str) -> tuple[int, ...]:
   m = _version_regex.match(v)
   if m is None:
     raise ValueError(f"Unable to parse version '{v}'")
@@ -1207,3 +1253,96 @@ def _parse_version(v: str) -> Tuple[int, ...]:
 
 def numpy_version():
   return _parse_version(np.__version__)
+
+def parameterized_filterable(*,
+    kwargs: Sequence[dict[str, Any]],
+    testcase_name: Optional[Callable[[dict[str, Any]], str]] = None,
+    one_containing: Optional[str] = None,
+):
+  """
+  Decorator for named parameterized tests, with filtering.
+
+  Works like parameterized.named_parameters, except that it supports the
+  `one_containing` option. This is useful to select only one of the tests,
+  and to leave the test name unchanged (helps with specifying the desired test
+  when debugging).
+
+  Args:
+    kwargs: Each entry is a set of kwargs to be passed to the test function.
+    testcase_name: Optionally, a function to construct the testcase_name from
+      one kwargs dict. If not given then kwarg may contain `testcase_name` and
+      if not, the test case name is constructed as `str(kwarg)`.
+      We sanitize the test names to work with -k test filters. See
+      `sanitize_test_name`.
+    one_containing: If given, then leave the test name unchanged, and use
+      only one `kwargs` whose `testcase_name` includes `one_containing`.
+  """
+  # Ensure that all kwargs contain a testcase_name
+  kwargs_with_testcase_name: Sequence[dict[str, Any]]
+  if testcase_name is not None:
+    kwargs_with_testcase_name = [
+      dict(testcase_name=sanitize_test_name(str(testcase_name(kw))), **kw)
+      for kw in kwargs]
+  else:
+    for kw in kwargs:
+      testcase_name = kw.get("testcase_name")
+      if testcase_name is None:
+        testcase_name = "_".join(f"{k}={str(kw[k])}"  # type: ignore
+                                 for k in sorted(kw.keys()))
+      kw["testcase_name"] = sanitize_test_name(testcase_name)  # type: ignore
+
+    kwargs_with_testcase_name = kwargs
+  if one_containing is not None:
+    filtered = tuple(kw for kw in kwargs_with_testcase_name
+                     if one_containing in kw["testcase_name"])
+    assert filtered, f"No testcase_name contains '{one_containing}'"
+    kw = filtered[0]
+    kw["testcase_name"] = ""
+    return parameterized.named_parameters([kw])
+  else:
+    return parameterized.named_parameters(*kwargs_with_testcase_name)
+
+@contextmanager
+def register_event_duration_listener(callback):
+  """Manages registering/unregistering an event duration listener callback."""
+  try:
+    monitoring.register_event_duration_secs_listener(callback)
+    yield
+  finally:
+    monitoring._unregister_event_duration_listener_by_callback(callback)
+
+
+@contextmanager
+def set_env(**kwargs):
+  """Context manager to temporarily set/unset one or more environment variables.
+
+  Example:
+
+    >>> import os
+    >>> os.environ['my_var'] = 'original'
+
+    >>> with set_env(my_var=None, other_var='some_value'):
+    ...   print("my_var is set:", 'my_var' in os.environ)
+    ...   print("other_var =", os.environ['other_var'])
+    ...
+    my_var is set: False
+    other_var = some_value
+
+    >>> os.environ['my_var']
+    'original'
+    >>> 'other_var' in os.environ
+    False
+  """
+  original = {key: os.environ.pop(key, None) for key in kwargs}
+  os.environ.update({k: v for k, v in kwargs.items() if v is not None})
+  try:
+    yield
+  finally:
+    _ = [os.environ.pop(key, None) for key in kwargs]
+    os.environ.update({k: v for k, v in original.items() if v is not None})
+
+def fwd_bwd_jaxprs(f, *example_args):
+  fwd_jaxpr, (y_shape, res_shape) = jax.make_jaxpr(
+      lambda *args: jax.vjp(f, *args), return_shape=True)(*example_args)
+  bwd_jaxpr = jax.make_jaxpr(lambda res, outs: res(outs))(res_shape, y_shape)
+  return fwd_jaxpr, bwd_jaxpr
