@@ -12,14 +12,13 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from collections.abc import Iterable, Iterator, Sequence
 import functools
 from functools import partial
 import itertools as it
 import logging
 import operator
-from typing import (Any, Callable, Generic, Iterable, Iterator, List,
-                    Optional, Sequence, Set, Tuple, TypeVar, overload,
-                    TYPE_CHECKING, cast)
+from typing import (Any, Callable, Generic, Optional, TypeVar, overload, TYPE_CHECKING, cast)
 import weakref
 
 import numpy as np
@@ -43,13 +42,13 @@ if TYPE_CHECKING:
   # to that used for builtins.zip in python/typeshed. This supports
   # return types matching input types for up to three arguments.
   @overload
-  def safe_zip(__arg1: Iterable[T1]) -> List[Tuple[T1]]: ...
+  def safe_zip(__arg1: Iterable[T1]) -> list[tuple[T1]]: ...
   @overload
-  def safe_zip(__arg1: Iterable[T1], __arg2: Iterable[T2]) -> List[Tuple[T1, T2]]: ...
+  def safe_zip(__arg1: Iterable[T1], __arg2: Iterable[T2]) -> list[tuple[T1, T2]]: ...
   @overload
-  def safe_zip(__arg1: Iterable[T1], __arg2: Iterable[T2], __arg3: Iterable[T3]) -> List[Tuple[T1, T2, T3]]: ...
+  def safe_zip(__arg1: Iterable[T1], __arg2: Iterable[T2], __arg3: Iterable[T3]) -> list[tuple[T1, T2, T3]]: ...
   @overload
-  def safe_zip(__arg1: Iterable[Any], __arg2: Iterable[Any], __arg3: Iterable[Any], __arg4: Iterable[Any], *args) -> List[Tuple[Any, ...]]: ...
+  def safe_zip(__arg1: Iterable[Any], __arg2: Iterable[Any], __arg3: Iterable[Any], __arg4: Iterable[Any], *args) -> list[tuple[Any, ...]]: ...
 
   def safe_zip(*args):
     args = list(map(list, args))
@@ -59,17 +58,7 @@ if TYPE_CHECKING:
     return list(zip(*args))
 
 else:
-  # TODO(phawkins): remove the hasattr condition after jaxlib 0.4.9 is the
-  # minimum
-  if hasattr(jaxlib_utils, 'safe_zip'):
-    safe_zip = jaxlib_utils.safe_zip
-  else:
-    def safe_zip(*args):
-      args = list(map(list, args))
-      n = len(args[0])
-      for arg in args[1:]:
-        assert len(arg) == n, f'length mismatch: {list(map(len, args))}'
-      return list(zip(*args))
+  safe_zip = jaxlib_utils.safe_zip
 
 
 if TYPE_CHECKING:
@@ -77,16 +66,16 @@ if TYPE_CHECKING:
   # to that used for builtins.map in python/typeshed. This supports
   # checking input types for the callable with up to three arguments.
   @overload
-  def safe_map(f: Callable[[T1], T], __arg1: Iterable[T1]) -> List[T]: ...
+  def safe_map(f: Callable[[T1], T], __arg1: Iterable[T1]) -> list[T]: ...
 
   @overload
-  def safe_map(f: Callable[[T1, T2], T], __arg1: Iterable[T1], __arg2: Iterable[T2]) -> List[T]: ...
+  def safe_map(f: Callable[[T1, T2], T], __arg1: Iterable[T1], __arg2: Iterable[T2]) -> list[T]: ...
 
   @overload
-  def safe_map(f: Callable[[T1, T2, T3], T], __arg1: Iterable[T1], __arg2: Iterable[T2], __arg3: Iterable[T3]) -> List[T]: ...
+  def safe_map(f: Callable[[T1, T2, T3], T], __arg1: Iterable[T1], __arg2: Iterable[T2], __arg3: Iterable[T3]) -> list[T]: ...
 
   @overload
-  def safe_map(f: Callable[..., T], __arg1: Iterable[Any], __arg2: Iterable[Any], __arg3: Iterable[Any], __arg4: Iterable[Any], *args) -> List[T]: ...
+  def safe_map(f: Callable[..., T], __arg1: Iterable[Any], __arg2: Iterable[Any], __arg3: Iterable[Any], __arg4: Iterable[Any], *args) -> list[T]: ...
 
   def safe_map(f, *args):
     args = list(map(list, args))
@@ -96,38 +85,28 @@ if TYPE_CHECKING:
     return list(map(f, *args))
 
 else:
-  # TODO(phawkins): remove the hasattr condition after jaxlib 0.4.9 is the
-  # minimum
-  if hasattr(jaxlib_utils, 'safe_map'):
-    safe_map = jaxlib_utils.safe_map
-  else:
-    def safe_map(f, *args):
-      args = list(map(list, args))
-      n = len(args[0])
-      for arg in args[1:]:
-        assert len(arg) == n, f'length mismatch: {list(map(len, args))}'
-      return list(map(f, *args))
+  safe_map = jaxlib_utils.safe_map
 
-def unzip2(xys: Iterable[Tuple[T1, T2]]
-    ) -> Tuple[Tuple[T1, ...], Tuple[T2, ...]]:
+def unzip2(xys: Iterable[tuple[T1, T2]]
+    ) -> tuple[tuple[T1, ...], tuple[T2, ...]]:
   """Unzip sequence of length-2 tuples into two tuples."""
   # Note: we deliberately don't use zip(*xys) because it is lazily evaluated,
   # is too permissive about inputs, and does not guarantee a length-2 output.
-  xs: List[T1] = []
-  ys: List[T2] = []
+  xs: list[T1] = []
+  ys: list[T2] = []
   for x, y in xys:
     xs.append(x)
     ys.append(y)
   return tuple(xs), tuple(ys)
 
-def unzip3(xyzs: Iterable[Tuple[T1, T2, T3]]
-    ) -> Tuple[Tuple[T1, ...], Tuple[T2, ...], Tuple[T3, ...]]:
+def unzip3(xyzs: Iterable[tuple[T1, T2, T3]]
+    ) -> tuple[tuple[T1, ...], tuple[T2, ...], tuple[T3, ...]]:
   """Unzip sequence of length-3 tuples into three tuples."""
   # Note: we deliberately don't use zip(*xyzs) because it is lazily evaluated,
   # is too permissive about inputs, and does not guarantee a length-3 output.
-  xs: List[T1] = []
-  ys: List[T2] = []
-  zs: List[T3] = []
+  xs: list[T1] = []
+  ys: list[T2] = []
+  zs: list[T3] = []
   for x, y, z in xyzs:
     xs.append(x)
     ys.append(y)
@@ -140,7 +119,7 @@ def subvals(lst, replace):
     lst[i] = v
   return tuple(lst)
 
-def split_list(args: Sequence[T], ns: Sequence[int]) -> List[List[T]]:
+def split_list(args: Sequence[T], ns: Sequence[int]) -> list[list[T]]:
   args = list(args)
   lists = []
   for n in ns:
@@ -149,14 +128,14 @@ def split_list(args: Sequence[T], ns: Sequence[int]) -> List[List[T]]:
   lists.append(args)
   return lists
 
-def partition_list(bs: Sequence[bool], l: Sequence[T]) -> Tuple[List[T], List[T]]:
+def partition_list(bs: Sequence[bool], l: Sequence[T]) -> tuple[list[T], list[T]]:
   assert len(bs) == len(l)
   lists = [], []  # type: ignore
   for b, x in zip(bs, l):
     lists[b].append(x)
   return lists
 
-def merge_lists(bs: Sequence[bool], l0: Sequence[T], l1: Sequence[T]) -> List[T]:
+def merge_lists(bs: Sequence[bool], l0: Sequence[T], l1: Sequence[T]) -> list[T]:
   assert sum(bs) == len(l1) and len(bs) - sum(bs) == len(l0)
   i0, i1 = iter(l0), iter(l1)
   out = [next(i1) if b else next(i0) for b in bs]
@@ -170,7 +149,7 @@ def split_dict(dct, names):
   assert not dct
   return lst
 
-def concatenate(xs: Iterable[Sequence[T]]) -> List[T]:
+def concatenate(xs: Iterable[Sequence[T]]) -> list[T]:
   """Concatenates/flattens a list of lists."""
   return list(it.chain.from_iterable(xs))
 
@@ -178,7 +157,7 @@ flatten = concatenate
 
 _unflatten_done = object()
 
-def unflatten(xs: Iterable[T], ns: Sequence[int]) -> List[List[T]]:
+def unflatten(xs: Iterable[T], ns: Sequence[int]) -> list[list[T]]:
   """Splits `xs` into subsequences of lengths `ns`.
 
   Unlike `split_list`, the `sum(ns)` must be equal to `len(xs)`."""
@@ -202,7 +181,7 @@ def curry(f):
   >>> curry(f)(2, 3, 4, 5)()
   26
   """
-  return partial(partial, f)
+  return wraps(f)(partial(partial, f))
 
 def toposort(end_nodes):
   if not end_nodes: return []
@@ -499,8 +478,8 @@ def distributed_debug_log(*pairs):
 
 
 class OrderedSet(Generic[T]):
-  elts_set: Set[T]
-  elts_list: List[T]
+  elts_set: set[T]
+  elts_list: list[T]
 
   def __init__(self):
     self.elts_set = set()
@@ -548,8 +527,8 @@ def _original_func(f):
   return f
 
 
-def set_module(module):
-  def wrapper(func):
+def set_module(module: str) -> Callable[[T], T]:
+  def wrapper(func: T) -> T:
     if module is not None:
       func.__module__ = module
     return func
@@ -605,3 +584,11 @@ else:
           "Decorator got wrong type: @use_cpp_method(is_enabled: bool=True)"
       )
     return decorator
+
+
+try:
+  # numpy 1.25.0 or newer
+  NumpyComplexWarning: type[Warning] = np.exceptions.ComplexWarning
+except AttributeError:
+  # legacy numpy
+  NumpyComplexWarning = np.ComplexWarning
