@@ -19,35 +19,32 @@ and provide some automatic type mapping logic for converting between Numpy and
 XLA. There are also a handful of related casting utilities.
 """
 
+from collections.abc import Mapping
+import dataclasses
 from functools import partial, lru_cache
+import glob
 import importlib
-import io
 import json
 import logging
 import os
+import pathlib
 import platform as py_platform
 import pkgutil
 import sys
 import threading
-from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Union
+from typing import Any, Callable, Optional, Union
 import warnings
-import numpy as np
 
-from jax._src import lib
 from jax._src import distributed
-from jax._src.config import flags, bool_env, config, int_env
+from jax._src import config as jax_config
+from jax._src.config import config
+from jax._src.lib import cuda_versions
 from jax._src.lib import xla_client
+from jax._src.lib import xla_extension
 from jax._src.lib import xla_extension_version
 from jax._src import traceback_util
 from jax._src import util
 
-iree: Optional[Any]
-
-try:
-  import jax._src.iree as iree  # type: ignore
-except (ModuleNotFoundError, ImportError):
-  iree = None
-
 logger = logging.getLogger(__name__)
 
 jax_plugins: Optional[Any]
@@ -64,120 +61,40 @@ traceback_util.register_exclusion(__file__)
 
 XlaBackend = xla_client.Client
 
-FLAGS = flags.FLAGS
-
 
 # TODO(phawkins): Remove jax_xla_backend.
-flags.DEFINE_string(
+_XLA_BACKEND = jax_config.DEFINE_string(
     'jax_xla_backend', '',
     'Deprecated, please use --jax_platforms instead.')
-flags.DEFINE_string(
+BACKEND_TARGET = jax_config.DEFINE_string(
     'jax_backend_target',
     os.getenv('JAX_BACKEND_TARGET', '').lower(),
     'Either "local" or "rpc:address" to connect to a remote service target.')
 # TODO(skye): warn when this is used once we test out --jax_platforms a bit
-flags.DEFINE_string(
+_PLATFORM_NAME = jax_config.DEFINE_string(
     'jax_platform_name',
     os.getenv('JAX_PLATFORM_NAME', '').lower(),
     'Deprecated, please use --jax_platforms instead.')
-flags.DEFINE_bool(
-    'jax_disable_most_optimizations',
-    bool_env('JAX_DISABLE_MOST_OPTIMIZATIONS', False),
-    'Try not to do much optimization work. This can be useful if the cost of '
-    'optimization is greater than that of running a less-optimized program.')
-flags.DEFINE_integer(
-    'jax_xla_profile_version', int_env('JAX_XLA_PROFILE_VERSION', 0),
-    'Optional profile version for XLA compilation. '
-    'This is meaningful only when XLA is configured to '
-    'support the remote compilation profile feature.')
-flags.DEFINE_string(
+CUDA_VISIBLE_DEVICES = jax_config.DEFINE_string(
     'jax_cuda_visible_devices', 'all',
     'Restricts the set of CUDA devices that JAX will use. Either "all", or a '
     'comma-separate list of integer device IDs.')
-flags.DEFINE_string(
+_ROCM_VISIBLE_DEVICES = jax_config.DEFINE_string(
     'jax_rocm_visible_devices', 'all',
     'Restricts the set of ROCM devices that JAX will use. Either "all", or a '
     'comma-separate list of integer device IDs.')
 
-def get_compile_options(
-    num_replicas: int,
-    num_partitions: int,
-    device_assignment=None,
-    use_spmd_partitioning: bool = True,
-    use_auto_spmd_partitioning: bool = False,
-    auto_spmd_partitioning_mesh_shape=[],
-    auto_spmd_partitioning_mesh_ids=[],
-    env_options_overrides: Optional[Dict[str, str]] = None,
-) -> xla_client.CompileOptions:
-  """Returns the compile options to use, as derived from flag values.
+_USE_MOCK_GPU_CLIENT = jax_config.DEFINE_bool(
+    name="use_mock_gpu_client",
+    default=False,
+    help="If True, use a mock GPU client instead of a real one.",
+)
 
-  Args:
-    num_replicas: Number of replicas for which to compile.
-    num_partitions: Number of partitions for which to compile.
-    device_assignment: Optional ndarray of jax devices indicating the assignment
-      of logical replicas to physical devices (default inherited from
-      xla_client.CompileOptions). Must be consistent with `num_replicas` and
-      `num_partitions`.
-    use_spmd_partitioning: boolean indicating whether to enable SPMD or MPMD
-      partitioning in XLA.
-    use_auto_spmd_partitioning: boolean indicating whether to automatically
-      generate XLA shardings for SPMD partitioner.
-    auto_spmd_partitioning_mesh_shape: device mesh shape used to create
-      auto_spmd_partitioning search space.
-    auto_spmd_partitioning_mesh_ids: device ids used to create
-      auto_spmd_partitioning search space.
-    env_options_overrides: dict of additional options parsed by the compiler
-  """
-  compile_options = xla_client.CompileOptions()
-  compile_options.num_replicas = num_replicas
-  compile_options.num_partitions = num_partitions
-  build_options = compile_options.executable_build_options
-  build_options.use_spmd_partitioning = use_spmd_partitioning
-  build_options.use_auto_spmd_partitioning = use_auto_spmd_partitioning
-  if use_auto_spmd_partitioning:
-    build_options.auto_spmd_partitioning_mesh_shape = auto_spmd_partitioning_mesh_shape
-    build_options.auto_spmd_partitioning_mesh_ids = auto_spmd_partitioning_mesh_ids
-  if device_assignment is not None:
-    logger.debug(
-        'get_compile_options: num_replicas=%s num_partitions=%s device_assignment=%s',
-        num_replicas, num_partitions, device_assignment)
-    device_assignment = np.array(device_assignment)
-
-    # Allow 1D device assignment if num_partitions is 1.
-    if (device_assignment.ndim == 1) and (num_partitions == 1):
-      device_assignment = device_assignment[:, None]
-
-    if num_replicas != device_assignment.shape[0]:
-      msg = 'device_assignment does not match num_replicas: {} vs {}.'
-      raise ValueError(msg.format(device_assignment, num_replicas))
-
-    if num_partitions != device_assignment.shape[1]:
-      msg = 'device_assignment does not match num_partitions: {} vs {}.'
-      raise ValueError(msg.format(device_assignment, num_partitions))
-
-    if device_assignment.dtype == object:
-      device_assignment = np.vectorize(lambda d: d.id, otypes=[int])(
-          device_assignment)
-    device_assignment = xla_client.DeviceAssignment.create(device_assignment)
-    assert device_assignment.replica_count() == num_replicas
-    assert device_assignment.computation_count() == num_partitions
-    compile_options.device_assignment = device_assignment
-
-  if env_options_overrides is not None:
-    compile_options.env_option_overrides = list(env_options_overrides.items())
-
-  debug_options = compile_options.executable_build_options.debug_options
-  if lib.cuda_path is not None:
-    debug_options.xla_gpu_cuda_data_dir = lib.cuda_path
-
-  if FLAGS.jax_disable_most_optimizations:
-
-    debug_options.xla_backend_optimization_level = 0
-    debug_options.xla_llvm_disable_expensive_passes = True
-    debug_options.xla_test_all_input_layouts = False
-
-  compile_options.profile_version = FLAGS.jax_xla_profile_version
-  return compile_options
+_MOCK_NUM_GPUS = jax_config.DEFINE_integer(
+    name="mock_num_gpus",
+    default=1,
+    help="Mock GPU client number of gpus.",
+)
 
 
 # Backends
@@ -202,89 +119,183 @@ def tpu_client_timer_callback(timer_secs: float) -> Optional[xla_client.Client]:
   return client
 
 
-# Backends, in increasing order of preference.
+# Backends
+#
 # We have no particular opinion about how "backends" relate to "devices". For
 # example, there could be multiple backends that provide the same kind of
 # device.
+
 BackendFactory = Callable[[], Optional[xla_client.Client]]
-_backend_factories: Dict[str, Tuple[BackendFactory, int]] = {}
+
+@dataclasses.dataclass
+class BackendRegistration:
+  factory: BackendFactory
+
+  # Priority of this backend when choosing a default backend. Higher = more
+  # preferred.
+  priority: int
+
+  # If this backend fails to initialize, should we log a user-visible warning?
+  # For plugins (e.g., TPU) we usually want a visible failure, because why
+  # install a plugin if you don't intend it to be used?
+  fail_quietly: bool = False
+
+  # Is this plugin experimental? If a plugin is deemed experimental, we issue
+  # a warning when it is initialized. This is mostly to set user expectations
+  # correctly: we don't want users to think that JAX is buggy because of a
+  # a buggy plugin.
+  experimental: bool = False
+
+_backend_factories: dict[str, BackendRegistration] = {}
 _default_backend: Optional[xla_client.Client] = None
-_backends : Dict[str, xla_client.Client] = {}
-_backends_errors : Dict[str, str] = {}
+_backends : dict[str, xla_client.Client] = {}
+_backend_errors : dict[str, str] = {}
 _backend_lock = threading.Lock()
+_plugins_registered: bool = False
+_plugin_lock = threading.Lock()
+
+# The set of known non-experimental plugins.
+#
+# If a plugin passes the JAX test suite, it can be added to the allowlist below.
+# Send a PR if you would like to be added.
+#
+# It is fine for a plugin not to implement every feature that JAX uses, provided
+# that a reasonable feature set is implemented and the plugin fails gracefully
+# for unimplemented features. Wrong outputs are not acceptable.
+_nonexperimental_plugins: set[str] = set()
 
 def register_backend_factory(name: str, factory: BackendFactory, *,
-                             priority: int = 0) -> None:
+                             priority: int = 0,
+                             fail_quietly: bool = True,
+                             experimental: bool = False) -> None:
   with _backend_lock:
     if name in _backends:
       raise RuntimeError(f"Backend {name} already initialized")
-  _backend_factories[name] = (factory, priority)
+  _backend_factories[name] = BackendRegistration(
+    factory, priority, fail_quietly, experimental)
 
 
-register_backend_factory('interpreter', xla_client.make_interpreter_client,
-                         priority=-100)
-register_backend_factory('cpu',
-                         partial(xla_client.make_cpu_client, use_tfrt=True),
-                         priority=0)
+register_backend_factory(
+    "cpu", xla_client.make_cpu_client, priority=0, fail_quietly=False
+)
+
+
+def _check_cuda_versions():
+  # TODO(phawkins): remove the test for None cuda_versions after jaxlib 0.4.17
+  # is the minimum.
+  if cuda_versions is None:
+    return
+
+  def _version_check(name, get_version, get_build_version,
+                     scale_for_comparison=1):
+    build_version = get_build_version()
+    try:
+      version = get_version()
+    except Exception as e:
+      raise RuntimeError(f"Unable to load {name}.") from e
+    if build_version // scale_for_comparison > version // scale_for_comparison:
+      raise RuntimeError(
+          f"Found {name} version {version}, but JAX was built against version "
+          f"{build_version}, which is newer. The copy of {name} that is "
+          "installed must be at least as new as the version against which JAX "
+          "was built."
+      )
+
+  _version_check("CUDA", cuda_versions.cuda_runtime_get_version,
+                 cuda_versions.cuda_runtime_build_version)
+  _version_check(
+      "cuDNN",
+      cuda_versions.cudnn_get_version,
+      cuda_versions.cudnn_build_version,
+      # NVIDIA promise both backwards and forwards compatibility for cuDNN patch
+      # versions: https://docs.nvidia.com/deeplearning/cudnn/developer-guide/index.html#api-compat
+      scale_for_comparison=100,
+  )
+  _version_check("cuFFT", cuda_versions.cufft_get_version,
+                 cuda_versions.cufft_build_version)
+  _version_check("cuSOLVER", cuda_versions.cusolver_get_version,
+                 cuda_versions.cusolver_build_version)
+  _version_check("cuPTI", cuda_versions.cupti_get_version,
+                 cuda_versions.cupti_build_version)
+  # TODO(phawkins): ideally we'd check cublas and cusparse here also, but their
+  # "get version" APIs require initializing those libraries, which we don't want
+  # to do here.
 
 
 def make_gpu_client(
-    *, platform_name: str, visible_devices_flag: str
+    *, platform_name: str, visible_devices_flag: jax_config.FlagHolder[str]
 ) -> xla_client.Client:
-  visible_devices = getattr(FLAGS, visible_devices_flag, "all")
+  visible_devices = visible_devices_flag.value
   allowed_devices = None
   if visible_devices != "all":
     allowed_devices = {int(x) for x in visible_devices.split(",")}
 
-  if xla_extension_version < 160:
-    return xla_client.make_gpu_client(
-        distributed_client=distributed.global_state.client,
-        node_id=distributed.global_state.process_id,
-        platform_name=platform_name,
-        allowed_devices=allowed_devices,
-    )
-  else:
-    # Remove `type: ignore` when the min jaxlib version (xla_extension_version)
-    # >= 160.
+  if platform_name == "cuda":
+    _check_cuda_versions()
+
+  if xla_extension_version <= 199:
     return xla_client.make_gpu_client(
         distributed_client=distributed.global_state.client,
         node_id=distributed.global_state.process_id,
         num_nodes=distributed.global_state.num_processes,
         platform_name=platform_name,
         allowed_devices=allowed_devices,
-    )  # type: ignore
+    )
+  use_mock_gpu_client = _USE_MOCK_GPU_CLIENT.value
+  num_nodes = (
+      _MOCK_NUM_GPUS.value
+      if use_mock_gpu_client
+      else distributed.global_state.num_processes
+  )
+
+  return xla_client.make_gpu_client(
+      distributed_client=distributed.global_state.client,
+      node_id=distributed.global_state.process_id,
+      num_nodes=num_nodes,
+      platform_name=platform_name,
+      allowed_devices=allowed_devices,
+      mock=use_mock_gpu_client,  # type: ignore[call-arg]
+  )
 
 
 if hasattr(xla_client, "make_gpu_client"):
   register_backend_factory(
-      'cuda', partial(make_gpu_client, platform_name='cuda',
-      visible_devices_flag='jax_cuda_visible_devices'),
-      priority=200)
+      "cuda",
+      partial(
+          make_gpu_client,
+          platform_name="cuda",
+          visible_devices_flag=CUDA_VISIBLE_DEVICES,
+      ),
+      priority=200,
+      fail_quietly=True,
+  )
   register_backend_factory(
-      'rocm', partial(make_gpu_client, platform_name='rocm',
-      visible_devices_flag='jax_rocm_visible_devices'),
-      priority=200)
+      "rocm",
+      partial(
+          make_gpu_client,
+          platform_name="rocm",
+          visible_devices_flag=_ROCM_VISIBLE_DEVICES,
+      ),
+      priority=200,
+      fail_quietly=True,
+  )
 
 
 if hasattr(xla_client, "make_tpu_client"):
+  # TODO(phawkins,skyewm): switch TPU plugin to use the PJRT plugin mechanism,
+  # and then fail loudly on initialization failure.
   register_backend_factory(
-    'tpu', partial(tpu_client_timer_callback, timer_secs=60.0), priority=300)
-
-if hasattr(xla_client, "make_plugin_device_client"):
-  # It is assumed that if jax has been built with a plugin client, then the
-  # user wants to use the plugin client by default. Therefore, it gets the
-  # highest priority.
-  register_backend_factory("plugin", xla_client.make_plugin_device_client,
-      priority=400)
+    'tpu', partial(tpu_client_timer_callback, timer_secs=60.0), priority=300,
+    fail_quietly=True)
 
 
 def _get_pjrt_plugin_names_and_library_paths(
     plugins_from_env: str,
-) -> Dict[str, str]:
+) -> dict[str, str]:
   """Gets the names and library paths of PJRT plugins to load from env var.
 
   Args:
-    plugins_from_env: plugin name and pathes from env var. It is in the format
+    plugins_from_env: plugin name and paths from env var. It is in the format
       of 'name1:path1,name2:path2' ('name1;path1,name2;path2' for windows).
 
   Returns:
@@ -309,7 +320,9 @@ def _get_pjrt_plugin_names_and_library_paths(
 
 def _get_pjrt_plugin_config(
     json_path: str,
-) -> Tuple[str, Optional[Mapping[str, Union[str, int, List[int], float]]]]:
+) -> tuple[
+    str, Optional[Mapping[str, Union[str, int, list[int], float, bool]]]
+]:
   """Gets PJRT plugin configuration from a json file.
 
   The json file needs to have a "library_path" field for the plugin library
@@ -318,7 +331,7 @@ def _get_pjrt_plugin_config(
   pairs. Please see xla_client._NameValueMapping for the supported types of
   values.
   """
-  with io.open(json_path, 'r') as f:
+  with open(json_path) as f:
     config = json.load(f)
   if 'library_path' not in config.keys():
     raise ValueError(
@@ -406,7 +419,7 @@ def register_plugin(
     *,
     priority: int = 400,
     library_path: Optional[str] = None,
-    options: Optional[Mapping[str, Union[str, int, List[int], float]]] = None,
+    options: Optional[Mapping[str, Union[str, int, list[int], float, bool]]] = None,
 ) -> None:
   """Registers a backend factory for the PJRT plugin.
 
@@ -419,20 +432,35 @@ def register_plugin(
     options: Optional. It is used when creating a PJRT plugin client.
   """
   def factory():
-    # Plugin may already be statically linked in some configurations.
-    if not xla_client.pjrt_plugin_loaded(plugin_name):
-      if library_path is None:
-        raise ValueError(
-            'The library path is None when trying to dynamically load the'
-            ' plugin.'
-        )
-      xla_client.load_pjrt_plugin_dynamically(plugin_name, library_path)
-    return xla_client.make_c_api_client(plugin_name, options)
+    if xla_extension_version >= 183:
+      if not xla_client.pjrt_plugin_initialized(plugin_name):
+        xla_client.initialize_pjrt_plugin(plugin_name)
+
+    if distributed.global_state.client is None:
+      return xla_client.make_c_api_client(plugin_name, options, None)
+    distribute_options = {
+        'node_id': distributed.global_state.process_id,
+        'num_nodes': distributed.global_state.num_processes,
+    }
+    if options is not None:
+      distribute_options.update(options)
+    return xla_client.make_c_api_client(
+        plugin_name, distribute_options, distributed.global_state.client
+    )
+
 
   logger.debug(
       'registering PJRT plugin %s from %s', plugin_name, library_path
   )
-  register_backend_factory(plugin_name, factory, priority=priority)
+  experimental = plugin_name not in _nonexperimental_plugins
+  register_backend_factory(plugin_name, factory, priority=priority,
+                           fail_quietly=False, experimental=experimental)
+  if library_path is not None:
+    if xla_extension_version >= 198:
+      return xla_client.load_pjrt_plugin_dynamically(plugin_name, library_path)
+    else:
+      xla_client.load_pjrt_plugin_dynamically(plugin_name, library_path)
+  return None
 
 
 def register_pjrt_plugin_factories_from_env() -> None:
@@ -464,22 +492,12 @@ def register_pjrt_plugin_factories_from_env() -> None:
     register_plugin(plugin_name, library_path=library_path, options=options)
 
 
-# Plugins in the namespace package `jax_plugins` will be imported.
-discover_pjrt_plugins()
-# Registers plugins names and paths set in env var PJRT_NAMES_AND_LIBRARY_PATHS,
-# in the format of 'name1:path1,name2:path2' ('name1;path1,name2;path2' for
-# windows).
-register_pjrt_plugin_factories_from_env()
-
-if iree is not None:
-  register_backend_factory("iree", iree.iree_client_factory, priority=-100)
-
 _platform_aliases = {
   "cuda": "gpu",
   "rocm": "gpu",
 }
 
-_alias_to_platforms: Dict[str, List[str]] = {}
+_alias_to_platforms: dict[str, list[str]] = {}
 for _platform, _alias in _platform_aliases.items():
   _alias_to_platforms.setdefault(_alias, []).append(_platform)
 
@@ -513,7 +531,7 @@ def canonicalize_platform(platform: str) -> str:
                      "Platforms present are: " + ",".join(b.keys()))
 
 
-def expand_platform_alias(platform: str) -> List[str]:
+def expand_platform_alias(platform: str) -> list[str]:
   """Expands, e.g., "gpu" to ["cuda", "rocm"].
 
   This is used for convenience reasons: we expect cuda and rocm to act similarly
@@ -524,10 +542,24 @@ def expand_platform_alias(platform: str) -> List[str]:
 def is_gpu(platform):
   return platform in ("cuda", "rocm")
 
-def backends() -> Dict[str, xla_client.Client]:
+def backends() -> dict[str, xla_client.Client]:
   global _backends
-  global _backends_errors
+  global _backend_errors
   global _default_backend
+  global _plugins_registered
+
+  # Needs a separate lock because register_backend_factory (called from
+  # register_plugin) requries to hold _backend_lock.
+  with _plugin_lock:
+    if not _plugins_registered:
+      # Plugins in the namespace package `jax_plugins` or have an entry-point
+      # under the `jax_plugins` group will be imported.
+      discover_pjrt_plugins()
+      # Registers plugins names and paths set in env var
+      # PJRT_NAMES_AND_LIBRARY_PATHS, in the format of 'name1:path1,name2:path2'
+      # ('name1;path1,name2;path2' for windows).
+      register_pjrt_plugin_factories_from_env()
+      _plugins_registered = True
 
   with _backend_lock:
     if _backends:
@@ -539,13 +571,19 @@ def backends() -> Dict[str, xla_client.Client]:
       for platform in jax_platforms:
         platforms.extend(expand_platform_alias(platform))
       priorities = range(len(platforms), 0, -1)
-      platforms_and_priorities = list(zip(platforms, priorities))
+      # If the user specified a list of platforms explicitly, always fail
+      # loudly.
+      fail_quietly_list = [False] * len(platforms)
+      platform_registrations = list(
+        zip(platforms, priorities, fail_quietly_list))
     else:
-      platforms_and_priorities = list(
-          (platform, priority) for platform, (_, priority)
-          in _backend_factories.items())
+      platform_registrations = list(
+          (platform, registration.priority, registration.fail_quietly)
+          for platform, registration
+          in _backend_factories.items()
+      )
     default_priority = -1000
-    for platform, priority in platforms_and_priorities:
+    for platform, priority, fail_quietly in platform_registrations:
       try:
         backend = _init_backend(platform)
         _backends[platform] = backend
@@ -554,55 +592,111 @@ def backends() -> Dict[str, xla_client.Client]:
           _default_backend = backend
           default_priority = priority
       except Exception as err:
-        if platform in ('cpu', 'interpreter'):
-          # We always expect the CPU and interpreter backends to initialize
-          # successfully.
-          raise
+        err_msg = f"Unable to initialize backend '{platform}': {err}"
+        if fail_quietly:
+          _backend_errors[platform] = str(err)
+          logger.info(err_msg)
         else:
-          # If the backend isn't built into the binary, or if it has no devices,
-          # we expect a RuntimeError.
-          err_msg = f"Unable to initialize backend '{platform}': {err}"
           if config.jax_platforms:
             err_msg += " (set JAX_PLATFORMS='' to automatically choose an available backend)"
-            raise RuntimeError(err_msg)
           else:
-            _backends_errors[platform] = str(err)
-            logger.info(err_msg)
-            continue
+            err_msg += " (you may need to uninstall the failing plugin package, or set JAX_PLATFORMS=cpu to skip this backend.)"
+          raise RuntimeError(err_msg)
+
     assert _default_backend is not None
-    # We don't warn about falling back to CPU on Mac OS, because we don't
-    # support anything else there at the moment and warning would be pointless.
-    if (py_platform.system() != "Darwin" and
-        _default_backend.platform == "cpu" and
-        FLAGS.jax_platform_name != 'cpu'):
-      logger.warning('No GPU/TPU found, falling back to CPU. '
-                      '(Set TF_CPP_MIN_LOG_LEVEL=0 and rerun for more info.)')
+    if not config.jax_platforms:
+      _suggest_missing_backends()
     return _backends
 
 
+# Code to suggest plugins that should be installed.
+#
+# Plugin vendors are welcome to add code to this list, assuming there's a
+# lightweight way to determine if hardware is present without requiring
+# the relevant plugin be installed.
+
+_GOOGLE_PCI_VENDOR_ID = '0x1ae0'
+_TPU_PCI_DEVICE_IDS = [
+    # TPU v2, v3
+    '0x0027',
+    # TPU v4
+    '0x005e',
+    # TPU v5e
+    '0x0063',
+    # Testing only
+    '0x0056',
+    '0x0062',
+]
+
+def _num_available_tpu_chips() -> int:
+  """Returns the number of TPU chips attached through PCI."""
+  num_chips = 0
+  for vendor_path in glob.glob('/sys/bus/pci/devices/*/vendor'):
+    vendor_id = pathlib.Path(vendor_path).read_text().strip()
+    if vendor_id != _GOOGLE_PCI_VENDOR_ID:
+      continue
+
+    device_path = os.path.join(os.path.dirname(vendor_path), 'device')
+    device_id = pathlib.Path(device_path).read_text().strip()
+    if device_id in _TPU_PCI_DEVICE_IDS:
+      num_chips += 1
+
+  return num_chips
+
+def _suggest_missing_backends():
+  if py_platform.system() != "Linux":
+    # If you're not using Linux (or WSL2), we don't have any suggestions at the
+    # moment.
+    return
+
+  assert _default_backend is not None
+  default_platform = _default_backend.platform
+  nvidia_gpu_devices = [
+    "/dev/nvidia0",
+    "/dev/dxg",  # WSL2
+  ]
+  if ("cuda" not in _backends and
+      any(os.path.exists(d) for d in nvidia_gpu_devices)):
+    if hasattr(xla_extension, "GpuAllocatorConfig") and "cuda" in _backend_errors:
+      err = _backend_errors["cuda"]
+      logger.warning(f"CUDA backend failed to initialize: {err} (Set "
+                     "TF_CPP_MIN_LOG_LEVEL=0 and rerun for more info.)")
+    else:
+      logger.warning("An NVIDIA GPU may be present on this machine, but a "
+                     "CUDA-enabled jaxlib is not installed. Falling back to "
+                     f"{default_platform}.")
+  elif "tpu" not in _backends and _num_available_tpu_chips() > 0:
+    logger.warning("A Google TPU may be present on this machine, but either a "
+                    "TPU-enabled jaxlib or libtpu is not installed. Falling "
+                    f"back to {default_platform}.")
+
+
 def _clear_backends() -> None:
   global _backends
-  global _backends_errors
+  global _backend_errors
   global _default_backend
 
   logger.info("Clearing JAX backend caches.")
   with _backend_lock:
     _backends = {}
-    _backends_errors = {}
+    _backend_errors = {}
     _default_backend = None
 
   get_backend.cache_clear()
 
 
 def _init_backend(platform: str) -> xla_client.Client:
-  factory, unused_priority = _backend_factories.get(platform, (None, None))
-  if factory is None:
+  registration = _backend_factories.get(platform, None)
+  if registration is None:
     raise RuntimeError(
         f"Backend '{platform}' is not in the list of known backends: "
         f"{list(_backend_factories.keys())}.")
 
+  if registration.experimental:
+    logger.warning(f"Platform '{platform}' is experimental and not all JAX "
+                   "functionality may be correctly supported!")
   logger.debug("Initializing backend '%s'", platform)
-  backend = factory()
+  backend = registration.factory()
   # TODO(skye): consider raising more descriptive errors directly from backend
   # factories instead of returning None.
   if backend is None:
@@ -625,17 +719,17 @@ def _get_backend_uncached(
   if platform is not None and not isinstance(platform, str):
     return platform
 
-  platform = (platform or FLAGS.jax_xla_backend or FLAGS.jax_platform_name
-              or None)
+  platform = (platform or _XLA_BACKEND.value or _PLATFORM_NAME.value or None)
 
   bs = backends()
   if platform is not None:
     platform = canonicalize_platform(platform)
     backend = bs.get(platform, None)
     if backend is None:
-      if platform in _backends_errors:
+      if platform in _backend_errors:
         raise RuntimeError(f"Backend '{platform}' failed to initialize: "
-                           f"{_backends_errors[platform]}")
+                           f"{_backend_errors[platform]}. "
+                           f'Available backends are {list(bs)}')
       raise RuntimeError(f"Unknown backend {platform}")
     return backend
   else:
@@ -690,7 +784,7 @@ def local_device_count(
 
 def devices(
     backend: Optional[Union[str, xla_client.Client]] = None
-) -> List[xla_client.Device]:
+) -> list[xla_client.Device]:
   """Returns a list of all devices for a given backend.
 
   .. currentmodule:: jaxlib.xla_extension
@@ -723,7 +817,7 @@ def default_backend() -> str:
 @lru_cache
 def local_devices(process_index: Optional[int] = None,
                   backend: Optional[Union[str, xla_client.Client]] = None,
-                  host_id: Optional[int] = None) -> List[xla_client.Device]:
+                  host_id: Optional[int] = None) -> list[xla_client.Device]:
   """Like :py:func:`jax.devices`, but only returns devices local to a given process.
 
   If ``process_index`` is ``None``, returns devices local to this process.
@@ -797,7 +891,7 @@ def host_count(backend: Optional[Union[str, xla_client.Client]] = None) -> int:
 # TODO: remove this sometime after jax 0.2.13 is released
 def host_ids(
     backend: Optional[Union[str, xla_client.Client]] = None
-) -> List[int]:
+) -> list[int]:
   warnings.warn(
       "jax.host_ids has been deprecated; please use range(jax.process_count()) "
       "instead. jax.host_ids will eventually be removed; please update your "
@@ -810,7 +904,7 @@ def using_pjrt_c_api(backend=None):
 
 
 # TODO(parkers): Get rid of this in favor of a generic way to get topologies.
-def make_pjrt_tpu_topology(topology_name=None, **kwargs):
+def make_pjrt_tpu_topology(topology_name='', **kwargs):
   # TODO(b/261484192): Make a system for lazily loading libtpu.so and call
   # that inside make_tfrt_tpu_c_api_device_topology.
   get_backend()  # Properly initialize libtpu.so.
