@@ -11,10 +11,11 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-"""GlobalDeviceArray serialization and deserialization."""
+"""Array serialization and deserialization."""
 
 import abc
 import asyncio
+from collections.abc import Awaitable, Sequence
 import itertools
 import logging
 from functools import partial
@@ -22,7 +23,7 @@ import os
 import re
 import time
 import threading
-from typing import Awaitable, Any, Callable, Dict, Optional, Sequence, Union
+from typing import Any, Callable, Optional, Union
 
 import jax
 from jax._src import distributed
@@ -39,6 +40,10 @@ TS_CONTEXT = ts.Context({'file_io_concurrency': {'limit': 128}})
 _REMOVED_VALUE = 'Value removed'
 _CHECKPOINT_SUCCESS = 'checkpoint_write_success'
 _module_unique_count = itertools.count()
+_DEFAULT_DRIVER = 'file'
+_DISTRIBUTED_SYSTEM_MSG = (
+    'Please initialize the distributed system via '
+    '`jax.distributed.initialize()` at the start of your program.')
 
 logger = logging.getLogger(__name__)
 
@@ -101,14 +106,14 @@ def get_tensorstore_spec(ckpt_path: str, ocdbt: bool = False):
     base_path = os.path.dirname(ckpt_path)
     spec['kvstore'] = {
         'driver': 'ocdbt',
-        'base': base_path if is_gcs_path else f'file://{base_path}',
+        'base': base_path if is_gcs_path else f'{_DEFAULT_DRIVER}://{base_path}',
         'path': os.path.basename(ckpt_path),
     }
   else:
     if is_gcs_path:
       spec['kvstore'] = _get_kvstore_for_gcs(ckpt_path)
     else:
-      spec['kvstore'] = {'driver': 'file', 'path': ckpt_path}
+      spec['kvstore'] = {'driver': _DEFAULT_DRIVER, 'path': ckpt_path}
 
   return spec
 
@@ -143,8 +148,11 @@ async def async_serialize(
 ):
   if (isinstance(arr_inp, array.ArrayImpl) and jax.process_count() > 1 and
       arr_inp.is_fully_addressable):
-    raise ValueError('Passing fully addressable Arrays to a multiprocess '
-                     'serialization is not allowed.')
+    raise ValueError(
+        f'Passing fully addressable arrays to a multiprocess '
+        f'serialization is not allowed, as this may lead to a race condition '
+        f'between processes. Serialization have failed for the array with '
+        f'the path "{tensorstore_spec["kvstore"]["path"]}".')
   # 'metadata' may not be present at the top level (for example, if we are using
   # a 'cast' driver).
   if not _spec_has_metadata(tensorstore_spec):
@@ -236,7 +244,7 @@ def estimate_read_memory_footprint(t: ts.TensorStore,
 
 async def async_deserialize(
     in_sharding: sharding_impls.XLACompatibleSharding,
-    tensorstore_spec: Union[ts.Spec, Dict[str, Any]],
+    tensorstore_spec: Union[ts.Spec, dict[str, Any]],
     global_shape: Optional[Sequence[int]] = None,
     dtype=None,
     byte_limiter: Optional[_LimitInFlightBytes] = None,
@@ -279,7 +287,7 @@ async def async_deserialize(
       # transfer instead of loading data. In the future, if memory pressure
       # becomes a problem, we can instead instrument  bytelimiter to
       # keep track of all in-flight tensors and only block_until_ready, if byte
-      # limiter hits the limit to get reduced memory usage, without loosing
+      # limiter hits the limit to get reduced memory usage, without losing
       # performance in common use cases.
       await byte_limiter.release_bytes(requested_bytes)
     return result
@@ -288,7 +296,7 @@ async def async_deserialize(
 
 
 def run_deserialization(shardings: Sequence[sharding.Sharding],
-                        tensorstore_specs: Sequence[Dict[str, Any]],
+                        tensorstore_specs: Sequence[dict[str, Any]],
                         global_shapes: Optional[Sequence[array.Shape]] = None,
                         dtypes: Optional[Sequence[typing.DTypeLike]] = None,
                         concurrent_gb: int = 32):
@@ -307,7 +315,7 @@ def run_deserialization(shardings: Sequence[sharding.Sharding],
   return asyncio.run(_run_deserializer())
 
 
-def _get_key(key: str):
+def _get_key(key: int):
   return f'tensorstore_checkpoint_{key}'
 
 
@@ -370,7 +378,7 @@ class GlobalAsyncCheckpointManagerBase(metaclass=abc.ABCMeta):
 
   @abc.abstractmethod
   def deserialize(self, shardings: Sequence[sharding.Sharding],
-                  tensorstore_specs: Sequence[Dict[str, Any]],
+                  tensorstore_specs: Sequence[dict[str, Any]],
                   global_shapes: Optional[Sequence[array.Shape]] = None,
                   dtypes: Optional[Sequence[typing.DTypeLike]] = None):
     """Deserializes GDAs from TensorStore."""
@@ -387,9 +395,7 @@ class AsyncManager:
     self._exception = None
 
     if jax.process_count() > 1 and distributed.global_state.client is None:
-      raise ValueError('Please initialize the distributed system via '
-                       '`jax.distributed.initialize()` at the start of your '
-                       'program.')
+      raise ValueError(_DISTRIBUTED_SYSTEM_MSG)
     if jax.process_count() > 1:
       self._client = distributed.global_state.client
     self._count = None
@@ -477,7 +483,7 @@ class GlobalAsyncCheckpointManager(AsyncManager, GlobalAsyncCheckpointManagerBas
   """Responsible for serializing GDAs via TensorStore."""
 
   def serialize(self, arrays, tensorstore_specs, *, on_commit_callback):
-    """Serializes GlobalDeviceArrays or Arrays via TensorStore asynchronously.
+    """Serializes Arrays or Arrays via TensorStore asynchronously.
 
     TensorStore writes to a storage layer in 2 steps:
     *  Reading/copying from the source after which the source can be modified.
@@ -489,7 +495,7 @@ class GlobalAsyncCheckpointManager(AsyncManager, GlobalAsyncCheckpointManagerBas
     finish in a separate thread allowing other computation to proceed.
 
     Args:
-      arrays: GlobalDeviceArrays or Arrays that should be serialized.
+      arrays: Arrays or Arrays that should be serialized.
       tensorstore_specs: TensorStore specs that are used to serialize GDAs or
         Arrays.
       on_commit_callback: This callback will be executed after all processes
@@ -518,11 +524,26 @@ class GlobalAsyncCheckpointManager(AsyncManager, GlobalAsyncCheckpointManagerBas
     # has finished writing.
     self._start_async_commit(on_commit_callback)
 
+  def serialize_with_paths(self, arrays: Sequence[jax.Array],
+                           paths: Sequence[str], *, on_commit_callback):
+    tspecs = jax.tree_map(get_tensorstore_spec, paths)
+    self.serialize(arrays, tspecs, on_commit_callback=on_commit_callback)
+
   def deserialize(self, shardings: Sequence[sharding.Sharding],
-                  tensorstore_specs: Sequence[Dict[str, Any]],
+                  tensorstore_specs: Sequence[dict[str, Any]],
                   global_shapes: Optional[Sequence[array.Shape]] = None,
                   dtypes: Optional[Sequence[typing.DTypeLike]] = None,
                   concurrent_gb: int = 32):
     self.wait_until_finished()
     return run_deserialization(shardings, tensorstore_specs,
                                global_shapes, dtypes, concurrent_gb)
+
+  def deserialize_with_paths(
+      self, shardings: Sequence[sharding.Sharding],
+      paths: Sequence[str],
+      global_shapes: Optional[Sequence[array.Shape]] = None,
+      dtypes: Optional[Sequence[typing.DTypeLike]] = None,
+      concurrent_gb: int = 32):
+    tspecs = jax.tree_map(get_tensorstore_spec, paths)
+    return self.deserialize(shardings, tspecs, global_shapes, dtypes,
+                            concurrent_gb)
