@@ -12,12 +12,14 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 """Workarounds for jax2tf transforms when XLA is not linked in."""
+
 import builtins
+from collections.abc import Sequence
 import dataclasses
 from functools import partial, wraps
 import math
 import string
-from typing import Any, Callable, Dict, Optional, Sequence, Tuple
+from typing import Any, Callable, Optional
 
 from jax._src import core
 from jax import lax
@@ -35,7 +37,7 @@ import tensorflow as tf  # type: ignore[import]
 # implementations are workarounds, making use of TF ops that do work when XLA is
 # not linked in. They are only used when the argument `enable_xla=False` when
 # calling jax2tf.convert().
-tf_impl_no_xla: Dict[core.Primitive, Callable[..., Any]] = {}
+tf_impl_no_xla: dict[core.Primitive, Callable[..., Any]] = {}
 
 
 TfVal = Any
@@ -70,7 +72,7 @@ def _invert_permutation(perm):
   return tuple(perm.index(i) for i in range(len(perm)))
 
 
-def _transpose_with_shape(x: TfVal, x_shape: core.Shape, permutation) -> Tuple[TfVal, core.Shape]:
+def _transpose_with_shape(x: TfVal, x_shape: core.Shape, permutation) -> tuple[TfVal, core.Shape]:
   """Computes transposition of x and its shape.
 
   x_shape matches x.shape in the known dimensions, and it has dimension
@@ -81,7 +83,7 @@ def _transpose_with_shape(x: TfVal, x_shape: core.Shape, permutation) -> Tuple[T
 
 def _transpose_for_tf_conv(lhs, lhs_shape: core.Shape,
                            rhs, rhs_shape: core.Shape, dimension_numbers):
-  """Tranposes lhs and rhs to respectively NHWC and HWIO so they can be passed to TF functions.
+  """Transposes lhs and rhs to respectively NHWC and HWIO so they can be passed to TF functions.
 
   The shapes passed in and returned may contain polynomials, and thus may
   be different than lhs.shape and rhs.shape.
@@ -89,7 +91,7 @@ def _transpose_for_tf_conv(lhs, lhs_shape: core.Shape,
   # TODO(marcvanzee): Add tests for this ops for shape polymorphism.
   lhs_perm, rhs_perm, _ = dimension_numbers
 
-  # TODO(marcvanzee): Consider merging tranposes if we want to optimize.
+  # TODO(marcvanzee): Consider merging transposes if we want to optimize.
   # For `lhs_perm` / `output_perm`, perm (0, 1, 2, 3) corresponds to "NCHW".
   lhs, lhs_shape = _transpose_with_shape(lhs, lhs_shape, lhs_perm)  # lhs --> "NCHW"
   if len(lhs_perm) == 3:
@@ -232,7 +234,7 @@ def _validate_conv_features(
   elif [is_depthwise, is_atrous, is_transpose].count(True) > 1:
     raise _conv_error(
         f"Can only do one of depthwise ({is_depthwise}), atrous ({is_atrous}) "
-        f"and tranposed convolutions ({is_transpose})")
+        f"and transposed convolutions ({is_transpose})")
 
   # We can implement batch grouping when there is a need for it.
   if batch_group_count != 1:
@@ -248,7 +250,7 @@ def _conv_general_dilated(
     lhs, rhs, *, window_strides, padding, lhs_dilation, rhs_dilation,
     dimension_numbers: lax.ConvDimensionNumbers, feature_group_count: int,
     batch_group_count: int,
-    precision: Optional[Tuple[PrecisionType, PrecisionType]],
+    precision: Optional[tuple[PrecisionType, PrecisionType]],
     preferred_element_type: Optional[DType],
     _in_avals: Sequence[core.ShapedArray], _out_aval: core.ShapedArray):
   """Implementation of lax.conv_general_dilated_p using XlaConv."""
@@ -288,7 +290,7 @@ def _conv_general_dilated(
   else:
     padding_type = pads_to_padtype(
       lhs_shape[1:3], rhs_dilated_shape, window_strides, padding)
-    # We only manually pad if we aren't using a tranposed convolutions.
+    # We only manually pad if we aren't using a transposed convolutions.
     if padding_type == "EXPLICIT":
       lhs, lhs_shape, padding = _check_pad_spatial_dims(lhs, lhs_shape, padding)
       padding_type = padding
@@ -358,7 +360,7 @@ tf_impl_no_xla[lax.conv_general_dilated_p] = _conv_general_dilated
 
 
 def _dot_general(lhs, rhs, *, dimension_numbers,
-                 precision: Optional[Tuple[PrecisionType, PrecisionType]],
+                 precision: Optional[tuple[PrecisionType, PrecisionType]],
                  preferred_element_type: Optional[DType],
                  _in_avals: Sequence[core.ShapedArray],
                  _out_aval: core.ShapedArray):
@@ -367,6 +369,9 @@ def _dot_general(lhs, rhs, *, dimension_numbers,
   del precision
   del preferred_element_type
 
+  lhs, rhs, convert_result = jax2tf._dot_general_convert_to_common_dtype(
+    lhs, _in_avals[0], rhs, _in_avals[1], _out_aval)
+
   (lhs_contracting, rhs_contracting), (lhs_batch, rhs_batch) = dimension_numbers
   lhs_ndim, rhs_ndim = len(lhs.shape), len(rhs.shape)
 
@@ -413,7 +418,7 @@ def _dot_general(lhs, rhs, *, dimension_numbers,
     if len(squeeze_idxs) != 0:
       assert all([result.shape[i] == 1 for i in squeeze_idxs])
       result = tf.squeeze(result, squeeze_idxs)
-    return result
+    return convert_result(result)
 
   new_id = iter(string.ascii_letters)
   lhs_axis_ids = [next(new_id) for _ in lhs.shape]
@@ -443,7 +448,7 @@ def _dot_general(lhs, rhs, *, dimension_numbers,
   assert lhs.dtype == rhs.dtype
   spec = "{},{}->{}".format("".join(lhs_axis_ids), "".join(rhs_axis_ids),
                             "".join(out_axis_ids))
-  return tf.linalg.einsum(spec, lhs, rhs)
+  return convert_result(tf.linalg.einsum(spec, lhs, rhs))
 
 
 tf_impl_no_xla[lax.dot_general_p] = _dot_general
@@ -641,6 +646,11 @@ def _reduce_monoid(operand, window_dimensions, window_strides, padding,
       has_only_spatial_dims=has_only_spatial_dims)
 
   def tf_pool(inputs, pooling_type):
+    if any(not core.is_constant_shape(s) for s in
+           (window_dimensions, window_strides, dilations)):
+      raise NotImplementedError(
+          f"TODO: use tf.nn.pool with dynamic shapes¨{window_dimensions=} "
+          f" {window_strides=} {dilations=}")
     result = tf.nn.pool(
         inputs,
         window_shape=window_dimensions,
@@ -672,8 +682,8 @@ def _reduce_monoid(operand, window_dimensions, window_strides, padding,
 def _reduce_window(*args, jaxpr, consts, window_dimensions,
                    window_strides, padding, base_dilation, window_dilation,
                    _in_avals: Sequence[core.ShapedArray],
-                   _out_aval: Tuple[core.ShapedArray, ...]
-                   ) -> Tuple[TfVal, ...]:
+                   _out_aval: tuple[core.ShapedArray, ...]
+                   ) -> tuple[TfVal, ...]:
   assert len(consts) == 0, "Reduction computation cannot have constants"
   operands, init_values = util.split_list(args, [len(args) // 2])
 
@@ -706,7 +716,7 @@ def _reduce_window(*args, jaxpr, consts, window_dimensions,
   }[computation_name]
   result = reduce_fn(result, init_value)
 
-  # The outut is expected to be wrapped in a tuple, and since we don't use
+  # The output is expected to be wrapped in a tuple, and since we don't use
   # variadic reductions, this tuple always contains a single element.
   return (result,)
 
@@ -860,7 +870,7 @@ def _pre_gather_for_multidim_indexing(args: GatherArgs):
           == start_index_map[0] and len(offset_dims) == len(op_shape) - 1):
     raise ValueError("unsupported dimension numbers")
   # We added a trailing dimension of size 1
-  if not core.symbolic_equal_dim(args.start_indices_shape[-1], 1):
+  if not core.definitely_equal(args.start_indices_shape[-1], 1):
     raise ValueError("start_indices shape[-1] should be 1")
   # Guess the axis
   axis = collapsed_slice_dims[0]
@@ -872,7 +882,7 @@ def _pre_gather_for_multidim_indexing(args: GatherArgs):
   if offset_dims != expected_offset_dims:
     raise ValueError("unsupported offset_dims")
   expected_slice_sizes = op_shape[:axis] + (1,) + op_shape[axis + 1:]  # type: ignore
-  if not core.symbolic_equal_shape(args.slice_sizes, expected_slice_sizes):
+  if not core.definitely_equal_shape(args.slice_sizes, expected_slice_sizes):
     raise ValueError("unsupported slice_sizes")
 
 
@@ -907,7 +917,7 @@ def _pre_gather_with_batch_dim(args: GatherArgs):
     raise ValueError("unsupported start_index_map")
 
   # The batch dims in `start_indices` and `operand` should agree.
-  if not core.symbolic_equal_dim(args.op_shape[0], args.start_indices_shape[0]):
+  if not core.definitely_equal(args.op_shape[0], args.start_indices_shape[0]):
     raise ValueError("Batch dimensions in operand and start_indices don't "
                      "agree")
 
@@ -923,7 +933,7 @@ def _pre_gather_with_batch_dims(args: GatherArgs):
     #   also need to re-work the output reshaping
     raise ValueError("only len(collapsed_slice_dims) == 0 is supported")
 
-  # NOTE: This supports higher dimensions than listed (the highest dimenison
+  # NOTE: This supports higher dimensions than listed (the highest dimension
   # in the tests is 3D so it is limited to that, but the implementation is
   # designed to handle higher dimensions (N-Dimensional)).
   if len(args.batch_dims) not in [1, 2, 3]:
@@ -948,7 +958,7 @@ def _gather_with_batch_dim(args: GatherArgs):
   return result
 
 
-def _gather_generate_indices(shape: Tuple[int, ...]):
+def _gather_generate_indices(shape: tuple[int, ...]):
   """
   Returns the indices of the according to `shape`:
     each element in the output is the index of an element of an array
@@ -1140,11 +1150,11 @@ tf_impl_no_xla[lax.dynamic_update_slice_p] = _dynamic_update_slice
 
 
 def shift_axes_forward(operand,
-                       axes: Tuple[int, ...],
+                       axes: tuple[int, ...],
                        inverse: bool = False,
                        forward: bool = True):
   """Shifts the tuple of axes to the front of an array"""
-  other_axes = tuple([i for i in range(len(operand.shape)) if i not in axes])
+  other_axes = tuple(i for i in range(len(operand.shape)) if i not in axes)
   fwd_order = axes + other_axes if forward else other_axes + axes
   order = fwd_order if not inverse else _invert_permutation(fwd_order)
   return tf.transpose(operand, order)
@@ -1199,8 +1209,8 @@ def convert_scatter_jax_to_tf(update_op, unsorted_segment_op=None):
     Wrapper around the scatter function.
     The underlying tf ops `tf.tensor_scatter_nd_update` and
     `tf.math.unsorted_segment_*` index from the front dimensions.
-    `tf.math.unsorted_segment_*` indexs to a depth 1 from the front.
-    `tf.tensor_scatter_nd_update` indexs from the front dimensions onwards,
+    `tf.math.unsorted_segment_*` indexes to a depth 1 from the front.
+    `tf.tensor_scatter_nd_update` indexes from the front dimensions onwards,
     with no ability to skip a dimension. This function shifts the axes to be
     indexed to the front then calls a front-specific implementation, then
     inverse-shifts the output.
