@@ -13,7 +13,8 @@
 # limitations under the License.
 """Tests for call_tf."""
 from functools import partial
-from typing import Callable, Dict, Tuple
+import os
+from typing import Callable
 import unittest
 
 from absl import logging
@@ -114,7 +115,7 @@ class CallTfTest(tf_test_util.JaxToTfTestCase):
   @_parameterized_jit
   def test_eval_pytree(self, with_jit=True):
 
-    def fun_tf(x: Dict, y: Tuple) -> Tuple:
+    def fun_tf(x: dict, y: tuple) -> tuple:
       return (x["first"] * x["second"], y[0] + y[1])
 
     x = dict(first=np.float32(3.), second=np.float32(4.))
@@ -349,7 +350,7 @@ class CallTfTest(tf_test_util.JaxToTfTestCase):
 
   @_parameterized_jit
   def test_with_multiple_capture(self, with_jit=True):
-    if jtu.device_under_test() == "gpu":
+    if jtu.test_device_matches(["gpu"]):
       raise unittest.SkipTest("Test fails on GPU")
     v2 = tf.Variable(2., dtype=np.float32)
     v3 = tf.Variable(3., dtype=np.float32)
@@ -372,7 +373,7 @@ class CallTfTest(tf_test_util.JaxToTfTestCase):
   @_parameterized_jit
   def test_grad_pytree(self, with_jit=False):
 
-    def fun_tf(x: Dict, y: Tuple) -> Tuple:
+    def fun_tf(x: dict, y: tuple) -> tuple:
       return x["first"] * x["second"] + 3. * y[0] + 4. * y[1]
 
     x = dict(first=np.float32(3.), second=np.float32(4.))
@@ -694,6 +695,48 @@ class CallTfTest(tf_test_util.JaxToTfTestCase):
     grad_res = grad_fun_jax(operand, indices)
     self.assertEqual(grad_res.shape, (100, 128))
 
+  def test_output_shape_dtype_none(self):
+    x = jnp.zeros((10), dtype=jnp.float32)
+
+    @tf.function(jit_compile=True, autograph=False)
+    def fun_tf(x):  # pylint: disable=unused-argument
+      return
+
+    fun_jax_1 = jax2tf.call_tf(fun_tf, output_shape_dtype=None)
+    fun_jax_2 = jax2tf.call_tf(fun_tf)
+    self.assertIsNone(fun_jax_1(x))
+    self.assertIsNone(fun_jax_2(x))
+    fun_jax_3 = jax2tf.call_tf(
+        fun_tf, output_shape_dtype=jax.ShapeDtypeStruct((10,), jnp.float32)
+    )
+    with self.assertRaisesRegex(
+        ValueError,
+        "The pytree of the TensorFlow function results does not match the"
+        " pytree of the declared output_shape_dtype",
+    ):
+      _ = fun_jax_3(x)
+
+  def test_output_shape_dtype_not_none(self):
+    x = jnp.zeros((10), dtype=jnp.float32)
+
+    @tf.function(jit_compile=True, autograph=False)
+    def fun_tf(x):
+      return x
+
+    fun_jax_1 = jax2tf.call_tf(
+        fun_tf, output_shape_dtype=jax.ShapeDtypeStruct((10,), jnp.float32)
+    )
+    fun_jax_2 = jax2tf.call_tf(fun_tf)
+    self.assertAllClose(fun_jax_1(x), fun_jax_2(x))
+
+    fun_jax_3 = jax2tf.call_tf(fun_tf, output_shape_dtype=None)
+    with self.assertRaisesRegex(
+        ValueError,
+        "The pytree of the TensorFlow function results does not match the"
+        " pytree of the declared output_shape_dtype",
+    ):
+      _ = fun_jax_3(x)
+
 
 class RoundTripToJaxTest(tf_test_util.JaxToTfTestCase):
   "Reloading output of jax2tf into JAX with call_tf"
@@ -899,7 +942,12 @@ class RoundTripToJaxTest(tf_test_util.JaxToTfTestCase):
 
       baseline = jax.tree_util.tree_map(conversion, baseline)
       candidate = jax.tree_util.tree_map(conversion, candidate)
-      self.assertAllClose(baseline, candidate)
+      tol = (
+          1e-2
+          if jtu.test_device_matches(["tpu"]) and dtype == np.float16
+          else None
+      )
+      self.assertAllClose(baseline, candidate, atol=tol, rtol=tol)
 
     # Eager mode
     assert_all_close_support_bfloat16(tf_f(x), tf_f_rt(x))
@@ -1247,6 +1295,12 @@ class RoundTripToTfTest(tf_test_util.JaxToTfTestCase):
   def test_several_round_trips(self,
                                f2_function=False, f2_saved_model=False,
                                f4_function=False, f4_saved_model=False):
+    if (f2_saved_model and
+        f4_saved_model and
+        not config.jax2tf_default_native_serialization):
+      # TODO: Getting error Found invalid capture Tensor("jax2tf_vjp/jax2tf_arg_0:0", shape=(), dtype=float32) when saving custom gradients
+      # when saving f4, but only with non-native serialization.
+      raise unittest.SkipTest("TODO: error invalid capture when saving custom gradients")
     x = np.array(.7, dtype=np.float32)
     # f(n)(x) = 2. * x^n
     def f(n):
@@ -1312,10 +1366,6 @@ class RoundTripToTfTest(tf_test_util.JaxToTfTestCase):
     x = jnp.array(3.0, dtype=jnp.float32)
     y = jnp.array(3.0, dtype=jnp.float32)
     z = jnp.array(5.0, dtype=jnp.float32)
-    output_shape_dtype = (
-        jax.ShapeDtypeStruct(x.shape, x.dtype),
-        jax.ShapeDtypeStruct(z.shape, z.dtype),
-    )
     f_jax = jax.jit(jax2tf.call_tf(tf_func_3, call_tf_graph=False))
     stablehlo_module = f_jax.lower(x, y, z).compiler_ir("stablehlo")
     self.assertNotIn("stablehlo.custom_call", str(stablehlo_module))
@@ -1324,7 +1374,6 @@ class RoundTripToTfTest(tf_test_util.JaxToTfTestCase):
         jax2tf.call_tf(
             tf_func_3,
             call_tf_graph=True,
-            output_shape_dtype=output_shape_dtype,
         )
     )
     with self.assertRaisesRegex(
@@ -1415,7 +1464,7 @@ class RoundTripToTfTest(tf_test_util.JaxToTfTestCase):
 
     # Check the xla_call_module version and function_list attributes.
     xla_call_module = xla_call_module_list[0]
-    self.assertEqual(xla_call_module.attr["version"].i, 5)
+    self.assertGreaterEqual(xla_call_module.attr["version"].i, 5)
     self.assertIn("function_list", str(xla_call_module.attr))
     xla_call_module_list.clear()
     called_index_list.clear()
@@ -1454,7 +1503,6 @@ class RoundTripToTfTest(tf_test_util.JaxToTfTestCase):
     jax_f = jax2tf.call_tf(
         tf.function(tf_f),
         call_tf_graph=True,
-        output_shape_dtype=jax.ShapeDtypeStruct((10,), jnp.float32),
     )
     tf_f_rt = jax2tf.convert(
         jax_f,
@@ -1485,7 +1533,6 @@ class RoundTripToTfTest(tf_test_util.JaxToTfTestCase):
     call_tf_print = jax2tf.call_tf(
         tf_print,
         call_tf_graph=True,
-        output_shape_dtype=None,
         ordered=True,
     )
 
@@ -1541,6 +1588,89 @@ class RoundTripToTfTest(tf_test_util.JaxToTfTestCase):
     )
     _, restored_model = tf_test_util.SaveAndLoadFunction(f_tf, input_args=[x])
 
+  @jtu.parameterized_filterable(
+    kwargs=[dict(poly=poly) for poly in [True, False]]
+  )
+  def test_call_tf_ordered_dead_inputs(self, *, poly: bool):
+    def f_jax(x1, x_dead, x3):
+      return (x1, jax2tf.call_tf(lambda x: tf.math.sin(x), ordered=True,
+                                 call_tf_graph=True)(x3))
+    if poly:
+      polymorphic_shapes = ["b", None, None]
+    else:
+      polymorphic_shapes = None
+    f_tf = jax2tf.convert(f_jax, polymorphic_shapes=polymorphic_shapes)
+    x1 = np.arange(3, dtype=np.float32)
+    x_dead = np.arange(4, dtype=np.float32)
+    x3 = np.arange(5, dtype=np.float32)
+    self.assertAllClose(f_jax(x1, x_dead, x3),
+                        f_tf(x1, x_dead, x3))
+
+  @parameterized.named_parameters([
+    dict(testcase_name=f"{ordered=}", ordered=ordered)
+    for ordered in [True, False]
+  ])
+  def test_call_tf_graph_polymorphic(self, ordered: bool):
+    @tf.function(jit_compile=True, autograph=False)
+    @partial(jax2tf.convert,
+      with_gradient=False,
+      native_serialization=True,
+      polymorphic_shapes=["(b)"])
+    @jax.jit
+    def tf_f_2(x):
+      tf_f = lambda x: print(tf.strings.length(tf.constant("hello, world")))
+      jax2tf.call_tf(tf_f,
+                     call_tf_graph=True,
+                     ordered=ordered)(x)
+      return x
+
+    x = np.arange(3, dtype=np.int32)
+    _ = tf.function(tf_f_2, autograph=False).get_concrete_function(x)
+
+  # TODO(b/293927250): call_tf_graph=True only accept concrete_function. The
+  # workaround here is to set `module.call=concrete_fn.`.
+  @unittest.skip(
+      "The root cause here is because the XLACallModule.function_list attribute"
+      " depends on JAX call_tf lowering. The 2nd time tf.SavedModel TF tracing"
+      " will not trigger call_tf tracing since it was already cached. The"
+      " solution is to create the `CallTFContext` to make TF tracing and JAX"
+      " tracing work together correctly."
+  )
+  def test_call_tf_graph_save_and_load(self):
+    def jax_func(x):
+      def func_tf(x):
+        return tf.math.sin(x)
+
+      return jnp.cos(
+          jax2tf.call_tf(func_tf, output_shape_dtype=x, call_tf_graph=True)(x)
+      )
+    data_inputs = (np.array([0.5, 0.7], dtype=np.float32),)
+
+    def tf_func(the_input):
+      res = jax2tf.convert(jax_func, native_serialization=True)(the_input)
+      return tf.identity(res, name="the_result")
+
+    jit_tf_func = tf.function(
+        tf_func,
+        autograph=False,
+        jit_compile=True,
+    )
+    # The next line is necessary to reproduce this issue. It trigger TF
+    # ConcreteFunction tracing. Otherwise, you will fail with another error
+    # `Found zero restored functions for caller function`.
+    _ = jit_tf_func.get_concrete_function(*data_inputs)
+    module = tf.Module()
+    module.call = jit_tf_func # Switching to concrete_function works.
+    root_dir = self.create_tempdir()
+    saved_model_dir = os.path.join(root_dir, "saved_model")
+    tf.saved_model.save(
+        module,
+        saved_model_dir,
+        options=tf.saved_model.SaveOptions(experimental_custom_gradients=False),
+    )
+    loaded_model = tf.saved_model.load(saved_model_dir)
+    res = loaded_model.call(*data_inputs)
+    self.assertAllClose(jax_func(*data_inputs), res)
 
 if __name__ == "__main__":
   absltest.main(testLoader=jtu.JaxTestLoader())
