@@ -13,13 +13,14 @@
 # limitations under the License.
 """See primitives_test docstring for how the Jax2TfLimitations are used."""
 
+from collections.abc import Sequence
 import itertools
-from typing import Any, Callable, Optional, Sequence, Union
+from typing import Any, Callable, Optional, Union
 
 import jax
 from jax import lax
 from jax import numpy as jnp
-from jax.config import config
+from jax import config
 from jax._src import test_util as jtu
 from jax._src import dtypes
 from jax.experimental.jax2tf.tests import primitive_harness
@@ -47,7 +48,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
       enabled: bool = True,
       # jax2tf specific
       modes=("eager", "graph", "compiled"),
-      native_lowering=FOR_NON_NATIVE,
+      native_serialization=FOR_NON_NATIVE,
       skip_tf_run=False,
       expect_tf_error: bool = True,
       skip_comparison=False,
@@ -57,7 +58,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
 
     Args :
       modes: one of "eager", "graph", "compiled"
-      for_native_lowering: A bitmask with some of {FOR_NATIVE, FOR_NON_NATIVE}
+      for_native_serialization: A bitmask with some of {FOR_NATIVE, FOR_NON_NATIVE}
         to specify how the limitation applies to native and non-native lowering.
       skip_tf_run: if set will skip the TF execution. Use this sparingly,
         prefer `expect_tf_error`. Use only when the test cannot recover from
@@ -83,7 +84,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
       modes = (modes,)
     assert all(m in ["eager", "graph", "compiled"] for m in modes), "Invalid modes: {modes}"
     self.modes = modes
-    self.native_lowering = native_lowering
+    self.native_serialization = native_serialization
     self.expect_tf_error = expect_tf_error
     self.skip_tf_run = skip_tf_run
     self.custom_assert = custom_assert
@@ -111,12 +112,12 @@ class Jax2TfLimitation(primitive_harness.Limitation):
       device: Optional[str] = None,
       mode: Optional[str] = None) -> bool:
     """Checks if this limitation is enabled for dtype and device and mode."""
-    native_lowering_mask = (
+    native_serialization_mask = (
         Jax2TfLimitation.FOR_NATIVE
-        if config.jax2tf_default_experimental_native_lowering
+        if config.jax2tf_default_native_serialization
         else Jax2TfLimitation.FOR_NON_NATIVE)
     return ((mode is None or mode in self.modes) and
-            (self.native_lowering & native_lowering_mask) and
+            (self.native_serialization & native_serialization_mask) and
             super().filter(device=device, dtype=dtype))
 
   @classmethod
@@ -141,7 +142,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
   # We keep here the explicit set of groups for which we don't have limitations
   harness_groups_no_limitations = {
       "abs", "add", "add_any", "and", "atan2", "bitcast_convert_type",
-      "broadcast", "broadcast_in_dim", "cbrt", "ceil", "clamp", "concatenate",
+      "broadcast", "broadcast_in_dim", "ceil", "clamp", "concatenate",
       "cos", "cosh", "complex", "conj", "convert_element_type", "cummax",
       "cummin", "device_put", "dynamic_slice", "dynamic_update_slice", "exp",
       "eq", "floor", "gather", "ge", "gt", "imag", "iota", "iota_2x32_shape",
@@ -200,20 +201,48 @@ class Jax2TfLimitation(primitive_harness.Limitation):
   @classmethod
   def acosh(cls, harness: primitive_harness.Harness):
     return [
-        custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu"), tol=1e-3),
+        custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu", "tpu"),
+                       tol=1e-3),
         custom_numeric(dtypes=[np.complex128], devices=("cpu", "gpu"), tol=1e-12),
         cls.helper_get_trig_custom_limitation(np.cosh)
     ]
 
   @classmethod
-  def approx_max_k(cls, harness: primitive_harness.Harness):
+  def approx_top_k(cls, harness: primitive_harness.Harness):
     supported_dtypes = jtu.supported_dtypes()
-    return Jax2TfLimitation(
-        "eager is not supported in CPU or GPU.",
-        dtypes=[t for t in [jnp.bfloat16, np.float16, np.float32]
-                if t in supported_dtypes],
-        devices=("cpu", "gpu", "tpu"),
-        modes=("graph", "compiled"))
+    def custom_assert(tst, result_jax, result_tf, *, args, tol, err_msg):
+      del tol, err_msg
+      # Tests only that the indices correspond to the returned values
+      jax_values, jax_indices = result_jax
+      tf_values, tf_indices = result_tf
+      operand, = args
+      def operand_values(indices):
+        if operand.ndim == 1:
+          return operand[indices]
+        elif operand.ndim == 2:
+          return operand[np.arange(operand.shape[0]).reshape((-1, 1)), indices]
+        else:
+          assert False
+      tst.assertAllClose(operand_values(jax_indices), jax_values)
+      tst.assertAllClose(operand_values(tf_indices), tf_values)
+
+    return [
+        missing_tf_kernel(
+            dtypes=[t for t in [jnp.bfloat16, np.float16, np.float32, np.float64]
+                    if t in supported_dtypes],
+            devices=("cpu", "gpu"),
+            modes=("graph", "eager")),
+        Jax2TfLimitation(
+            "compilation not supported for float64.",
+            dtypes=[np.float64],
+            devices=("cpu", "gpu"),
+            modes=("compiled",)),
+        custom_numeric(
+            dtypes=[t for t in [jnp.bfloat16, np.float16, np.float32, np.float64]
+                    if t in supported_dtypes],
+            devices=("cpu", "gpu"),
+            modes=("eager", "graph"),
+            custom_assert=custom_assert)]
 
   @classmethod
   def argmax(cls, harness: primitive_harness.Harness):
@@ -237,6 +266,8 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     return [
         custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu"), tol=1e-4,
                        modes=("eager", "graph", "compiled")),
+        custom_numeric(dtypes=[np.complex64], devices=("tpu", "gpu"), tol=2e-4,
+                       modes=("eager", "graph", "compiled")),
         custom_numeric(dtypes=[np.complex128], devices=("cpu", "gpu"), tol=1e-12,
                        modes=("eager", "graph", "compiled")),
         cls.helper_get_trig_custom_limitation(np.sin)
@@ -245,7 +276,8 @@ class Jax2TfLimitation(primitive_harness.Limitation):
   @classmethod
   def asinh(cls, harness: primitive_harness.Harness):
     return [
-        custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu"), tol=1e-3),
+        custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu", "tpu"),
+                       tol=1e-3),
         custom_numeric(dtypes=[np.complex128], devices=("cpu", "gpu"), tol=1e-12),
         cls.helper_get_trig_custom_limitation(np.sinh)
     ]
@@ -254,6 +286,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
   def atan(cls, harness: primitive_harness.Harness):
     return [
         custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu"), tol=1e-5),
+        custom_numeric(dtypes=[np.complex64], devices=("tpu"), tol=1e-3),
         custom_numeric(dtypes=[np.complex128], devices=("cpu", "gpu"), tol=1e-12),
         cls.helper_get_trig_custom_limitation(np.tan)
     ]
@@ -280,6 +313,12 @@ class Jax2TfLimitation(primitive_harness.Limitation):
   def bessel_i1e(cls, harness: primitive_harness.Harness):
     return cls.bessel_i0e(harness)
 
+  @classmethod
+  def cbrt(cls, harness: primitive_harness.Harness):
+    return [
+        custom_numeric(dtypes=[np.float32], devices=("tpu"), tol=1e-5),
+    ]
+
   @classmethod
   def cholesky(cls, harness: primitive_harness.Harness):
 
@@ -318,7 +357,18 @@ class Jax2TfLimitation(primitive_harness.Limitation):
 
   @classmethod
   def conv_general_dilated(cls, harness: primitive_harness.Harness):
+    prefer_elem = harness.params["preferred_element_type"]
     return [
+        Jax2TfLimitation(
+          "Non-deterministic NaN for conv_general_dilated with preferred_element_type",
+          dtypes=[
+            jnp.int32, np.int16, np.int64
+          ],
+          devices=["cpu", "gpu", "tpu"],
+          modes=("eager", "graph", "compiled"),
+          enabled=(prefer_elem is not None
+                   and prefer_elem in [jnp.bfloat16, np.float16, np.float32, np.float64]),
+          skip_comparison=True),
         # Even in compiled mode, for GPU we see a bit of discrepancy but
         # very minor.
         custom_numeric(dtypes=[np.float32], devices="gpu",
@@ -327,7 +377,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
         custom_numeric(dtypes=[np.float32], devices="cpu",
                        modes=("eager", "graph", "compiled"),
                        tol=1e-4,
-                       native_lowering=Jax2TfLimitation.FOR_NATIVE | Jax2TfLimitation.FOR_NON_NATIVE),
+                       native_serialization=Jax2TfLimitation.FOR_NATIVE | Jax2TfLimitation.FOR_NON_NATIVE),
         custom_numeric(description="higher numeric inaccuracy when `enable_xla=False`",
                        modes=("eager", "graph", "compiled"),
                        enabled=(not harness.params["enable_xla"]),
@@ -457,6 +507,40 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             devices=["cpu", "gpu", "tpu"],
             enabled=prefer_elem and np.dtype(harness.dtype) < np.dtype(prefer_elem),
             skip_comparison=True),
+        # TODO(necula): look into this, but this is only for non-native serialization
+        Jax2TfLimitation(
+            "Errors when lhs_dtype != rhs_dtype for non-native serialization with 64-bit types",
+            devices=["cpu", "gpu", "tpu"],
+            enabled=(harness.dtype != harness.params["rhs_dtype"] and
+                     (harness.dtype in [np.int64, np.uint64, np.float64] or
+                      harness.params["rhs_dtype"] in [np.int64, np.uint64, np.float64])),
+            skip_comparison=True),
+      # TODO(necula): look into this, but this is only for non-native serialization and enable_xla=False
+      Jax2TfLimitation(
+        "Errors for non-native serialization with enable_xla=False for certain input dtype combinations",
+        devices=["cpu", "gpu", "tpu"],
+        enabled=(not harness.params["enable_xla"] and
+                 (harness.dtype in [np.int16, np.uint32, np.uint16] or
+                  harness.params["rhs_dtype"] in [np.int16, np.uint32, np.uint16] or
+                  # Some combinations end up being widened to a larger type that is not
+                  # supported
+                  (harness.dtype, harness.params["rhs_dtype"]) in [
+                    (np.float16, jnp.bfloat16),
+                    (np.int32, np.float16),
+                    (np.int8, np.float16),
+                    (np.int8, np.uint8),
+                  ])),
+        skip_comparison=True,
+        skip_tf_run=True),
+        # TODO(necula): look into this, but this is only for non-native serialization
+        Jax2TfLimitation(
+            "Crash when lhs_dtype != rhs_dtype for non-native serialization on TPU for complex numbers",
+            devices=["tpu"],
+            enabled=(harness.dtype != harness.params["rhs_dtype"] and
+                     (harness.dtype in [np.complex64, np.complex128] or
+                      harness.params["rhs_dtype"] in [np.complex64, np.complex128])),
+            skip_comparison=True,
+            skip_tf_run=True),
         # JAX performs float16 matmuls in float32 on CPU, so the JAX result
         # may be more precise.
         custom_numeric(dtypes=[np.float16], devices=["cpu"], tol=1e-2,
@@ -649,12 +733,12 @@ class Jax2TfLimitation(primitive_harness.Limitation):
   def fft(cls, harness):
     return [
         Jax2TfLimitation(
-            "TF function not compileable",
+            "TF function not compilableble",
             devices=("cpu", "gpu"),
             dtypes=[np.float64],
             modes="compiled"),
         Jax2TfLimitation(
-            "TF function not compileable for IFFT and IRFFT",
+            "TF function not compilableble for IFFT and IRFFT",
             devices=("cpu", "gpu"),
             dtypes=[np.complex128],
             modes="compiled",
@@ -707,6 +791,8 @@ class Jax2TfLimitation(primitive_harness.Limitation):
       tst.assertAllClose(
           np.full((nr_special_cases,), 0., dtype=dtype),
           result_tf[special_cases])
+      if harness.dtype == np.float32:
+        tol = 1e-5
       # non-special cases are equal
       tst.assertAllClose(
           result_jax[~special_cases],
@@ -887,7 +973,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
                 "JAX always returns NaN, while TF returns the value NaN is compared with."
             ),
             modes=("eager", "graph", "compiled"),
-            native_lowering=Jax2TfLimitation.FOR_NON_NATIVE),
+            native_serialization=Jax2TfLimitation.FOR_NON_NATIVE),
         # TODO(b/269996580)
         custom_numeric(
             custom_assert=custom_assert,
@@ -898,7 +984,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
                 "different behavior of NaN propagation through min/max."
             ),
             modes=("eager", "graph", "compiled"),
-            native_lowering=Jax2TfLimitation.FOR_NATIVE)
+            native_serialization=Jax2TfLimitation.FOR_NATIVE)
     ]
 
   @classmethod
@@ -917,7 +1003,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
                 "JAX always returns NaN, while TF returns the value NaN is compared with."
             ),
             modes=("eager", "graph", "compiled"),
-            native_lowering=Jax2TfLimitation.FOR_NON_NATIVE),
+            native_serialization=Jax2TfLimitation.FOR_NON_NATIVE),
         # TODO(b/269996580)
         custom_numeric(
             custom_assert=custom_assert,
@@ -928,7 +1014,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
                 "different behavior of NaN propagation through min/max."
             ),
             modes=("eager", "graph", "compiled"),
-            native_lowering=Jax2TfLimitation.FOR_NATIVE)
+            native_serialization=Jax2TfLimitation.FOR_NATIVE)
     ]
 
   @classmethod
@@ -994,6 +1080,10 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             expect_tf_error=False,
             skip_comparison=True,
             enabled=not harness.params["enable_xla"]),
+      custom_numeric(devices="cpu", dtypes=[np.float32],
+                     modes=("eager", "graph", "compiled",), tol=1e-5),
+      custom_numeric(devices=("cpu", "gpu"), dtypes=[np.float16],
+                     modes=("eager", "graph", "compiled",), tol=5e-3),
     ]
 
   @classmethod
@@ -1049,6 +1139,9 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             expect_tf_error=True,
             skip_comparison=True,
             enabled=("modes_out_of_bounds" in harness.name and not harness.params["enable_xla"])),
+      custom_numeric(modes=("eager", "graph", "compiled"),
+                     dtypes=[np.float16], tol=5e-3,
+                     enabled=(not harness.params["enable_xla"])),
     ]
 
   @classmethod
@@ -1145,10 +1238,10 @@ class Jax2TfLimitation(primitive_harness.Limitation):
       # should also consider the gap between it and zero. Note that this code
       # relies on the singular values being in descending order.
       def compute_absolute_gap(s, m, n):
-        forward_appendant = np.Inf if m == n else 0
+        forward_appendant = np.inf if m == n else 0
         forward_diff = jnp.diff(s, axis=-1, append=forward_appendant)
         backward_diff = jnp.diff(
-            s[..., ::-1], axis=-1, append=np.Inf)[..., ::-1]
+            s[..., ::-1], axis=-1, append=np.inf)[..., ::-1]
         absolute_gap = jnp.minimum(jnp.abs(forward_diff),
                                    jnp.abs(backward_diff))
         return absolute_gap
@@ -1277,6 +1370,14 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             devices=("cpu", "gpu"),
             modes=("eager", "graph", "compiled"),
             enabled=(compute_uv == True)),
+        custom_numeric(
+            tol=1e-5,
+            description="custom numeric comparison when !compute_uv on TPU",
+            dtypes=[np.float32, np.complex64],
+            custom_assert=custom_assert,
+            devices=("tpu"),
+            modes=("eager", "graph", "compiled"),
+            enabled=not compute_uv),
         custom_numeric(
             tol=1e-2,
             description="custom numeric comparison when compute_uv on TPU",
@@ -1338,7 +1439,8 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             dtypes=[np.float16],
             devices=("gpu", "cpu"),
             modes=("eager", "graph")),
-        custom_numeric(dtypes=[np.float32], tol=5e-3)
+        custom_numeric(dtypes=[np.float32], tol=5e-3,
+                       modes=("eager", "graph", "compiled"))
     ]
 
   @classmethod
@@ -1357,7 +1459,7 @@ def custom_numeric(
     devices=("cpu", "gpu", "tpu"),
     custom_assert=None,
     enabled=True,
-    native_lowering=Jax2TfLimitation.FOR_NON_NATIVE,
+    native_serialization=Jax2TfLimitation.FOR_NON_NATIVE,
     tol=None) -> Jax2TfLimitation:
 
   return Jax2TfLimitation(
@@ -1368,15 +1470,14 @@ def custom_numeric(
       modes=modes,
       custom_assert=custom_assert,
       enabled=enabled,
-      native_lowering=native_lowering,
+      native_serialization=native_serialization,
       tol=tol)
 
 def custom_random_keys_output():
   def custom_assert(tst, result_jax, result_tf, *, args, tol, err_msg):
-    # TODO(frostig): Don't need this conditional once we always
-    # enable_custom_prng. We can even assert the isinstance instead.
+    # Here we handle both new-style and old-style keys; see JEP 9263
     def unwrap_keys(keys):
-      if isinstance(keys, jax.random.KeyArray):
+      if jax.dtypes.issubdtype(keys.dtype, jax.dtypes.prng_key):
         return jax._src.prng.random_unwrap(keys)
       else:
         return keys
@@ -1391,11 +1492,13 @@ def custom_random_keys_output():
 
 
 def missing_tf_kernel(*,
-                      description="op not defined for dtype",
-                      dtypes,
-                      modes=("eager", "graph", "compiled"),
-                      devices=("cpu", "gpu", "tpu"),
-                      enabled=True) -> Jax2TfLimitation:
+    description="op not defined for dtype",
+    dtypes,
+    modes=("eager", "graph", "compiled"),
+    devices=("cpu", "gpu", "tpu"),
+    native_serialization = Jax2TfLimitation.FOR_NON_NATIVE,
+    enabled=True) -> Jax2TfLimitation:
 
   return Jax2TfLimitation(
-      description, dtypes=dtypes, devices=devices, modes=modes, enabled=enabled)
+      description, dtypes=dtypes, devices=devices, modes=modes, enabled=enabled,
+      native_serialization=native_serialization)
