@@ -19,7 +19,7 @@ import contextlib
 import math
 import os
 import re
-from typing import Callable, Dict, Optional, Tuple
+from typing import Callable, Optional
 import unittest
 
 from absl import logging
@@ -35,11 +35,11 @@ from jax._src import core
 from jax._src import source_info_util
 from jax._src import test_util as jtu
 from jax._src.lib import xla_client
-from jax._src.lib.mlir.dialects import stablehlo
+from jax._src.lib.mlir.dialects import hlo
 import jax._src.xla_bridge
 from jax import config
 from jax.experimental import jax2tf
-from jax.experimental.jax2tf import jax_export
+from jax.experimental.export import export
 from jax.experimental.jax2tf.tests import tf_test_util
 from jax.experimental.maps import xmap
 from jax.experimental.shard_map import shard_map
@@ -131,7 +131,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_pytrees(self):
     # Take and return pytrees
-    def f_jax(x: Tuple[float, Dict[str, float]]) -> Tuple[float, Dict[str, float]]:
+    def f_jax(x: tuple[float, dict[str, float]]) -> tuple[float, dict[str, float]]:
       x_a, x_dict = x
       return x_a * 2., {k: v * 3. for k, v in x_dict.items()}
 
@@ -189,7 +189,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     f_tf = tf.function(lambda x: x + x)
     self.assertEqual(f_tf(jnp.ones([])).numpy(), 2.)
 
-    # Test with ShardedDeviceArray.
+    # Test with a PmapSharding-sharded Array.
     n = jax.local_device_count()
     mk_sharded = lambda f: jax.pmap(lambda x: x)(f([n]))
     f_tf = tf.function(lambda x: x)
@@ -320,9 +320,25 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.assertAllClose(5., tape.gradient(v, x))
     self.assertAllClose(4., tape.gradient(v, y))
 
+  def test_higher_order_gradients(self):
+    f = lambda x: x ** 3
+    f_tf = jax2tf.convert(f)
+    x = tf.Variable(4.0, dtype=tf.float32)  # Create a Tensorflow variable initialized to 4.0
+    with tf.GradientTape() as t2:
+      with tf.GradientTape() as t1:
+        y = f_tf(x)
+
+      # Compute the gradient inside the outer `t2` context manager
+      # which means the gradient computation is differentiable as well.
+      dy_dx = t1.gradient(y, x)
+    d2y_dx2 = t2.gradient(dy_dx, x)
+
+    self.assertAllClose(np.float32(48.), dy_dx.numpy())
+    self.assertAllClose(np.float32(24.), d2y_dx2.numpy())
+
   @jtu.sample_product(with_function=[False, True])
   def test_gradients_pytree(self, with_function=False):
-    def f(xy: Tuple[float, float]) -> Dict[str, float]:
+    def f(xy: tuple[float, float]) -> dict[str, float]:
       x, y = xy
       return dict(one=x * x, two=x * y)
 
@@ -1132,18 +1148,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
     self.ConvertAndCompare(randint)
 
-  def test_error_disallowed_custom_call(self):
-    if jtu.device_under_test() != "cpu":
-      self.skipTest("Test intended for CPU only")
-    # For now triangular_solve on CPU uses the unsupported "blas_strsm" target
-    a = np.arange(16, dtype=np.float32).reshape((4, 4))
-    b = np.arange(4, dtype=np.float32).reshape((4, 1))
-    with self.assertRaisesRegex(ValueError,
-        "Cannot serialize code with custom calls whose targets .*"):
-      jax2tf.convert(
-          lambda a, b: jax.lax.linalg.triangular_solve(a, b, left_side=True),
-          native_serialization=True)(a, b)
-
   def test_op_metadata_simple(self):
     self.skipTest("include_xla_op_metadata not yet enabled")
     # A simple example
@@ -1534,22 +1538,21 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
         stack.enter_context(mesh)
       # Run the JAX native version, to check it works, and to fill caches.
       _ = func_to_convert(*args)
-      exported = jax_export.export(
+      exported = export.export(
           func_to_convert,
-          lowering_platform='tpu',
-          strict_checks=True
+          lowering_platform='tpu'
       )(*(core.ShapedArray(a.shape, a.dtype) for a in args))
 
     if transform1 == "shard_map":
-      self.assertIn("stablehlo.all_gather", str(exported.mlir_module))
+      self.assertIn("stablehlo.all_gather", str(exported.mlir_module()))
     else:
-      self.assertIn("stablehlo.reduce_window", str(exported.mlir_module))
+      self.assertIn("stablehlo.reduce_window", str(exported.mlir_module()))
 
   def test_cross_platform_error(self):
     f_tf = jax2tf.convert(jnp.sin, native_serialization=True,
                           native_serialization_platforms=('tpu',))
     x = np.float32(.5)
-    if jtu.device_under_test() == "tpu":
+    if jtu.test_device_matches(["tpu"]):
       self.assertAllClose(jnp.sin(x), f_tf(x))
     else:
       # We can construct the tf.Graph
@@ -1574,7 +1577,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       self.assertAllClose(jnp.sin(x), f_tf(x))
 
     f_tf = jax2tf.convert(jnp.sin,
-                          native_serialization_strict_checks=False)
+                          native_serialization_disabled_checks=(
+                            jax2tf.DisabledSafetyCheck.platform(),))
     self.assertAllClose(jnp.sin(x), f_tf(x))
 
   def test_native_serialization_grad(self):
@@ -1599,7 +1603,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     logging.info("Found graph_def: %s", graph_def)
     self.assertLen(re.findall(r'op:\s*"XlaCallModule"', str(graph_def)), 2)
 
-    if jtu.device_under_test() != "tpu":
+    if not jtu.test_device_matches(["tpu"]):
       with self.assertRaisesRegex(
           tf.errors.NotFoundError,
           r"The current platform .* is not among the platforms required by the module: \[TPU\]"):
@@ -1625,7 +1629,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
   def test_tuple_args(self):
     # On TPU if we have more than 2000 arguments, we pass them as a tuple.
     # This is a compiler option, and should have no effect on lowering.
-    if jtu.device_under_test() != "tpu":
+    if not jtu.test_device_matches(["tpu"]):
       raise unittest.SkipTest("Test enabled on TPU only")
     def f_jax(*many_args):
       acc = 0.
@@ -1660,125 +1664,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     f_tf_nested = jax2tf.convert(f_jax_nested, native_serialization=True)
     self.assertAllClose(res, f_tf_nested(inputs))
 
-def get_serialized_computation(
-    f_jax: Callable,
-    *args,
-    abstracted_axes: Optional[Tuple[Dict[int, str]]] = None,
-    use_pjit: bool = False,
-    in_shardings = None,
-    out_shardings = None) -> Tuple[str, int]:
-  if use_pjit:
-    assert not abstracted_axes
-    lowered = pjit.pjit(
-        f_jax, in_shardings=in_shardings, out_shardings=out_shardings
-  ).lower(*args)
-  else:
-    lowered = jax.jit(f_jax, abstracted_axes=abstracted_axes).lower(*args)
-  mlir_module = lowered._lowering.stablehlo()
-  xla_call_module_version = 5
-  mlir_str = mlir.module_to_bytecode(mlir_module)
-  if stablehlo.get_api_version() < 4:
-    target_version = stablehlo.get_earliest_forward_compatible_version()
-  else:
-    # See comments next to the usage of stablehlo.get_minimum_version() in
-    # jax_export.py for an explanation how it works.
-    target_version = stablehlo.get_minimum_version()
-  mlir_module_serialized = xla_client._xla.mlir.serialize_portable_artifact(
-    mlir_str, target_version)
-  return mlir_module_serialized, xla_call_module_version
-
-class XlaCallModuleTest(tf_test_util.JaxToTfTestCase):
-  """Unit tests for XlaCallModule. Will move these eventually to TF."""
-  def test_simple(self):
-
-    def f_jax(x):
-      return jnp.sin(x)
-
-    x = np.ones((2, 3), dtype=np.float32)
-
-    jax_res = f_jax(x)
-    module, version = get_serialized_computation(f_jax, x)
-    res = tfxla.call_module([x],
-                            version=version,
-                            module=module,
-                            Tout=[jax_res.dtype],
-                            Sout=[jax_res.shape])
-    self.assertAllClose(tf.nest.map_structure(lambda t: t.numpy(), res),
-                        [jax_res])
-
-  def test_while(self):
-    # With nested computation
-    def f_jax(count, x):
-      return lax.while_loop(lambda carry: carry[0] < count, lambda carry:
-                            (carry[0] + 1, carry[1] + 1.), (0, x))[1]
-
-    count = np.int32(5)
-    x = np.ones((2, 3), dtype=np.float32)
-
-    jax_res = f_jax(count, x)
-    module, version = get_serialized_computation(f_jax, count, x)
-    res = tfxla.call_module([count, x],
-                            version=version,
-                            module=module,
-                            Tout=[jax_res.dtype],
-                            Sout=[jax_res.shape])
-    self.assertAllClose(tf.nest.map_structure(lambda t: t.numpy(), res),
-                        [jax_res])
-
-  def test_multiple_args_results(self):
-
-    def f_jax(x1, x2):
-      return (jnp.sin(x1), jnp.cos(x2))
-
-    x1 = np.ones((2, 3), dtype=np.float32)
-    x2 = np.ones((3, 4), dtype=np.float32)
-
-    jax_res = f_jax(x1, x2)
-    module, version = get_serialized_computation(f_jax, x1, x2)
-    def f_tf(x1_tf, x2_tf):
-      return tfxla.call_module([x1_tf, x2_tf],
-                               version=version,
-                               module=module,
-                               Tout=[jax_res[0].dtype, jax_res[1].dtype],
-                               Sout=[jax_res[0].shape, jax_res[1].shape])
-
-    res = tf.function(f_tf, jit_compile=True, autograph=False)(x1, x2)
-    self.assertAllClose(tf.nest.map_structure(lambda t: t.numpy(), res),
-                        jax_res)
-
-  @jtu.with_mesh([("x", 2)])
-  def test_pjit_basic1D(self):
-
-    def func_jax(x, y):
-      return x + y
-
-    shape = (8, 10)
-    x = np.arange(math.prod(shape), dtype=np.float32).reshape(shape)
-    in_axis_resources = (P("x"), P("x"))
-    out_axis_resources = None
-    res_jax = pjit.pjit(
-        func_jax,
-        in_shardings=in_axis_resources,
-        out_shardings=out_axis_resources,
-    )(x, x)
-    module, version = get_serialized_computation(
-        func_jax,
-        x,
-        x,
-        use_pjit=True,
-        in_shardings=in_axis_resources,
-        out_shardings=out_axis_resources)
-
-    def f_tf(x_tf, y_tf):
-      return tfxla.call_module([x_tf, y_tf],
-                               version=version,
-                               module=module,
-                               Tout=[x.dtype],
-                               Sout=[x.shape])
-
-    res_tf = tf.function(f_tf, jit_compile=True, autograph=False)(x, x)[0]
-    self.assertAllClose(res_tf.numpy(), res_jax)
-
 
 @jtu.with_config(jax_enable_custom_prng=True)
 class Jax2tfWithCustomPRNGTest(tf_test_util.JaxToTfTestCase):
@@ -1808,6 +1693,15 @@ class Jax2tfWithCustomPRNGTest(tf_test_util.JaxToTfTestCase):
     jax_result = func()
     self.assertEqual(tf_result, jax_result)
 
+class Jax2TfVersioningTest(tf_test_util.JaxToTfTestCase):
+  # Use a separate test case with the default jax_serialization_version
+  def setUp(self):
+    self.use_max_serialization_version = False
+    super().setUp()
+
+  def test_simple(self):
+    self.ConvertAndCompare(jnp.sin, 0.7)
+
 
 if __name__ == "__main__":
   # TODO: Remove once tensorflow is 2.10.0 everywhere.
