@@ -19,15 +19,17 @@ too many or too few limitations.
 """
 
 import collections
+from collections.abc import Sequence
 import datetime
 import logging
 import os
-from typing import Any, Dict, Sequence, Tuple
+from typing import Any
 import unittest
 
 from absl.testing import absltest
 
 from jax._src import test_util as jtu
+from jax._src import maps  # Needed for config flags.
 from jax import config
 
 import numpy as np
@@ -38,6 +40,7 @@ config.parse_flags_with_absl()
 from jax.experimental.jax2tf.tests import primitive_harness
 
 
+@jtu.with_config(jax_legacy_prng_key='allow')
 class JaxPrimitiveTest(jtu.JaxTestCase):
 
   # This test runs for all primitive harnesses. For each primitive "xxx" the
@@ -47,7 +50,7 @@ class JaxPrimitiveTest(jtu.JaxTestCase):
   # If you want to run this test for only one harness, add parameter
   # `one_containing="foo"` to parameterized below.
   @primitive_harness.parameterized(primitive_harness.all_harnesses,
-                                   #one_containing="gather_from_slicing_name",
+                                   #one_containing="",
                                    include_jax_unimpl=True)
   @jtu.ignore_warning(category=UserWarning,
                       message="Using reduced precision for gradient.*")
@@ -90,13 +93,13 @@ class JaxPrimitiveTest(jtu.JaxTestCase):
     harnesses = primitive_harness.all_harnesses
     print(f"Found {len(harnesses)} harnesses")
 
-    harness_groups: Dict[str, Sequence[primitive_harness.Harness]] = collections.defaultdict(list)
+    harness_groups: dict[str, Sequence[primitive_harness.Harness]] = collections.defaultdict(list)
 
     def unique_hash(h: primitive_harness.Harness, l: primitive_harness.Limitation):
       return (h.group_name, l.description, l.devices,
               tuple(np.dtype(d).name for d in l.dtypes))
 
-    unique_limitations: Dict[Any, Tuple[primitive_harness.Harness,
+    unique_limitations: dict[Any, tuple[primitive_harness.Harness,
                                         primitive_harness.Limitation]] = {}
 
     for h in harnesses:
