@@ -12,9 +12,12 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 """All the models to convert."""
+
+from collections.abc import Sequence
 import dataclasses
 import functools
-from typing import Any, Callable, Dict, Optional, Sequence, Union
+from typing import Any, Callable, Optional, Union
+import re
 
 import numpy as np
 import jraph
@@ -40,7 +43,7 @@ import tensorflow as tf
 class ModelHarness:
   name: str
   apply: Callable[..., Any]
-  variables: Dict[str, Any]
+  variables: dict[str, Any]
   inputs: Sequence[np.ndarray]
   rtol: float = 1e-4
   polymorphic_shapes: Optional[Sequence[Union[str, None]]] = None
@@ -65,13 +68,17 @@ class ModelHarness:
 
 
 ##### All harnesses in this file.
-ALL_HARNESSES: Dict[str, Callable[[str], ModelHarness]] = {}
+ALL_HARNESSES: dict[str, Callable[[str], ModelHarness]] = {}
 
 
 def _make_harness(harness_fn, name, poly_shapes=None, tensor_specs=None):
-  """Partially apply harness in order to create variables lazily."""
+  """Partially apply harness in order to create variables lazily.
+
+  Note: quotes and commas are stripped from `name` to ensure they can be passed
+        through the command-line.
+  """
   if poly_shapes:
-    name += "_" + str(poly_shapes).replace("'", "").replace('"', "")
+    name += "_" + re.sub(r"(?:'|\"|,)", "", str(poly_shapes))
   if tensor_specs:
     tensor_specs = [tf.TensorSpec(spec, dtype) for spec, dtype in tensor_specs]
   partial_fn = functools.partial(
