@@ -14,7 +14,7 @@
 """Tests for JAX primitive coverage.
 
 The bulk of the testing is done by `test_prim`, which is parameterized by
-about 2000+ test harnesses. See `primitive_harness.py` docstring for a
+about 3500+ test harnesses. See `primitive_harness.py` docstring for a
 description of test harnesses. That module contains also the definitions
 of all the test harnesses, and a specification of which are only partially
 implemented for JAX.
@@ -53,7 +53,7 @@ not need limitations, then it must be listed in the
 
 import datetime
 import os
-from typing import Any, Dict, Tuple
+from typing import Any
 import unittest
 
 from absl import logging
@@ -131,21 +131,7 @@ class JaxPrimitiveTest(tf_test_util.JaxToTfTestCase):
       raise unittest.SkipTest(
           f"TODO(b/272239584): custom call target not guaranteed stable: {target}")
     if config.jax2tf_default_native_serialization:
-      if device == "cpu":
-        if "cholesky_shape" in harness.fullname:
-          skipCustomCallTest("lapack_spotrf, lapack_dpotrf, lapack_zpotrf, lapack_cpotrf")
-        if "eig_shape" in harness.fullname:
-          skipCustomCallTest("lapack_cgeev, lapack_sgeev, lapack_dgeev, lapack_zgeev")
-        if "lu_shape" in harness.fullname:
-          skipCustomCallTest("lapack_zgetrf, lapack_sgetrf")
-        if "svd_shape" in harness.fullname:
-          skipCustomCallTest("lapack_sgesdd, lapack_zgesdd, lapack_cgesdd")
-        if "triangular_solve_" in harness.fullname:
-          skipCustomCallTest("blas_ctrsm, blas_dtrsm, blas_ztrsm, blas_strsm")
-        if "custom_linear_solve" in harness.fullname:
-          skipCustomCallTest("lapack_sgetrf, lapack_dgetrf")
-
-      elif device == "gpu":
+      if device == "gpu":
         if "custom_linear_solve_" in harness.fullname:
           skipCustomCallTest("cusolver_geqrf, cublas_geqrf_batched")
         if "svd_shape" in harness.fullname:
@@ -162,7 +148,7 @@ class JaxPrimitiveTest(tf_test_util.JaxToTfTestCase):
       # TODO(b/264596006): custom calls are not registered properly with TF in OSS
       if (config.jax2tf_default_native_serialization and
           "does not work with custom calls" in str(e)):
-        logging.warning("Supressing error %s", e)
+        logging.warning("Suppressing error %s", e)
         raise unittest.SkipTest("b/264596006: custom calls in native serialization fail in TF")
       else:
         raise e
@@ -218,7 +204,7 @@ class JaxPrimitiveTest(tf_test_util.JaxToTfTestCase):
       return (h.group_name, l.description, l.devices,
               tuple(np.dtype(d).name for d in l.dtypes), l.modes)
 
-    unique_limitations: Dict[Any, Tuple[primitive_harness.Harness, Jax2TfLimitation]] = {}
+    unique_limitations: dict[Any, tuple[primitive_harness.Harness, Jax2TfLimitation]] = {}
     for h in harnesses:
       for l in h.jax_unimplemented:
         if l.enabled:
