@@ -12,12 +12,13 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from collections.abc import Sequence
 import contextlib
 import dataclasses
+import functools
 import re
 import os
-
-from typing import Any, Callable, List, Optional, Sequence, Tuple
+from typing import Any, Callable, Optional
 
 from absl.testing import absltest
 from absl import logging
@@ -30,14 +31,16 @@ from jax import tree_util
 
 from jax import config
 from jax.experimental import jax2tf
+from jax.experimental.export import export
 from jax._src import xla_bridge
 import numpy as np
 import tensorflow as tf  # type: ignore[import]
 from tensorflow.compiler.xla import xla_data_pb2  # type: ignore[import]
+from tensorflow.compiler.tf2xla.python import xla as tfxla  # type: ignore[import]
 
 DType = Any
 
-def _make_tf_input_signature(*tf_args) -> List[tf.TensorSpec]:
+def _make_tf_input_signature(*tf_args) -> list[tf.TensorSpec]:
   # tf_args can be PyTrees
   def _make_one_array_signature(tf_arg):
     return tf.TensorSpec(np.shape(tf_arg), jax2tf.dtype_of_val(tf_arg))
@@ -91,7 +94,7 @@ def SaveAndLoadFunction(f_tf: Callable, *,
                         input_signature: Optional[Sequence[tf.TensorSpec]] = None,
                         input_args: Optional[Sequence[Any]] = None,
                         variables: Sequence[tf.Variable] = (),
-                        save_gradients=True) -> Tuple[Callable, tf.train.Checkpoint]:
+                        save_gradients=True) -> tuple[Callable, tf.train.Checkpoint]:
   # Roundtrip through saved model on disk. Return the Checkpoint also
   # for the cases when there are variables. If you don't pass input_signature
   # then it is created from the input_args.
@@ -151,8 +154,12 @@ def ComputeTfValueAndGrad(tf_f: Callable, tf_args: Sequence,
 
 
 @jtu.with_config(jax_numpy_rank_promotion="allow",
-                 jax_numpy_dtype_promotion='standard')
+                 jax_numpy_dtype_promotion='standard',
+                 jax_legacy_prng_key="allow")
 class JaxToTfTestCase(jtu.JaxTestCase):
+  # We want most tests to use the maximum available version, from the locally
+  # installed tfxla module and export.
+  use_max_serialization_version = True
 
   def setUp(self):
     super().setUp()
@@ -167,6 +174,25 @@ class JaxToTfTestCase(jtu.JaxTestCase):
     self.assertEqual(jtu.device_under_test().upper(),
                      self.tf_default_device.device_type)
 
+    # We run the tests using the maximum version supported, even though
+    # the default serialization version may be held back for a while to
+    # ensure compatibility
+    version = config.jax_serialization_version
+    self.addCleanup(functools.partial(config.update,
+                                      "jax_serialization_version", version))
+    if self.use_max_serialization_version:
+      # Use the largest supported by both export and tfxla.call_module
+      version = min(export.maximum_supported_serialization_version,
+                    tfxla.call_module_maximum_supported_version())
+      self.assertGreaterEqual(version,
+                              export.minimum_supported_serialization_version)
+      config.update("jax_serialization_version", version)
+    logging.info(
+      "Using JAX serialization version %s (export.max_version %s, tf.XlaCallModule max version %s)",
+      version,
+      export.maximum_supported_serialization_version,
+      tfxla.call_module_maximum_supported_version())
+
     with contextlib.ExitStack() as stack:
       stack.enter_context(tf.device(self.tf_default_device))
       self.addCleanup(stack.pop_all().close)
@@ -213,7 +239,7 @@ class JaxToTfTestCase(jtu.JaxTestCase):
 
     func_tf = jax2tf.convert(func_jax, enable_xla=enable_xla)
 
-    unexpected_successes: List[str] = []
+    unexpected_successes: list[str] = []
     # Run the "compiled" mode first, it is most important
     for mode in ("compiled", "eager", "graph"):
       if mode == "graph" and jtu.device_under_test() == "tpu":
