@@ -15,9 +15,10 @@
 """COO (coordinate format) matrix object and associated primitives."""
 from __future__ import annotations
 
+from collections.abc import Sequence
 from functools import partial
 import operator
-from typing import Any, Dict, NamedTuple, Optional, Sequence, Tuple
+from typing import Any, NamedTuple
 import warnings
 
 import numpy as np
@@ -40,7 +41,7 @@ import jax.numpy as jnp
 
 
 Dtype = Any
-Shape = Tuple[int, ...]
+Shape = tuple[int, ...]
 
 class COOInfo(NamedTuple):
   shape: Shape
@@ -63,7 +64,7 @@ class COO(JAXSparse):
   data: jax.Array
   row: jax.Array
   col: jax.Array
-  shape: Tuple[int, int]
+  shape: tuple[int, int]
   nse = property(lambda self: self.data.size)
   dtype = property(lambda self: self.data.dtype)
   _info = property(lambda self: COOInfo(
@@ -73,7 +74,7 @@ class COO(JAXSparse):
   _rows_sorted: bool
   _cols_sorted: bool
 
-  def __init__(self, args: Tuple[Array, Array, Array], *, shape: Shape,
+  def __init__(self, args: tuple[Array, Array, Array], *, shape: Shape,
                rows_sorted: bool = False, cols_sorted: bool = False):
     self.data, self.row, self.col = map(jnp.asarray, args)
     self._rows_sorted = rows_sorted
@@ -81,7 +82,7 @@ class COO(JAXSparse):
     super().__init__(args, shape=shape)
 
   @classmethod
-  def fromdense(cls, mat: Array, *, nse: Optional[int] = None, index_dtype: DTypeLike = np.int32) -> COO:
+  def fromdense(cls, mat: Array, *, nse: int | None = None, index_dtype: DTypeLike = np.int32) -> COO:
     return coo_fromdense(mat, nse=nse, index_dtype=index_dtype)
 
   def _sort_indices(self) -> COO:
@@ -98,7 +99,7 @@ class COO(JAXSparse):
                           rows_sorted=True)
 
   @classmethod
-  def _empty(cls, shape: Sequence[int], *, dtype: Optional[DTypeLike] = None,
+  def _empty(cls, shape: Sequence[int], *, dtype: DTypeLike | None = None,
              index_dtype: DTypeLike = 'int32') -> COO:
     """Create an empty COO instance. Public method is sparse.empty()."""
     shape = tuple(shape)
@@ -110,7 +111,7 @@ class COO(JAXSparse):
                cols_sorted=True)
 
   @classmethod
-  def _eye(cls, N: int, M: int, k: int, *, dtype: Optional[DTypeLike] = None,
+  def _eye(cls, N: int, M: int, k: int, *, dtype: DTypeLike | None = None,
            index_dtype: DTypeLike = 'int32') -> COO:
     if k > 0:
       diag_size = min(N, M - k)
@@ -132,13 +133,13 @@ class COO(JAXSparse):
   def todense(self) -> Array:
     return coo_todense(self)
 
-  def transpose(self, axes: Optional[Tuple[int, ...]] = None) -> COO:
+  def transpose(self, axes: tuple[int, ...] | None = None) -> COO:
     if axes is not None:
       raise NotImplementedError("axes argument to transpose()")
     return COO((self.data, self.col, self.row), shape=self.shape[::-1],
                rows_sorted=self._cols_sorted, cols_sorted=self._rows_sorted)
 
-  def tree_flatten(self) -> Tuple[Tuple[Array, Array, Array], Dict[str, Any]]:
+  def tree_flatten(self) -> tuple[tuple[Array, Array, Array], dict[str, Any]]:
     return (self.data, self.row, self.col), self._info._asdict()
 
   @classmethod
@@ -268,7 +269,7 @@ if gpu_sparse.rocm_is_supported:
 coo_fromdense_p = core.Primitive('coo_fromdense')
 coo_fromdense_p.multiple_results = True
 
-def coo_fromdense(mat: Array, *, nse: Optional[int] = None, index_dtype: DTypeLike = jnp.int32) -> COO:
+def coo_fromdense(mat: Array, *, nse: int | None = None, index_dtype: DTypeLike = jnp.int32) -> COO:
   """Create a COO-format sparse matrix from a dense matrix.
 
   Args:
@@ -286,7 +287,7 @@ def coo_fromdense(mat: Array, *, nse: Optional[int] = None, index_dtype: DTypeLi
   return COO(_coo_fromdense(mat, nse=nse_int, index_dtype=index_dtype),
              shape=mat.shape, rows_sorted=True)
 
-def _coo_fromdense(mat: Array, *, nse: int, index_dtype: DTypeLike = jnp.int32) -> Tuple[Array, Array, Array]:
+def _coo_fromdense(mat: Array, *, nse: int, index_dtype: DTypeLike = jnp.int32) -> tuple[Array, Array, Array]:
   """Create COO-format sparse matrix from a dense matrix.
 
   Args:
