@@ -47,14 +47,16 @@ Array([-1.2655463 , -0.52060574, -0.14522289, -0.10817424,
        -0.15574613], dtype=float32)
 """
 
+from collections.abc import Sequence
 import functools
-from typing import (
-  Any, Callable, Dict, NamedTuple, List, Optional, Sequence, Tuple)
+from typing import Any, Callable, NamedTuple, Optional
 
 import numpy as np
 
+import jax
 from jax import lax
 from jax._src import core
+from jax._src.custom_derivatives import lift_jvp
 from jax._src import linear_util as lu
 from jax._src import pjit
 from jax._src import sharding_impls
@@ -71,8 +73,8 @@ from jax._src.numpy import lax_numpy
 from jax.experimental import sparse
 from jax.experimental.sparse import BCOO, BCSR
 
-sparse_rules_bcoo : Dict[core.Primitive, Callable] = {}
-sparse_rules_bcsr : Dict[core.Primitive, Callable] = {}
+sparse_rules_bcoo : dict[core.Primitive, Callable] = {}
+sparse_rules_bcsr : dict[core.Primitive, Callable] = {}
 
 _zero_preserving_linear_unary_primitives = [
   lax.copy_p,
@@ -99,7 +101,7 @@ _zero_preserving_unary_primitives = [
   lax.convert_element_type_p,
 ]
 
-_densifying_primitives : List[core.Primitive] = [
+_densifying_primitives : list[core.Primitive] = [
   lax.acos_p,
   lax.acosh_p,
   lax.bessel_i0e_p,
@@ -135,7 +137,7 @@ class SparsifyEnv:
   that may be shared between one or more SparsifyValue objects, which
   represent sparse or dense arrays via indices into the list of buffers.
   """
-  _buffers : List[Array]
+  _buffers : list[Array]
 
   def __init__(self, bufs=()):
     self._buffers = list(bufs)
@@ -193,12 +195,12 @@ class SparsifyEnv:
 
 
 class SparsifyValue(NamedTuple):
-  shape: Tuple[int, ...]
-  data_ref: Optional[int]
-  indices_ref: Optional[int] = None
-  indptr_ref: Optional[int] = None
-  indices_sorted: Optional[bool] = False
-  unique_indices: Optional[bool] = False
+  shape: tuple[int, ...]
+  data_ref: int | None
+  indices_ref: int | None = None
+  indptr_ref: int | None = None
+  indices_sorted: bool | None = False
+  unique_indices: bool | None = False
 
   @property
   def ndim(self):
@@ -343,6 +345,11 @@ class SparseTrace(core.Trace):
     setnewattr(self.main, 'spenv', SparsifyEnv(bufs_out))
     return [SparseTracer(self, spvalue=spvalue) for spvalue in out_spvalues()]
 
+  def process_custom_jvp_call(self, primitive, fun, jvp, tracers, *, symbolic_zeros):
+    # TODO(jakevdp): handle the jvp here
+    del primitive, jvp, symbolic_zeros
+    return fun.call_wrapped(*tracers)
+
 @lu.transformation_with_aux
 def sparsify_subtrace(main, spvalues, *bufs):
   setnewattr(main, 'spenv', SparsifyEnv(bufs))
@@ -353,7 +360,7 @@ def sparsify_subtrace(main, spvalues, *bufs):
   buffers = popattr(main, 'spenv')._buffers
   yield buffers, [out._spvalue for out in out_traces]
 
-def sparsify_fun(wrapped_fun, args: List[ArrayOrSparse]):
+def sparsify_fun(wrapped_fun, args: list[ArrayOrSparse]):
   with core.new_main(SparseTrace) as main:
     spenv = SparsifyEnv()
     spvalues = arrays_to_spvalues(spenv, args)
@@ -383,7 +390,7 @@ def eval_sparse(
     spvalues: Sequence[SparsifyValue],  # mix of sparse and dense pointers into spenv
     spenv: SparsifyEnv,
 ) -> Sequence[SparsifyValue]:
-  env : Dict[core.Var, SparsifyValue] = {}
+  env : dict[core.Var, SparsifyValue] = {}
 
   def read(var: core.Atom) -> SparsifyValue:
     # all literals are dense
@@ -435,7 +442,7 @@ def sparsify_raw(f):
 
   def wrapped(
       spenv: SparsifyEnv, *spvalues: SparsifyValue, **params: Any
-  ) -> Tuple[Sequence[SparsifyValue], pytree.PyTreeDef]:
+  ) -> tuple[Sequence[SparsifyValue], pytree.PyTreeDef]:
     spvalues_flat, in_tree = tree_flatten(spvalues, is_leaf=_is_spvalue)
     in_avals_flat = spvalues_to_avals(spenv, spvalues_flat)
     wrapped_fun, out_tree = flatten_fun_nokwargs(lu.wrap_init(f, params), in_tree)
@@ -720,7 +727,7 @@ def _sparsify_jaxpr(spenv, jaxpr, *spvalues):
   #   shared data & indices when generating the sparsified jaxpr. The
   #   current approach produces valid sparsified while loops, but they
   #   don't work in corner cases (see associated TODO in sparsify_test.py)
-  out_tree: Optional[pytree.PyTreeDef] = None
+  out_tree: pytree.PyTreeDef | None = None
 
   @lu.wrap_init
   def wrapped(*args_flat):
@@ -818,10 +825,10 @@ def _scan_sparse(spenv, *spvalues, jaxpr, num_consts, num_carry, **params):
   # params['linear'] has one entry per arg; expand it to match the sparsified args.
   const_linear, carry_linear, xs_linear = split_list(
     params.pop('linear'), [num_consts, num_carry])
-  sp_linear = tuple([
+  sp_linear = (
     *_duplicate_for_sparse_spvalues(const_spvalues, const_linear),
     *_duplicate_for_sparse_spvalues(carry_spvalues, carry_linear),
-    *_duplicate_for_sparse_spvalues(xs_spvalues, xs_linear)])
+    *_duplicate_for_sparse_spvalues(xs_spvalues, xs_linear))
 
   out = lax.scan_p.bind(*consts, *carry, *xs, jaxpr=sp_jaxpr, linear=sp_linear,
                         num_consts=len(consts), num_carry=len(carry), **params)
@@ -851,6 +858,25 @@ def _todense_sparse_rule(spenv, spvalue, *, tree):
   return (spenv.dense(out),)
 
 sparse_rules_bcoo[sparse.todense_p] = _todense_sparse_rule
+sparse_rules_bcsr[sparse.todense_p] = _todense_sparse_rule
+
+def _custom_jvp_sparse_rule(spenv, *spvalues, **params):
+  call_jaxpr = params.pop('call_jaxpr')
+  jvp_jaxpr_thunk = params.pop('jvp_jaxpr_thunk')
+  num_consts = params.pop('num_consts')
+  sp_call_jaxpr, out_tree = _sparsify_jaxpr(spenv, call_jaxpr, *spvalues)
+  @lu.wrap_init
+  def fun(*arrs):
+    sparrs = arrays_to_spvalues(spenv, arrs)
+    out = eval_sparse(call_jaxpr.jaxpr, call_jaxpr.consts, sparrs, spenv)
+    return spvalues_to_arrays(spenv, out)
+  jvp = lift_jvp(num_consts, jvp_jaxpr_thunk)
+  invals = spvalues_to_arrays(spenv, spvalues)
+  outvals = jax.custom_derivatives.custom_jvp_call_p.bind(fun, jvp, *invals, **params)
+  return arrays_to_spvalues(spenv, outvals)
+
+sparse_rules_bcoo[jax.custom_derivatives.custom_jvp_call_p] = _custom_jvp_sparse_rule
+sparse_rules_bcsr[jax.custom_derivatives.custom_jvp_call_p] = _custom_jvp_sparse_rule
 
 
 # ------------------------------------------------------------------------------
