@@ -21,7 +21,7 @@ from jax._src.interpreters.batching import (
   BatchTrace as BatchTrace,
   BatchTracer as BatchTracer,
   BatchingRule as BatchingRule,
-  ConcatAxis as ConcatAxis,
+  RaggedAxis as RaggedAxis,
   Elt as Elt,
   FromEltHandler as FromEltHandler,
   GetIdx as GetIdx,
@@ -29,9 +29,9 @@ from jax._src.interpreters.batching import (
   MakeIotaHandler as MakeIotaHandler,
   MapSpec as MapSpec,
   NotMapped as NotMapped,
-  Pile as Pile,
-  PileAxis as PileAxis,
-  PileTy as PileTy,
+  Jumble as Jumble,
+  JumbleAxis as JumbleAxis,
+  JumbleTy as JumbleTy,
   ToEltHandler as ToEltHandler,
   Vmappable as Vmappable,
   Zero as Zero,
@@ -60,17 +60,14 @@ from jax._src.interpreters.batching import (
   matchaxis as matchaxis,
   moveaxis as moveaxis,
   not_mapped as not_mapped,
-  pile_axis as pile_axis,
+  jumble_axis as jumble_axis,
   primitive_batchers as primitive_batchers,
-  reassemble_concat_axes as reassemble_concat_axes,
   reducer_batcher as reducer_batcher,
   register_vmappable as register_vmappable,
-  segment_sum as segment_sum,
   spec_types as spec_types,
   spmd_axis_primitive_batchers as spmd_axis_primitive_batchers,
   to_elt as to_elt,
   to_elt_handlers as to_elt_handlers,
-  unpack_concat_axes as unpack_concat_axes,
   unregister_vmappable as unregister_vmappable,
   vectorized_batcher as vectorized_batcher,
   vmappables as vmappables,
