@@ -93,6 +93,8 @@ from jax._src.lax.lax import (
   eq_p as eq_p,
   exp as exp,
   exp_p as exp_p,
+  exp2 as exp2,
+  exp2_p as exp2_p,
   expand_dims as expand_dims,
   expm1 as expm1,
   expm1_p as expm1_p,
@@ -244,9 +246,13 @@ from jax._src.lax.special import (
   igamma_p as igamma_p,
   lgamma as lgamma,
   lgamma_p as lgamma_p,
+  polygamma as polygamma,
+  polygamma_p as polygamma_p,
   random_gamma_grad as random_gamma_grad,
   random_gamma_grad_p as random_gamma_grad_p,
   regularized_incomplete_beta_p as regularized_incomplete_beta_p,
+  zeta as zeta,
+  zeta_p as zeta_p,
 )
 from jax._src.lax.slicing import (
   GatherDimensionNumbers as GatherDimensionNumbers,
@@ -368,7 +374,6 @@ from jax._src.lax.ann import (
 from jax._src.ad_util import stop_gradient_p as stop_gradient_p
 from jax.lax import linalg as linalg
 
-from jax._src.pjit import with_sharding_constraint
-from jax._src.dispatch import device_put_p
-
-from math import prod  # TODO(phawkins): remove this accidental export
+from jax._src.pjit import with_sharding_constraint as with_sharding_constraint
+from jax._src.pjit import sharding_constraint_p as sharding_constraint_p
+from jax._src.dispatch import device_put_p as device_put_p
