@@ -15,6 +15,8 @@
 # Note: import <name> as <name> is required for names to be exported.
 # See PEP 484 & https://github.com/google/jax/issues/7570
 
+import numpy as _numpy
+
 from jax.numpy import fft as fft
 from jax.numpy import linalg as linalg
 
@@ -22,9 +24,6 @@ from jax._src.basearray import Array as ndarray
 
 from jax._src.numpy.lax_numpy import (
     ComplexWarning as ComplexWarning,
-    NINF as NINF,
-    NZERO as NZERO,
-    PZERO as PZERO,
     allclose as allclose,
     angle as angle,
     append as append,
@@ -107,8 +106,11 @@ from jax._src.numpy.lax_numpy import (
     float16 as float16,
     float32 as float32,
     float64 as float64,
+    float8_e4m3b11fnuz as float8_e4m3b11fnuz,
     float8_e4m3fn as float8_e4m3fn,
+    float8_e4m3fnuz as float8_e4m3fnuz,
     float8_e5m2 as float8_e5m2,
+    float8_e5m2fnuz as float8_e5m2fnuz,
     float_ as float_,
     floating as floating,
     fmax as fmax,
@@ -142,6 +144,7 @@ from jax._src.numpy.lax_numpy import (
     inf as inf,
     inner as inner,
     insert as insert,
+    int4 as int4,
     int8 as int8,
     int16 as int16,
     int32 as int32,
@@ -156,7 +159,6 @@ from jax._src.numpy.lax_numpy import (
     isrealobj as isrealobj,
     isscalar as isscalar,
     issubdtype as issubdtype,
-    issubsctype as issubsctype,
     iterable as iterable,
     ix_ as ix_,
     kaiser as kaiser,
@@ -204,7 +206,6 @@ from jax._src.numpy.lax_numpy import (
     rot90 as rot90,
     round as round,
     round_ as round_,
-    row_stack as row_stack,
     save as save,
     savez as savez,
     searchsorted as searchsorted,
@@ -225,7 +226,7 @@ from jax._src.numpy.lax_numpy import (
     tensordot as tensordot,
     tile as tile,
     trace as trace,
-    trapz as trapz,
+    trapz as _deprecated_trapz,
     transpose as transpose,
     tri as tri,
     tril as tril,
@@ -237,6 +238,7 @@ from jax._src.numpy.lax_numpy import (
     triu_indices_from as triu_indices_from,
     trunc as trunc,
     uint as uint,
+    uint4 as uint4,
     uint8 as uint8,
     uint16 as uint16,
     uint32 as uint32,
@@ -254,22 +256,6 @@ from jax._src.numpy.lax_numpy import (
     zeros_like as zeros_like,
 )
 
-# TODO(phawkins): make this import unconditional after increasing the ml_dtypes
-# minimum version.
-import jax._src.numpy.lax_numpy
-if hasattr(jax._src.numpy.lax_numpy, "int4"):
-  from jax._src.numpy.lax_numpy import (
-    int4 as int4,
-    uint4 as uint4,
-  )
-# TODO(jakevdp): make this import unconditional after increasing the minimum
-# version for ml_dtypes and jaxlib
-if hasattr(jax._src.numpy.lax_numpy, "float8_e4m3b11fnuz"):
-  from jax._src.numpy.lax_numpy import (
-    float8_e4m3b11fnuz as float8_e4m3b11fnuz,
-  )
-
-
 from jax._src.numpy.index_tricks import (
   c_ as c_,
   index_exp as index_exp,
@@ -327,7 +313,7 @@ from jax._src.numpy.reductions import (
 )
 
 from jax._src.numpy.setops import (
-    in1d as in1d,
+    in1d as _deprecated_in1d,
     intersect1d as intersect1d,
     isin as isin,
     setdiff1d as setdiff1d,
@@ -427,6 +413,11 @@ from jax._src.numpy.ufuncs import (
     true_divide as true_divide,
 )
 
+from jax._src.numpy.ufunc_api import (
+    frompyfunc as frompyfunc,
+    ufunc as ufunc,
+)
+
 from jax._src.numpy.vectorize import vectorize as vectorize
 
 # Dynamically register numpy-style methods on JAX arrays.
@@ -438,39 +429,56 @@ del register_jax_array_methods
 # Deprecations
 
 _deprecations = {
-    # Added March 14, 2023:
-    "DeviceArray": (
-        "jax.numpy.DeviceArray is deprecated. Use jax.Array.",
-        ndarray,
+    # Added August 10, 2023:
+    "NINF": (
+        "jax.numpy.NINF is deprecated. Use -jax.numpy.inf instead.",
+        -inf,
     ),
-    # Added June 2, 2023:
-    "alltrue": (
-        "jax.numpy.alltrue is deprecated. Use jax.numpy.all",
-        all,
+    "NZERO": (
+        "jax.numpy.NZERO is deprecated. Use -0.0 instead.",
+        -0.0,
     ),
-    "cumproduct": (
-        "jax.numpy.cumproduct is deprecated. Use jax.numpy.cumprod",
-        cumprod,
+    "PZERO": (
+        "jax.numpy.PZERO is deprecated. Use 0.0 instead.",
+        0.0,
     ),
-    "product": (
-        "jax.numpy.product is deprecated. Use jax.numpy.prod",
-        prod,
+    # Added Aug 17, 2023:
+    "issubsctype": (
+        "jax.numpy.issubsctype is deprecated. In most cases, jax.numpy.issubdtype can be used instead.",
+        _numpy.core.numerictypes.issubsctype,
     ),
-    "sometrue": (
-        "jax.numpy.sometrue is deprecated. Use jax.numpy.any",
-        any,
+    # Added Aug 22, 2023
+    "row_stack": (
+        "jax.numpy.row_stack is deprecated. Use jax.numpy.vstack instead.",
+        vstack,
+    ),
+    # Added Aug 23, 2023
+    "in1d": (
+        "jax.numpy.in1d is deprecated. Use jax.numpy.isin instead.",
+        _deprecated_in1d,
+    ),
+    # Added Aug 24, 2023
+    "trapz": (
+        "jax.numpy.trapz is deprecated. Use jax.scipy.integrate.trapezoid instead.",
+        _deprecated_trapz,
     ),
 }
 
 import typing
 if typing.TYPE_CHECKING:
-  from jax._src.basearray import Array as DeviceArray
-  alltrue = all
-  cumproduct = cumprod
-  product = prod
-  sometrue = any
+  row_stack = vstack
+  NINF = -inf
+  NZERO = -0.0
+  PZERO = 0.0
+  issubsctype = _numpy.core.numerictypes.issubsctype
+  in1d = _deprecated_in1d
+  trapz = _deprecated_trapz
 else:
   from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
   __getattr__ = _deprecation_getattr(__name__, _deprecations)
   del _deprecation_getattr
 del typing
+del _numpy
+
+del _deprecated_in1d
+del _deprecated_trapz
